// WARNING!
// There is a static array that could potentially go out of bounds.
// Its size is set by defined by Scanline::MAX_CUT_LINES.

#pragma once
#include <map>
using namespace std;

#include "common.h"
#include "vect.h"
#include "list.h"
#include "array.h"
#include "array2d.h"

struct Line
{
#ifdef USE_ONLY_LINES
	static const int num_pts = 2;
	vect2d p[2];
#elif defined(USE_ONLY_QUADRATICS)
	static const int num_pts = 3;
	vect2d p[3];
#elif defined(USE_ONLY_CUBICS)
	static const int num_pts = 4;
	vect2d p[4];
#else
	int num_pts;
	vect2d p[4];
#endif
};

void rasterize(Array2D<double> &img, Array<Line> &bound);





#ifdef SORT_LINES

struct LineRaster
{
	LineRaster *next;
	LineRaster *next_aae;

#ifdef USE_ONLY_LINES
	static const int num_pts = 2;
	vect2d px[2], py[2]; // control points of curve segments stepping in the y and x directions
#elif defined(USE_ONLY_QUADRATICS)
	static const int num_pts = 3;
	vect2d px[3], py[3]; // control points of curve segments stepping in the y and x directions
#elif defined(USE_ONLY_CUBICS)
	static const int num_pts = 4;
	vect2d px[4], py[4]; // control points of curve segments stepping in the y and x directions
#else
	int num_pts;
	vect2d px[4], py[4]; // control points of curve segments stepping in the y and x directions
#endif

	int end_y;
	int start, end;

	vect2d high_y;
	union
	{
		struct
		{
			double dx, dy; // deriv wrt to x,y in the linear case
		} line;
		struct
		{
			// mid_y is used to store the mid point of the remainder after the first cut 
			// (because the first segment could be arbitrarily short)
			vect2d mid_x, mid_y; 
		} quad;
		struct
		{
			// mid_y is used to store the mid point of the remainder after the first cut 
			// (because the first segment could be arbitrarily short)
			vect2d mid_x[2], mid_y[2]; 
		} cube;
	};

	int flip_x, flip_y;
};

struct EdgeTable
{
	Array< LineRaster* > lists;
	Array< LineRaster > heap;
	int start, end;
};

#else


struct LineStep
{
	LineStep *next;

#ifdef USE_ONLY_LINES
	static const int num_pts = 2;
	vect2d px[1], py[2]; // control points of curve segments stepping in the y and x directions
#elif defined(USE_ONLY_QUADRATICS)
	static const int num_pts = 3;
	vect2d px[1], py[3]; // control points of curve segments stepping in the y and x directions
#elif defined(USE_ONLY_CUBICS)
	static const int num_pts = 4;
	vect2d px[1], py[4]; // control points of curve segments stepping in the y and x directions
#else
	int num_pts;
	vect2d px[1], py[4]; // control points of curve segments stepping in the y and x directions
#endif

	int end_y;

	vect2d high_y;
	union
	{
		struct
		{
			double dx, dy; // deriv wrt to x,y in the linear case
		} line;
		struct
		{
			// mid_y is used to store the mid point of the remainder after the first cut 
			// (because the first segment could be arbitrarily short)
			vect2d mid_x, mid_y; 
		} quad;
		struct
		{
			// mid_y is used to store the mid point of the remainder after the first cut 
			// (because the first segment could be arbitrarily short)
			vect2d mid_x[2], mid_y[2]; 
		} cube;
	};

	int flip_x, flip_y;
};

struct LineRaster
{
	LineRaster *next;

#ifdef USE_ONLY_LINES
	static const int num_pts = 2;
	vect2d px[2]; // control points of curve segments stepping in the y and x directions
#elif defined(USE_ONLY_QUADRATICS)
	static const int num_pts = 3;
	vect2d px[3]; // control points of curve segments stepping in the y and x directions
#elif defined(USE_ONLY_CUBICS)
	static const int num_pts = 4;
	vect2d px[4]; // control points of curve segments stepping in the y and x directions
#else
	int num_pts;
	vect2d px[4]; // control points of curve segments stepping in the y and x directions
#endif

	int flip_x, flip_y;
};

struct EdgeTable
{
	Array<LineStep*> lists; // the curves that start on this scanline (pointers to heap)
	Array<LineStep> heap; // stores the lines in the lists
	int start, end;
};

struct Scanline
{
	static const int MAX_CUT_LINES = 1000000;
	Array<LineRaster*> lists; // the curves that start on this scanline (pointers to heap)
	Array<LineRaster> heap; // stores the lines in the lists
	int start, end;
};

#endif
