#include <stdio.h>
#include <string>
#include <fstream>
#include "rasterizer.h"
#include "common.h"

void poly_to_bound(Array<Line> &lines, Array<vect2d> &poly)
{
	for (int i = 0; i < poly.s; i++)
	{
		Line c;
#ifdef USE_MIXED_CURVES
		c.num_pts = 2;
#endif
		c.p[0] = poly[i];
		c.p[1] = poly[(i+1)%poly.s];
		lines.push_back(c);
	}
}

void create_test_boundary(Array<Line> &lines, const int img_size)
{
	Array<vect2d> poly;
	vect2d pt;

#ifdef USE_FILE
#define TEST_NAME FILE
#endif

#ifdef TEST_HALF
#define TEST_NAME half
	// half is a polygon
	pt.set(0, 0); poly.push_back(pt);
	pt.set(.5, 0); poly.push_back(pt);
	pt.set(.5, 1); poly.push_back(pt);
	pt.set(0, 1); poly.push_back(pt);
	poly_to_bound(lines, poly);
#endif

#ifdef TEST3A
#define TEST_NAME radialA
	// radial polygon test
	//int n = 512;
	int n = 128;
	const double pi = 3.1415926535;
	for (int i = 0; i < n/4; i++)
	{
		double t1 = i * 2.0 * pi / n;
		double t2 = (i + .5) * 2.0 * pi / n;

		poly.s = 0;
		pt.set(0, 0); poly.push_back(pt);
		pt.set(cos(t1)*.98, sin(t1)*.98); poly.push_back(pt);
		pt.set(cos(t2)*.98, sin(t2)*.98); poly.push_back(pt);
		poly_to_bound(lines, poly);
	}
#endif
	
#ifdef TEST5A
#define TEST_NAME quadZoneA
	// Quadratic zone plate
	Array<Line> proto;
	Line b;
#ifdef USE_MIXED_CURVES
	b.num_pts = 3;
#endif

	//b.p[0].set(1,0);
	//b.p[1].set(1,1);
	//b.p[2].set(0,1);
	//proto.push_back(b);
	
	const double s2 = sqrt(.5);
	const double S2 = sqrt(2.0);

	b.p[0].set(1,0);
	b.p[1].set(1,S2-1);
	b.p[2].set(s2,s2);
	proto.push_back(b);
	
	b.p[0].set(s2,s2);
	b.p[1].set(S2-1,1);
	b.p[2].set(0,1);
	proto.push_back(b);

	//for (int i = 1; i < 64/8; i+=2)
	for (int i = 1; i < 64; i+=2)
	{
		//double w = .085;
		double w = .003*4;
		double factor1 = sqrt(i*w + w*w*i*i*.25);
		double factor2 = sqrt((i+1)*w + w*w*(i+1)*(i+1)*.25);
		if (sqrt((i+2)*w + w*w*(i+2)*(i+2)*.25) > 1)
			break;

		Line a;

		// outside
		a = proto[0];
		a.num_pts = 3;
		for (int k = 0; k < 3; k++)
			a.p[k] = a.p[k] * factor2;
		lines.push_back(a);

		a = proto[1];
		a.num_pts = 3;
		for (int k = 0; k < 3; k++)
			a.p[k] = a.p[k] * factor2;
		lines.push_back(a);

		a.num_pts = 2;
		a.p[0] = proto[1].p[2] * factor2;
		a.p[1] = proto[1].p[2] * factor1;
		lines.push_back(a);


		// inside
		a = proto[1];
		a.num_pts = 3;
		for (int k = 0; k < 3; k++)
			a.p[k] = a.p[k] * factor1;
		swap(a.p[0], a.p[2]);
		lines.push_back(a);

		a = proto[0];
		a.num_pts = 3;
		for (int k = 0; k < 3; k++)
			a.p[k] = a.p[k] * factor1;
		swap(a.p[0], a.p[2]);
		lines.push_back(a);

		a.num_pts = 2;
		a.p[0] = proto[0].p[0] * factor1;
		a.p[1] = proto[0].p[0] * factor2;
		lines.push_back(a);
	}
#endif

#ifdef TEST1
#define TEST_NAME small
	// small
	Line l;
	l.num_pts = 2;
	l.p[0].set(.5,.5);
	l.p[1].set(.75,.505);
	lines.push_back(l);

	/*// box
	pt.set(0+.25,0+.25); poly.push_back(pt);
	pt.set(1-.25,0+.25); poly.push_back(pt);
	pt.set(1-.25,1-.25); poly.push_back(pt);
	poly_to_bound(lines, poly);*/
	
	//Line b;
	/*b.num_pts = 3;
	b.p[0].set(.9,.1);
	b.p[1].set(.9,.9);
	b.p[2].set(.1,.9);
	lines.push_back(b);*/

	//b.num_pts = 3;
	//b.p[0].set(.1,.5);
	//b.p[1].set(.1,.1);
	//b.p[2].set(.5,.1);
	//lines.push_back(b);

	//b.num_pts = 3;
	//b.p[0].set(.5,.1);
	//b.p[1].set(.9,.1);
	//b.p[2].set(.9,.5);
	//lines.push_back(b);

	//b.num_pts = 3;
	//b.p[0].set(.45,.1);
	//b.p[1].set(1.2,1.2);
	//b.p[2].set(.1,.45);
	//lines.push_back(b);
	//
	//b.num_pts = 2;
	//b.p[0].set(.1,.45);
	//b.p[1].set(.45,.1);
	//lines.push_back(b);

	/*b.num_pts = 4;
	b.p[0].set(.1,.5);
	b.p[1].set(.1,.1);
	b.p[2].set(.9,.1);
	b.p[3].set(.9,.5);
	lines.push_back(b);*/

	/*b.num_pts = 4;
	b.p[0].set(.1,.1);
	b.p[1].set(.9,.1);
	b.p[2].set(.1,.9);
	b.p[3].set(.1,.1);
	lines.push_back(b);*/

	/*b.num_pts = 4;
	b.p[0].set(.1,.1);
	b.p[1].set(.9,.1);
	b.p[2].set(.1,.9);
	b.p[3].set(.9,.9);
	lines.push_back(b);*/
#endif
	
#ifdef TEST2
#define TEST_NAME concave
	// concave poly
	pt.set(.10242, .13534); poly.push_back(pt);
	pt.set(.73242, .66432); poly.push_back(pt);
	pt.set(.91243, .17742); poly.push_back(pt);
	pt.set(.83151, .88431); poly.push_back(pt);
	pt.set(.14864, .72564); poly.push_back(pt);
	pt.set(.23354, .44168); poly.push_back(pt);
	poly_to_bound(lines, poly);
#endif
	
#ifdef TEST3
#define TEST_NAME radial
	// radial polygon test
	//int n = 512;
	int n = 128;
	const double pi = 3.1415926535;
	for (int i = 0/* + n/4-1*/; i < n/*/4+1*/; i++)
	{
		double t1 = i * 2.0 * pi / n;
		double t2 = (i + .5) * 2.0 * pi / n;

		poly.s = 0;
		pt.set(.5, .5); poly.push_back(pt);
		pt.set(.5 + cos(t1)*.49, .5 + sin(t1)*.49); poly.push_back(pt);
		pt.set(.5 + cos(t2)*.49, .5 + sin(t2)*.49); poly.push_back(pt);
		poly_to_bound(lines, poly);
	}
#endif
	
#ifdef TEST4
#define TEST_NAME linePerspective
	// Linear test pattern
	//int fake_img_size = 512;
	int fake_img_size = 128;
	double px = 1.0 / fake_img_size;
	int n = fake_img_size * 2;
	double dx = 1.0/n;
	double dx2 = dx * 8;
	for (int i = 0; i < n; i++)
	{
		poly.s = 0;

		if ((i+.5)*dx2 <= 1)
		{
			pt.set(i*dx, 0); poly.push_back(pt);
			pt.set((i+.5)*dx, 0); poly.push_back(pt);
			pt.set((i+.5)*dx2, 1); poly.push_back(pt);
			pt.set(i*dx2, 1); poly.push_back(pt);
			poly_to_bound(lines, poly);
		}
		else
		{
			pt.set(i*dx, 0); poly.push_back(pt);
			pt.set((i+.5)*dx, 0); poly.push_back(pt);

			{
				double x1 = (i+.5)*dx;
				double x2 = (i+.5)*dx2;
				double y1 = 0;
				double y2 = 1;
				double t = (1 - x1) / (x2 - x1);
				if (y1 + t*(y2 - y1) <= 0)
					continue;
				pt.set(1, y1 + t*(y2 - y1)); poly.push_back(pt);
			}
			{
				double x1 = i*dx;
				double x2 = i*dx2;
				double y1 = 0;
				double y2 = 1;
				double t = (1 - x1) / (x2 - x1);
				if (y1 + t*(y2 - y1) <= 0)
					continue;
				pt.set(1, y1 + t*(y2 - y1)); poly.push_back(pt);
			}

			poly_to_bound(lines, poly);
		}
	}
#endif
	
#ifdef TEST5
#define TEST_NAME quadZone
	// Quadratic zone plate
	Array<Line> proto;
	Line b;
#ifdef USE_MIXED_CURVES
	b.num_pts = 3;
#endif

	//b.p[0].set(0,.5);
	//b.p[1].set(0,0);
	//b.p[2].set(.5,0);
	//proto.push_back(b);

	//b.p[0].set(.5,0);
	//b.p[1].set(1,0);
	//b.p[2].set(1,.5);
	//proto.push_back(b);

	//b.p[0].set(1,.5);
	//b.p[1].set(1,1);
	//b.p[2].set(.5,1);
	//proto.push_back(b);

	//b.p[0].set(.5,1);
	//b.p[1].set(0,1);
	//b.p[2].set(0,.5);
	//proto.push_back(b);

	
	const double s2 = sqrt(.5);
	const double S2 = sqrt(2.0);

	b.p[0].set(1,0);
	b.p[1].set(1,S2-1);
	b.p[2].set(s2,s2);
	proto.push_back(b);
	
	b.p[0].set(s2,s2);
	b.p[1].set(S2-1,1);
	b.p[2].set(0,1);
	proto.push_back(b);

	b.p[0].set(-s2,s2);
	b.p[1].set(-1,S2-1);
	b.p[2].set(-1,0);
	proto.push_back(b);
	
	b.p[0].set(0,1);
	b.p[1].set(-S2+1,1);
	b.p[2].set(-s2,s2);
	proto.push_back(b);


	b.p[2].set(1,0);
	b.p[1].set(1,-S2+1);
	b.p[0].set(s2,-s2);
	proto.push_back(b);
	
	b.p[2].set(s2,-s2);
	b.p[1].set(S2-1,-1);
	b.p[0].set(0,-1);
	proto.push_back(b);

	b.p[2].set(-s2,-s2);
	b.p[1].set(-1,-S2+1);
	b.p[0].set(-1,0);
	proto.push_back(b);
	
	b.p[2].set(0,-1);
	b.p[1].set(-S2+1,-1);
	b.p[0].set(-s2,-s2);
	proto.push_back(b);
	
	vect2d center; center = .5;
	for (int i = 1; i < 250; i++)
	//for (int i = 1; i < 64; i++)
	{
		double w = .003;
		//double w = .003*4;
		double factor = sqrt(i*w + w*w*i*i*.25);
		if ((i % 2) == 0)
			if (sqrt((i+1)*w + w*w*(i+1)*(i+1)*.25) > 1)
				break;

		if ((i % 2) == 0)
		{
			for (int j = proto.s-1; j >= 0; j--)
			{
				Line c = proto[j];

				for (int k = 0; k < c.num_pts; k++)
					c.p[k] = center + c.p[k] * (factor * .5);

				for (int k = 0; k < c.num_pts/2; k++)
					swap(c.p[k], c.p[c.num_pts-k-1]);

				lines.push_back(c);
			}
		}
		else
		{
			for (int j = 0; j < proto.s; j++)
			{
				Line c = proto[j];

				for (int k = 0; k < c.num_pts; k++)
					c.p[k] = center + c.p[k] * (factor * .5);

				lines.push_back(c);
			}
		}
	}
#endif

#ifdef TEST6
#define TEST_NAME clover
	// Cubic zone plate
	Array<Line> proto;
	Line b;
#ifdef USE_MIXED_CURVES
	b.num_pts = 4;
#endif

	b.p[0].set(.05,0);
	b.p[1].set(1,0);
	b.p[2].set(0,1);
	b.p[3].set(0,.05);
	proto.push_back(b);

	b.p[0].set(0,.05);
	b.p[1].set(0,1);
	b.p[2].set(-1,0);
	b.p[3].set(-.05,0);
	proto.push_back(b);

	b.p[0].set(-.05,0);
	b.p[1].set(-1,0);
	b.p[2].set(0,-1);
	b.p[3].set(0,-.05);
	proto.push_back(b);

	b.p[0].set(0,-.05);
	b.p[1].set(0,-1);
	b.p[2].set(1,0);
	b.p[3].set(.05,0);
	proto.push_back(b);
	
	vect2d center; center = .5;
	for (int i = 1; i < 200; i++)
	//for (int i = 1; i < 4; i++)
	{
		double w = .004;
		//double w = .004*50;
		double factor = sqrt(i*w + w*w*i*i*.25);
		if ((i % 2) == 0)
			if (sqrt((i+1)*w + w*w*(i+1)*(i+1)*.25) > 1)
				break;

		if ((i % 2) == 0)
		{
			for (int j = proto.s-1; j >= 0; j--)
			{
				Line c = proto[j];

				for (int k = 0; k < c.num_pts; k++)
					c.p[k] = center + c.p[k] * factor;

				for (int k = 0; k < c.num_pts/2; k++)
					swap(c.p[k], c.p[c.num_pts-k-1]);

				lines.push_back(c);
			}
		}
		else
		{
			for (int j = 0; j < proto.s; j++)
			{
				Line c = proto[j];

				for (int k = 0; k < c.num_pts; k++)
					c.p[k] = center + c.p[k] * factor;

				lines.push_back(c);
			}
		}
	}
#endif

	// scale up
	const double FILTER_OFFSET = -1 + .5*(FILTER_SIZE - 1) - floor(.5*(FILTER_SIZE - 1));

	for (int i = 0; i < lines.s; i++)
	{
		Line &c = lines[i];
		for (int j = 0; j < c.num_pts; j++)
		{
			c.p[j][0] = c.p[j][0] * img_size + FILTER_SIZE + FILTER_OFFSET + 1;
			c.p[j][1] = c.p[j][1] * img_size + FILTER_SIZE + FILTER_OFFSET + 1;
		}
	}

	// save the path
	char fn[1024];
	sprintf(fn, "path_%s.txt", STR(TEST_NAME));
	FILE *f = fopen(fn, "wb");

	fprintf(f, "%d %d %d\n\n", img_size, img_size, lines.s);

	for (int i = 0; i < lines.s; i++)
	{
		Line &l = lines[i];
		fprintf(f, "%d  ", l.num_pts);
		for (int j = 0; j < l.num_pts; j++)
		{
			double x = l.p[j][0] - (FILTER_SIZE + FILTER_OFFSET + 1);
			double y = l.p[j][1] - (FILTER_SIZE + FILTER_OFFSET + 1);
			fprintf(f, "%18.16g %18.16g  ", x, y);
		}
		fprintf(f, "\n");
	}

	fclose(f);
}
