#pragma once

#include "common.h"
#include "rasterizer.h"
#include "array.h"
#include "array2d.h"
#include <string>
#include <fstream>

using namespace std;

struct Value
{
	double full, part;
	void operator=(double v)
	{
		full = part = v;
	}
	__forceinline double sum()
	{
		return full+part;
	}
};


#ifdef FILTER0
struct FilterBox
{
	static const int w = 1, h = 1;
	static double precalc[h];
	static const double threshhold;

	__forceinline void add_to_vals(double *pixel[h], double (&propagate)[w][h], LineRaster &c)
	{
		if (c.num_pts == 2)
		{
#if defined(USE_ONLY_LINES) || defined(USE_MIXED_CURVES)
			const double p00 = c.px[0].v[0];
			const double p01 = c.px[0].v[1];
			const double p10 = c.px[1].v[0];
			const double p11 = c.px[1].v[1];

			const double p = -((p00 + p10)*(p01 - p11))*.5;
			const double f = (-p01 + p11);

			if (c.flip_x != c.flip_y)
			{
				pixel[0][0] -= p;
				propagate[0][0] -= f;
			}
			else
			{
				pixel[0][0] += p;
				propagate[0][0] += f;
			}
#endif
		}
		else if (c.num_pts == 3)
		{
#if defined(USE_ONLY_QUADRATICS) || defined(USE_MIXED_CURVES)
			const double p00 = c.px[0].v[0];
			const double p01 = c.px[0].v[1];
			const double p10 = c.px[1].v[0];
			const double p11 = c.px[1].v[1];
			const double p20 = c.px[2].v[0];
			const double p21 = c.px[2].v[1];

			const double p = ((1.0/6.0)*((-2.0*p11*p20)+(-1.0*p01*((2.0*p10)+p20))+(2.0*p10*p21)+(3.0*p20*p21)+(p00*((-3.0*p01)+(2.0*p11)+p21))));
			const double f = ((-1.0*p01)+p21);

			if (c.flip_x != c.flip_y)
			{
				pixel[0][0] -= p;
				propagate[0][0] -= f;
			}
			else
			{
				pixel[0][0] += p;
				propagate[0][0] += f;
			}
#endif
		}
		else// if (c.num_pts == 4)
		{
#if defined(USE_ONLY_CUBICS) || defined(USE_MIXED_CURVES)
#ifdef USE_CUBICS
			const double p00 = c.px[0].v[0];
			const double p01 = c.px[0].v[1];
			const double p10 = c.px[1].v[0];
			const double p11 = c.px[1].v[1];
			const double p20 = c.px[2].v[0];
			const double p21 = c.px[2].v[1];
			const double p30 = c.px[3].v[0];
			const double p31 = c.px[3].v[1];

			const double p = ((1.0/20.0)*((-3.0*p11*p20)+(-3.0*p11*p30)+(-6.0*p21*p30)+(-1.0*p01*((6.0*p10)+(3.0*p20)+p30))+(6.0*p20*p31)+(10.0*p30*p31)+(3.0*p10*(p21+p31))+(p00*((-10.0*p01)+(6.0*p11)+(3.0*p21)+p31))));
			const double f = ((-1.0*p01)+p31);

			if (c.flip_x != c.flip_y)
			{
				pixel[0][0] -= p;
				propagate[0][0] -= f;
			}
			else
			{
				pixel[0][0] += p;
				propagate[0][0] += f;
			}
#endif
#endif
		}
	}
};
#endif



#ifdef FILTER1
struct FilterTent
{
	static const int w = 2, h = 2;
	static double precalc[h];
	static const double threshhold;

	__forceinline void add_to_vals(double *pixel[h], double (&propagate)[w][h], LineRaster &c)
	{
		int x, y;

		if (c.num_pts == 2)
		{
#if defined(USE_ONLY_LINES) || defined(USE_MIXED_CURVES)
			const double p00 = c.px[0].v[0];
			const double p01 = c.px[0].v[1];
			const double p10 = c.px[1].v[0];
			const double p11 = c.px[1].v[1];

			double p002, p012, p102, p112;
#if 0
			int sign = c.flip_x != c.flip_y ? -1 : 1;

			p002 = p00*p00;
			p012 = p01*p01;
			p102 = p10*p10;
			p112 = p11*p11;

			x = 0; y = 0;
			pixel[y][-x] += ((1.0/24.0)*((-1.0*p01)+p11)*((2.0*p00*p10*(p01+p11))+(p002*((3.0*p01)+p11))+(p102*(p01+(3.0*p11))))*sign);
			propagate[x][y] += ((-1.0/4.0)*(p01+(-1.0*p11))*(p01+p11)*sign);
			x = 1; y = 0;
			pixel[y][-x] += ((1.0/24.0)*(p01+(-1.0*p11))*((p002*((3.0*p01)+p11))+(2.0*p00*((p01*(-4.0+p10))+((-2.0+p10)*p11)))+(p10*((p01*(-4.0+p10))+((-8.0+(3.0*p10))*p11))))*sign);
			propagate[x][y] += ((-1.0/4.0)*(p01+(-1.0*p11))*(p01+p11)*sign);
			x = 0; y = 1;
			pixel[y][-x] += ((1.0/24.0)*(p01+(-1.0*p11))*((2.0*p00*p10*(-2.0+p01+p11))+(p002*(-4.0+(3.0*p01)+p11))+(p102*(-4.0+p01+(3.0*p11))))*sign);
			propagate[x][y] += ((1.0/4.0)*(p01+(-1.0*p11))*(-2.0+p01+p11)*sign);
			x = 1; y = 1;
			pixel[y][-x] += ((-1.0/24.0)*(p01+(-1.0*p11))*((2.0*p00*(6.0+(p01*(-4.0+p10))+(p10*(-2.0+p11))+(-2.0*p11)))+(p002*(-4.0+(3.0*p01)+p11))+(p10*(12.0+(p01*(-4.0+p10))+(-8.0*p11)+(p10*(-4.0+(3.0*p11))))))*sign);
			propagate[x][y] += ((1.0/4.0)*(p01+(-1.0*p11))*(-2.0+p01+p11)*sign);

#else
			int sign = c.flip_x != c.flip_y ? -1 : 1;
			int special = 0;
			if (p00 == 0) special += 1 * 1;
			if (p10 == 0) special += 1 * 2;
			if (p01 == 0) special += 1 * 2*2;
			if (p01 == 1) special += 2 * 2*2;
			if (p11 == 0) special += 1 * 2*2*3;
			if (p11 == 1) special += 2 * 2*2*3;

switch (special)
{
case 0:
p002 = p00*p00;
p012 = p01*p01;
p102 = p10*p10;
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*((-1.0*p01)+p11)*((2.0*p00*p10*(p01+p11))+(p002*((3.0*p01)+p11))+(p102*(p01+(3.0*p11))))*sign);
propagate[x][y] += ((-1.0/4.0)*(p01+(-1.0*p11))*(p01+p11)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*(p01+(-1.0*p11))*((p002*((3.0*p01)+p11))+(2.0*p00*((p01*(-4.0+p10))+((-2.0+p10)*p11)))+(p10*((p01*(-4.0+p10))+((-8.0+(3.0*p10))*p11))))*sign);
propagate[x][y] += ((-1.0/4.0)*(p01+(-1.0*p11))*(p01+p11)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*(p01+(-1.0*p11))*((2.0*p00*p10*(-2.0+p01+p11))+(p002*(-4.0+(3.0*p01)+p11))+(p102*(-4.0+p01+(3.0*p11))))*sign);
propagate[x][y] += ((1.0/4.0)*(p01+(-1.0*p11))*(-2.0+p01+p11)*sign);
x = 1; y = 1;
pixel[y][-x] += ((-1.0/24.0)*(p01+(-1.0*p11))*((2.0*p00*(6.0+(p01*(-4.0+p10))+(p10*(-2.0+p11))+(-2.0*p11)))+(p002*(-4.0+(3.0*p01)+p11))+(p10*(12.0+(p01*(-4.0+p10))+(-8.0*p11)+(p10*(-4.0+(3.0*p11))))))*sign);
propagate[x][y] += ((1.0/4.0)*(p01+(-1.0*p11))*(-2.0+p01+p11)*sign);
break;
case 1:
p012 = p01*p01;
p102 = p10*p10;
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*p102*((-1.0*p01)+p11)*(p01+(3.0*p11))*sign);
propagate[x][y] += ((-1.0/4.0)*(p01+(-1.0*p11))*(p01+p11)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*p10*(p01+(-1.0*p11))*((p01*(-4.0+p10))+((-8.0+(3.0*p10))*p11))*sign);
propagate[x][y] += ((-1.0/4.0)*(p01+(-1.0*p11))*(p01+p11)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*p102*(p01+(-1.0*p11))*(-4.0+p01+(3.0*p11))*sign);
propagate[x][y] += ((1.0/4.0)*(p01+(-1.0*p11))*(-2.0+p01+p11)*sign);
x = 1; y = 1;
pixel[y][-x] += ((-1.0/24.0)*p10*(p01+(-1.0*p11))*(12.0+(p01*(-4.0+p10))+(-8.0*p11)+(p10*(-4.0+(3.0*p11))))*sign);
propagate[x][y] += ((1.0/4.0)*(p01+(-1.0*p11))*(-2.0+p01+p11)*sign);
break;
case 2:
p002 = p00*p00;
p012 = p01*p01;
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*((-1.0*p01)+p11)*(p01+(3.0*p11)+(2.0*p00*(p01+p11))+(p002*((3.0*p01)+p11)))*sign);
propagate[x][y] += ((-1.0/4.0)*(p01+(-1.0*p11))*(p01+p11)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*(p01+(-1.0*p11))*((3.0*(-1.0+(-2.0*p00)+p002)*p01)+((-5.0+(-2.0*p00)+p002)*p11))*sign);
propagate[x][y] += ((-1.0/4.0)*(p01+(-1.0*p11))*(p01+p11)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*(p01+(-1.0*p11))*(-4.0+p01+(3.0*p11)+(2.0*p00*(-2.0+p01+p11))+(p002*(-4.0+(3.0*p01)+p11)))*sign);
propagate[x][y] += ((1.0/4.0)*(p01+(-1.0*p11))*(-2.0+p01+p11)*sign);
x = 1; y = 1;
pixel[y][-x] += ((-1.0/24.0)*(p01+(-1.0*p11))*(8.0+(-3.0*p01)+(-5.0*p11)+(-2.0*p00*(-4.0+(3.0*p01)+p11))+(p002*(-4.0+(3.0*p01)+p11)))*sign);
propagate[x][y] += ((1.0/4.0)*(p01+(-1.0*p11))*(-2.0+p01+p11)*sign);
break;
case 3:
p012 = p01*p01;
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*((-1.0*p01)+p11)*(p01+(3.0*p11))*sign);
propagate[x][y] += ((-1.0/4.0)*(p01+(-1.0*p11))*(p01+p11)*sign);
x = 1; y = 0;
pixel[y][-x] += ((-1.0/24.0)*(p01+(-1.0*p11))*((3.0*p01)+(5.0*p11))*sign);
propagate[x][y] += ((-1.0/4.0)*(p01+(-1.0*p11))*(p01+p11)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*(p01+(-1.0*p11))*(-4.0+p01+(3.0*p11))*sign);
propagate[x][y] += ((1.0/4.0)*(p01+(-1.0*p11))*(-2.0+p01+p11)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*(p01+(-1.0*p11))*(-8.0+(3.0*p01)+(5.0*p11))*sign);
propagate[x][y] += ((1.0/4.0)*(p01+(-1.0*p11))*(-2.0+p01+p11)*sign);
break;
case 4:
p002 = p00*p00;
p102 = p10*p10;
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*(p002+(2.0*p00*p10)+(3.0*p102))*p112*sign);
propagate[x][y] += ((1.0/4.0)*p112*sign);
x = 1; y = 0;
pixel[y][-x] += ((-1.0/24.0)*(p002+(2.0*p00*(-2.0+p10))+(p10*(-8.0+(3.0*p10))))*p112*sign);
propagate[x][y] += ((1.0/4.0)*p112*sign);
x = 0; y = 1;
pixel[y][-x] += ((-1.0/24.0)*p11*((p002*(-4.0+p11))+(2.0*p00*p10*(-2.0+p11))+(p102*(-4.0+(3.0*p11))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p11)*p11*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*p11*((2.0*p00*(6.0+(p10*(-2.0+p11))+(-2.0*p11)))+(p002*(-4.0+p11))+(p10*(12.0+(-8.0*p11)+(p10*(-4.0+(3.0*p11))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p11)*p11*sign);
break;
case 5:
p102 = p10*p10;
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/8.0)*p102*p112*sign);
propagate[x][y] += ((1.0/4.0)*p112*sign);
x = 1; y = 0;
pixel[y][-x] += ((-1.0/24.0)*p10*(-8.0+(3.0*p10))*p112*sign);
propagate[x][y] += ((1.0/4.0)*p112*sign);
x = 0; y = 1;
pixel[y][-x] += ((-1.0/24.0)*p102*p11*(-4.0+(3.0*p11))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p11)*p11*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*p10*p11*(12.0+(-8.0*p11)+(p10*(-4.0+(3.0*p11))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p11)*p11*sign);
break;
case 6:
p002 = p00*p00;
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*(3.0+(2.0*p00)+p002)*p112*sign);
propagate[x][y] += ((1.0/4.0)*p112*sign);
x = 1; y = 0;
pixel[y][-x] += ((-1.0/24.0)*(-5.0+(-2.0*p00)+p002)*p112*sign);
propagate[x][y] += ((1.0/4.0)*p112*sign);
x = 0; y = 1;
pixel[y][-x] += ((-1.0/24.0)*p11*(-4.0+(p002*(-4.0+p11))+(2.0*p00*(-2.0+p11))+(3.0*p11))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p11)*p11*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*(8.0+(-2.0*p00*(-4.0+p11))+(p002*(-4.0+p11))+(-5.0*p11))*p11*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p11)*p11*sign);
break;
case 7:
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/8.0)*p112*sign);
propagate[x][y] += ((1.0/4.0)*p112*sign);
x = 1; y = 0;
pixel[y][-x] += ((5.0/24.0)*p112*sign);
propagate[x][y] += ((1.0/4.0)*p112*sign);
x = 0; y = 1;
pixel[y][-x] += ((-1.0/24.0)*p11*(-4.0+(3.0*p11))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p11)*p11*sign);
x = 1; y = 1;
pixel[y][-x] += ((-1.0/24.0)*p11*(-8.0+(5.0*p11))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p11)*p11*sign);
break;
case 8:
p002 = p00*p00;
p102 = p10*p10;
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*(-1.0+p11)*((2.0*p00*p10*(1.0+p11))+(p002*(3.0+p11))+(p102*(1.0+(3.0*p11))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p11)*(1.0+p11)*sign);
x = 1; y = 0;
pixel[y][-x] += ((-1.0/24.0)*(-1.0+p11)*((p002*(3.0+p11))+(2.0*p00*(-4.0+p10+(-2.0*p11)+(p10*p11)))+(p10*(-4.0+p10+(-8.0*p11)+(3.0*p10*p11))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p11)*(1.0+p11)*sign);
x = 0; y = 1;
pixel[y][-x] += ((-1.0/24.0)*(p002+(2.0*p00*p10)+(3.0*p102))*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p11)*(-1.0+p11))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*(p002+(2.0*p00*(-2.0+p10))+(p10*(-8.0+(3.0*p10))))*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p11)*(-1.0+p11))*sign);
break;
case 9:
p102 = p10*p10;
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*p102*(-1.0+p11)*(1.0+(3.0*p11))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p11)*(1.0+p11)*sign);
x = 1; y = 0;
pixel[y][-x] += ((-1.0/24.0)*p10*(-1.0+p11)*(-4.0+p10+(-8.0*p11)+(3.0*p10*p11))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p11)*(1.0+p11)*sign);
x = 0; y = 1;
pixel[y][-x] += ((-1.0/8.0)*p102*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p11)*(-1.0+p11))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*p10*(-8.0+(3.0*p10))*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p11)*(-1.0+p11))*sign);
break;
case 10:
p002 = p00*p00;
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*(-1.0+p11)*(1.0+(3.0*p11)+(2.0*p00*(1.0+p11))+(p002*(3.0+p11)))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p11)*(1.0+p11)*sign);
x = 1; y = 0;
pixel[y][-x] += ((-1.0/24.0)*(-1.0+p11)*(-3.0+(-5.0*p11)+(-2.0*p00*(3.0+p11))+(p002*(3.0+p11)))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p11)*(1.0+p11)*sign);
x = 0; y = 1;
pixel[y][-x] += ((-1.0/24.0)*(3.0+(2.0*p00)+p002)*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p11)*(-1.0+p11))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*(-5.0+(-2.0*p00)+p002)*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p11)*(-1.0+p11))*sign);
break;
case 11:
p112 = p11*p11;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*(-1.0+p11)*(1.0+(3.0*p11))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p11)*(1.0+p11)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*(-1.0+p11)*(3.0+(5.0*p11))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p11)*(1.0+p11)*sign);
x = 0; y = 1;
pixel[y][-x] += ((-1.0/8.0)*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p11)*(-1.0+p11))*sign);
x = 1; y = 1;
pixel[y][-x] += ((-5.0/24.0)*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p11)*(-1.0+p11))*sign);
break;
case 12:
p002 = p00*p00;
p012 = p01*p01;
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += ((-1.0/24.0)*p012*((3.0*p002)+(2.0*p00*p10)+p102)*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*p012*((3.0*p002)+(2.0*p00*(-4.0+p10))+((-4.0+p10)*p10))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*p01*((p002*(-4.0+(3.0*p01)))+(2.0*p00*(-2.0+p01)*p10)+((-4.0+p01)*p102))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((-1.0/24.0)*p01*((p002*(-4.0+(3.0*p01)))+(2.0*p00*(6.0+(p01*(-4.0+p10))+(-2.0*p10)))+(((p01*(-4.0+p10))+(-4.0*(-3.0+p10)))*p10))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 13:
p012 = p01*p01;
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += ((-1.0/24.0)*p012*p102*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*p012*(-4.0+p10)*p10*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*(-4.0+p01)*p01*p102*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((-1.0/24.0)*p01*((p01*(-4.0+p10))+(-4.0*(-3.0+p10)))*p10*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 14:
p002 = p00*p00;
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((-1.0/24.0)*(1.0+(2.0*p00)+(3.0*p002))*p012*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/8.0)*(-1.0+(-2.0*p00)+p002)*p012*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*p01*(-4.0+(2.0*p00*(-2.0+p01))+p01+(p002*(-4.0+(3.0*p01))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((-1.0/24.0)*p01*(8.0+(p00*(8.0+(-6.0*p01)))+(-3.0*p01)+(p002*(-4.0+(3.0*p01))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 15:
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((-1.0/24.0)*p012*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((-1.0/8.0)*p012*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*(-4.0+p01)*p01*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*p01*(-8.0+(3.0*p01))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 16:
p002 = p00*p00;
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
break;
case 17:
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
break;
case 18:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
break;
case 19:
x = 0; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
break;
case 20:
p002 = p00*p00;
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*((-3.0*p002)+(-2.0*p00*p10)+(-1.0*p102))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*((3.0*p002)+(2.0*p00*(-4.0+p10))+((-4.0+p10)*p10))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*((-1.0*p002)+(-2.0*p00*p10)+(-3.0*p102))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*(p002+(2.0*p00*(-2.0+p10))+(p10*(-8.0+(3.0*p10))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 21:
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += ((-1.0/24.0)*p102*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*(-4.0+p10)*p10*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((-1.0/8.0)*p102*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*p10*(-8.0+(3.0*p10))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 22:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*(-1.0+(-2.0*p00)+(-3.0*p002))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/8.0)*(-1.0+(-2.0*p00)+p002)*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*(-3.0+(-2.0*p00)+(-1.0*p002))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*(-5.0+(-2.0*p00)+p002)*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 23:
x = 0; y = 0;
pixel[y][-x] += ((-1.0/24.0)*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((-1.0/8.0)*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((-1.0/8.0)*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((-5.0/24.0)*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 24:
p002 = p00*p00;
p012 = p01*p01;
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += ((-1.0/24.0)*(-1.0+p01)*((p002*(1.0+(3.0*p01)))+(2.0*p00*(1.0+p01)*p10)+((3.0+p01)*p102))*sign);
propagate[x][y] += ((-1.0/4.0)*(-1.0+p01)*(1.0+p01)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*(-1.0+p01)*((p002*(1.0+(3.0*p01)))+(2.0*p00*(-2.0+(p01*(-4.0+p10))+p10))+(p10*(-8.0+(p01*(-4.0+p10))+(3.0*p10))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-1.0+p01)*(1.0+p01)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*((-1.0+p01)*(-1.0+p01))*((3.0*p002)+(2.0*p00*p10)+p102)*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((-1.0/24.0)*((-1.0+p01)*(-1.0+p01))*((3.0*p002)+(2.0*p00*(-4.0+p10))+((-4.0+p10)*p10))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 25:
p012 = p01*p01;
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += ((-1.0/24.0)*(-1.0+p01)*(3.0+p01)*p102*sign);
propagate[x][y] += ((-1.0/4.0)*(-1.0+p01)*(1.0+p01)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*(-1.0+p01)*p10*(-8.0+(p01*(-4.0+p10))+(3.0*p10))*sign);
propagate[x][y] += ((-1.0/4.0)*(-1.0+p01)*(1.0+p01)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*((-1.0+p01)*(-1.0+p01))*p102*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((-1.0/24.0)*((-1.0+p01)*(-1.0+p01))*(-4.0+p10)*p10*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 26:
p002 = p00*p00;
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((-1.0/24.0)*(-1.0+p01)*(3.0+p01+(2.0*p00*(1.0+p01))+(p002*(1.0+(3.0*p01))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-1.0+p01)*(1.0+p01)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*(-1.0+p01)*(-5.0+(-3.0*p01)+(-2.0*p00*(1.0+(3.0*p01)))+(p002*(1.0+(3.0*p01))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-1.0+p01)*(1.0+p01)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*(1.0+(2.0*p00)+(3.0*p002))*((-1.0+p01)*(-1.0+p01))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((-1.0/8.0)*(-1.0+(-2.0*p00)+p002)*((-1.0+p01)*(-1.0+p01))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 27:
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((-1.0/24.0)*(-1.0+p01)*(3.0+p01)*sign);
propagate[x][y] += ((-1.0/4.0)*(-1.0+p01)*(1.0+p01)*sign);
x = 1; y = 0;
pixel[y][-x] += ((-1.0/24.0)*(-1.0+p01)*(5.0+(3.0*p01))*sign);
propagate[x][y] += ((-1.0/4.0)*(-1.0+p01)*(1.0+p01)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*((-1.0+p01)*(-1.0+p01))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/8.0)*((-1.0+p01)*(-1.0+p01))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 28:
p002 = p00*p00;
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*(p002+(2.0*p00*p10)+(3.0*p102))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*((4.0*p00)+(-1.0*p002)+(8.0*p10)+(-2.0*p00*p10)+(-3.0*p102))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*((3.0*p002)+(2.0*p00*p10)+p102)*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/24.0)*((-3.0*p002)+(-2.0*p00*(-4.0+p10))+(-1.0*(-4.0+p10)*p10))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 29:
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += ((1.0/8.0)*p102*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*(8.0+(-3.0*p10))*p10*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*p102*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((-1.0/24.0)*(-4.0+p10)*p10*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 30:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += ((1.0/24.0)*(3.0+(2.0*p00)+p002)*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/24.0)*(5.0+(2.0*p00)+(-1.0*p002))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*(1.0+(2.0*p00)+(3.0*p002))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/8.0)*(1.0+(2.0*p00)+(-1.0*p002))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 31:
x = 0; y = 0;
pixel[y][-x] += ((1.0/8.0)*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((5.0/24.0)*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/24.0)*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/8.0)*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 32:
p002 = p00*p00;
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
break;
case 33:
p102 = p10*p10;
x = 0; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
break;
case 34:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
break;
case 35:
x = 0; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += 0.0;
propagate[x][y] += 0.0;
break;
}
#endif

#endif

		}
		else if (c.num_pts == 3)
		{
#if defined(USE_ONLY_QUADRATICS) || defined(USE_MIXED_CURVES)
			const double p00 = c.px[0].v[0];
			const double p01 = c.px[0].v[1];
			const double p10 = c.px[1].v[0];
			const double p11 = c.px[1].v[1];
			const double p20 = c.px[2].v[0];
			const double p21 = c.px[2].v[1];

			const double p102 = p10*p10;
			const double p112 = p11*p11;
			
			double p002, p012, p202, p212;

#if 0
			int sign = c.flip_x != c.flip_y ? -1 : 1;
p002 = p00*p00;
p012 = p01*p01;
p202 = p20*p20;
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-16.0*p10*p112*p20)+(-20.0*p112*p202)+(-1.0*p012*((20.0*p102)+(12.0*p10*p20)+(3.0*p202)))+(16.0*p102*p11*p21)+(-60.0*p11*p202*p21)+(20.0*p102*p212)+(60.0*p10*p20*p212)+(105.0*p202*p212)+(p002*((-105.0*p012)+(20.0*p112)+(12.0*p11*p21)+(3.0*p212)+(10.0*p01*((6.0*p11)+p21))))+(-2.0*p01*((8.0*p102*p11)+(4.0*p10*p20*((3.0*p11)+p21))+(p202*((6.0*p11)+(5.0*p21)))))+(2.0*p00*((-5.0*p012*((6.0*p10)+p20))+(p20*p21*((4.0*p11)+(5.0*p21)))+(p01*((-4.0*p11*p20)+(4.0*p10*p21)))+(2.0*p10*((4.0*p112)+(6.0*p11*p21)+(3.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p212)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-56.0*p112*p20)+(16.0*p10*p112*p20)+(20.0*p112*p202)+(p012*((20.0*p102)+(4.0*p10*(-28.0+(3.0*p20)))+(p20*(-28.0+(3.0*p20)))))+(56.0*p10*p11*p21)+(-16.0*p102*p11*p21)+(-112.0*p11*p20*p21)+(60.0*p11*p202*p21)+(112.0*p10*p212)+(-20.0*p102*p212)+(280.0*p20*p212)+(-60.0*p10*p20*p212)+(-105.0*p202*p212)+(p002*((105.0*p012)+(-20.0*p112)+(-12.0*p11*p21)+(-3.0*p212)+(-10.0*p01*((6.0*p11)+p21))))+(2.0*p00*(((28.0+(-8.0*p10))*p112)+(5.0*p012*(-28.0+(6.0*p10)+p20))+(-4.0*p11*(-7.0+(3.0*p10)+p20)*p21)+((14.0+(-6.0*p10)+(-5.0*p20))*p212)+(2.0*p01*((2.0*p11*(14.0+p20))+((7.0+(-2.0*p10))*p21)))))+(2.0*p01*((8.0*p102*p11)+(4.0*p10*((p11*(-7.0+(3.0*p20)))+(p20*p21)))+(p20*((p11*(-28.0+(6.0*p20)))+((-14.0+(5.0*p20))*p21))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p212)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-56.0*p10*p11*p20)+(16.0*p10*p112*p20)+(-112.0*p11*p202)+(20.0*p112*p202)+(p012*((20.0*p102)+(12.0*p10*p20)+(3.0*p202)))+(56.0*p102*p21)+(-16.0*p102*p11*p21)+(112.0*p10*p20*p21)+(140.0*p202*p21)+(60.0*p11*p202*p21)+(-20.0*p102*p212)+(-60.0*p10*p20*p212)+(-105.0*p202*p212)+(p002*((105.0*p012)+(-20.0*p112)+((28.0+(-3.0*p21))*p21)+(-10.0*p01*(14.0+(6.0*p11)+p21))+(-4.0*p11*(-28.0+(3.0*p21)))))+(2.0*p01*((4.0*p102*(-7.0+(2.0*p11)))+(4.0*p10*p20*(-7.0+(3.0*p11)+p21))+(p202*(-14.0+(6.0*p11)+(5.0*p21)))))+(2.0*p00*((5.0*p012*((6.0*p10)+p20))+(p20*(14.0+(-4.0*p11)+(-5.0*p21))*p21)+(-2.0*p01*(((7.0+(-2.0*p11))*p20)+(2.0*p10*(14.0+p21))))+(-2.0*p10*((4.0*p112)+(p21*(-14.0+(3.0*p21)))+(2.0*p11*(-7.0+(3.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p21)*p21))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-280.0*p11*p20)+(56.0*p10*p11*p20)+(56.0*p112*p20)+(-16.0*p10*p112*p20)+(112.0*p11*p202)+(-20.0*p112*p202)+(p012*((-20.0*p102)+((28.0+(-3.0*p20))*p20)+(-4.0*p10*(-28.0+(3.0*p20)))))+(280.0*p10*p21)+(-56.0*p102*p21)+(-56.0*p10*p11*p21)+(16.0*p102*p11*p21)+(420.0*p20*p21)+(-112.0*p10*p20*p21)+(112.0*p11*p20*p21)+(-140.0*p202*p21)+(-60.0*p11*p202*p21)+(-112.0*p10*p212)+(20.0*p102*p212)+(-280.0*p20*p212)+(60.0*p10*p20*p212)+(105.0*p202*p212)+(p002*((-105.0*p012)+(20.0*p112)+(10.0*p01*(14.0+(6.0*p11)+p21))+(4.0*p11*(-28.0+(3.0*p21)))+(p21*(-28.0+(3.0*p21)))))+(-2.0*p01*((4.0*p102*(-7.0+(2.0*p11)))+(4.0*p10*(35.0+(p11*(-7.0+(3.0*p20)))+(p20*(-7.0+p21))))+(p20*(70.0+(-14.0*p20)+(p11*(-28.0+(6.0*p20)))+(-14.0*p21)+(5.0*p20*p21)))))+(-2.0*p00*(((28.0+(-8.0*p10))*p112)+(5.0*p012*(-28.0+(6.0*p10)+p20))+(((p10*(28.0+(-6.0*p21)))+(p20*(14.0+(-5.0*p21)))+(14.0*(-5.0+p21)))*p21)+(2.0*p01*((2.0*p11*(14.0+p20))+(-2.0*p10*(14.0+p21))+(7.0*(15.0+(-1.0*p20)+p21))))+(-4.0*p11*(35.0+((-7.0+p20)*p21)+(p10*(-7.0+(3.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p21)*p21))*sign);
#else
			int sign = c.flip_x != c.flip_y ? -1 : 1;
			int special = 0;
			if (p00 == 0) special += 1 * 1;
			if (p20 == 0) special += 1 * 2;
			if (p01 == 0) special += 1 * 2*2;
			if (p01 == 1) special += 2 * 2*2;
			if (p21 == 0) special += 1 * 2*2*3;
			if (p21 == 1) special += 2 * 2*2*3;
			
switch (special)
{
case 0:
p002 = p00*p00;
p012 = p01*p01;
p202 = p20*p20;
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-16.0*p10*p112*p20)+(-20.0*p112*p202)+(-1.0*p012*((20.0*p102)+(12.0*p10*p20)+(3.0*p202)))+(16.0*p102*p11*p21)+(-60.0*p11*p202*p21)+(20.0*p102*p212)+(60.0*p10*p20*p212)+(105.0*p202*p212)+(p002*((-105.0*p012)+(20.0*p112)+(12.0*p11*p21)+(3.0*p212)+(10.0*p01*((6.0*p11)+p21))))+(-2.0*p01*((8.0*p102*p11)+(4.0*p10*p20*((3.0*p11)+p21))+(p202*((6.0*p11)+(5.0*p21)))))+(2.0*p00*((-5.0*p012*((6.0*p10)+p20))+(p20*p21*((4.0*p11)+(5.0*p21)))+(p01*((-4.0*p11*p20)+(4.0*p10*p21)))+(2.0*p10*((4.0*p112)+(6.0*p11*p21)+(3.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p212)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-56.0*p112*p20)+(16.0*p10*p112*p20)+(20.0*p112*p202)+(p012*((20.0*p102)+(4.0*p10*(-28.0+(3.0*p20)))+(p20*(-28.0+(3.0*p20)))))+(56.0*p10*p11*p21)+(-16.0*p102*p11*p21)+(-112.0*p11*p20*p21)+(60.0*p11*p202*p21)+(112.0*p10*p212)+(-20.0*p102*p212)+(280.0*p20*p212)+(-60.0*p10*p20*p212)+(-105.0*p202*p212)+(p002*((105.0*p012)+(-20.0*p112)+(-12.0*p11*p21)+(-3.0*p212)+(-10.0*p01*((6.0*p11)+p21))))+(2.0*p00*(((28.0+(-8.0*p10))*p112)+(5.0*p012*(-28.0+(6.0*p10)+p20))+(-4.0*p11*(-7.0+(3.0*p10)+p20)*p21)+((14.0+(-6.0*p10)+(-5.0*p20))*p212)+(2.0*p01*((2.0*p11*(14.0+p20))+((7.0+(-2.0*p10))*p21)))))+(2.0*p01*((8.0*p102*p11)+(4.0*p10*((p11*(-7.0+(3.0*p20)))+(p20*p21)))+(p20*((p11*(-28.0+(6.0*p20)))+((-14.0+(5.0*p20))*p21))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p212)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-56.0*p10*p11*p20)+(16.0*p10*p112*p20)+(-112.0*p11*p202)+(20.0*p112*p202)+(p012*((20.0*p102)+(12.0*p10*p20)+(3.0*p202)))+(56.0*p102*p21)+(-16.0*p102*p11*p21)+(112.0*p10*p20*p21)+(140.0*p202*p21)+(60.0*p11*p202*p21)+(-20.0*p102*p212)+(-60.0*p10*p20*p212)+(-105.0*p202*p212)+(p002*((105.0*p012)+(-20.0*p112)+((28.0+(-3.0*p21))*p21)+(-10.0*p01*(14.0+(6.0*p11)+p21))+(-4.0*p11*(-28.0+(3.0*p21)))))+(2.0*p01*((4.0*p102*(-7.0+(2.0*p11)))+(4.0*p10*p20*(-7.0+(3.0*p11)+p21))+(p202*(-14.0+(6.0*p11)+(5.0*p21)))))+(2.0*p00*((5.0*p012*((6.0*p10)+p20))+(p20*(14.0+(-4.0*p11)+(-5.0*p21))*p21)+(-2.0*p01*(((7.0+(-2.0*p11))*p20)+(2.0*p10*(14.0+p21))))+(-2.0*p10*((4.0*p112)+(p21*(-14.0+(3.0*p21)))+(2.0*p11*(-7.0+(3.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p21)*p21))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-280.0*p11*p20)+(56.0*p10*p11*p20)+(56.0*p112*p20)+(-16.0*p10*p112*p20)+(112.0*p11*p202)+(-20.0*p112*p202)+(p012*((-20.0*p102)+((28.0+(-3.0*p20))*p20)+(-4.0*p10*(-28.0+(3.0*p20)))))+(280.0*p10*p21)+(-56.0*p102*p21)+(-56.0*p10*p11*p21)+(16.0*p102*p11*p21)+(420.0*p20*p21)+(-112.0*p10*p20*p21)+(112.0*p11*p20*p21)+(-140.0*p202*p21)+(-60.0*p11*p202*p21)+(-112.0*p10*p212)+(20.0*p102*p212)+(-280.0*p20*p212)+(60.0*p10*p20*p212)+(105.0*p202*p212)+(p002*((-105.0*p012)+(20.0*p112)+(10.0*p01*(14.0+(6.0*p11)+p21))+(4.0*p11*(-28.0+(3.0*p21)))+(p21*(-28.0+(3.0*p21)))))+(-2.0*p01*((4.0*p102*(-7.0+(2.0*p11)))+(4.0*p10*(35.0+(p11*(-7.0+(3.0*p20)))+(p20*(-7.0+p21))))+(p20*(70.0+(-14.0*p20)+(p11*(-28.0+(6.0*p20)))+(-14.0*p21)+(5.0*p20*p21)))))+(-2.0*p00*(((28.0+(-8.0*p10))*p112)+(5.0*p012*(-28.0+(6.0*p10)+p20))+(((p10*(28.0+(-6.0*p21)))+(p20*(14.0+(-5.0*p21)))+(14.0*(-5.0+p21)))*p21)+(2.0*p01*((2.0*p11*(14.0+p20))+(-2.0*p10*(14.0+p21))+(7.0*(15.0+(-1.0*p20)+p21))))+(-4.0*p11*(35.0+((-7.0+p20)*p21)+(p10*(-7.0+(3.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p21)*p21))*sign);
break;
case 1:
p012 = p01*p01;
p202 = p20*p20;
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-1.0*p012*((20.0*p102)+(12.0*p10*p20)+(3.0*p202)))+(4.0*p102*p21*((4.0*p11)+(5.0*p21)))+(-5.0*p202*((4.0*p112)+(12.0*p11*p21)+(-21.0*p212)))+(p10*((-16.0*p112*p20)+(60.0*p20*p212)))+(-2.0*p01*((8.0*p102*p11)+(4.0*p10*p20*((3.0*p11)+p21))+(p202*((6.0*p11)+(5.0*p21))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p212)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p112*p20*(-14.0+(4.0*p10)+(5.0*p20)))+(p012*((20.0*p102)+(4.0*p10*(-28.0+(3.0*p20)))+(p20*(-28.0+(3.0*p20)))))+(-4.0*p11*((-14.0*p10)+(4.0*p102)+((28.0+(-15.0*p20))*p20))*p21)+(((-20.0*p102)+(p10*(112.0+(-60.0*p20)))+(35.0*(8.0+(-3.0*p20))*p20))*p212)+(2.0*p01*((8.0*p102*p11)+(4.0*p10*((-7.0*p11)+(3.0*p11*p20)+(p20*p21)))+(p20*((-28.0*p11)+(6.0*p11*p20)+(-14.0*p21)+(5.0*p20*p21))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p212)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((p012*((20.0*p102)+(12.0*p10*p20)+(3.0*p202)))+(-4.0*p102*p21*(-14.0+(4.0*p11)+(5.0*p21)))+(4.0*p10*p20*((-14.0*p11)+(4.0*p112)+((28.0+(-15.0*p21))*p21)))+(2.0*p01*((4.0*p102*(-7.0+(2.0*p11)))+(4.0*p10*p20*(-7.0+(3.0*p11)+p21))+(p202*(-14.0+(6.0*p11)+(5.0*p21)))))+(p202*((20.0*p112)+(35.0*(4.0+(-3.0*p21))*p21)+(4.0*p11*(-28.0+(15.0*p21))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p21)*p21))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p112*p20*(-14.0+(4.0*p10)+(5.0*p20)))+(p012*((-20.0*p102)+((28.0+(-3.0*p20))*p20)+(-4.0*p10*(-28.0+(3.0*p20)))))+(-4.0*p11*((2.0*(7.0+(-2.0*p10))*p10*p21)+(-14.0*p20*(-5.0+p10+(2.0*p21)))+(p202*(-28.0+(15.0*p21)))))+(-2.0*p01*((4.0*p102*(-7.0+(2.0*p11)))+(4.0*p10*(35.0+(p11*(-7.0+(3.0*p20)))+(p20*(-7.0+p21))))+(p20*(70.0+(-14.0*p20)+(p11*(-28.0+(6.0*p20)))+(-14.0*p21)+(5.0*p20*p21)))))+(p21*((4.0*p102*(-14.0+(5.0*p21)))+(35.0*p20*(12.0+(-8.0*p21)+(p20*(-4.0+(3.0*p21)))))+(4.0*p10*(70.0+(-28.0*p21)+(p20*(-28.0+(15.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p21)*p21))*sign);
break;
case 2:
p002 = p00*p00;
p012 = p01*p01;
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-1.0*p012*(3.0+(105.0*p002)+(12.0*p10)+(20.0*p102)+(10.0*p00*(1.0+(6.0*p10)))))+(4.0*(-1.0+p00)*(5.0+(5.0*p00)+(4.0*p10))*p112)+(4.0*(-15.0+(2.0*p00)+(3.0*p002)+(6.0*p00*p10)+(4.0*p102))*p11*p21)+((105.0+(10.0*p00)+(3.0*p002)+(60.0*p10)+(12.0*p00*p10)+(20.0*p102))*p212)+(2.0*p01*((2.0*(-3.0+(-2.0*p00)+(15.0*p002)+(-6.0*p10)+(-4.0*p102))*p11)+((-1.0+p00)*(5.0+(5.0*p00)+(4.0*p10))*p21))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p212)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((5.0*p012*(-5.0+(21.0*p002)+(-20.0*p10)+(4.0*p102)+(6.0*p00*(-9.0+(2.0*p10)))))+(-4.0*(-1.0+p00)*(-9.0+(5.0*p00)+(4.0*p10))*p112)+(-4.0*(13.0+(3.0*p002)+(6.0*p00*(-2.0+p10))+(-14.0*p10)+(4.0*p102))*p11*p21)+(-1.0*(-175.0+(3.0*p002)+(-52.0*p10)+(20.0*p102)+(6.0*p00*(-3.0+(2.0*p10))))*p212)+(-2.0*p01*((2.0*(11.0+(-30.0*p00)+(15.0*p002)+(8.0*p10)+(-4.0*p102))*p11)+((-1.0+p00)*(-9.0+(5.0*p00)+(4.0*p10))*p21))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p212)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((p012*(3.0+(105.0*p002)+(12.0*p10)+(20.0*p102)+(10.0*p00*(1.0+(6.0*p10)))))+(-4.0*(-1.0+p00)*(5.0+(5.0*p00)+(4.0*p10))*p112)+(-2.0*p01*(14.0+(p102*(28.0+(-8.0*p11)))+(-6.0*p11)+(-5.0*p21)+(-4.0*p10*(-7.0+(3.0*p11)+p21))+(5.0*p002*(14.0+(6.0*p11)+p21))+(2.0*p00*(7.0+(-2.0*p11)+(2.0*p10*(14.0+p21))))))+(-4.0*p11*(28.0+(14.0*p10)+(-15.0*p21)+(4.0*p102*p21)+(p002*(-28.0+(3.0*p21)))+(2.0*p00*(p21+(p10*(-7.0+(3.0*p21)))))))+(-1.0*p21*((p002*(-28.0+(3.0*p21)))+(35.0*(-4.0+(3.0*p21)))+(4.0*p102*(-14.0+(5.0*p21)))+(4.0*p10*(-28.0+(15.0*p21)))+(2.0*p00*(-14.0+(5.0*p21)+(p10*(-28.0+(6.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p21)*p21))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-5.0*p012*(-5.0+(21.0*p002)+(-20.0*p10)+(4.0*p102)+(6.0*p00*(-9.0+(2.0*p10)))))+(4.0*(-1.0+p00)*(-9.0+(5.0*p00)+(4.0*p10))*p112)+(2.0*p01*(-56.0+(p102*(28.0+(-8.0*p11)))+(22.0*p11)+(4.0*p10*(-28.0+(4.0*p11)+(-1.0*p21)))+(9.0*p21)+(5.0*p002*(14.0+(6.0*p11)+p21))+(2.0*p00*(-98.0+(-30.0*p11)+(-7.0*p21)+(2.0*p10*(14.0+p21))))))+(4.0*p11*(-42.0+(-14.0*p10*(-1.0+p21))+(13.0*p21)+(4.0*p102*p21)+(p002*(-28.0+(3.0*p21)))+(2.0*p00*(35.0+(-6.0*p21)+(p10*(-7.0+(3.0*p21)))))))+(p21*((p10*(168.0+(-52.0*p21)))+(35.0*(8.0+(-5.0*p21)))+(p002*(-28.0+(3.0*p21)))+(4.0*p102*(-14.0+(5.0*p21)))+(2.0*p00*(56.0+(-9.0*p21)+(p10*(-28.0+(6.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p21)*p21))*sign);
break;
case 3:
p012 = p01*p01;
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-1.0*p012*(3.0+(12.0*p10)+(20.0*p102)))+(-4.0*(5.0+(4.0*p10))*p112)+(4.0*(-15.0+(4.0*p102))*p11*p21)+(5.0*(21.0+(12.0*p10)+(4.0*p102))*p212)+(-2.0*p01*((2.0*(3.0+(6.0*p10)+(4.0*p102))*p11)+((5.0+(4.0*p10))*p21))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p212)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((5.0*p012*(-5.0+(-20.0*p10)+(4.0*p102)))+(4.0*(-9.0+(4.0*p10))*p112)+(-4.0*(13.0+(-14.0*p10)+(4.0*p102))*p11*p21)+((175.0+(52.0*p10)+(-20.0*p102))*p212)+(2.0*p01*((2.0*(-11.0+(-8.0*p10)+(4.0*p102))*p11)+((-9.0+(4.0*p10))*p21))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p212)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((p012*(3.0+(12.0*p10)+(20.0*p102)))+(4.0*(5.0+(4.0*p10))*p112)+(((p10*(112.0+(-60.0*p21)))+(p102*(56.0+(-20.0*p21)))+(35.0*(4.0+(-3.0*p21))))*p21)+(-4.0*p11*(28.0+(14.0*p10)+(-15.0*p21)+(4.0*p102*p21)))+(2.0*p01*(-14.0+(6.0*p11)+(4.0*p102*(-7.0+(2.0*p11)))+(5.0*p21)+(4.0*p10*(-7.0+(3.0*p11)+p21)))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p21)*p21))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-5.0*p012*(-5.0+(-20.0*p10)+(4.0*p102)))+(-4.0*(-9.0+(4.0*p10))*p112)+(4.0*p11*(-42.0+(-14.0*p10*(-1.0+p21))+(13.0*p21)+(4.0*p102*p21)))+(-2.0*p01*(56.0+(-22.0*p11)+(4.0*p102*(-7.0+(2.0*p11)))+(-9.0*p21)+(4.0*p10*(28.0+(-4.0*p11)+p21))))+(p21*((p10*(168.0+(-52.0*p21)))+(35.0*(8.0+(-5.0*p21)))+(4.0*p102*(-14.0+(5.0*p21))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p21)*p21))*sign);
break;
case 4:
p002 = p00*p00;
p202 = p20*p20;
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*p21*((4.0*p11)+(5.0*p21)))+(-5.0*p202*((4.0*p112)+(12.0*p11*p21)+(-21.0*p212)))+(p002*((20.0*p112)+(12.0*p11*p21)+(3.0*p212)))+(p10*((-16.0*p112*p20)+(60.0*p20*p212)))+(2.0*p00*((p20*p21*((4.0*p11)+(5.0*p21)))+(2.0*p10*((4.0*p112)+(6.0*p11*p21)+(3.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*p212*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p112*p20*(-14.0+(4.0*p10)+(5.0*p20)))+(-4.0*p11*((-14.0*p10)+(4.0*p102)+((28.0+(-15.0*p20))*p20))*p21)+(((-20.0*p102)+(p10*(112.0+(-60.0*p20)))+(35.0*(8.0+(-3.0*p20))*p20))*p212)+(-1.0*p002*((20.0*p112)+(12.0*p11*p21)+(3.0*p212)))+(-2.0*p00*((4.0*(-7.0+(2.0*p10))*p112)+(4.0*p11*(-7.0+(3.0*p10)+p20)*p21)+((-14.0+(6.0*p10)+(5.0*p20))*p212))))*sign);
propagate[x][y] += ((1.0/4.0)*p212*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*p21*(-14.0+(4.0*p11)+(5.0*p21)))+(4.0*p10*p20*((-14.0*p11)+(4.0*p112)+((28.0+(-15.0*p21))*p21)))+(p002*((-20.0*p112)+((28.0+(-3.0*p21))*p21)+(-4.0*p11*(-28.0+(3.0*p21)))))+(p202*((20.0*p112)+(35.0*(4.0+(-3.0*p21))*p21)+(4.0*p11*(-28.0+(15.0*p21)))))+(-2.0*p00*((p20*p21*(-14.0+(4.0*p11)+(5.0*p21)))+(2.0*p10*((4.0*p112)+(p21*(-14.0+(3.0*p21)))+(2.0*p11*(-7.0+(3.0*p21))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p21)*p21*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p112*p20*(-14.0+(4.0*p10)+(5.0*p20)))+(p002*((20.0*p112)+(4.0*p11*(-28.0+(3.0*p21)))+(p21*(-28.0+(3.0*p21)))))+(-4.0*p11*((2.0*(7.0+(-2.0*p10))*p10*p21)+(-14.0*p20*(-5.0+p10+(2.0*p21)))+(p202*(-28.0+(15.0*p21)))))+(2.0*p00*((4.0*(-7.0+(2.0*p10))*p112)+(4.0*p11*(35.0+((-7.0+p20)*p21)+(p10*(-7.0+(3.0*p21)))))+(p21*(70.0+(-14.0*p20)+(-14.0*p21)+(5.0*p20*p21)+(p10*(-28.0+(6.0*p21)))))))+(p21*((4.0*p102*(-14.0+(5.0*p21)))+(35.0*p20*(12.0+(-8.0*p21)+(p20*(-4.0+(3.0*p21)))))+(4.0*p10*(70.0+(-28.0*p21)+(p20*(-28.0+(15.0*p21))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p21)*p21*sign);
break;
case 5:
p202 = p20*p20;
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*p21*((4.0*p11)+(5.0*p21)))+(-5.0*p202*((4.0*p112)+(12.0*p11*p21)+(-21.0*p212)))+(p10*((-16.0*p112*p20)+(60.0*p20*p212))))*sign);
propagate[x][y] += ((1.0/4.0)*p212*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p112*p20*(-14.0+(4.0*p10)+(5.0*p20)))+(-4.0*p11*((-14.0*p10)+(4.0*p102)+((28.0+(-15.0*p20))*p20))*p21)+(((-20.0*p102)+(p10*(112.0+(-60.0*p20)))+(35.0*(8.0+(-3.0*p20))*p20))*p212))*sign);
propagate[x][y] += ((1.0/4.0)*p212*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*p21*(-14.0+(4.0*p11)+(5.0*p21)))+(4.0*p10*p20*((-14.0*p11)+(4.0*p112)+((28.0+(-15.0*p21))*p21)))+(p202*((20.0*p112)+(35.0*(4.0+(-3.0*p21))*p21)+(4.0*p11*(-28.0+(15.0*p21))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p21)*p21*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p112*p20*(-14.0+(4.0*p10)+(5.0*p20)))+(-4.0*p11*((2.0*(7.0+(-2.0*p10))*p10*p21)+(-14.0*p20*(-5.0+p10+(2.0*p21)))+(p202*(-28.0+(15.0*p21)))))+(p21*((4.0*p102*(-14.0+(5.0*p21)))+(35.0*p20*(12.0+(-4.0*p20)+(-8.0*p21)+(3.0*p20*p21)))+(4.0*p10*(70.0+(-28.0*p20)+(-28.0*p21)+(15.0*p20*p21))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p21)*p21*sign);
break;
case 6:
p002 = p00*p00;
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*(-1.0+p00)*(5.0+(5.0*p00)+(4.0*p10))*p112)+(4.0*(-15.0+(3.0*p002)+(4.0*p102)+(p00*(2.0+(6.0*p10))))*p11*p21)+(((3.0*p002)+(2.0*p00*(5.0+(6.0*p10)))+(5.0*(21.0+(12.0*p10)+(4.0*p102))))*p212))*sign);
propagate[x][y] += ((1.0/4.0)*p212*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-4.0*(-1.0+p00)*(-9.0+(5.0*p00)+(4.0*p10))*p112)+(-4.0*(13.0+(3.0*p002)+(6.0*p00*(-2.0+p10))+(-14.0*p10)+(4.0*p102))*p11*p21)+((175.0+(-3.0*p002)+(52.0*p10)+(-20.0*p102)+(-6.0*p00*(-3.0+(2.0*p10))))*p212))*sign);
propagate[x][y] += ((1.0/4.0)*p212*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*(-1.0+p00)*(5.0+(5.0*p00)+(4.0*p10))*p112)+(-4.0*p11*(28.0+(14.0*p10)+(-15.0*p21)+(4.0*p102*p21)+(p002*(-28.0+(3.0*p21)))+(2.0*p00*(p21+(p10*(-7.0+(3.0*p21)))))))+(-1.0*p21*((p002*(-28.0+(3.0*p21)))+(35.0*(-4.0+(3.0*p21)))+(4.0*p102*(-14.0+(5.0*p21)))+(4.0*p10*(-28.0+(15.0*p21)))+(2.0*p00*(-14.0+(5.0*p21)+(p10*(-28.0+(6.0*p21))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p21)*p21*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*(-1.0+p00)*(-9.0+(5.0*p00)+(4.0*p10))*p112)+(4.0*p11*(-42.0+(-14.0*p10*(-1.0+p21))+(13.0*p21)+(4.0*p102*p21)+(p002*(-28.0+(3.0*p21)))+(2.0*p00*(35.0+(-6.0*p21)+(p10*(-7.0+(3.0*p21)))))))+(p21*((p10*(168.0+(-52.0*p21)))+(35.0*(8.0+(-5.0*p21)))+(p002*(-28.0+(3.0*p21)))+(4.0*p102*(-14.0+(5.0*p21)))+(2.0*p00*(56.0+(-9.0*p21)+(p10*(-28.0+(6.0*p21))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p21)*p21*sign);
break;
case 7:
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-4.0*(5.0+(4.0*p10))*p112)+(4.0*(-15.0+(4.0*p102))*p11*p21)+(5.0*(21.0+(12.0*p10)+(4.0*p102))*p212))*sign);
propagate[x][y] += ((1.0/4.0)*p212*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*(-9.0+(4.0*p10))*p112)+(-4.0*(13.0+(-14.0*p10)+(4.0*p102))*p11*p21)+((175.0+(52.0*p10)+(-20.0*p102))*p212))*sign);
propagate[x][y] += ((1.0/4.0)*p212*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*(5.0+(4.0*p10))*p112)+(((p10*(112.0+(-60.0*p21)))+(p102*(56.0+(-20.0*p21)))+(35.0*(4.0+(-3.0*p21))))*p21)+(-4.0*p11*(28.0+(14.0*p10)+(-15.0*p21)+(4.0*p102*p21))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p21)*p21*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*(-9.0+(4.0*p10))*p112)+(4.0*p11*(-42.0+(-14.0*p10*(-1.0+p21))+(13.0*p21)+(4.0*p102*p21)))+(p21*((p10*(168.0+(-52.0*p21)))+(35.0*(8.0+(-5.0*p21)))+(4.0*p102*(-14.0+(5.0*p21))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p21)*p21*sign);
break;
case 8:
p002 = p00*p00;
p202 = p20*p20;
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(-1.0+p21)*(5.0+(4.0*p11)+(5.0*p21)))+(-1.0*p202*(3.0+(12.0*p11)+(20.0*p112)+(10.0*p21)+(60.0*p11*p21)+(-105.0*p212)))+(-4.0*p10*p20*(3.0+(6.0*p11)+(4.0*p112)+(2.0*p21)+(-15.0*p212)))+(p002*(-105.0+(20.0*p112)+(10.0*p21)+(3.0*p212)+(12.0*p11*(5.0+p21))))+(2.0*p00*((p20*(-1.0+p21)*(5.0+(4.0*p11)+(5.0*p21)))+(2.0*p10*(-15.0+(4.0*p112)+(2.0*p21)+(6.0*p11*p21)+(3.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p212)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(-1.0+p21)*(5.0+(4.0*p11)+(5.0*p21)))+(-1.0*p002*(-105.0+(20.0*p112)+(10.0*p21)+(3.0*p212)+(12.0*p11*(5.0+p21))))+(4.0*p10*((4.0*p112*p20)+(2.0*p11*(-7.0+(3.0*p20)+(7.0*p21)))+(p20*(3.0+(2.0*p21)+(-15.0*p212)))+(28.0*(-1.0+p212))))+(p20*((4.0*p112*(-14.0+(5.0*p20)))+(4.0*p11*(-14.0+(3.0*p20)+(-28.0*p21)+(15.0*p20*p21)))+(p20*(3.0+(10.0*p21)+(-105.0*p212)))+(28.0*(-1.0+(-1.0*p21)+(10.0*p212)))))+(-2.0*p00*(140.0+(-28.0*p112)+(-5.0*p20)+(-14.0*p21)+(-14.0*p212)+(5.0*p20*p212)+(2.0*p10*(-15.0+(4.0*p112)+(2.0*p21)+(6.0*p11*p21)+(3.0*p212)))+(-4.0*p11*(p20+(-1.0*p20*p21)+(7.0*(2.0+p21)))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p212)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(-1.0+p21)*(-9.0+(4.0*p11)+(5.0*p21)))+(4.0*p10*p20*(-11.0+(-8.0*p11)+(4.0*p112)+(30.0*p21)+(-15.0*p212)))+(-1.0*p002*(35.0+(20.0*p112)+(-18.0*p21)+(3.0*p212)+(4.0*p11*(-13.0+(3.0*p21)))))+(5.0*p202*(-5.0+(4.0*p112)+(30.0*p21)+(-21.0*p212)+(4.0*p11*(-5.0+(3.0*p21)))))+(-2.0*p00*((p20*(-1.0+p21)*(-9.0+(4.0*p11)+(5.0*p21)))+(2.0*p10*(13.0+(-14.0*p11)+(4.0*p112)+(-12.0*p21)+(6.0*p11*p21)+(3.0*p212))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p21)*(-1.0+p21))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(-1.0+p21)*(-9.0+(4.0*p11)+(5.0*p21)))+(p002*(35.0+(20.0*p112)+(-18.0*p21)+(3.0*p212)+(4.0*p11*(-13.0+(3.0*p21)))))+(-4.0*p10*((4.0*p112*p20)+(-2.0*p11*(7.0+(4.0*p20)+(-7.0*p21)))+(p20*(-11.0+(30.0*p21)+(-15.0*p212)))+(14.0*(3.0+(-5.0*p21)+(2.0*p212)))))+(2.0*p00*(-70.0+(-28.0*p112)+(9.0*p20)+(56.0*p21)+(-14.0*p20*p21)+(-14.0*p212)+(5.0*p20*p212)+(4.0*p11*(21.0+(-1.0*p20)+(-7.0*p21)+(p20*p21)))+(2.0*p10*(13.0+(4.0*p112)+(-12.0*p21)+(3.0*p212)+(2.0*p11*(-7.0+(3.0*p21)))))))+(p20*((p112*(56.0+(-20.0*p20)))+(-56.0*(2.0+(-8.0*p21)+(5.0*p212)))+(5.0*p20*(5.0+(-30.0*p21)+(21.0*p212)))+(-4.0*p11*((-28.0*(-2.0+p21))+(5.0*p20*(-5.0+(3.0*p21))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p21)*(-1.0+p21))*sign);
break;
case 9:
p202 = p20*p20;
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(-1.0+p21)*(5.0+(4.0*p11)+(5.0*p21)))+(-4.0*p10*p20*(3.0+(6.0*p11)+(4.0*p112)+(2.0*p21)+(-15.0*p212)))+(-1.0*p202*(3.0+(20.0*p112)+(10.0*p21)+(-105.0*p212)+(12.0*p11*(1.0+(5.0*p21))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p212)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(-1.0+p21)*(5.0+(4.0*p11)+(5.0*p21)))+(4.0*p10*((4.0*p112*p20)+(2.0*p11*(-7.0+(3.0*p20)+(7.0*p21)))+(p20*(3.0+(2.0*p21)+(-15.0*p212)))+(28.0*(-1.0+p212))))+(p20*((4.0*p112*(-14.0+(5.0*p20)))+(4.0*p11*(-14.0+(3.0*p20)+(-28.0*p21)+(15.0*p20*p21)))+(p20*(3.0+(10.0*p21)+(-105.0*p212)))+(28.0*(-1.0+(-1.0*p21)+(10.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p212)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(-1.0+p21)*(-9.0+(4.0*p11)+(5.0*p21)))+(4.0*p10*p20*(-11.0+(-8.0*p11)+(4.0*p112)+(30.0*p21)+(-15.0*p212)))+(5.0*p202*(-5.0+(4.0*p112)+(30.0*p21)+(-21.0*p212)+(4.0*p11*(-5.0+(3.0*p21))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p21)*(-1.0+p21))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(-1.0+p21)*(-9.0+(4.0*p11)+(5.0*p21)))+(-4.0*p10*((4.0*p112*p20)+(-2.0*p11*(7.0+(4.0*p20)+(-7.0*p21)))+(p20*(-11.0+(30.0*p21)+(-15.0*p212)))+(14.0*(3.0+(-5.0*p21)+(2.0*p212)))))+(p20*((p112*(56.0+(-20.0*p20)))+(-56.0*(2.0+(-8.0*p21)+(5.0*p212)))+(5.0*p20*(5.0+(-30.0*p21)+(21.0*p212)))+(-4.0*p11*((-28.0*(-2.0+p21))+(5.0*p20*(-5.0+(3.0*p21))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p21)*(-1.0+p21))*sign);
break;
case 10:
p002 = p00*p00;
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*(-3.0+(-12.0*p11)+(-20.0*p112)+(-10.0*p21)+(-60.0*p11*p21)+(105.0*p212)+(4.0*p102*(-1.0+p21)*(5.0+(4.0*p11)+(5.0*p21)))+(-4.0*p10*(3.0+(6.0*p11)+(4.0*p112)+(2.0*p21)+(-15.0*p212)))+(p002*(-105.0+(20.0*p112)+(10.0*p21)+(3.0*p212)+(12.0*p11*(5.0+p21))))+(2.0*p00*(((-1.0+p21)*(5.0+(4.0*p11)+(5.0*p21)))+(2.0*p10*(-15.0+(4.0*p112)+(2.0*p21)+(6.0*p11*p21)+(3.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p212)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*(-25.0+(-44.0*p11)+(-36.0*p112)+(-18.0*p21)+(-52.0*p11*p21)+(175.0*p212)+(-4.0*p102*(-1.0+p21)*(5.0+(4.0*p11)+(5.0*p21)))+(-1.0*p002*(-105.0+(20.0*p112)+(10.0*p21)+(3.0*p212)+(12.0*p11*(5.0+p21))))+(4.0*p10*(-25.0+(4.0*p112)+(2.0*p21)+(13.0*p212)+(2.0*p11*(-4.0+(7.0*p21)))))+(-2.0*p00*(135.0+(-28.0*p112)+(-14.0*p21)+(-9.0*p212)+(-12.0*p11*(5.0+(2.0*p21)))+(2.0*p10*(-15.0+(4.0*p112)+(2.0*p21)+(6.0*p11*p21)+(3.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p212)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(-1.0+p21)*(-9.0+(4.0*p11)+(5.0*p21)))+(4.0*p10*(-11.0+(-8.0*p11)+(4.0*p112)+(30.0*p21)+(-15.0*p212)))+(-1.0*p002*(35.0+(20.0*p112)+(-18.0*p21)+(3.0*p212)+(4.0*p11*(-13.0+(3.0*p21)))))+(5.0*(-5.0+(4.0*p112)+(30.0*p21)+(-21.0*p212)+(4.0*p11*(-5.0+(3.0*p21)))))+(-2.0*p00*(((-1.0+p21)*(-9.0+(4.0*p11)+(5.0*p21)))+(2.0*p10*(13.0+(-14.0*p11)+(4.0*p112)+(-12.0*p21)+(6.0*p11*p21)+(3.0*p212))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p21)*(-1.0+p21))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*(-87.0+(-124.0*p11)+(36.0*p112)+(298.0*p21)+(52.0*p11*p21)+(-175.0*p212)+(4.0*p102*(-1.0+p21)*(-9.0+(4.0*p11)+(5.0*p21)))+(-4.0*p10*(31.0+(-22.0*p11)+(4.0*p112)+(-40.0*p21)+(14.0*p11*p21)+(13.0*p212)))+(p002*(35.0+(20.0*p112)+(-18.0*p21)+(3.0*p212)+(4.0*p11*(-13.0+(3.0*p21)))))+(2.0*p00*(-61.0+(80.0*p11)+(-28.0*p112)+(42.0*p21)+(-24.0*p11*p21)+(-9.0*p212)+(2.0*p10*(13.0+(-14.0*p11)+(4.0*p112)+(-12.0*p21)+(6.0*p11*p21)+(3.0*p212))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p21)*(-1.0+p21))*sign);
break;
case 11:
p212 = p21*p21;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*(-3.0+(-20.0*p112)+(-10.0*p21)+(105.0*p212)+(-12.0*p11*(1.0+(5.0*p21)))+(4.0*p102*(-1.0+p21)*(5.0+(4.0*p11)+(5.0*p21)))+(-4.0*p10*(3.0+(6.0*p11)+(4.0*p112)+(2.0*p21)+(-15.0*p212))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p212)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*(-25.0+(-44.0*p11)+(-36.0*p112)+(-18.0*p21)+(-52.0*p11*p21)+(175.0*p212)+(-4.0*p102*(-1.0+p21)*(5.0+(4.0*p11)+(5.0*p21)))+(4.0*p10*(-25.0+(4.0*p112)+(2.0*p21)+(13.0*p212)+(2.0*p11*(-4.0+(7.0*p21))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p212)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(-1.0+p21)*(-9.0+(4.0*p11)+(5.0*p21)))+(4.0*p10*(-11.0+(-8.0*p11)+(4.0*p112)+(30.0*p21)+(-15.0*p212)))+(5.0*(-5.0+(4.0*p112)+(30.0*p21)+(-21.0*p212)+(4.0*p11*(-5.0+(3.0*p21))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p21)*(-1.0+p21))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*(-87.0+(36.0*p112)+(298.0*p21)+(-175.0*p212)+(4.0*p102*(-1.0+p21)*(-9.0+(4.0*p11)+(5.0*p21)))+(4.0*p11*(-31.0+(13.0*p21)))+(-4.0*p10*(31.0+(4.0*p112)+(-40.0*p21)+(13.0*p212)+(2.0*p11*(-11.0+(7.0*p21))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p21)*(-1.0+p21))*sign);
break;
case 12:
p002 = p00*p00;
p012 = p01*p01;
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((p002*((-105.0*p012)+(60.0*p01*p11)+(20.0*p112)))+(-4.0*p112*p20*((4.0*p10)+(5.0*p20)))+(-4.0*p01*p11*((4.0*p102)+(6.0*p10*p20)+(3.0*p202)))+(-1.0*p012*((20.0*p102)+(12.0*p10*p20)+(3.0*p202)))+(-2.0*p00*((-8.0*p10*p112)+(4.0*p01*p11*p20)+(5.0*p012*((6.0*p10)+p20)))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((5.0*p002*((21.0*p012)+(-12.0*p01*p11)+(-4.0*p112)))+(4.0*p112*p20*(-14.0+(4.0*p10)+(5.0*p20)))+(2.0*p00*((4.0*(7.0+(-2.0*p10))*p112)+(4.0*p01*p11*(14.0+p20))+(5.0*p012*(-28.0+(6.0*p10)+p20))))+(p012*((20.0*p102)+(4.0*p10*(-28.0+(3.0*p20)))+(p20*(-28.0+(3.0*p20)))))+(4.0*p01*p11*((4.0*p102)+(p20*(-14.0+(3.0*p20)))+(2.0*p10*(-7.0+(3.0*p20))))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((p002*((105.0*p012)+(4.0*(28.0+(-5.0*p11))*p11)+(-20.0*p01*(7.0+(3.0*p11)))))+(4.0*p11*p20*((2.0*p10*(-7.0+(2.0*p11)))+((-28.0+(5.0*p11))*p20)))+(p012*((20.0*p102)+(12.0*p10*p20)+(3.0*p202)))+(4.0*p01*((2.0*p102*(-7.0+(2.0*p11)))+(2.0*p10*(-7.0+(3.0*p11))*p20)+((-7.0+(3.0*p11))*p202)))+(2.0*p00*((4.0*p10*(7.0+(-2.0*p11))*p11)+(5.0*p012*((6.0*p10)+p20))+(-2.0*p01*((28.0*p10)+(7.0*p20)+(-2.0*p11*p20))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((p002*((-105.0*p012)+(20.0*p01*(7.0+(3.0*p11)))+(4.0*p11*(-28.0+(5.0*p11)))))+(-4.0*p11*p20*(70.0+(-14.0*p11)+(2.0*p10*(-7.0+(2.0*p11)))+(-28.0*p20)+(5.0*p11*p20)))+(p012*((-20.0*p102)+((28.0+(-3.0*p20))*p20)+(-4.0*p10*(-28.0+(3.0*p20)))))+(-4.0*p01*((2.0*p102*(-7.0+(2.0*p11)))+(p20*(35.0+(-14.0*p11)+(-7.0*p20)+(3.0*p11*p20)))+(2.0*p10*(35.0+(-7.0*p11)+(-7.0*p20)+(3.0*p11*p20)))))+(-2.0*p00*((4.0*p11*(-35.0+(7.0*p10)+(7.0*p11)+(-2.0*p10*p11)))+(5.0*p012*(-28.0+(6.0*p10)+p20))+(p01*(210.0+(-56.0*p10)+(56.0*p11)+(-14.0*p20)+(4.0*p11*p20))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 13:
p012 = p01*p01;
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p112*p20*((4.0*p10)+(5.0*p20)))+(-4.0*p01*p11*((4.0*p102)+(6.0*p10*p20)+(3.0*p202)))+(-1.0*p012*((20.0*p102)+(12.0*p10*p20)+(3.0*p202))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p112*p20*(-14.0+(4.0*p10)+(5.0*p20)))+(p012*((20.0*p102)+(4.0*p10*(-28.0+(3.0*p20)))+(p20*(-28.0+(3.0*p20)))))+(4.0*p01*p11*((4.0*p102)+(p20*(-14.0+(3.0*p20)))+(2.0*p10*(-7.0+(3.0*p20))))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*p11*p20*((2.0*p10*(-7.0+(2.0*p11)))+((-28.0+(5.0*p11))*p20)))+(p012*((20.0*p102)+(12.0*p10*p20)+(3.0*p202)))+(4.0*p01*((2.0*p102*(-7.0+(2.0*p11)))+(2.0*p10*(-7.0+(3.0*p11))*p20)+((-7.0+(3.0*p11))*p202))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p11*p20*(70.0+(-14.0*p11)+(2.0*p10*(-7.0+(2.0*p11)))+(-28.0*p20)+(5.0*p11*p20)))+(p012*((-20.0*p102)+((28.0+(-3.0*p20))*p20)+(-4.0*p10*(-28.0+(3.0*p20)))))+(-4.0*p01*((2.0*p102*(-7.0+(2.0*p11)))+(p20*(35.0+(-14.0*p11)+(-7.0*p20)+(3.0*p11*p20)))+(2.0*p10*(35.0+(-7.0*p11)+(-7.0*p20)+(3.0*p11*p20))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 14:
p002 = p00*p00;
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-1.0*p012*(3.0+(105.0*p002)+(12.0*p10)+(20.0*p102)+(10.0*p00*(1.0+(6.0*p10)))))+(-4.0*p01*(3.0+(2.0*p00)+(-15.0*p002)+(6.0*p10)+(4.0*p102))*p11)+(4.0*(-1.0+p00)*(5.0+(5.0*p00)+(4.0*p10))*p112))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((5.0*p012*(-5.0+(21.0*p002)+(-20.0*p10)+(4.0*p102)+(6.0*p00*(-9.0+(2.0*p10)))))+(-4.0*p01*(11.0+(-30.0*p00)+(15.0*p002)+(8.0*p10)+(-4.0*p102))*p11)+(-4.0*(-1.0+p00)*(-9.0+(5.0*p00)+(4.0*p10))*p112))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((p012*(3.0+(105.0*p002)+(12.0*p10)+(20.0*p102)+(10.0*p00*(1.0+(6.0*p10)))))+(-4.0*p01*(7.0+(p10*(14.0+(-6.0*p11)))+(p102*(14.0+(-4.0*p11)))+(p00*(7.0+(28.0*p10)+(-2.0*p11)))+(-3.0*p11)+(5.0*p002*(7.0+(3.0*p11)))))+(-4.0*(-1.0+p00)*p11*(-28.0+(5.0*p11)+(2.0*p10*(-7.0+(2.0*p11)))+(p00*(-28.0+(5.0*p11))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-5.0*p012*(-5.0+(21.0*p002)+(-20.0*p10)+(4.0*p102)+(6.0*p00*(-9.0+(2.0*p10)))))+(4.0*p01*(-28.0+(p00*(-98.0+(28.0*p10)+(-30.0*p11)))+(p102*(14.0+(-4.0*p11)))+(8.0*p10*(-7.0+p11))+(11.0*p11)+(5.0*p002*(7.0+(3.0*p11)))))+(4.0*(-1.0+p00)*p11*(42.0+(-9.0*p11)+(2.0*p10*(-7.0+(2.0*p11)))+(p00*(-28.0+(5.0*p11))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 15:
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-1.0*p012*(3.0+(12.0*p10)+(20.0*p102)))+(-4.0*p01*(3.0+(6.0*p10)+(4.0*p102))*p11)+(-4.0*(5.0+(4.0*p10))*p112))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((5.0*p012*(-5.0+(-20.0*p10)+(4.0*p102)))+(4.0*p01*(-11.0+(-8.0*p10)+(4.0*p102))*p11)+(4.0*(-9.0+(4.0*p10))*p112))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((p012*(3.0+(12.0*p10)+(20.0*p102)))+(4.0*p11*(-28.0+(5.0*p11)+(2.0*p10*(-7.0+(2.0*p11)))))+(4.0*p01*(-7.0+(3.0*p11)+(2.0*p102*(-7.0+(2.0*p11)))+(2.0*p10*(-7.0+(3.0*p11))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-5.0*p012*(-5.0+(-20.0*p10)+(4.0*p102)))+(4.0*p11*(-42.0+(p10*(14.0+(-4.0*p11)))+(9.0*p11)))+(-4.0*p01*(28.0+(-8.0*p10*(-7.0+p11))+(-11.0*p11)+(2.0*p102*(-7.0+(2.0*p11))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 16:
p002 = p00*p00;
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((1.0/210.0)*p112*(p00+(-1.0*p20))*((5.0*p00)+(4.0*p10)+(5.0*p20))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((-1.0/210.0)*p112*(p00+(-1.0*p20))*(-14.0+(5.0*p00)+(4.0*p10)+(5.0*p20))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((-1.0/210.0)*p11*(p00+(-1.0*p20))*((2.0*p10*(-7.0+(2.0*p11)))+(p00*(-28.0+(5.0*p11)))+((-28.0+(5.0*p11))*p20))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((1.0/210.0)*p11*(p00+(-1.0*p20))*(70.0+(-14.0*p11)+(2.0*p10*(-7.0+(2.0*p11)))+(p00*(-28.0+(5.0*p11)))+(-28.0*p20)+(5.0*p11*p20))*sign);
propagate[x][y] += 0.0;
break;
case 17:
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((-1.0/210.0)*p112*p20*((4.0*p10)+(5.0*p20))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((1.0/210.0)*p112*p20*(-14.0+(4.0*p10)+(5.0*p20))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((1.0/210.0)*p11*p20*((2.0*p10*(-7.0+(2.0*p11)))+((-28.0+(5.0*p11))*p20))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((1.0/210.0)*p11*p20*(-70.0+(p10*(14.0+(-4.0*p11)))+(p11*(14.0+(-5.0*p20)))+(28.0*p20))*sign);
propagate[x][y] += 0.0;
break;
case 18:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += ((1.0/210.0)*(-1.0+p00)*(5.0+(5.0*p00)+(4.0*p10))*p112*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((-1.0/210.0)*(-1.0+p00)*(-9.0+(5.0*p00)+(4.0*p10))*p112*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((-1.0/210.0)*(-1.0+p00)*p11*(-28.0+(5.0*p11)+(2.0*p10*(-7.0+(2.0*p11)))+(p00*(-28.0+(5.0*p11))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((1.0/210.0)*(-1.0+p00)*p11*(42.0+(-9.0*p11)+(2.0*p10*(-7.0+(2.0*p11)))+(p00*(-28.0+(5.0*p11))))*sign);
propagate[x][y] += 0.0;
break;
case 19:
x = 0; y = 0;
pixel[y][-x] += ((-1.0/210.0)*(5.0+(4.0*p10))*p112*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((1.0/210.0)*(-9.0+(4.0*p10))*p112*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((1.0/210.0)*p11*(-28.0+(5.0*p11)+(2.0*p10*(-7.0+(2.0*p11))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((1.0/210.0)*p11*(-42.0+(p10*(14.0+(-4.0*p11)))+(9.0*p11))*sign);
propagate[x][y] += 0.0;
break;
case 20:
p002 = p00*p00;
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(5.0+(4.0*p11)))+(5.0*p002*(-21.0+(12.0*p11)+(4.0*p112)))+(-4.0*p10*(3.0+(6.0*p11)+(4.0*p112))*p20)+(-1.0*(3.0+(12.0*p11)+(20.0*p112))*p202)+(2.0*p00*((p10*(-30.0+(8.0*p112)))+(-1.0*(5.0+(4.0*p11))*p20))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(5.0+(4.0*p11)))+(-5.0*p002*(-21.0+(12.0*p11)+(4.0*p112)))+(2.0*p00*((28.0*p112)+(p10*(30.0+(-8.0*p112)))+(5.0*(-28.0+p20))+(4.0*p11*(14.0+p20))))+(4.0*p10*(-28.0+(3.0*p20)+(4.0*p112*p20)+(2.0*p11*(-7.0+(3.0*p20)))))+(p20*(-28.0+(3.0*p20)+(4.0*p11*(-14.0+(3.0*p20)))+(4.0*p112*(-14.0+(5.0*p20))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(-9.0+(4.0*p11)))+(p002*(-35.0+(52.0*p11)+(-20.0*p112)))+(4.0*p10*(-11.0+(-8.0*p11)+(4.0*p112))*p20)+(5.0*(-5.0+(-20.0*p11)+(4.0*p112))*p202)+(-2.0*p00*((p10*(26.0+(-28.0*p11)+(8.0*p112)))+((9.0+(-4.0*p11))*p20))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(-9.0+(4.0*p11)))+(p002*(35.0+(-52.0*p11)+(20.0*p112)))+(2.0*p00*(-70.0+(84.0*p11)+(-28.0*p112)+(p10*(26.0+(-28.0*p11)+(8.0*p112)))+(9.0*p20)+(-4.0*p11*p20)))+(-4.0*p10*(42.0+(-11.0*p20)+(4.0*p112*p20)+(-2.0*p11*(7.0+(4.0*p20)))))+(p20*(-112.0+(p112*(56.0+(-20.0*p20)))+(25.0*p20)+(4.0*p11*(-56.0+(25.0*p20))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 21:
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(5.0+(4.0*p11)))+(-4.0*p10*(3.0+(6.0*p11)+(4.0*p112))*p20)+(-1.0*(3.0+(12.0*p11)+(20.0*p112))*p202))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(5.0+(4.0*p11)))+(4.0*p10*(-28.0+(3.0*p20)+(4.0*p112*p20)+(2.0*p11*(-7.0+(3.0*p20)))))+(p20*(-28.0+(3.0*p20)+(4.0*p11*(-14.0+(3.0*p20)))+(4.0*p112*(-14.0+(5.0*p20))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(-9.0+(4.0*p11)))+(4.0*p10*(-11.0+(-8.0*p11)+(4.0*p112))*p20)+(5.0*(-5.0+(-20.0*p11)+(4.0*p112))*p202))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(-9.0+(4.0*p11)))+(-4.0*p10*(42.0+(-11.0*p20)+(4.0*p112*p20)+(-2.0*p11*(7.0+(4.0*p20)))))+(p20*(-112.0+(p112*(56.0+(-20.0*p20)))+(25.0*p20)+(4.0*p11*(-56.0+(25.0*p20))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 22:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*(-3.0+(-12.0*p11)+(-20.0*p112)+(-4.0*p102*(5.0+(4.0*p11)))+(-4.0*p10*(3.0+(6.0*p11)+(4.0*p112)))+(5.0*p002*(-21.0+(12.0*p11)+(4.0*p112)))+(2.0*p00*(-5.0+(-4.0*p11)+(p10*(-30.0+(8.0*p112))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*(-25.0+(-44.0*p11)+(-36.0*p112)+(4.0*p102*(5.0+(4.0*p11)))+(4.0*p10*(-25.0+(-8.0*p11)+(4.0*p112)))+(-5.0*p002*(-21.0+(12.0*p11)+(4.0*p112)))+(p00*(-270.0+(120.0*p11)+(56.0*p112)+(p10*(60.0+(-16.0*p112))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(-9.0+(4.0*p11)))+(p002*(-35.0+(52.0*p11)+(-20.0*p112)))+(5.0*(-5.0+(-20.0*p11)+(4.0*p112)))+(4.0*p10*(-11.0+(-8.0*p11)+(4.0*p112)))+(-2.0*p00*(9.0+(-4.0*p11)+(p10*(26.0+(-28.0*p11)+(8.0*p112))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*(-87.0+(-124.0*p11)+(36.0*p112)+(-4.0*p102*(-9.0+(4.0*p11)))+(-4.0*p10*(31.0+(-22.0*p11)+(4.0*p112)))+(p002*(35.0+(-52.0*p11)+(20.0*p112)))+(2.0*p00*(-61.0+(80.0*p11)+(-28.0*p112)+(p10*(26.0+(-28.0*p11)+(8.0*p112))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 23:
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*(-3.0+(-12.0*p11)+(-20.0*p112)+(-4.0*p102*(5.0+(4.0*p11)))+(-4.0*p10*(3.0+(6.0*p11)+(4.0*p112))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*(-25.0+(-44.0*p11)+(-36.0*p112)+(4.0*p102*(5.0+(4.0*p11)))+(4.0*p10*(-25.0+(-8.0*p11)+(4.0*p112))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(-9.0+(4.0*p11)))+(5.0*(-5.0+(-20.0*p11)+(4.0*p112)))+(4.0*p10*(-11.0+(-8.0*p11)+(4.0*p112))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*(-87.0+(-124.0*p11)+(36.0*p112)+(-4.0*p102*(-9.0+(4.0*p11)))+(-4.0*p10*(31.0+(-22.0*p11)+(4.0*p112))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 24:
p002 = p00*p00;
p012 = p01*p01;
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-4.0*(-1.0+p01)*p102*(5.0+(5.0*p01)+(4.0*p11)))+(p002*(3.0+(-105.0*p012)+(12.0*p11)+(20.0*p112)+(10.0*p01*(1.0+(6.0*p11)))))+(-4.0*p10*(-15.0+(2.0*p01)+(3.0*p012)+(6.0*p01*p11)+(4.0*p112))*p20)+(-1.0*(-105.0+(10.0*p01)+(3.0*p012)+(60.0*p11)+(12.0*p01*p11)+(20.0*p112))*p202)+(2.0*p00*((p10*(6.0+(4.0*p01)+(-30.0*p012)+(12.0*p11)+(8.0*p112)))+(-1.0*(-1.0+p01)*(5.0+(5.0*p01)+(4.0*p11))*p20))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*(-1.0+p01)*p102*(5.0+(5.0*p01)+(4.0*p11)))+(p002*(-3.0+(105.0*p012)+(-12.0*p11)+(-20.0*p112)+(-10.0*p01*(1.0+(6.0*p11)))))+(2.0*p00*(14.0+(28.0*p11)+(28.0*p112)+(-2.0*p10*(3.0+(6.0*p11)+(4.0*p112)))+(-5.0*p20)+(-4.0*p11*p20)+(5.0*p012*(-28.0+(6.0*p10)+p20))+(p01*(14.0+(-4.0*p10)+(4.0*p11*(14.0+p20))))))+(4.0*p10*(28.0+(14.0*p11)+(-15.0*p20)+(4.0*p112*p20)+(p012*(-28.0+(3.0*p20)))+(2.0*p01*(p20+(p11*(-7.0+(3.0*p20)))))))+(p20*((35.0*(8.0+(-3.0*p20)))+(p012*(-28.0+(3.0*p20)))+(4.0*p112*(-14.0+(5.0*p20)))+(4.0*p11*(-28.0+(15.0*p20)))+(2.0*p01*(-14.0+(5.0*p20)+(p11*(-28.0+(6.0*p20))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*(-1.0+p01)*p102*(-9.0+(5.0*p01)+(4.0*p11)))+(5.0*p002*(5.0+(21.0*p012)+(20.0*p11)+(-4.0*p112)+(-6.0*p01*(5.0+(2.0*p11)))))+(4.0*p10*(13.0+(3.0*p012)+(6.0*p01*(-2.0+p11))+(-14.0*p11)+(4.0*p112))*p20)+((35.0+(3.0*p012)+(-52.0*p11)+(20.0*p112)+(6.0*p01*(-3.0+(2.0*p11))))*p202)+(2.0*p00*((2.0*p10*(11.0+(-30.0*p01)+(15.0*p012)+(8.0*p11)+(-4.0*p112)))+((-1.0+p01)*(-9.0+(5.0*p01)+(4.0*p11))*p20))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*(-1.0+p01)*p102*(-9.0+(5.0*p01)+(4.0*p11)))+(-5.0*p002*(5.0+(21.0*p012)+(20.0*p11)+(-4.0*p112)+(-6.0*p01*(5.0+(2.0*p11)))))+(-2.0*p00*(-56.0+(-112.0*p11)+(28.0*p112)+(p10*(22.0+(16.0*p11)+(-8.0*p112)))+(9.0*p20)+(-4.0*p11*p20)+(5.0*p012*(-28.0+(6.0*p10)+p20))+(p01*(224.0+(-60.0*p10)+(56.0*p11)+(-14.0*p20)+(4.0*p11*p20)))))+(-4.0*p10*(-42.0+(-14.0*p11*(-1.0+p20))+(13.0*p20)+(4.0*p112*p20)+(p012*(-28.0+(3.0*p20)))+(2.0*p01*(35.0+(-6.0*p20)+(p11*(-7.0+(3.0*p20)))))))+(-1.0*p20*((p11*(168.0+(-52.0*p20)))+(35.0*(-4.0+p20))+(p012*(-28.0+(3.0*p20)))+(4.0*p112*(-14.0+(5.0*p20)))+(2.0*p01*(56.0+(-9.0*p20)+(p11*(-28.0+(6.0*p20))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 25:
p012 = p01*p01;
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-4.0*(-1.0+p01)*p102*(5.0+(5.0*p01)+(4.0*p11)))+(-4.0*p10*(-15.0+(3.0*p012)+(4.0*p112)+(p01*(2.0+(6.0*p11))))*p20)+(-1.0*((3.0*p012)+(2.0*p01*(5.0+(6.0*p11)))+(5.0*(-21.0+(12.0*p11)+(4.0*p112))))*p202))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*(-1.0+p01)*p102*(5.0+(5.0*p01)+(4.0*p11)))+(4.0*p10*(28.0+(14.0*p11)+(-15.0*p20)+(4.0*p112*p20)+(p012*(-28.0+(3.0*p20)))+(2.0*p01*(p20+(p11*(-7.0+(3.0*p20)))))))+(p20*((35.0*(8.0+(-3.0*p20)))+(p012*(-28.0+(3.0*p20)))+(4.0*p112*(-14.0+(5.0*p20)))+(4.0*p11*(-28.0+(15.0*p20)))+(2.0*p01*(-14.0+(5.0*p20)+(p11*(-28.0+(6.0*p20))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*(-1.0+p01)*p102*(-9.0+(5.0*p01)+(4.0*p11)))+(4.0*p10*(13.0+(3.0*p012)+(6.0*p01*(-2.0+p11))+(-14.0*p11)+(4.0*p112))*p20)+((35.0+(3.0*p012)+(-52.0*p11)+(20.0*p112)+(6.0*p01*(-3.0+(2.0*p11))))*p202))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*(-1.0+p01)*p102*(-9.0+(5.0*p01)+(4.0*p11)))+(-4.0*p10*(-42.0+(-14.0*p11*(-1.0+p20))+(13.0*p20)+(4.0*p112*p20)+(p012*(-28.0+(3.0*p20)))+(2.0*p01*(35.0+(-6.0*p20)+(p11*(-7.0+(3.0*p20)))))))+(p20*((p112*(56.0+(-20.0*p20)))+(p012*(28.0+(-3.0*p20)))+(-35.0*(-4.0+p20))+(4.0*p11*(-42.0+(13.0*p20)))+(-2.0*p01*(56.0+(-9.0*p20)+(p11*(-28.0+(6.0*p20))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 26:
p002 = p00*p00;
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*(105.0+(60.0*p10)+(20.0*p102)+(-1.0*p012*(3.0+(12.0*p10)+(20.0*p102)))+(-60.0*p11)+(16.0*p102*p11)+(-20.0*p112)+(-16.0*p10*p112)+(-2.0*p01*(5.0+(6.0*p11)+(8.0*p102*p11)+(4.0*p10*(1.0+(3.0*p11)))))+(p002*(3.0+(-105.0*p012)+(12.0*p11)+(20.0*p112)+(10.0*p01*(1.0+(6.0*p11)))))+(2.0*p00*(5.0+(-5.0*p012*(1.0+(6.0*p10)))+(4.0*p01*(p10+(-1.0*p11)))+(4.0*p11)+(2.0*p10*(3.0+(6.0*p11)+(4.0*p112))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*(175.0+(52.0*p10)+(-20.0*p102)+(5.0*p012*(-5.0+(-20.0*p10)+(4.0*p102)))+(-52.0*p11)+(56.0*p10*p11)+(-16.0*p102*p11)+(-36.0*p112)+(16.0*p10*p112)+(2.0*p01*(-9.0+(p10*(4.0+(-16.0*p11)))+(-22.0*p11)+(8.0*p102*p11)))+(p002*(-3.0+(105.0*p012)+(-12.0*p11)+(-20.0*p112)+(-10.0*p01*(1.0+(6.0*p11)))))+(2.0*p00*(9.0+(15.0*p012*(-9.0+(2.0*p10)))+(24.0*p11)+(28.0*p112)+(p01*(14.0+(-4.0*p10)+(60.0*p11)))+(-2.0*p10*(3.0+(6.0*p11)+(4.0*p112))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*(35.0+(52.0*p10)+(36.0*p102)+(p012*(3.0+(12.0*p10)+(20.0*p102)))+(-52.0*p11)+(-56.0*p10*p11)+(-16.0*p102*p11)+(20.0*p112)+(16.0*p10*p112)+(2.0*p01*(-9.0+(12.0*p10*(-2.0+p11))+(6.0*p11)+(4.0*p102*(-7.0+(2.0*p11)))))+(5.0*p002*(5.0+(21.0*p012)+(20.0*p11)+(-4.0*p112)+(-6.0*p01*(5.0+(2.0*p11)))))+(2.0*p00*(9.0+(5.0*p012*(1.0+(6.0*p10)))+(-2.0*p01*(7.0+(30.0*p10)+(-2.0*p11)))+(-4.0*p11)+(p10*(22.0+(16.0*p11)+(-8.0*p112))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*(105.0+(116.0*p10)+(-36.0*p102)+(-5.0*p012*(-5.0+(-20.0*p10)+(4.0*p102)))+(-116.0*p11)+(16.0*p102*p11)+(36.0*p112)+(-16.0*p10*p112)+(-5.0*p002*(5.0+(21.0*p012)+(20.0*p11)+(-4.0*p112)+(-6.0*p01*(5.0+(2.0*p11)))))+(p01*(-94.0+(44.0*p11)+(-8.0*p102*(-7.0+(2.0*p11)))+(8.0*p10*(-29.0+(4.0*p11)))))+(-2.0*p00*(-47.0+(15.0*p012*(-9.0+(2.0*p10)))+(-30.0*p01*(-7.0+(2.0*p10)+(-2.0*p11)))+(-116.0*p11)+(28.0*p112)+(p10*(22.0+(16.0*p11)+(-8.0*p112))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 27:
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-1.0*p012*(3.0+(12.0*p10)+(20.0*p102)))+(4.0*p102*(5.0+(4.0*p11)))+(p10*(60.0+(-16.0*p112)))+(-5.0*(-21.0+(12.0*p11)+(4.0*p112)))+(-2.0*p01*(5.0+(6.0*p11)+(8.0*p102*p11)+(4.0*p10*(1.0+(3.0*p11))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*(175.0+(5.0*p012*(-5.0+(-20.0*p10)+(4.0*p102)))+(-52.0*p11)+(-36.0*p112)+(-4.0*p102*(5.0+(4.0*p11)))+(2.0*p01*(-9.0+(p10*(4.0+(-16.0*p11)))+(-22.0*p11)+(8.0*p102*p11)))+(4.0*p10*(13.0+(14.0*p11)+(4.0*p112))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*(35.0+(p012*(3.0+(12.0*p10)+(20.0*p102)))+(-52.0*p11)+(20.0*p112)+(-4.0*p102*(-9.0+(4.0*p11)))+(4.0*p10*(13.0+(-14.0*p11)+(4.0*p112)))+(2.0*p01*(-9.0+(12.0*p10*(-2.0+p11))+(6.0*p11)+(4.0*p102*(-7.0+(2.0*p11))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*(105.0+(-5.0*p012*(-5.0+(-20.0*p10)+(4.0*p102)))+(-116.0*p11)+(36.0*p112)+(4.0*p102*(-9.0+(4.0*p11)))+(-4.0*p10*(-29.0+(4.0*p112)))+(p01*(-94.0+(44.0*p11)+(-8.0*p102*(-7.0+(2.0*p11)))+(8.0*p10*(-29.0+(4.0*p11))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 28:
p002 = p00*p00;
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(5.0+(4.0*p11)))+(p002*(3.0+(12.0*p11)+(20.0*p112)))+(4.0*p10*(15.0+(-4.0*p112))*p20)+(-5.0*(-21.0+(12.0*p11)+(4.0*p112))*p202)+(2.0*p00*((2.0*p10*(3.0+(6.0*p11)+(4.0*p112)))+((5.0+(4.0*p11))*p20))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(5.0+(4.0*p11)))+(-1.0*p002*(3.0+(12.0*p11)+(20.0*p112)))+(-2.0*p00*(-14.0+(-28.0*p112)+(2.0*p10*(3.0+(6.0*p11)+(4.0*p112)))+(4.0*p11*(-7.0+p20))+(5.0*p20)))+(4.0*p10*(28.0+(14.0*p11)+(-15.0*p20)+(4.0*p112*p20)))+(p20*((35.0*(8.0+(-3.0*p20)))+(4.0*p112*(-14.0+(5.0*p20)))+(4.0*p11*(-28.0+(15.0*p20))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(-9.0+(4.0*p11)))+(-5.0*p002*(-5.0+(-20.0*p11)+(4.0*p112)))+(4.0*p10*(13.0+(-14.0*p11)+(4.0*p112))*p20)+((35.0+(-52.0*p11)+(20.0*p112))*p202)+(p00*((p10*(44.0+(32.0*p11)+(-16.0*p112)))+(2.0*(9.0+(-4.0*p11))*p20))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(-9.0+(4.0*p11)))+(5.0*p002*(-5.0+(-20.0*p11)+(4.0*p112)))+(-4.0*p10*(-42.0+(-14.0*p11*(-1.0+p20))+(13.0*p20)+(4.0*p112*p20)))+(2.0*p00*(56.0+(-28.0*p112)+(2.0*p10*(-11.0+(-8.0*p11)+(4.0*p112)))+(-9.0*p20)+(4.0*p11*(28.0+p20))))+(p20*((p112*(56.0+(-20.0*p20)))+(-35.0*(-4.0+p20))+(4.0*p11*(-42.0+(13.0*p20))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 29:
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(5.0+(4.0*p11)))+(4.0*p10*(15.0+(-4.0*p112))*p20)+(-5.0*(-21.0+(12.0*p11)+(4.0*p112))*p202))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(5.0+(4.0*p11)))+(4.0*p10*(28.0+(14.0*p11)+(-15.0*p20)+(4.0*p112*p20)))+(p20*((35.0*(8.0+(-3.0*p20)))+(4.0*p112*(-14.0+(5.0*p20)))+(4.0*p11*(-28.0+(15.0*p20))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*((-4.0*p102*(-9.0+(4.0*p11)))+(4.0*p10*(13.0+(-14.0*p11)+(4.0*p112))*p20)+((35.0+(-52.0*p11)+(20.0*p112))*p202))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(-9.0+(4.0*p11)))+(-4.0*p10*(-42.0+(-14.0*p11*(-1.0+p20))+(13.0*p20)+(4.0*p112*p20)))+(p20*((p112*(56.0+(-20.0*p20)))+(-35.0*(-4.0+p20))+(4.0*p11*(-42.0+(13.0*p20))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 30:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(5.0+(4.0*p11)))+(p10*(60.0+(-16.0*p112)))+(-5.0*(-21.0+(12.0*p11)+(4.0*p112)))+(p002*(3.0+(12.0*p11)+(20.0*p112)))+(2.0*p00*(5.0+(4.0*p11)+(2.0*p10*(3.0+(6.0*p11)+(4.0*p112))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*(175.0+(-52.0*p11)+(-36.0*p112)+(-4.0*p102*(5.0+(4.0*p11)))+(4.0*p10*(13.0+(14.0*p11)+(4.0*p112)))+(-1.0*p002*(3.0+(12.0*p11)+(20.0*p112)))+(-2.0*p00*(-9.0+(-24.0*p11)+(-28.0*p112)+(2.0*p10*(3.0+(6.0*p11)+(4.0*p112))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*(35.0+(-52.0*p11)+(20.0*p112)+(-4.0*p102*(-9.0+(4.0*p11)))+(-5.0*p002*(-5.0+(-20.0*p11)+(4.0*p112)))+(4.0*p10*(13.0+(-14.0*p11)+(4.0*p112)))+(p00*(18.0+(-8.0*p11)+(p10*(44.0+(32.0*p11)+(-16.0*p112))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*(105.0+(-116.0*p11)+(36.0*p112)+(4.0*p102*(-9.0+(4.0*p11)))+(-4.0*p10*(-29.0+(4.0*p112)))+(5.0*p002*(-5.0+(-20.0*p11)+(4.0*p112)))+(2.0*p00*(47.0+(116.0*p11)+(-28.0*p112)+(2.0*p10*(-11.0+(-8.0*p11)+(4.0*p112))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 31:
x = 0; y = 0;
pixel[y][-x] += ((1.0/840.0)*((4.0*p102*(5.0+(4.0*p11)))+(p10*(60.0+(-16.0*p112)))+(-5.0*(-21.0+(12.0*p11)+(4.0*p112))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/840.0)*(175.0+(-52.0*p11)+(-36.0*p112)+(-4.0*p102*(5.0+(4.0*p11)))+(4.0*p10*(13.0+(14.0*p11)+(4.0*p112))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/840.0)*(35.0+(-52.0*p11)+(20.0*p112)+(-4.0*p102*(-9.0+(4.0*p11)))+(4.0*p10*(13.0+(-14.0*p11)+(4.0*p112))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/840.0)*(105.0+(-116.0*p11)+(36.0*p112)+(4.0*p102*(-9.0+(4.0*p11)))+(-4.0*p10*(-29.0+(4.0*p112))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 32:
p002 = p00*p00;
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((1.0/210.0)*(-1.0+p11)*(p00+(-1.0*p20))*((2.0*p10*(5.0+(2.0*p11)))+(p00*(23.0+(5.0*p11)))+((23.0+(5.0*p11))*p20))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((-1.0/210.0)*(-1.0+p11)*(p00+(-1.0*p20))*(-56.0+(-14.0*p11)+(2.0*p10*(5.0+(2.0*p11)))+(p00*(23.0+(5.0*p11)))+(23.0*p20)+(5.0*p11*p20))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((-1.0/210.0)*((-1.0+p11)*(-1.0+p11))*(p00+(-1.0*p20))*((5.0*p00)+(4.0*p10)+(5.0*p20))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((1.0/210.0)*((-1.0+p11)*(-1.0+p11))*(p00+(-1.0*p20))*(-14.0+(5.0*p00)+(4.0*p10)+(5.0*p20))*sign);
propagate[x][y] += 0.0;
break;
case 33:
p202 = p20*p20;
x = 0; y = 0;
pixel[y][-x] += ((-1.0/210.0)*(-1.0+p11)*p20*((2.0*p10*(5.0+(2.0*p11)))+((23.0+(5.0*p11))*p20))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((1.0/210.0)*(-1.0+p11)*p20*(-56.0+(2.0*p10*(5.0+(2.0*p11)))+(23.0*p20)+(p11*(-14.0+(5.0*p20))))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((1.0/210.0)*((-1.0+p11)*(-1.0+p11))*p20*((4.0*p10)+(5.0*p20))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((-1.0/210.0)*((-1.0+p11)*(-1.0+p11))*p20*(-14.0+(4.0*p10)+(5.0*p20))*sign);
propagate[x][y] += 0.0;
break;
case 34:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += ((1.0/210.0)*(-1.0+p00)*(-1.0+p11)*(23.0+(5.0*p11)+(2.0*p10*(5.0+(2.0*p11)))+(p00*(23.0+(5.0*p11))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((-1.0/210.0)*(-1.0+p00)*(-1.0+p11)*(-33.0+(-9.0*p11)+(2.0*p10*(5.0+(2.0*p11)))+(p00*(23.0+(5.0*p11))))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((-1.0/210.0)*(-1.0+p00)*(5.0+(5.0*p00)+(4.0*p10))*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((1.0/210.0)*(-1.0+p00)*(-9.0+(5.0*p00)+(4.0*p10))*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += 0.0;
break;
case 35:
x = 0; y = 0;
pixel[y][-x] += ((-1.0/210.0)*(-1.0+p11)*(23.0+(5.0*p11)+(2.0*p10*(5.0+(2.0*p11))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((1.0/210.0)*(-1.0+p11)*(-33.0+(-9.0*p11)+(2.0*p10*(5.0+(2.0*p11))))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((1.0/210.0)*(5.0+(4.0*p10))*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((-1.0/210.0)*(-9.0+(4.0*p10))*((-1.0+p11)*(-1.0+p11))*sign);
propagate[x][y] += 0.0;
break;
}
#endif

#endif
		}
		else// if (c.num_pts == 4)
		{
#if defined(USE_ONLY_CUBICS) || defined(USE_MIXED_CURVES)
#ifdef USE_CUBICS
			const double p00 = c.px[0].v[0];
			const double p01 = c.px[0].v[1];
			const double p10 = c.px[1].v[0];
			const double p11 = c.px[1].v[1];
			const double p20 = c.px[2].v[0];
			const double p21 = c.px[2].v[1];
			const double p30 = c.px[3].v[0];
			const double p31 = c.px[3].v[1];

			const double p102 = p10*p10;
			const double p112 = p11*p11;
			const double p202 = p20*p20;
			const double p212 = p21*p21;

			double p002, p012, p302, p312;

#if 0
			int sign = c.flip_x != c.flip_y ? -1 : 1;
			int special = 0;
			if (p00 == 0) special += 1 * 1;
			if (p30 == 0) special += 1 * 2;
			if (p01 == 0) special += 1 * 2*2;
			if (p01 == 1) special += 2 * 2*2;
			if (p31 == 0) special += 1 * 2*2*3;
			if (p31 == 1) special += 2 * 2*2*3;

switch (special)
{
case 0:
p002 = p00*p00;
p012 = p01*p01;
p302 = p30*p30;
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(-2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-1.0*p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(-22.0*((27.0*p112*p20)+(-27.0*p10*p11*p21)+(27.0*p11*p20*p21)+(-27.0*p10*p212)+(18.0*p112*p30)+(45.0*p11*p21*p30)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(-18.0*p10*p11*p31)+(-45.0*p10*p21*p31)+(-45.0*p20*p21*p31)+(30.0*p11*p30*p31)+(105.0*p21*p30*p31)+(-30.0*p10*p312)+(-105.0*p20*p312)+(-280.0*p30*p312)+(p01*((45.0*p10*p11)+(45.0*p11*p20)+(18.0*p20*p21)+(15.0*p11*p30)+(12.0*p21*p30)+(-3.0*p10*p31)+(3.0*p20*p31)+(5.0*p30*p31)))+(p00*((280.0*p012)+(-45.0*p112)+(-45.0*p11*p21)+(-18.0*p212)+(-12.0*p11*p31)+(-15.0*p21*p31)+(-5.0*p312)+(-5.0*p01*((21.0*p11)+(6.0*p21)+p31))))))+(2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-1.0*p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-22.0*((45.0*p01*p102)+(45.0*p01*p10*p20)+(27.0*p10*p11*p20)+(18.0*p01*p202)+(27.0*p11*p202)+(-27.0*p102*p21)+(-27.0*p10*p20*p21)+(12.0*p01*p10*p30)+(18.0*p10*p11*p30)+(15.0*p01*p20*p30)+(45.0*p11*p20*p30)+(45.0*p20*p21*p30)+(5.0*p01*p302)+(30.0*p11*p302)+(105.0*p21*p302)+(5.0*p002*((28.0*p01)+(-21.0*p11)+(-6.0*p21)+(-1.0*p31)))+(-18.0*p102*p31)+(-45.0*p10*p20*p31)+(-45.0*p202*p31)+(-30.0*p10*p30*p31)+(-105.0*p20*p30*p31)+(-140.0*p302*p31)+(p00*((-18.0*p20*p21)+(3.0*p11*p30)+(-3.0*p21*p30)+(5.0*p01*((21.0*p10)+(6.0*p20)+p30))+(-12.0*p20*p31)+(-5.0*p30*p31)+(-15.0*p10*((3.0*p11)+(3.0*p21)+p31))))))+(2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-22.0*((126.0*p11*p20)+(-27.0*p10*p11*p20)+(-27.0*p11*p202)+(-126.0*p10*p21)+(27.0*p102*p21)+(27.0*p10*p20*p21)+(126.0*p11*p30)+(-18.0*p10*p11*p30)+(-45.0*p11*p20*p30)+(252.0*p21*p30)+(-45.0*p20*p21*p30)+(-30.0*p11*p302)+(-105.0*p21*p302)+(-1.0*p01*((45.0*p102)+(18.0*p202)+(3.0*p10*(-84.0+(15.0*p20)+(4.0*p30)))+(3.0*p20*(-42.0+(5.0*p30)))+(p30*(-42.0+(5.0*p30)))))+(-126.0*p10*p31)+(18.0*p102*p31)+(-252.0*p20*p31)+(45.0*p10*p20*p31)+(45.0*p202*p31)+(-420.0*p30*p31)+(30.0*p10*p30*p31)+(105.0*p20*p30*p31)+(140.0*p302*p31)+(5.0*p002*((-28.0*p01)+(21.0*p11)+(6.0*p21)+p31))+(p00*((-126.0*p21)+(45.0*p10*p21)+(18.0*p20*p21)+(3.0*p11*(-84.0+(15.0*p10)+(-1.0*p30)))+(3.0*p21*p30)+(-5.0*p01*(-84.0+(21.0*p10)+(6.0*p20)+p30))+(-42.0*p31)+(15.0*p10*p31)+(12.0*p20*p31)+(5.0*p30*p31)))))+(-2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(22.0*((27.0*p112*p20)+(-27.0*p10*p11*p21)+(27.0*p11*p20*p21)+(-27.0*p10*p212)+(18.0*p112*p30)+(45.0*p11*p21*p30)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(-18.0*p10*p11*p31)+(-45.0*p10*p21*p31)+(-45.0*p20*p21*p31)+(30.0*p11*p30*p31)+(105.0*p21*p30*p31)+(-30.0*p10*p312)+(-105.0*p20*p312)+(-280.0*p30*p312)+(p01*((45.0*p10*p11)+(45.0*p11*p20)+(18.0*p20*p21)+(15.0*p11*p30)+(12.0*p21*p30)+(-3.0*p10*p31)+(3.0*p20*p31)+(5.0*p30*p31)))+(p00*((280.0*p012)+(-45.0*p112)+(-45.0*p11*p21)+(-18.0*p212)+(-12.0*p11*p31)+(-15.0*p21*p31)+(-5.0*p312)+(-5.0*p01*((21.0*p11)+(6.0*p21)+p31))))))+(-2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31))*sign);
break;
case 1:
p012 = p01*p01;
p302 = p30*p30;
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-22.0*((-27.0*p10*p212)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(9.0*p112*((3.0*p20)+(2.0*p30)))+(-45.0*p10*p21*p31)+(-45.0*p20*p21*p31)+(105.0*p21*p30*p31)+(-30.0*p10*p312)+(-105.0*p20*p312)+(-280.0*p30*p312)+(p01*((18.0*p20*p21)+(12.0*p21*p30)+(15.0*p11*((3.0*p20)+p30))+(p10*((45.0*p11)+(-3.0*p31)))+(3.0*p20*p31)+(5.0*p30*p31)))+(3.0*p11*((-9.0*p10*p21)+(9.0*p20*p21)+(15.0*p21*p30)+(-6.0*p10*p31)+(10.0*p30*p31)))))+(2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-594.0*p11*p202)+(135.0*p112*p202)+(162.0*p11*p202*p21)+(-990.0*p11*p20*p30)+(162.0*p112*p20*p30)+(-990.0*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(252.0*p20*p212*p30)+(-660.0*p11*p302)+(63.0*p112*p302)+(-2310.0*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(990.0*p202*p31)+(-252.0*p202*p21*p31)+(2310.0*p20*p30*p31)+(168.0*p11*p20*p30*p31)+(3080.0*p302*p31)+(252.0*p11*p302*p31)+(1260.0*p21*p302*p31)+(-378.0*p202*p312)+(-1260.0*p20*p30*p312)+(-2310.0*p302*p312)+(-9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(2.0*p01*((9.0*p102*(-55.0+(14.0*p11)+(-1.0*p31)))+(9.0*p202*(-22.0+(9.0*p11)+(5.0*p21)+p31))+(3.0*p20*p30*(-55.0+(20.0*p11)+(18.0*p21)+(7.0*p31)))+(p302*(-55.0+(15.0*p11)+(21.0*p21)+(14.0*p31)))+(3.0*p10*((3.0*p20*(-55.0+(21.0*p11)+(6.0*p21)))+(2.0*p30*(-22.0+(9.0*p11)+(5.0*p21)+p31))))))+(6.0*p10*((3.0*p112*((9.0*p20)+(5.0*p30)))+(-2.0*p30*p31*(-55.0+(14.0*p21)+(21.0*p31)))+(-3.0*p11*((22.0*p30)+(-6.0*p21*p30)+(p20*(33.0+(6.0*p31)))))+(-3.0*p20*((9.0*p212)+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((2772.0*p10*p21)+(-594.0*p102*p21)+(-594.0*p10*p20*p21)+(-594.0*p10*p212)+(135.0*p102*p212)+(162.0*p10*p20*p212)+(-5544.0*p21*p30)+(990.0*p20*p21*p30)+(990.0*p212*p30)+(-252.0*p20*p212*p30)+(2310.0*p21*p302)+(-378.0*p212*p302)+(-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*(-22.0+p30))+(5.0*(-22.0+p30)*p30)+(42.0*p10*(-55.0+(6.0*p20)+p30))))+(-9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(2772.0*p10*p31)+(-396.0*p102*p31)+(5544.0*p20*p31)+(-990.0*p10*p20*p31)+(-990.0*p202*p31)+(-990.0*p10*p21*p31)+(162.0*p102*p21*p31)+(-990.0*p20*p21*p31)+(378.0*p10*p20*p21*p31)+(252.0*p202*p21*p31)+(9240.0*p30*p31)+(-660.0*p10*p30*p31)+(-2310.0*p20*p30*p31)+(2310.0*p21*p30*p31)+(168.0*p10*p21*p30*p31)+(-3080.0*p302*p31)+(-1260.0*p21*p302*p31)+(-660.0*p10*p312)+(63.0*p102*p312)+(-2310.0*p20*p312)+(252.0*p10*p20*p312)+(378.0*p202*p312)+(-6160.0*p30*p312)+(252.0*p10*p30*p312)+(1260.0*p20*p30*p312)+(2310.0*p302*p312)+(-2.0*p01*((9.0*p102*(-55.0+(14.0*p11)+(-1.0*p31)))+(9.0*p202*(-22.0+(9.0*p11)+(5.0*p21)+p31))+(3.0*p10*(924.0+(3.0*p20*(-55.0+(6.0*p21)))+(-44.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))+(11.0*p31)+(2.0*p30*p31)))+(3.0*p20*(462.0+(-55.0*p30)+(6.0*p21*(-11.0+(3.0*p30)))+(5.0*p11*(-33.0+(4.0*p30)))+(-11.0*p31)+(7.0*p30*p31)))+(p30*(462.0+(15.0*p11*(-11.0+p30))+(-55.0*p30)+(3.0*p21*(-44.0+(7.0*p30)))+(-55.0*p31)+(14.0*p30*p31)))))+(-6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)+(22.0*p31)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30)+(-110.0*p31)+(42.0*p30*p31)))+(p20*(462.0+(-99.0*p21)+(-165.0*p30)+(63.0*p21*p30)+(28.0*p30*p31)+(-9.0*p10*(11.0+(2.0*p31))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31))*sign);
break;
case 2:
p002 = p00*p00;
p012 = p01*p01;
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))))+(-2.0*p00*((9.0*p112)+(-126.0*p10*p112)+(14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-9.0*p212)+(-81.0*p10*p212)+(-45.0*p20*p212)+(-6.0*p11*p31)+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-21.0*p21*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-14.0*p312)+(-15.0*p10*p312)+(-21.0*p20*p312)+(3.0*p01*((7.0*p11*(1.0+(4.0*p20)))+(2.0*p21)+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31)))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((21.0*p012*(-5.0+(110.0*p002)+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)+(4.0*p00*(-73.0+(15.0*p10)+(3.0*p20)))))+(-9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(-612.0*p212)+(378.0*p00*p212)+(-63.0*p002*p212)+(594.0*p10*p212)+(-162.0*p00*p10*p212)+(-135.0*p102*p212)+(252.0*p20*p212)+(-90.0*p00*p20*p212)+(-162.0*p10*p20*p212)+(-1050.0*p21*p31)+(288.0*p00*p21*p31)+(-30.0*p002*p21*p31)+(822.0*p10*p21*p31)+(-120.0*p00*p10*p21*p31)+(-162.0*p102*p21*p31)+(990.0*p20*p21*p31)+(-108.0*p00*p20*p21*p31)+(-378.0*p10*p20*p21*p31)+(-252.0*p202*p21*p31)+(3850.0*p312)+(82.0*p00*p312)+(-5.0*p002*p312)+(408.0*p10*p312)+(-30.0*p00*p10*p312)+(-63.0*p102*p312)+(1050.0*p20*p312)+(-42.0*p00*p20*p312)+(-252.0*p10*p20*p312)+(-378.0*p202*p312)+(-6.0*p11*((3.0*(41.0+(14.0*p002)+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202)+(p00*(-55.0+(21.0*p10)+(6.0*p20))))*p21)+((68.0+(7.0*p002)+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p31)))+(-2.0*p01*((3.0*p11*(50.0+(210.0*p002)+(-42.0*p102)+(145.0*p20)+(-27.0*p202)+(-28.0*p00*(14.0+p20))+(-21.0*p10*(-7.0+(3.0*p20)))))+(3.0*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20))))*p21)+((41.0+(14.0*p002)+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202)+(p00*(-55.0+(21.0*p10)+(6.0*p20))))*p31))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(-1.0*p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))))+(-22.0*((5.0*p01)+(12.0*p01*p10)+(45.0*p01*p102)+(30.0*p11)+(18.0*p10*p11)+(15.0*p01*p20)+(45.0*p01*p10*p20)+(45.0*p11*p20)+(27.0*p10*p11*p20)+(18.0*p01*p202)+(27.0*p11*p202)+(105.0*p21)+(-27.0*p102*p21)+(45.0*p20*p21)+(-27.0*p10*p20*p21)+(5.0*p002*((28.0*p01)+(-21.0*p11)+(-6.0*p21)+(-1.0*p31)))+(-140.0*p31)+(-30.0*p10*p31)+(-18.0*p102*p31)+(-105.0*p20*p31)+(-45.0*p10*p20*p31)+(-45.0*p202*p31)+(p00*((3.0*p11)+(5.0*p01*(1.0+(21.0*p10)+(6.0*p20)))+(-3.0*p21)+(-18.0*p20*p21)+(-5.0*p31)+(-12.0*p20*p31)+(-15.0*p10*((3.0*p11)+(3.0*p21)+p31))))))+(2.0*p00*((9.0*p112)+(-126.0*p10*p112)+(14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-9.0*p212)+(-81.0*p10*p212)+(-45.0*p20*p212)+(-6.0*p11*p31)+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-21.0*p21*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-14.0*p312)+(-15.0*p10*p312)+(-21.0*p20*p312)+(3.0*p01*((7.0*p11*(1.0+(4.0*p20)))+(2.0*p21)+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31)))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-21.0*p012*(-5.0+(110.0*p002)+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)+(4.0*p00*(-73.0+(15.0*p10)+(3.0*p20)))))+(9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(-3234.0*p21)+(2706.0*p00*p21)+(-660.0*p002*p21)+(2772.0*p10*p21)+(-990.0*p00*p10*p21)+(-594.0*p102*p21)+(990.0*p20*p21)+(-396.0*p00*p20*p21)+(-594.0*p10*p20*p21)+(612.0*p212)+(-378.0*p00*p212)+(63.0*p002*p212)+(-594.0*p10*p212)+(162.0*p00*p10*p212)+(135.0*p102*p212)+(-252.0*p20*p212)+(90.0*p00*p20*p212)+(162.0*p10*p20*p212)+(6160.0*p31)+(814.0*p00*p31)+(-110.0*p002*p31)+(2112.0*p10*p31)+(-330.0*p00*p10*p31)+(-396.0*p102*p31)+(3234.0*p20*p31)+(-264.0*p00*p20*p31)+(-990.0*p10*p20*p31)+(-990.0*p202*p31)+(1050.0*p21*p31)+(-288.0*p00*p21*p31)+(30.0*p002*p21*p31)+(-822.0*p10*p21*p31)+(120.0*p00*p10*p21*p31)+(162.0*p102*p21*p31)+(-990.0*p20*p21*p31)+(108.0*p00*p20*p21*p31)+(378.0*p10*p20*p21*p31)+(252.0*p202*p21*p31)+(-3850.0*p312)+(-82.0*p00*p312)+(5.0*p002*p312)+(-408.0*p10*p312)+(30.0*p00*p10*p312)+(63.0*p102*p312)+(-1050.0*p20*p312)+(42.0*p00*p20*p312)+(252.0*p10*p20*p312)+(378.0*p202*p312)+(2.0*p01*(-407.0+(150.0*p11)+(-1221.0*p20)+(435.0*p11*p20)+(198.0*p202)+(-81.0*p11*p202)+(111.0*p21)+(144.0*p20*p21)+(-45.0*p202*p21)+(41.0*p31)+(12.0*p20*p31)+(-9.0*p202*p31)+(9.0*p102*(55.0+(-14.0*p11)+p31))+(14.0*p002*(110.0+(45.0*p11)+(9.0*p21)+p31))+(-3.0*p10*(880.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))+(13.0*p31)))+(p00*(-4565.0+(330.0*p20)+(-84.0*p11*(14.0+p20))+(-336.0*p21)+(-55.0*p31)+(6.0*p20*p31)+(21.0*p10*(55.0+(4.0*p21)+p31))))))+(6.0*p11*(-352.0+(-297.0*p20)+(99.0*p202)+(123.0*p21)+(36.0*p20*p21)+(-27.0*p202*p21)+(68.0*p31)+(-28.0*p20*p31)+(7.0*p002*(-55.0+(6.0*p21)+p31))+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31)))))+(p00*(935.0+(3.0*(-55.0+(6.0*p20))*p21)+(-42.0*p31)+(10.0*p20*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31))*sign);
break;
case 3:
p012 = p01*p01;
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-612.0*p212)+(594.0*p10*p212)+(-135.0*p102*p212)+(252.0*p20*p212)+(-162.0*p10*p20*p212)+(-1050.0*p21*p31)+(822.0*p10*p21*p31)+(-162.0*p102*p21*p31)+(990.0*p20*p21*p31)+(-378.0*p10*p20*p21*p31)+(-252.0*p202*p21*p31)+(3850.0*p312)+(408.0*p10*p312)+(-63.0*p102*p312)+(1050.0*p20*p312)+(-252.0*p10*p20*p312)+(-378.0*p202*p312)+(2.0*p01*((3.0*p11*(-50.0+(42.0*p102)+(-145.0*p20)+(27.0*p202)+(21.0*p10*(-7.0+(3.0*p20)))))+(3.0*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((-41.0+(39.0*p10)+(-9.0*p102)+(-12.0*p20)+(9.0*p202))*p31)))+(-6.0*p11*((3.0*(41.0+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202))*p21)+((68.0+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20))))*p31))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-22.0*((3.0*p11*(10.0+(6.0*p10)+(15.0*p20)+(9.0*p10*p20)+(9.0*p202)))+(p01*(5.0+(45.0*p102)+(15.0*p20)+(18.0*p202)+(3.0*p10*(4.0+(15.0*p20)))))+(105.0*p21)+(-27.0*p102*p21)+(45.0*p20*p21)+(-27.0*p10*p20*p21)+(-140.0*p31)+(-30.0*p10*p31)+(-18.0*p102*p31)+(-105.0*p20*p31)+(-45.0*p10*p20*p31)+(-45.0*p202*p31)))+(2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(-3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-3234.0*p21)+(2772.0*p10*p21)+(-594.0*p102*p21)+(990.0*p20*p21)+(-594.0*p10*p20*p21)+(612.0*p212)+(-594.0*p10*p212)+(135.0*p102*p212)+(-252.0*p20*p212)+(162.0*p10*p20*p212)+(6160.0*p31)+(2112.0*p10*p31)+(-396.0*p102*p31)+(3234.0*p20*p31)+(-990.0*p10*p20*p31)+(-990.0*p202*p31)+(1050.0*p21*p31)+(-822.0*p10*p21*p31)+(162.0*p102*p21*p31)+(-990.0*p20*p21*p31)+(378.0*p10*p20*p21*p31)+(252.0*p202*p21*p31)+(-3850.0*p312)+(-408.0*p10*p312)+(63.0*p102*p312)+(-1050.0*p20*p312)+(252.0*p10*p20*p312)+(378.0*p202*p312)+(-2.0*p01*(407.0+(1221.0*p20)+(-198.0*p202)+(3.0*p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-111.0*p21)+(-144.0*p20*p21)+(45.0*p202*p21)+(9.0*p102*(-55.0+(14.0*p11)+(-1.0*p31)))+(-41.0*p31)+(-12.0*p20*p31)+(9.0*p202*p31)+(3.0*p10*(880.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))+(13.0*p31)))))+(6.0*p11*(-352.0+(p202*(99.0+(-27.0*p21)))+(123.0*p21)+(p20*(-297.0+(36.0*p21)+(-28.0*p31)))+(68.0*p31)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31))*sign);
break;
case 4:
p002 = p00*p00;
p302 = p30*p30;
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((p002*((378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))))+(2.0*p00*((45.0*p20*p212)+(-9.0*p112*p30)+(9.0*p212*p30)+(54.0*p20*p21*p31)+(21.0*p21*p30*p31)+(21.0*p20*p312)+(14.0*p30*p312)+(6.0*p11*((9.0*p20*p21)+(5.0*p20*p31)+(p30*p31)))+(3.0*p10*((42.0*p112)+(27.0*p212)+(20.0*p21*p31)+(5.0*p312)+(9.0*p11*((7.0*p21)+(2.0*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))))*sign);
propagate[x][y] += ((1.0/4.0)*p312*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((594.0*p10*p212)+(-135.0*p102*p212)+(-162.0*p10*p20*p212)+(-990.0*p212*p30)+(252.0*p20*p212*p30)+(378.0*p212*p302)+(9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(990.0*p10*p21*p31)+(-162.0*p102*p21*p31)+(990.0*p20*p21*p31)+(-378.0*p10*p20*p21*p31)+(-252.0*p202*p21*p31)+(-2310.0*p21*p30*p31)+(-168.0*p10*p21*p30*p31)+(1260.0*p21*p302*p31)+(660.0*p10*p312)+(-63.0*p102*p312)+(2310.0*p20*p312)+(-252.0*p10*p20*p312)+(-378.0*p202*p312)+(6160.0*p30*p312)+(-252.0*p10*p30*p312)+(-1260.0*p20*p30*p312)+(-2310.0*p302*p312)+(-1.0*p002*((378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))))+(-6.0*p11*((-27.0*p202*p21)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(-3.0*p10*((p21*(33.0+(6.0*p30)))+(22.0*p31)+(-6.0*p20*p31)))+(p30*((165.0*p21)+(-42.0*p21*p30)+(110.0*p31)+(-42.0*p30*p31)))+(p20*((99.0*p21)+(-63.0*p21*p30)+(-28.0*p30*p31)))))+(-2.0*p00*((9.0*p112*(-55.0+(14.0*p10)+(-1.0*p30)))+(9.0*p212*(-22.0+(9.0*p10)+(5.0*p20)+p30))+(3.0*p21*(-55.0+(20.0*p10)+(18.0*p20)+(7.0*p30))*p31)+((-55.0+(15.0*p10)+(21.0*p20)+(14.0*p30))*p312)+(3.0*p11*((3.0*(-55.0+(21.0*p10)+(6.0*p20))*p21)+(2.0*(-22.0+(9.0*p10)+(5.0*p20)+p30)*p31))))))*sign);
propagate[x][y] += ((1.0/4.0)*p312*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-594.0*p11*p202)+(135.0*p112*p202)+(162.0*p11*p202*p21)+(-990.0*p11*p20*p30)+(162.0*p112*p20*p30)+(-990.0*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(252.0*p20*p212*p30)+(-660.0*p11*p302)+(63.0*p112*p302)+(-2310.0*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(990.0*p202*p31)+(-252.0*p202*p21*p31)+(2310.0*p20*p30*p31)+(168.0*p11*p20*p30*p31)+(3080.0*p302*p31)+(252.0*p11*p302*p31)+(1260.0*p21*p302*p31)+(-378.0*p202*p312)+(-1260.0*p20*p30*p312)+(-2310.0*p302*p312)+(-1.0*p002*((378.0*p112)+(63.0*p212)+(30.0*p21*(-22.0+p31))+(5.0*(-22.0+p31)*p31)+(42.0*p11*(-55.0+(6.0*p21)+p31))))+(-9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(6.0*p10*((3.0*p112*((9.0*p20)+(5.0*p30)))+(-2.0*p30*p31*(-55.0+(14.0*p21)+(21.0*p31)))+(-3.0*p11*((22.0*p30)+(-6.0*p21*p30)+(p20*(33.0+(6.0*p31)))))+(-3.0*p20*((9.0*p212)+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31)))))))+(-2.0*p00*((3.0*p10*((42.0*p112)+(27.0*p212)+(5.0*(-11.0+p31)*p31)+(5.0*p21*(-33.0+(4.0*p31)))+(3.0*p11*(-55.0+(21.0*p21)+(6.0*p31)))))+(3.0*p20*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(p30*((-9.0*p112)+(9.0*p212)+(p11*(33.0+(6.0*p31)))+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((2772.0*p10*p21)+(-594.0*p102*p21)+(-594.0*p10*p20*p21)+(-594.0*p10*p212)+(135.0*p102*p212)+(162.0*p10*p20*p212)+(-5544.0*p21*p30)+(990.0*p20*p21*p30)+(990.0*p212*p30)+(-252.0*p20*p212*p30)+(2310.0*p21*p302)+(-378.0*p212*p302)+(-9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(2772.0*p10*p31)+(-396.0*p102*p31)+(5544.0*p20*p31)+(-990.0*p10*p20*p31)+(-990.0*p202*p31)+(-990.0*p10*p21*p31)+(162.0*p102*p21*p31)+(-990.0*p20*p21*p31)+(378.0*p10*p20*p21*p31)+(252.0*p202*p21*p31)+(9240.0*p30*p31)+(-660.0*p10*p30*p31)+(-2310.0*p20*p30*p31)+(2310.0*p21*p30*p31)+(168.0*p10*p21*p30*p31)+(-3080.0*p302*p31)+(-1260.0*p21*p302*p31)+(-660.0*p10*p312)+(63.0*p102*p312)+(-2310.0*p20*p312)+(252.0*p10*p20*p312)+(378.0*p202*p312)+(-6160.0*p30*p312)+(252.0*p10*p30*p312)+(1260.0*p20*p30*p312)+(2310.0*p302*p312)+(p002*((378.0*p112)+(63.0*p212)+(30.0*p21*(-22.0+p31))+(5.0*(-22.0+p31)*p31)+(42.0*p11*(-55.0+(6.0*p21)+p31))))+(-6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)+(22.0*p31)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30)+(-110.0*p31)+(42.0*p30*p31)))+(p20*(462.0+(-99.0*p21)+(-165.0*p30)+(63.0*p21*p30)+(28.0*p30*p31)+(-9.0*p10*(11.0+(2.0*p31)))))))+(2.0*p00*((9.0*p112*(-55.0+(14.0*p10)+(-1.0*p30)))+(9.0*p212*(-22.0+(9.0*p10)+(5.0*p20)+p30))+(3.0*p21*(462.0+(-11.0*p30)+(-55.0*p31)+(7.0*p30*p31)+(6.0*p20*(-11.0+(3.0*p31)))+(5.0*p10*(-33.0+(4.0*p31)))))+(3.0*p11*(924.0+(3.0*(-55.0+(6.0*p20))*p21)+(11.0*p30)+(-44.0*p31)+(10.0*p20*p31)+(2.0*p30*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31)))))+(p31*(462.0+(-55.0*p30)+(15.0*p10*(-11.0+p31))+(-55.0*p31)+(14.0*p30*p31)+(3.0*p20*(-44.0+(7.0*p31))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31*sign);
break;
case 5:
p302 = p30*p30;
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/6160.0)*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))*sign);
propagate[x][y] += ((1.0/4.0)*p312*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(-9.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)))+(-6.0*p10*((9.0*(-11.0+(3.0*p20))*p212)+(p21*(-165.0+(63.0*p20)+(28.0*p30))*p31)+(2.0*(-55.0+(21.0*p20)+(21.0*p30))*p312)))+(-2.0*((-9.0*p212*p30*(-55.0+(14.0*p20)+(21.0*p30)))+(3.0*p21*((-165.0*p20)+(42.0*p202)+(35.0*(11.0+(-6.0*p30))*p30))*p31)+(7.0*((27.0*p202)+(55.0*p30*(-8.0+(3.0*p30)))+(15.0*p20*(-11.0+(6.0*p30))))*p312)))+(-6.0*p11*((-27.0*p202*p21)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(-3.0*p10*((p21*(33.0+(6.0*p30)))+(22.0*p31)+(-6.0*p20*p31)))+(p30*((165.0*p21)+(-42.0*p21*p30)+(110.0*p31)+(-42.0*p30*p31)))+(p20*((99.0*p21)+(-63.0*p21*p30)+(-28.0*p30*p31))))))*sign);
propagate[x][y] += ((1.0/4.0)*p312*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-990.0*p20*p21*p30)+(252.0*p20*p212*p30)+(-2310.0*p21*p302)+(378.0*p212*p302)+(9.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(990.0*p202*p31)+(-252.0*p202*p21*p31)+(2310.0*p20*p30*p31)+(3080.0*p302*p31)+(1260.0*p21*p302*p31)+(-378.0*p202*p312)+(-1260.0*p20*p30*p312)+(-2310.0*p302*p312)+(-9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(2.0*p302*(-55.0+(21.0*p21)+(21.0*p31)))+(p20*p30*(-165.0+(63.0*p21)+(28.0*p31)))))+(6.0*p10*((3.0*p112*((9.0*p20)+(5.0*p30)))+(-2.0*p30*p31*(-55.0+(14.0*p21)+(21.0*p31)))+(-3.0*p11*((22.0*p30)+(-6.0*p21*p30)+(p20*(33.0+(6.0*p31)))))+(-3.0*p20*((9.0*p212)+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p31)))+(p31*(-44.0+(7.0*p31)))))+(-6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)+(22.0*p31)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30)+(-110.0*p31)+(42.0*p30*p31)))+(p20*(462.0+(-99.0*p21)+(-165.0*p30)+(63.0*p21*p30)+(28.0*p30*p31)+(-9.0*p10*(11.0+(2.0*p31)))))))+(6.0*p10*((9.0*(-11.0+(3.0*p20))*p212)+(p31*(462.0+(-165.0*p20)+(-110.0*p30)+(-110.0*p31)+(42.0*p20*p31)+(42.0*p30*p31)))+(p21*(462.0+(-165.0*p31)+(28.0*p30*p31)+(9.0*p20*(-11.0+(7.0*p31)))))))+(2.0*((-9.0*p212*p30*(-55.0+(14.0*p20)+(21.0*p30)))+(-3.0*p21*((3.0*(55.0+(-14.0*p20))*p20*p31)+(35.0*p302*(-11.0+(6.0*p31)))+(-11.0*p30*(-84.0+(15.0*p20)+(35.0*p31)))))+(p31*((9.0*p202*(-55.0+(21.0*p31)))+(385.0*p30*(12.0+(-4.0*p30)+(-8.0*p31)+(3.0*p30*p31)))+(21.0*p20*(132.0+(-55.0*p30)+(-55.0*p31)+(30.0*p30*p31))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31*sign);
break;
case 6:
p002 = p00*p00;
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(-7.0+(42.0*p002)+(p00*(-2.0+(28.0*p10)))+(-18.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(9.0*(-42.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(1.0+(9.0*p10)+(5.0*p20))))*p212)+(6.0*((5.0*p002)+(27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(p00*(7.0+(20.0*p10)+(18.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(((5.0*p002)+(p00*(28.0+(30.0*p10)+(42.0*p20)))+(21.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))))*p312)+(6.0*p11*((3.0*(-14.0+(14.0*p002)+(-6.0*p10)+(21.0*p00*p10)+(9.0*p102)+(-21.0*p20)+(6.0*p00*p20)+(-9.0*p202))*p21)+((-42.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(1.0+(9.0*p10)+(5.0*p20))))*p31))))*sign);
propagate[x][y] += ((1.0/4.0)*p312*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(-9.0*(68.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20)))+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p212)+(-6.0*(175.0+(5.0*p002)+(27.0*p102)+(-165.0*p20)+(42.0*p202)+(2.0*p00*(-24.0+(10.0*p10)+(9.0*p20)))+(p10*(-137.0+(63.0*p20))))*p21*p31)+(-1.0*((5.0*p002)+(63.0*p102)+(12.0*p10*(-34.0+(21.0*p20)))+(p00*(-82.0+(30.0*p10)+(42.0*p20)))+(14.0*(-275.0+(-75.0*p20)+(27.0*p202))))*p312)+(-6.0*p11*((3.0*(41.0+(14.0*p002)+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202)+(p00*(-55.0+(21.0*p10)+(6.0*p20))))*p21)+((68.0+(7.0*p002)+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p31))))*sign);
propagate[x][y] += ((1.0/4.0)*p312*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*(-7.0+(42.0*p002)+(p00*(-2.0+(28.0*p10)))+(-18.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(-9.0*(-42.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(1.0+(9.0*p10)+(5.0*p20))))*p212)+(-6.0*p21*(385.0+(165.0*p20)+(5.0*p002*(-22.0+p31))+(-210.0*p31)+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(p10*((-99.0*p20)+(28.0*p31)+(63.0*p20*p31)))+(p00*(-11.0+(-66.0*p20)+(7.0*p31)+(18.0*p20*p31)+(5.0*p10*(-33.0+(4.0*p31)))))))+(-6.0*p11*(110.0+(165.0*p20)+(99.0*p202)+(-42.0*p21)+(-63.0*p20*p21)+(-27.0*p202*p21)+(-42.0*p31)+(-28.0*p20*p31)+(7.0*p002*(-55.0+(6.0*p21)+p31))+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(33.0*p20)+(-6.0*p21)+(6.0*p20*p31)))+(p00*(11.0+(18.0*p20*p21)+(2.0*p31)+(10.0*p20*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31)))))))+(-1.0*p31*(-3080.0+(-2310.0*p20)+(-990.0*p202)+(5.0*p002*(-22.0+p31))+(2310.0*p31)+(1260.0*p20*p31)+(378.0*p202*p31)+(9.0*p102*(-44.0+(7.0*p31)))+(2.0*p00*(-55.0+(15.0*p10*(-11.0+p31))+(14.0*p31)+(3.0*p20*(-44.0+(7.0*p31)))))+(6.0*p10*(-110.0+(42.0*p31)+(3.0*p20*(-55.0+(14.0*p31))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(9.0*(68.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20)))+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p212)+(6.0*p21*(-539.0+(165.0*p20)+(5.0*p002*(-22.0+p31))+(175.0*p31)+(-165.0*p20*p31)+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(p10*(462.0+(-99.0*p20)+(-137.0*p31)+(63.0*p20*p31)))+(p00*(451.0+(-66.0*p20)+(-48.0*p31)+(18.0*p20*p31)+(5.0*p10*(-33.0+(4.0*p31)))))))+(6.0*p11*(-352.0+(-297.0*p20)+(99.0*p202)+(123.0*p21)+(36.0*p20*p21)+(-27.0*p202*p21)+(68.0*p31)+(-28.0*p20*p31)+(7.0*p002*(-55.0+(6.0*p21)+p31))+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31)))))+(p00*(935.0+(3.0*(-55.0+(6.0*p20))*p21)+(-42.0*p31)+(10.0*p20*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31)))))))+(p31*(6160.0+(3234.0*p20)+(-990.0*p202)+(5.0*p002*(-22.0+p31))+(-3850.0*p31)+(-1050.0*p20*p31)+(378.0*p202*p31)+(9.0*p102*(-44.0+(7.0*p31)))+(p00*(814.0+(30.0*p10*(-11.0+p31))+(-82.0*p31)+(6.0*p20*(-44.0+(7.0*p31)))))+(6.0*p10*(352.0+(-68.0*p31)+(3.0*p20*(-55.0+(14.0*p31))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31*sign);
break;
case 7:
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/6160.0)*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31))))*sign);
propagate[x][y] += ((1.0/4.0)*p312*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-9.0*(68.0+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(-6.0*(175.0+(-137.0*p10)+(27.0*p102)+(-165.0*p20)+(63.0*p10*p20)+(42.0*p202))*p21*p31)+((3850.0+(408.0*p10)+(-63.0*p102)+(1050.0*p20)+(-252.0*p10*p20)+(-378.0*p202))*p312)+(-6.0*p11*((3.0*(41.0+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202))*p21)+((68.0+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31))))*sign);
propagate[x][y] += ((1.0/4.0)*p312*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-9.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(-6.0*p11*(110.0+(p202*(99.0+(-27.0*p21)))+(-42.0*p21)+(p20*(165.0+(-63.0*p21)+(-28.0*p31)))+(-42.0*p31)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-6.0*p21)+(p20*(33.0+(6.0*p31)))))))+(-6.0*p21*(385.0+(165.0*p20)+(-210.0*p31)+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(p10*((28.0*p31)+(9.0*p20*(-11.0+(7.0*p31)))))))+(-1.0*p31*((770.0*(-4.0+(3.0*p31)))+(210.0*p20*(-11.0+(6.0*p31)))+(9.0*p102*(-44.0+(7.0*p31)))+(18.0*p202*(-55.0+(21.0*p31)))+(6.0*p10*(-110.0+(42.0*p31)+(3.0*p20*(-55.0+(14.0*p31))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(9.0*(68.0+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20))))*p212)+(6.0*p11*(-352.0+(p202*(99.0+(-27.0*p21)))+(123.0*p21)+(p20*(-297.0+(36.0*p21)+(-28.0*p31)))+(68.0*p31)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31)))))))+(6.0*p21*((-165.0*p20*(-1.0+p31))+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(7.0*(-77.0+(25.0*p31)))+(p10*(462.0+(-137.0*p31)+(9.0*p20*(-11.0+(7.0*p31)))))))+(p31*((-770.0*(-8.0+(5.0*p31)))+(9.0*p102*(-44.0+(7.0*p31)))+(18.0*p202*(-55.0+(21.0*p31)))+(-42.0*p20*(-77.0+(25.0*p31)))+(6.0*p10*(352.0+(-68.0*p31)+(3.0*p20*(-55.0+(14.0*p31))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31*sign);
break;
case 8:
p002 = p00*p00;
p302 = p30*p30;
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-63.0*p202)+(-162.0*p11*p202)+(-135.0*p112*p202)+(-90.0*p202*p21)+(-162.0*p11*p202*p21)+(-30.0*p20*p30)+(-120.0*p11*p20*p30)+(-162.0*p112*p20*p30)+(-108.0*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(-252.0*p20*p212*p30)+(-5.0*p302)+(-30.0*p11*p302)+(-63.0*p112*p302)+(-42.0*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(-18.0*p202*p31)+(252.0*p202*p21*p31)+(-42.0*p20*p30*p31)+(-168.0*p11*p20*p30*p31)+(-28.0*p302*p31)+(-252.0*p11*p302*p31)+(-1260.0*p21*p302*p31)+(378.0*p202*p312)+(1260.0*p20*p30*p312)+(2310.0*p302*p312)+(p002*(-2310.0+(378.0*p112)+(63.0*p212)+(28.0*p31)+(5.0*p312)+(42.0*p11*(30.0+(6.0*p21)+p31))+(6.0*p21*(42.0+(5.0*p31)))))+(9.0*p102*(-42.0+(15.0*p212)+(2.0*p31)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(-6.0*p10*((p30*(7.0+(15.0*p112)+(10.0*p21)+(18.0*p11*(1.0+p21))+(2.0*p31)+(-28.0*p21*p31)+(-42.0*p312)))+(3.0*p20*(14.0+(9.0*p112)+(6.0*p21)+(-9.0*p212)+(p11*(21.0+(-6.0*p31)))+(-21.0*p21*p31)+(-14.0*p312)))))+(2.0*p00*((3.0*p10*(-210.0+(42.0*p112)+(27.0*p212)+(7.0*p31)+(5.0*p312)+(9.0*p11*((7.0*p21)+(2.0*p31)))+(4.0*p21*(7.0+(5.0*p31)))))+(3.0*p20*(-42.0+(15.0*p212)+(2.0*p31)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(p30*((-9.0*p112)+(9.0*p212)+(3.0*p11*(-7.0+(2.0*p31)))+(3.0*p21*(-2.0+(7.0*p31)))+(14.0*(-1.0+p312)))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p312)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-660.0*p20)+(-990.0*p11*p20)+(-594.0*p112*p20)+(63.0*p202)+(162.0*p11*p202)+(135.0*p112*p202)+(-396.0*p20*p21)+(-594.0*p11*p20*p21)+(90.0*p202*p21)+(162.0*p11*p202*p21)+(-110.0*p30)+(-330.0*p11*p30)+(-396.0*p112*p30)+(30.0*p20*p30)+(120.0*p11*p20*p30)+(162.0*p112*p20*p30)+(-264.0*p21*p30)+(-990.0*p11*p21*p30)+(108.0*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(-990.0*p212*p30)+(252.0*p20*p212*p30)+(5.0*p302)+(30.0*p11*p302)+(63.0*p112*p302)+(42.0*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(-66.0*p20*p31)+(18.0*p202*p31)+(990.0*p20*p21*p31)+(-252.0*p202*p21*p31)+(-110.0*p30*p31)+(-660.0*p11*p30*p31)+(42.0*p20*p30*p31)+(168.0*p11*p20*p30*p31)+(-2310.0*p21*p30*p31)+(28.0*p302*p31)+(252.0*p11*p302*p31)+(1260.0*p21*p302*p31)+(2310.0*p20*p312)+(-378.0*p202*p312)+(6160.0*p30*p312)+(-1260.0*p20*p30*p312)+(-2310.0*p302*p312)+(-1.0*p002*(-2310.0+(378.0*p112)+(63.0*p212)+(28.0*p31)+(5.0*p312)+(42.0*p11*(30.0+(6.0*p21)+p31))+(6.0*p21*(42.0+(5.0*p31)))))+(-9.0*p102*(-42.0+(15.0*p212)+(2.0*p31)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(6.0*p10*(-385.0+(99.0*p212)+(7.0*p30)+(10.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(11.0*p31)+(165.0*p21*p31)+(2.0*p30*p31)+(-28.0*p21*p30*p31)+(110.0*p312)+(-42.0*p30*p312)+(3.0*p11*(-55.0+(33.0*p21)+(6.0*p30)+(6.0*p21*p30)+(p20*(21.0+(-6.0*p31)))+(22.0*p31)))+(-3.0*p20*(-14.0+(-6.0*p21)+(9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(-2.0*p00*(3080.0+(-126.0*p20)+(-330.0*p21)+(-198.0*p212)+(45.0*p20*p212)+(-14.0*p30)+(-6.0*p21*p30)+(9.0*p212*p30)+(-9.0*p112*(55.0+p30))+(-55.0*p31)+(6.0*p20*p31)+(-165.0*p21*p31)+(54.0*p20*p21*p31)+(21.0*p21*p30*p31)+(-55.0*p312)+(21.0*p20*p312)+(14.0*p30*p312)+(3.0*p10*(-210.0+(42.0*p112)+(27.0*p212)+(7.0*p31)+(5.0*p312)+(9.0*p11*((7.0*p21)+(2.0*p31)))+(4.0*p21*(7.0+(5.0*p31)))))+(-3.0*p11*(385.0+(165.0*p21)+(7.0*p30)+(44.0*p31)+(-2.0*p30*p31)+(-2.0*p20*(-14.0+(9.0*p21)+(5.0*p31))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p312)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*(770.0+(378.0*p112)+(63.0*p212)+(-82.0*p31)+(5.0*p312)+(42.0*p11*(-25.0+(6.0*p21)+p31))+(6.0*p21*(-68.0+(5.0*p31)))))+(-2.0*p00*((p30*(41.0+(-9.0*p112)+(-39.0*p21)+(9.0*p212)+(-55.0*p31)+(21.0*p21*p31)+(14.0*p312)+(6.0*p11*(2.0+p31))))+(3.0*p20*(68.0+(-66.0*p21)+(15.0*p212)+(-42.0*p31)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(3.0*p10*(175.0+(42.0*p112)+(-137.0*p21)+(27.0*p212)+(-48.0*p31)+(20.0*p21*p31)+(5.0*p312)+(3.0*p11*(-55.0+(21.0*p21)+(6.0*p31)))))))+(-3.0*((-3.0*p202*(-37.0+(15.0*p112)+(6.0*p11*(-8.0+(3.0*p21)))+(p21*(10.0+(-28.0*p31)))+(112.0*p31)+(-42.0*p312)))+(-7.0*p302*(-5.0+(3.0*p112)+(18.0*p212)+(148.0*p31)+(-110.0*p312)+(6.0*p11*(-5.0+(2.0*p21)+(2.0*p31)))+(12.0*p21*(-9.0+(5.0*p31)))))+(3.0*p102*(68.0+(15.0*p212)+(-42.0*p31)+(7.0*p312)+(6.0*p21*(-11.0+(3.0*p31)))+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(-2.0*p20*p30*(-50.0+(27.0*p112)+(-147.0*p21)+(42.0*p212)+(392.0*p31)+(-210.0*p312)+(p11*(-145.0+(63.0*p21)+(28.0*p31)))))+(p10*((2.0*p30*(37.0+(-15.0*p112)+(-10.0*p21)+(-6.0*p11*(-8.0+(3.0*p21)))+(-112.0*p31)+(28.0*p21*p31)+(42.0*p312)))+(-6.0*p20*(-41.0+(9.0*p112)+(39.0*p21)+(-9.0*p212)+(55.0*p31)+(-21.0*p21*p31)+(-14.0*p312)+(-6.0*p11*(2.0+p31)))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p31)*(-1.0+p31))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-2112.0*p20)+(-1782.0*p11*p20)+(594.0*p112*p20)+(333.0*p202)+(432.0*p11*p202)+(-135.0*p112*p202)+(396.0*p20*p21)+(594.0*p11*p20*p21)+(-90.0*p202*p21)+(-162.0*p11*p202*p21)+(-814.0*p30)+(-2442.0*p11*p30)+(396.0*p112*p30)+(300.0*p20*p30)+(870.0*p11*p20*p30)+(-162.0*p112*p20*p30)+(-5280.0*p21*p30)+(990.0*p11*p21*p30)+(882.0*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(990.0*p212*p30)+(-252.0*p20*p212*p30)+(105.0*p302)+(630.0*p11*p302)+(-63.0*p112*p302)+(2268.0*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(5610.0*p20*p31)+(-1008.0*p202*p31)+(-990.0*p20*p21*p31)+(252.0*p202*p21*p31)+(9350.0*p30*p31)+(660.0*p11*p30*p31)+(-2352.0*p20*p30*p31)+(-168.0*p11*p20*p30*p31)+(2310.0*p21*p30*p31)+(-3108.0*p302*p31)+(-252.0*p11*p302*p31)+(-1260.0*p21*p302*p31)+(-2310.0*p20*p312)+(378.0*p202*p312)+(-6160.0*p30*p312)+(1260.0*p20*p30*p312)+(2310.0*p302*p312)+(p002*(770.0+(378.0*p112)+(63.0*p212)+(-82.0*p31)+(5.0*p312)+(42.0*p11*(-25.0+(6.0*p21)+p31))+(6.0*p21*(-68.0+(5.0*p31)))))+(9.0*p102*(68.0+(15.0*p212)+(-42.0*p31)+(7.0*p312)+(6.0*p21*(-11.0+(3.0*p31)))+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(-6.0*p10*(539.0+(-462.0*p21)+(99.0*p212)+(-37.0*p30)+(10.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(-451.0*p31)+(165.0*p21*p31)+(112.0*p30*p31)+(-28.0*p21*p30*p31)+(110.0*p312)+(-42.0*p30*p312)+(-3.0*p20*(41.0+(-39.0*p21)+(9.0*p212)+(-55.0*p31)+(21.0*p21*p31)+(14.0*p312)))+(-3.0*p11*(55.0+(-33.0*p21)+(16.0*p30)+(-6.0*p21*p30)+(-22.0*p31)+(6.0*p20*(2.0+p31))))))+(2.0*p00*(-1540.0+(204.0*p20)+(1056.0*p21)+(-198.0*p20*p21)+(-198.0*p212)+(45.0*p20*p212)+(41.0*p30)+(-39.0*p21*p30)+(9.0*p212*p30)+(-9.0*p112*(55.0+p30))+(407.0*p31)+(-126.0*p20*p31)+(-165.0*p21*p31)+(54.0*p20*p21*p31)+(-55.0*p30*p31)+(21.0*p21*p30*p31)+(-55.0*p312)+(21.0*p20*p312)+(14.0*p30*p312)+(3.0*p11*(539.0+(-165.0*p21)+(4.0*p30)+(-44.0*p31)+(2.0*p30*p31)+(2.0*p20*(-14.0+(9.0*p21)+(5.0*p31)))))+(3.0*p10*(175.0+(42.0*p112)+(-137.0*p21)+(27.0*p212)+(-48.0*p31)+(20.0*p21*p31)+(5.0*p312)+(3.0*p11*(-55.0+(21.0*p21)+(6.0*p31))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p31)*(-1.0+p31))*sign);
break;
case 9:
p302 = p30*p30;
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(7.0+(15.0*p112)+(18.0*p11*(1.0+p21))+(p21*(10.0+(-28.0*p31)))+(2.0*p31)+(-42.0*p312)))+(9.0*p102*(-42.0+(15.0*p212)+(2.0*p31)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(-6.0*p20*p30*(5.0+(27.0*p112)+(18.0*p21)+(42.0*p212)+(7.0*p31)+(-210.0*p312)+(p11*(20.0+(63.0*p21)+(28.0*p31)))))+(-1.0*p302*(5.0+(63.0*p112)+(378.0*p212)+(28.0*p31)+(-2310.0*p312)+(42.0*p21*(1.0+(30.0*p31)))+(6.0*p11*(5.0+(42.0*p21)+(42.0*p31)))))+(-6.0*p10*((p30*(7.0+(15.0*p112)+(10.0*p21)+(18.0*p11*(1.0+p21))+(2.0*p31)+(-28.0*p21*p31)+(-42.0*p312)))+(3.0*p20*(14.0+(9.0*p112)+(6.0*p21)+(-9.0*p212)+(p11*(21.0+(-6.0*p31)))+(-21.0*p21*p31)+(-14.0*p312))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p312)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(7.0+(15.0*p112)+(18.0*p11*(1.0+p21))+(p21*(10.0+(-28.0*p31)))+(2.0*p31)+(-42.0*p312)))+(-9.0*p102*(-42.0+(15.0*p212)+(2.0*p31)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(p30*(-110.0+(5.0*p30)+(9.0*p112*(-44.0+(7.0*p30)))+(18.0*p212*(-55.0+(21.0*p30)))+(-110.0*p31)+(28.0*p30*p31)+(6160.0*p312)+(-2310.0*p30*p312)+(6.0*p11*(-55.0+(5.0*p30)+(3.0*p21*(-55.0+(14.0*p30)))+(-110.0*p31)+(42.0*p30*p31)))+(6.0*p21*(-44.0+(7.0*p30)+(-385.0*p31)+(210.0*p30*p31)))))+(6.0*p10*(-385.0+(99.0*p212)+(7.0*p30)+(10.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(11.0*p31)+(165.0*p21*p31)+(2.0*p30*p31)+(-28.0*p21*p30*p31)+(110.0*p312)+(-42.0*p30*p312)+(3.0*p11*(-55.0+(33.0*p21)+(6.0*p30)+(6.0*p21*p30)+(p20*(21.0+(-6.0*p31)))+(22.0*p31)))+(-3.0*p20*(-14.0+(-6.0*p21)+(9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(6.0*p20*(-110.0+(5.0*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(-11.0*p31)+(7.0*p30*p31)+(385.0*p312)+(-210.0*p30*p312)+(3.0*p21*(-22.0+(6.0*p30)+(55.0*p31)))+(p11*(-165.0+(9.0*p21*(-11.0+(7.0*p30)))+(4.0*p30*(5.0+(7.0*p31))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p312)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/6160.0)*((3.0*p202*(-37.0+(15.0*p112)+(6.0*p11*(-8.0+(3.0*p21)))+(p21*(10.0+(-28.0*p31)))+(112.0*p31)+(-42.0*p312)))+(7.0*p302*(-5.0+(3.0*p112)+(18.0*p212)+(148.0*p31)+(-110.0*p312)+(6.0*p11*(-5.0+(2.0*p21)+(2.0*p31)))+(12.0*p21*(-9.0+(5.0*p31)))))+(-3.0*p102*(68.0+(15.0*p212)+(-42.0*p31)+(7.0*p312)+(6.0*p21*(-11.0+(3.0*p31)))+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(2.0*p20*p30*(-50.0+(27.0*p112)+(-147.0*p21)+(42.0*p212)+(392.0*p31)+(-210.0*p312)+(p11*(-145.0+(63.0*p21)+(28.0*p31)))))+(2.0*p10*((p30*(-37.0+(15.0*p112)+(10.0*p21)+(6.0*p11*(-8.0+(3.0*p21)))+(112.0*p31)+(-28.0*p21*p31)+(-42.0*p312)))+(3.0*p20*(-41.0+(9.0*p112)+(39.0*p21)+(-9.0*p212)+(55.0*p31)+(-21.0*p21*p31)+(-14.0*p312)+(-6.0*p11*(2.0+p31)))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p31)*(-1.0+p31))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(-37.0+(15.0*p112)+(6.0*p11*(-8.0+(3.0*p21)))+(p21*(10.0+(-28.0*p31)))+(112.0*p31)+(-42.0*p312)))+(9.0*p102*(68.0+(15.0*p212)+(-42.0*p31)+(7.0*p312)+(6.0*p21*(-11.0+(3.0*p31)))+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(-6.0*p20*(352.0+(-50.0*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(-3.0*p21*(22.0+(49.0*p30)+(-55.0*p31)))+(-935.0*p31)+(392.0*p30*p31)+(385.0*p312)+(-210.0*p30*p312)+(p11*(297.0+(-145.0*p30)+(9.0*p21*(-11.0+(7.0*p30)))+(28.0*p30*p31)))))+(-6.0*p10*(539.0+(-462.0*p21)+(99.0*p212)+(-37.0*p30)+(10.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(-451.0*p31)+(165.0*p21*p31)+(112.0*p30*p31)+(-28.0*p21*p30*p31)+(110.0*p312)+(-42.0*p30*p312)+(-3.0*p20*(41.0+(-39.0*p21)+(9.0*p212)+(-55.0*p31)+(21.0*p21*p31)+(14.0*p312)))+(-3.0*p11*(55.0+(-33.0*p21)+(16.0*p30)+(-6.0*p21*p30)+(-22.0*p31)+(6.0*p20*(2.0+p31))))))+(p30*(-814.0+(p212*(990.0+(-378.0*p30)))+(105.0*p30)+(-9.0*p112*(-44.0+(7.0*p30)))+(9350.0*p31)+(-3108.0*p30*p31)+(-6160.0*p312)+(2310.0*p30*p312)+(-6.0*p11*(407.0+(3.0*p21*(-55.0+(14.0*p30)))+(-110.0*p31)+(21.0*p30*(-5.0+(2.0*p31)))))+(-6.0*p21*(880.0+(-385.0*p31)+(42.0*p30*(-9.0+(5.0*p31))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p31)*(-1.0+p31))*sign);
break;
case 10:
p002 = p00*p00;
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*(-5.0+(-30.0*p11)+(-63.0*p112)+(-30.0*p20)+(-120.0*p11*p20)+(-162.0*p112*p20)+(-63.0*p202)+(-162.0*p11*p202)+(-135.0*p112*p202)+(-42.0*p21)+(-252.0*p11*p21)+(-108.0*p20*p21)+(-378.0*p11*p20*p21)+(-90.0*p202*p21)+(-162.0*p11*p202*p21)+(-378.0*p212)+(-252.0*p20*p212)+(-28.0*p31)+(-252.0*p11*p31)+(-42.0*p20*p31)+(-168.0*p11*p20*p31)+(-18.0*p202*p31)+(-1260.0*p21*p31)+(252.0*p202*p21*p31)+(2310.0*p312)+(1260.0*p20*p312)+(378.0*p202*p312)+(p002*(-2310.0+(378.0*p112)+(63.0*p212)+(28.0*p31)+(5.0*p312)+(42.0*p11*(30.0+(6.0*p21)+p31))+(6.0*p21*(42.0+(5.0*p31)))))+(9.0*p102*(-42.0+(15.0*p212)+(2.0*p31)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(-6.0*p10*(7.0+(3.0*p112*(5.0+(9.0*p20)))+(10.0*p21)+(2.0*p31)+(-28.0*p21*p31)+(-42.0*p312)+(9.0*p11*(2.0+(7.0*p20)+(2.0*p21)+(-2.0*p20*p31)))+(-3.0*p20*(-14.0+(-6.0*p21)+(9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(2.0*p00*(-14.0+(-9.0*p112)+(-126.0*p20)+(-6.0*p21)+(9.0*p212)+(45.0*p20*p212)+(6.0*p20*p31)+(21.0*p21*p31)+(54.0*p20*p21*p31)+(14.0*p312)+(21.0*p20*p312)+(3.0*p10*(-210.0+(42.0*p112)+(27.0*p212)+(7.0*p31)+(5.0*p312)+(9.0*p11*((7.0*p21)+(2.0*p31)))+(4.0*p21*(7.0+(5.0*p31)))))+(3.0*p11*(-7.0+(2.0*p31)+(2.0*p20*(-14.0+(9.0*p21)+(5.0*p31))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p312)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*(-105.0+(-300.0*p11)+(-333.0*p112)+(-630.0*p20)+(-870.0*p11*p20)+(-432.0*p112*p20)+(63.0*p202)+(162.0*p11*p202)+(135.0*p112*p202)+(-222.0*p21)+(-738.0*p11*p21)+(-288.0*p20*p21)+(-216.0*p11*p20*p21)+(90.0*p202*p21)+(162.0*p11*p202*p21)+(-612.0*p212)+(252.0*p20*p212)+(-82.0*p31)+(-408.0*p11*p31)+(-24.0*p20*p31)+(168.0*p11*p20*p31)+(18.0*p202*p31)+(-1050.0*p21*p31)+(990.0*p20*p21*p31)+(-252.0*p202*p21*p31)+(3850.0*p312)+(1050.0*p20*p312)+(-378.0*p202*p312)+(-1.0*p002*(-2310.0+(378.0*p112)+(63.0*p212)+(28.0*p31)+(5.0*p312)+(42.0*p11*(30.0+(6.0*p21)+p31))+(6.0*p21*(42.0+(5.0*p31)))))+(-9.0*p102*(-42.0+(15.0*p212)+(2.0*p31)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(6.0*p10*(-378.0+(3.0*p112*(5.0+(9.0*p20)))+(10.0*p21)+(99.0*p212)+(13.0*p31)+(137.0*p21*p31)+(68.0*p312)+(3.0*p11*(-49.0+(39.0*p21)+(p20*(21.0+(-6.0*p31)))+(22.0*p31)))+(-3.0*p20*(-14.0+(-6.0*p21)+(9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(-2.0*p00*(3066.0+(-504.0*p112)+(-126.0*p20)+(-336.0*p21)+(-189.0*p212)+(45.0*p20*p212)+(-55.0*p31)+(6.0*p20*p31)+(-144.0*p21*p31)+(54.0*p20*p21*p31)+(-41.0*p312)+(21.0*p20*p312)+(3.0*p10*(-210.0+(42.0*p112)+(27.0*p212)+(7.0*p31)+(5.0*p312)+(9.0*p11*((7.0*p21)+(2.0*p31)))+(4.0*p21*(7.0+(5.0*p31)))))+(-3.0*p11*(392.0+(165.0*p21)+(42.0*p31)+(-2.0*p20*(-14.0+(9.0*p21)+(5.0*p31))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p312)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*(770.0+(378.0*p112)+(63.0*p212)+(-82.0*p31)+(5.0*p312)+(42.0*p11*(-25.0+(6.0*p21)+p31))+(6.0*p21*(-68.0+(5.0*p31)))))+(-2.0*p00*(41.0+(-9.0*p112)+(204.0*p20)+(-39.0*p21)+(-198.0*p20*p21)+(9.0*p212)+(45.0*p20*p212)+(-55.0*p31)+(-126.0*p20*p31)+(21.0*p21*p31)+(54.0*p20*p21*p31)+(14.0*p312)+(21.0*p20*p312)+(6.0*p11*(2.0+p31+(p20*(-14.0+(9.0*p21)+(5.0*p31)))))+(3.0*p10*(175.0+(42.0*p112)+(-137.0*p21)+(27.0*p212)+(-48.0*p31)+(20.0*p21*p31)+(5.0*p312)+(3.0*p11*(-55.0+(21.0*p21)+(6.0*p31)))))))+(-3.0*(35.0+(100.0*p20)+(111.0*p202)+(-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(756.0*p21)+(294.0*p20*p21)+(-30.0*p202*p21)+(-126.0*p212)+(-84.0*p20*p212)+(-1036.0*p31)+(-784.0*p20*p31)+(-336.0*p202*p31)+(-420.0*p21*p31)+(84.0*p202*p21*p31)+(770.0*p312)+(420.0*p20*p312)+(126.0*p202*p312)+(3.0*p102*(68.0+(15.0*p212)+(-42.0*p31)+(7.0*p312)+(6.0*p21*(-11.0+(3.0*p31)))+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(-2.0*p11*((9.0*p202*(-8.0+(3.0*p21)))+(21.0*(-5.0+(2.0*p21)+(2.0*p31)))+(p20*(-145.0+(63.0*p21)+(28.0*p31)))))+(p10*(74.0+(-6.0*p112*(5.0+(9.0*p20)))+(-20.0*p21)+(-224.0*p31)+(56.0*p21*p31)+(84.0*p312)+(6.0*p20*(41.0+(-39.0*p21)+(9.0*p212)+(-55.0*p31)+(21.0*p21*p31)+(14.0*p312)))+(12.0*p11*(8.0+(-3.0*p21)+(3.0*p20*(2.0+p31)))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p31)*(-1.0+p31))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*(-709.0+(-1812.0*p11)+(333.0*p112)+(-1812.0*p20)+(-912.0*p11*p20)+(432.0*p112*p20)+(333.0*p202)+(432.0*p11*p202)+(-135.0*p112*p202)+(-3012.0*p21)+(738.0*p11*p21)+(1278.0*p20*p21)+(216.0*p11*p20*p21)+(-90.0*p202*p21)+(-162.0*p11*p202*p21)+(612.0*p212)+(-252.0*p20*p212)+(6242.0*p31)+(408.0*p11*p31)+(3258.0*p20*p31)+(-168.0*p11*p20*p31)+(-1008.0*p202*p31)+(1050.0*p21*p31)+(-990.0*p20*p21*p31)+(252.0*p202*p21*p31)+(-3850.0*p312)+(-1050.0*p20*p312)+(378.0*p202*p312)+(p002*(770.0+(378.0*p112)+(63.0*p212)+(-82.0*p31)+(5.0*p312)+(42.0*p11*(-25.0+(6.0*p21)+p31))+(6.0*p21*(-68.0+(5.0*p31)))))+(9.0*p102*(68.0+(15.0*p212)+(-42.0*p31)+(7.0*p312)+(6.0*p21*(-11.0+(3.0*p31)))+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(-6.0*p10*(502.0+(3.0*p112*(5.0+(9.0*p20)))+(-452.0*p21)+(99.0*p212)+(-339.0*p31)+(137.0*p21*p31)+(68.0*p312)+(-3.0*p20*(41.0+(-39.0*p21)+(9.0*p212)+(-55.0*p31)+(21.0*p21*p31)+(14.0*p312)))+(-3.0*p11*(71.0+(-39.0*p21)+(-22.0*p31)+(6.0*p20*(2.0+p31))))))+(2.0*p00*(-1499.0+(-504.0*p112)+(204.0*p20)+(1017.0*p21)+(-198.0*p20*p21)+(-189.0*p212)+(45.0*p20*p212)+(352.0*p31)+(-126.0*p20*p31)+(-144.0*p21*p31)+(54.0*p20*p21*p31)+(-41.0*p312)+(21.0*p20*p312)+(3.0*p11*(543.0+(-165.0*p21)+(-42.0*p31)+(2.0*p20*(-14.0+(9.0*p21)+(5.0*p31)))))+(3.0*p10*(175.0+(42.0*p112)+(-137.0*p21)+(27.0*p212)+(-48.0*p31)+(20.0*p21*p31)+(5.0*p312)+(3.0*p11*(-55.0+(21.0*p21)+(6.0*p31))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p31)*(-1.0+p31))*sign);
break;
case 11:
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*(-5.0+(-30.0*p20)+(-63.0*p202)+(-9.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-42.0*p21)+(-108.0*p20*p21)+(-90.0*p202*p21)+(-378.0*p212)+(-252.0*p20*p212)+(-28.0*p31)+(-42.0*p20*p31)+(-18.0*p202*p31)+(-1260.0*p21*p31)+(252.0*p202*p21*p31)+(2310.0*p312)+(1260.0*p20*p312)+(378.0*p202*p312)+(9.0*p102*(-42.0+(15.0*p212)+(2.0*p31)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(-6.0*p11*(5.0+(42.0*p21)+(27.0*p202*(1.0+p21))+(42.0*p31)+(p20*(20.0+(63.0*p21)+(28.0*p31)))))+(-6.0*p10*(7.0+(3.0*p112*(5.0+(9.0*p20)))+(10.0*p21)+(2.0*p31)+(-28.0*p21*p31)+(-42.0*p312)+(9.0*p11*(2.0+(7.0*p20)+(2.0*p21)+(-2.0*p20*p31)))+(-3.0*p20*(-14.0+(-6.0*p21)+(9.0*p212)+(21.0*p21*p31)+(14.0*p312))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p312)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*(-105.0+(-630.0*p20)+(63.0*p202)+(9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)))+(-222.0*p21)+(-288.0*p20*p21)+(90.0*p202*p21)+(-612.0*p212)+(252.0*p20*p212)+(-82.0*p31)+(-24.0*p20*p31)+(18.0*p202*p31)+(-1050.0*p21*p31)+(990.0*p20*p21*p31)+(-252.0*p202*p21*p31)+(3850.0*p312)+(1050.0*p20*p312)+(-378.0*p202*p312)+(-9.0*p102*(-42.0+(15.0*p212)+(2.0*p31)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(6.0*p11*(-50.0+(-123.0*p21)+(27.0*p202*(1.0+p21))+(-68.0*p31)+(p20*(-145.0+(-36.0*p21)+(28.0*p31)))))+(6.0*p10*(-378.0+(3.0*p112*(5.0+(9.0*p20)))+(10.0*p21)+(99.0*p212)+(13.0*p31)+(137.0*p21*p31)+(68.0*p312)+(3.0*p11*(-49.0+(39.0*p21)+(p20*(21.0+(-6.0*p31)))+(22.0*p31)))+(-3.0*p20*(-14.0+(-6.0*p21)+(9.0*p212)+(21.0*p21*p31)+(14.0*p312))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-1.0+p312)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/6160.0)*(-35.0+(-100.0*p20)+(-111.0*p202)+(3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-756.0*p21)+(-294.0*p20*p21)+(30.0*p202*p21)+(126.0*p212)+(84.0*p20*p212)+(1036.0*p31)+(784.0*p20*p31)+(336.0*p202*p31)+(420.0*p21*p31)+(-84.0*p202*p21*p31)+(-770.0*p312)+(-420.0*p20*p312)+(-126.0*p202*p312)+(-3.0*p102*(68.0+(15.0*p212)+(-42.0*p31)+(7.0*p312)+(6.0*p21*(-11.0+(3.0*p31)))+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(2.0*p11*((9.0*p202*(-8.0+(3.0*p21)))+(21.0*(-5.0+(2.0*p21)+(2.0*p31)))+(p20*(-145.0+(63.0*p21)+(28.0*p31)))))+(-1.0*p10*(74.0+(-6.0*p112*(5.0+(9.0*p20)))+(-20.0*p21)+(-224.0*p31)+(56.0*p21*p31)+(84.0*p312)+(6.0*p20*(41.0+(-39.0*p21)+(9.0*p212)+(-55.0*p31)+(21.0*p21*p31)+(14.0*p312)))+(12.0*p11*(8.0+(-3.0*p21)+(3.0*p20*(2.0+p31)))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p31)*(-1.0+p31))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*(-709.0+(-1812.0*p20)+(333.0*p202)+(-9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)))+(-3012.0*p21)+(1278.0*p20*p21)+(-90.0*p202*p21)+(612.0*p212)+(-252.0*p20*p212)+(-6.0*p11*(302.0+(-123.0*p21)+(9.0*p202*(-8.0+(3.0*p21)))+(-4.0*p20*(-38.0+(9.0*p21)+(-7.0*p31)))+(-68.0*p31)))+(6242.0*p31)+(3258.0*p20*p31)+(-1008.0*p202*p31)+(1050.0*p21*p31)+(-990.0*p20*p21*p31)+(252.0*p202*p21*p31)+(-3850.0*p312)+(-1050.0*p20*p312)+(378.0*p202*p312)+(9.0*p102*(68.0+(15.0*p212)+(-42.0*p31)+(7.0*p312)+(6.0*p21*(-11.0+(3.0*p31)))+(2.0*p11*(-14.0+(9.0*p21)+(5.0*p31)))))+(-6.0*p10*(502.0+(3.0*p112*(5.0+(9.0*p20)))+(-452.0*p21)+(99.0*p212)+(-339.0*p31)+(137.0*p21*p31)+(68.0*p312)+(-3.0*p20*(41.0+(-39.0*p21)+(9.0*p212)+(-55.0*p31)+(21.0*p21*p31)+(14.0*p312)))+(-3.0*p11*(71.0+(-39.0*p21)+(-22.0*p31)+(6.0*p20*(2.0+p31)))))))*sign);
propagate[x][y] += ((-1.0/4.0)*((-1.0+p31)*(-1.0+p31))*sign);
break;
case 12:
p002 = p00*p00;
p012 = p01*p01;
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-21.0*p002*((110.0*p012)+(-12.0*p01*((5.0*p11)+p21))+(-3.0*((6.0*p112)+(4.0*p11*p21)+p212))))+(-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-2.0*p00*((14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(3.0*p01*((2.0*p21*((-14.0*p10)+p30))+(7.0*p11*((4.0*p20)+p30))))+(-9.0*((6.0*p11*p20*p21)+(p10*((14.0*p112)+(21.0*p11*p21)+(9.0*p212)))+(-1.0*p112*p30)+(p212*((5.0*p20)+p30))))))+(9.0*((3.0*p102*p21*((6.0*p11)+(5.0*p21)))+(-14.0*p212*p30*((2.0*p20)+(3.0*p30)))+(-1.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*p21*((9.0*p202)+(21.0*p20*p30)+(14.0*p302)))+(-2.0*p10*((-9.0*p20*p212)+(6.0*p11*p21*p30)+(p112*((9.0*p20)+(5.0*p30)))))))+(-6.0*p01*((42.0*p102*p11)+(p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(p21*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*p21*((9.0*p20)+(5.0*p30))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((21.0*p002*((110.0*p012)+(-12.0*p01*((5.0*p11)+p21))+(-3.0*((6.0*p112)+(4.0*p11*p21)+p212))))+(p012*((378.0*p102)+(63.0*p202)+(30.0*p20*(-22.0+p30))+(5.0*(-22.0+p30)*p30)+(42.0*p10*(-55.0+(6.0*p20)+p30))))+(2.0*p00*((14.0*p012*(-220.0+(45.0*p10)+(9.0*p20)+p30))+(3.0*p01*((2.0*p21*(55.0+(-14.0*p10)+p30))+(7.0*p11*(55.0+(4.0*p20)+p30))))+(-9.0*((p11*(-55.0+(21.0*p10)+(6.0*p20))*p21)+(p112*(-55.0+(14.0*p10)+(-1.0*p30)))+(p212*(-22.0+(9.0*p10)+(5.0*p20)+p30))))))+(6.0*p01*((42.0*p102*p11)+(p11*((27.0*p202)+(5.0*(-11.0+p30)*p30)+(5.0*p20*(-33.0+(4.0*p30)))))+(p10*((2.0*p21*((9.0*p20)+(5.0*p30)))+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))+(p21*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p30)))+(p30*(-44.0+(7.0*p30)))))))+(9.0*((-2.0*p11*p21*((9.0*p102)+(-9.0*p202)+(p20*(33.0+(-21.0*p30)))+((55.0+(-14.0*p30))*p30)+(-3.0*p10*(11.0+(2.0*p30)))))+(p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(p212*((-15.0*p102)+(p10*(66.0+(-18.0*p20)))+(2.0*p30*(-55.0+(14.0*p20)+(21.0*p30))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((p002*((2310.0*p012)+(-28.0*p01*(110.0+(45.0*p11)+(9.0*p21)))+(-3.0*((126.0*p112)+(14.0*p11*(-55.0+(6.0*p21)))+(p21*(-220.0+(21.0*p21)))))))+(p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(2.0*p01*((9.0*p102*(-55.0+(14.0*p11)))+(9.0*p202*(-22.0+(9.0*p11)+(5.0*p21)))+(3.0*p20*(-55.0+(20.0*p11)+(18.0*p21))*p30)+((-55.0+(15.0*p11)+(21.0*p21))*p302)+(3.0*p10*((3.0*p20*(-55.0+(21.0*p11)+(6.0*p21)))+(2.0*(-22.0+(9.0*p11)+(5.0*p21))*p30)))))+(2.0*p00*((14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(p01*((6.0*(-55.0+(14.0*p11))*p20)+(-21.0*p10*(55.0+(4.0*p21)))+((-55.0+(21.0*p11)+(6.0*p21))*p30)))+(-3.0*((3.0*p20*p21*(-22.0+(6.0*p11)+(5.0*p21)))+(3.0*p10*((14.0*p112)+(p21*(-55.0+(9.0*p21)))+(p11*(-55.0+(21.0*p21)))))+(((11.0*p11)+(-3.0*p112)+(p21*(-11.0+(3.0*p21))))*p30)))))+(-3.0*((9.0*p102*p21*(-22.0+(6.0*p11)+(5.0*p21)))+(-2.0*p21*p30*((3.0*p20*(-55.0+(14.0*p21)))+(7.0*(-55.0+(9.0*p21))*p30)))+(-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(3.0*p20*(-55.0+(21.0*p21))*p30)+(2.0*(-55.0+(21.0*p21))*p302)))+(-6.0*p10*((3.0*p20*(11.0+(-3.0*p21))*p21)+(p112*((9.0*p20)+(5.0*p30)))+(p11*((-33.0*p20)+(-22.0*p30)+(6.0*p21*p30))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*((2310.0*p012)+(-28.0*p01*(110.0+(45.0*p11)+(9.0*p21)))+(-3.0*((126.0*p112)+(14.0*p11*(-55.0+(6.0*p21)))+(p21*(-220.0+(21.0*p21)))))))+(-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*(-22.0+p30))+(5.0*(-22.0+p30)*p30)+(42.0*p10*(-55.0+(6.0*p20)+p30))))+(-2.0*p01*((9.0*p102*(-55.0+(14.0*p11)))+(9.0*p202*(-22.0+(9.0*p11)+(5.0*p21)))+(3.0*p20*(462.0+(-55.0*p30)+(6.0*p21*(-11.0+(3.0*p30)))+(5.0*p11*(-33.0+(4.0*p30)))))+(3.0*p10*(924.0+(3.0*p20*(-55.0+(6.0*p21)))+(-44.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))+(p30*(462.0+(15.0*p11*(-11.0+p30))+(-55.0*p30)+(3.0*p21*(-44.0+(7.0*p30)))))))+(-3.0*((3.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(2.0*p11*((-27.0*p102*p21)+(9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p20*(-154.0+(33.0*p10)+(33.0*p21)+(55.0*p30)+(-21.0*p21*p30)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30)))))+(p21*((-9.0*p102*(-22.0+(5.0*p21)))+(-6.0*p10*(154.0+(-33.0*p21)+(p20*(-33.0+(9.0*p21)))))+(2.0*p30*(924.0+(-165.0*p21)+(3.0*p20*(-55.0+(14.0*p21)))+(-385.0*p30)+(63.0*p21*p30)))))))+(-2.0*p00*((14.0*p012*(-220.0+(45.0*p10)+(9.0*p20)+p30))+(p01*(4620.0+(-330.0*p20)+(330.0*p21)+(-21.0*p10*(55.0+(4.0*p21)))+(-55.0*p30)+(6.0*p21*p30)+(21.0*p11*(55.0+(4.0*p20)+p30))))+(-3.0*((3.0*p112*(-55.0+(14.0*p10)+(-1.0*p30)))+(p21*(462.0+(-66.0*p21)+(3.0*p20*(-22.0+(5.0*p21)))+(3.0*p10*(-55.0+(9.0*p21)))+(-11.0*p30)+(3.0*p21*p30)))+(p11*((3.0*(-55.0+(6.0*p20))*p21)+(3.0*p10*(-55.0+(21.0*p21)))+(11.0*(84.0+p30)))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 13:
p012 = p01*p01;
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(9.0*((3.0*p102*p21*((6.0*p11)+(5.0*p21)))+(-14.0*p212*p30*((2.0*p20)+(3.0*p30)))+(-1.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*p21*((9.0*p202)+(21.0*p20*p30)+(14.0*p302)))+(-2.0*p10*((-9.0*p20*p212)+(6.0*p11*p21*p30)+(p112*((9.0*p20)+(5.0*p30)))))))+(-6.0*p01*((42.0*p102*p11)+(p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(p21*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*p21*((9.0*p20)+(5.0*p30))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*(-22.0+p30))+(5.0*(-22.0+p30)*p30)+(42.0*p10*(-55.0+(6.0*p20)+p30))))+(6.0*p01*((42.0*p102*p11)+(p11*((27.0*p202)+(5.0*(-11.0+p30)*p30)+(5.0*p20*(-33.0+(4.0*p30)))))+(p10*((2.0*p21*((9.0*p20)+(5.0*p30)))+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))+(p21*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p30)))+(p30*(-44.0+(7.0*p30)))))))+(-9.0*((2.0*p11*p21*((9.0*p102)+(-9.0*p202)+(p20*(33.0+(-21.0*p30)))+((55.0+(-14.0*p30))*p30)+(-3.0*p10*(11.0+(2.0*p30)))))+(-1.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(p212*((15.0*p102)+(6.0*p10*(-11.0+(3.0*p20)))+(-2.0*p30*(-55.0+(14.0*p20)+(21.0*p30))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(2.0*p01*((9.0*p102*(-55.0+(14.0*p11)))+(9.0*p202*(-22.0+(9.0*p11)+(5.0*p21)))+(3.0*p20*(-55.0+(20.0*p11)+(18.0*p21))*p30)+((-55.0+(15.0*p11)+(21.0*p21))*p302)+(3.0*p10*((3.0*p20*(-55.0+(21.0*p11)+(6.0*p21)))+(2.0*(-22.0+(9.0*p11)+(5.0*p21))*p30)))))+(-3.0*((9.0*p102*p21*(-22.0+(6.0*p11)+(5.0*p21)))+(-2.0*p21*p30*((3.0*p20*(-55.0+(14.0*p21)))+(7.0*(-55.0+(9.0*p21))*p30)))+(-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(3.0*p20*(-55.0+(21.0*p21))*p30)+(2.0*(-55.0+(21.0*p21))*p302)))+(-6.0*p10*((3.0*p20*(11.0+(-3.0*p21))*p21)+(p112*((9.0*p20)+(5.0*p30)))+(p11*((-33.0*p20)+(-22.0*p30)+(6.0*p21*p30))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*(-22.0+p30))+(5.0*(-22.0+p30)*p30)+(42.0*p10*(-55.0+(6.0*p20)+p30))))+(-2.0*p01*((9.0*p102*(-55.0+(14.0*p11)))+(9.0*p202*(-22.0+(9.0*p11)+(5.0*p21)))+(3.0*p20*(462.0+(-55.0*p30)+(6.0*p21*(-11.0+(3.0*p30)))+(5.0*p11*(-33.0+(4.0*p30)))))+(3.0*p10*(924.0+(3.0*p20*(-55.0+(6.0*p21)))+(-44.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))+(p30*(462.0+(15.0*p11*(-11.0+p30))+(-55.0*p30)+(3.0*p21*(-44.0+(7.0*p30)))))))+(3.0*((-3.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(p21*((9.0*p102*(-22.0+(5.0*p21)))+(6.0*p10*(154.0+(-33.0*p21)+(p20*(-33.0+(9.0*p21)))))+(2.0*p30*(-924.0+(p20*(165.0+(-42.0*p21)))+(165.0*p21)+(385.0*p30)+(-63.0*p21*p30)))))+(-2.0*p11*((-27.0*p102*p21)+(9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p20*(-154.0+(33.0*p10)+(33.0*p21)+(55.0*p30)+(-21.0*p21*p30)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 14:
p002 = p00*p00;
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(2310.0*p002)+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))+(28.0*p00*(1.0+(45.0*p10)+(9.0*p20)))))+(6.0*p01*((p11*(-5.0+(210.0*p002)+(-18.0*p10)+(-42.0*p102)+(-20.0*p20)+(-63.0*p10*p20)+(-27.0*p202)+(-7.0*p00*(1.0+(4.0*p20)))))+((-7.0+(42.0*p002)+(p00*(-2.0+(28.0*p10)))+(-18.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20))))*p21)))+(9.0*((p112*(-7.0+(42.0*p002)+(p00*(-2.0+(28.0*p10)))+(-18.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(2.0*p11*(-14.0+(14.0*p002)+(-6.0*p10)+(21.0*p00*p10)+(9.0*p102)+(-21.0*p20)+(6.0*p00*p20)+(-9.0*p202))*p21)+((-42.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(1.0+(9.0*p10)+(5.0*p20))))*p212))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/6160.0)*((7.0*p012*(-5.0+(110.0*p002)+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)+(4.0*p00*(-73.0+(15.0*p10)+(3.0*p20)))))+(p01*((2.0*p11*(-50.0+(-210.0*p002)+(42.0*p102)+(-145.0*p20)+(27.0*p202)+(28.0*p00*(14.0+p20))+(21.0*p10*(-7.0+(3.0*p20)))))+(2.0*(-37.0+(-42.0*p002)+(-28.0*p00*(-4.0+p10))+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)))+(-3.0*((p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(2.0*p11*(41.0+(14.0*p002)+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202)+(p00*(-55.0+(21.0*p10)+(6.0*p20))))*p21)+((68.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20)))+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p212))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((p012*(5.0+(2310.0*p002)+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))+(28.0*p00*(1.0+(45.0*p10)+(9.0*p20)))))+(-2.0*p01*(55.0+(-15.0*p11)+(-9.0*p102*(-55.0+(14.0*p11)))+(165.0*p20)+(-60.0*p11*p20)+(198.0*p202)+(-81.0*p11*p202)+(-21.0*p21)+(-54.0*p20*p21)+(-45.0*p202*p21)+(14.0*p002*(110.0+(45.0*p11)+(9.0*p21)))+(p00*(55.0+(330.0*p20)+(-21.0*p11*(1.0+(4.0*p20)))+(-6.0*p21)+(21.0*p10*(55.0+(4.0*p21)))))+(-3.0*p10*(-44.0+(9.0*p11*(2.0+(7.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))))))+(-3.0*((3.0*p112*(-7.0+(42.0*p002)+(p00*(-2.0+(28.0*p10)))+(-18.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(p21*(770.0+(330.0*p20)+(-126.0*p21)+(-84.0*p20*p21)+(18.0*p10*p20*(-11.0+(3.0*p21)))+(9.0*p102*(-22.0+(5.0*p21)))+(p002*(-220.0+(21.0*p21)))+(2.0*p00*(-11.0+(-66.0*p20)+(3.0*p21)+(15.0*p20*p21)+(3.0*p10*(-55.0+(9.0*p21)))))))+(2.0*p11*(110.0+(165.0*p20)+(99.0*p202)+(3.0*p10*(22.0+(33.0*p20)+(-6.0*p21)))+(-42.0*p21)+(27.0*p102*p21)+(-63.0*p20*p21)+(-27.0*p202*p21)+(7.0*p002*(-55.0+(6.0*p21)))+(p00*(11.0+(18.0*p20*p21)+(3.0*p10*(-55.0+(21.0*p21))))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-21.0*p012*(-5.0+(110.0*p002)+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)+(4.0*p00*(-73.0+(15.0*p10)+(3.0*p20)))))+(2.0*p01*(-407.0+(150.0*p11)+(-9.0*p102*(-55.0+(14.0*p11)))+(-1221.0*p20)+(435.0*p11*p20)+(198.0*p202)+(-81.0*p11*p202)+(111.0*p21)+(144.0*p20*p21)+(-45.0*p202*p21)+(14.0*p002*(110.0+(45.0*p11)+(9.0*p21)))+(p00*(-4565.0+(330.0*p20)+(-84.0*p11*(14.0+p20))+(-336.0*p21)+(21.0*p10*(55.0+(4.0*p21)))))+(-3.0*p10*((21.0*p11*(-7.0+(3.0*p20)))+(10.0*(88.0+p21))+(3.0*p20*(-55.0+(6.0*p21)))))))+(3.0*((3.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(p21*(-1078.0+(330.0*p20)+(204.0*p21)+(-84.0*p20*p21)+(9.0*p102*(-22.0+(5.0*p21)))+(p002*(-220.0+(21.0*p21)))+(6.0*p10*(154.0+(-33.0*p20)+(-33.0*p21)+(9.0*p20*p21)))+(2.0*p00*(451.0+(-66.0*p20)+(-63.0*p21)+(15.0*p20*p21)+(3.0*p10*(-55.0+(9.0*p21)))))))+(2.0*p11*(-352.0+(-297.0*p20)+(99.0*p202)+(3.0*p10*(22.0+(33.0*p20)+(-39.0*p21)))+(123.0*p21)+(27.0*p102*p21)+(36.0*p20*p21)+(-27.0*p202*p21)+(7.0*p002*(-55.0+(6.0*p21)))+(p00*(935.0+(3.0*(-55.0+(6.0*p20))*p21)+(3.0*p10*(-55.0+(21.0*p21))))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 15:
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-6.0*p01*((p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+((7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)))+(9.0*((-1.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-2.0*p11*(14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/6160.0)*((7.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(2.0*p01*((p11*(-50.0+(42.0*p102)+(-145.0*p20)+(27.0*p202)+(21.0*p10*(-7.0+(3.0*p20)))))+((-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)))+(-3.0*((p112*(37.0+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(2.0*p11*(41.0+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202))*p21)+((68.0+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20))))*p212))))*sign);
propagate[x][y] += ((-1.0/4.0)*p012*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(2.0*p11*(110.0+(p20*(165.0+(-63.0*p21)))+(p202*(99.0+(-27.0*p21)))+(3.0*p10*(22.0+(33.0*p20)+(-6.0*p21)))+(-42.0*p21)+(27.0*p102*p21)))+(p21*(770.0+(p20*(330.0+(-84.0*p21)))+(-126.0*p21)+(18.0*p10*p20*(-11.0+(3.0*p21)))+(9.0*p102*(-22.0+(5.0*p21)))))))+(2.0*p01*(-55.0+(9.0*p102*(-55.0+(14.0*p11)))+(-165.0*p20)+(-198.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(-44.0+(9.0*p11*(2.0+(7.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-2.0*p01*(407.0+(9.0*p102*(-55.0+(14.0*p11)))+(1221.0*p20)+(-198.0*p202)+(3.0*p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-111.0*p21)+(-144.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*((21.0*p11*(-7.0+(3.0*p20)))+(10.0*(88.0+p21))+(3.0*p20*(-55.0+(6.0*p21)))))))+(3.0*((-3.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(2.0*p11*(-352.0+(3.0*p10*(22.0+(33.0*p20)+(-39.0*p21)))+(p202*(99.0+(-27.0*p21)))+(123.0*p21)+(27.0*p102*p21)+(9.0*p20*(-33.0+(4.0*p21)))))+(p21*(-1078.0+(p20*(330.0+(-84.0*p21)))+(204.0*p21)+(9.0*p102*(-22.0+(5.0*p21)))+(6.0*p10*(154.0+(-33.0*p21)+(p20*(-33.0+(9.0*p21))))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(-2.0+p01)*p01*sign);
break;
case 16:
p002 = p00*p00;
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((3.0/6160.0)*((-15.0*p112*p202)+(-18.0*p11*p202*p21)+(3.0*p102*p21*((6.0*p11)+(5.0*p21)))+(7.0*p002*((6.0*p112)+(4.0*p11*p21)+p212))+(-18.0*p112*p20*p30)+(-42.0*p11*p20*p21*p30)+(-28.0*p20*p212*p30)+(-7.0*p112*p302)+(-28.0*p11*p21*p302)+(-42.0*p212*p302)+(2.0*p00*((6.0*p11*p20*p21)+(p10*((14.0*p112)+(21.0*p11*p21)+(9.0*p212)))+(-1.0*p112*p30)+(p212*((5.0*p20)+p30))))+(-2.0*p10*((-9.0*p20*p212)+(6.0*p11*p21*p30)+(p112*((9.0*p20)+(5.0*p30))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((-3.0/6160.0)*((7.0*p002*((6.0*p112)+(4.0*p11*p21)+p212))+(2.0*p00*((p11*(-55.0+(21.0*p10)+(6.0*p20))*p21)+(p112*(-55.0+(14.0*p10)+(-1.0*p30)))+(p212*(-22.0+(9.0*p10)+(5.0*p20)+p30))))+(2.0*p11*p21*((9.0*p102)+(-9.0*p202)+(p20*(33.0+(-21.0*p30)))+((55.0+(-14.0*p30))*p30)+(-3.0*p10*(11.0+(2.0*p30)))))+(-1.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(p212*((15.0*p102)+(6.0*p10*(-11.0+(3.0*p20)))+(-2.0*p30*(-55.0+(14.0*p20)+(21.0*p30))))))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((1.0/6160.0)*((-198.0*p11*p202)+(45.0*p112*p202)+(54.0*p11*p202*p21)+(-9.0*p102*p21*(-22.0+(6.0*p11)+(5.0*p21)))+(p002*((-126.0*p112)+(p11*(770.0+(-84.0*p21)))+((220.0+(-21.0*p21))*p21)))+(-330.0*p11*p20*p30)+(54.0*p112*p20*p30)+(-330.0*p20*p21*p30)+(126.0*p11*p20*p21*p30)+(84.0*p20*p212*p30)+(-220.0*p11*p302)+(21.0*p112*p302)+(-770.0*p21*p302)+(84.0*p11*p21*p302)+(126.0*p212*p302)+(-2.0*p00*((3.0*p20*p21*(-22.0+(6.0*p11)+(5.0*p21)))+(3.0*p10*((14.0*p112)+(p21*(-55.0+(9.0*p21)))+(p11*(-55.0+(21.0*p21)))))+(((11.0*p11)+(-3.0*p112)+(p21*(-11.0+(3.0*p21))))*p30)))+(6.0*p10*((3.0*p20*(11.0+(-3.0*p21))*p21)+(p112*((9.0*p20)+(5.0*p30)))+(p11*((-33.0*p20)+(-22.0*p30)+(6.0*p21*p30))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((1.0/6160.0)*((p002*((126.0*p112)+(14.0*p11*(-55.0+(6.0*p21)))+(p21*(-220.0+(21.0*p21)))))+(-3.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(p21*((9.0*p102*(-22.0+(5.0*p21)))+(6.0*p10*(154.0+(-33.0*p21)+(p20*(-33.0+(9.0*p21)))))+(2.0*p30*(-924.0+(p20*(165.0+(-42.0*p21)))+(165.0*p21)+(385.0*p30)+(-63.0*p21*p30)))))+(-2.0*p11*((-27.0*p102*p21)+(9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p20*(-154.0+(33.0*p10)+(33.0*p21)+(55.0*p30)+(-21.0*p21*p30)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30)))))+(2.0*p00*((3.0*p112*(-55.0+(14.0*p10)+(-1.0*p30)))+(p21*(462.0+(-66.0*p21)+(3.0*p20*(-22.0+(5.0*p21)))+(3.0*p10*(-55.0+(9.0*p21)))+(-11.0*p30)+(3.0*p21*p30)))+(p11*((3.0*(-55.0+(6.0*p20))*p21)+(3.0*p10*(-55.0+(21.0*p21)))+(11.0*(84.0+p30)))))))*sign);
propagate[x][y] += 0.0;
break;
case 17:
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((3.0/6160.0)*((3.0*p102*p21*((6.0*p11)+(5.0*p21)))+(-14.0*p212*p30*((2.0*p20)+(3.0*p30)))+(-1.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*p21*((9.0*p202)+(21.0*p20*p30)+(14.0*p302)))+(-2.0*p10*((-9.0*p20*p212)+(6.0*p11*p21*p30)+(p112*((9.0*p20)+(5.0*p30))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((3.0/6160.0)*((-2.0*p11*p21*((9.0*p102)+(-9.0*p202)+(p20*(33.0+(-21.0*p30)))+((55.0+(-14.0*p30))*p30)+(-3.0*p10*(11.0+(2.0*p30)))))+(p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(p212*((-15.0*p102)+(p10*(66.0+(-18.0*p20)))+(2.0*p30*(-55.0+(14.0*p20)+(21.0*p30))))))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((1.0/6160.0)*((-9.0*p102*p21*(-22.0+(6.0*p11)+(5.0*p21)))+(2.0*p21*p30*((3.0*p20*(-55.0+(14.0*p21)))+(7.0*(-55.0+(9.0*p21))*p30)))+(3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(2.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(3.0*p20*(-55.0+(21.0*p21))*p30)+(2.0*(-55.0+(21.0*p21))*p302)))+(6.0*p10*((3.0*p20*(11.0+(-3.0*p21))*p21)+(p112*((9.0*p20)+(5.0*p30)))+(p11*((-33.0*p20)+(-22.0*p30)+(6.0*p21*p30))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((1.0/6160.0)*((-3.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(p21*((9.0*p102*(-22.0+(5.0*p21)))+(6.0*p10*(154.0+(-33.0*p21)+(p20*(-33.0+(9.0*p21)))))+(2.0*p30*(-924.0+(p20*(165.0+(-42.0*p21)))+(165.0*p21)+(385.0*p30)+(-63.0*p21*p30)))))+(-2.0*p11*((-27.0*p102*p21)+(9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p20*(-154.0+(33.0*p10)+(33.0*p21)+(55.0*p30)+(-21.0*p21*p30)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30))))))*sign);
propagate[x][y] += 0.0;
break;
case 18:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += ((3.0/6160.0)*((p112*(-7.0+(42.0*p002)+(p00*(-2.0+(28.0*p10)))+(-18.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(2.0*p11*(-14.0+(14.0*p002)+(-6.0*p10)+(21.0*p00*p10)+(9.0*p102)+(-21.0*p20)+(6.0*p00*p20)+(-9.0*p202))*p21)+((-42.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(1.0+(9.0*p10)+(5.0*p20))))*p212))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((-3.0/6160.0)*((p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(2.0*p11*(41.0+(14.0*p002)+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202)+(p00*(-55.0+(21.0*p10)+(6.0*p20))))*p21)+((68.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20)))+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p212))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((1.0/6160.0)*((p112*(21.0+(-126.0*p002)+(p00*(6.0+(-84.0*p10)))+(54.0*p20)+(45.0*p202)+(6.0*p10*(5.0+(9.0*p20)))))+(p21*(-770.0+(-330.0*p20)+(p002*(220.0+(-21.0*p21)))+(18.0*p10*p20*(11.0+(-3.0*p21)))+(126.0*p21)+(84.0*p20*p21)+(-9.0*p102*(-22.0+(5.0*p21)))+(p00*(22.0+(132.0*p20)+(-6.0*p21)+(-30.0*p20*p21)+(-6.0*p10*(-55.0+(9.0*p21)))))))+(-2.0*p11*(110.0+(165.0*p20)+(99.0*p202)+(3.0*p10*(22.0+(33.0*p20)+(-6.0*p21)))+(-42.0*p21)+(27.0*p102*p21)+(-63.0*p20*p21)+(-27.0*p202*p21)+(7.0*p002*(-55.0+(6.0*p21)))+(p00*(11.0+(18.0*p20*p21)+(3.0*p10*(-55.0+(21.0*p21))))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((1.0/6160.0)*((3.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(p21*(-1078.0+(330.0*p20)+(204.0*p21)+(-84.0*p20*p21)+(9.0*p102*(-22.0+(5.0*p21)))+(p002*(-220.0+(21.0*p21)))+(6.0*p10*(154.0+(-33.0*p20)+(-33.0*p21)+(9.0*p20*p21)))+(2.0*p00*(451.0+(-66.0*p20)+(-63.0*p21)+(15.0*p20*p21)+(3.0*p10*(-55.0+(9.0*p21)))))))+(2.0*p11*(-352.0+(-297.0*p20)+(99.0*p202)+(3.0*p10*(22.0+(33.0*p20)+(-39.0*p21)))+(123.0*p21)+(27.0*p102*p21)+(36.0*p20*p21)+(-27.0*p202*p21)+(7.0*p002*(-55.0+(6.0*p21)))+(p00*(935.0+(3.0*(-55.0+(6.0*p20))*p21)+(3.0*p10*(-55.0+(21.0*p21))))))))*sign);
propagate[x][y] += 0.0;
break;
case 19:
x = 0; y = 0;
pixel[y][-x] += ((-3.0/6160.0)*((p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(2.0*p11*(14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p21)+((42.0+(-15.0*p102)+(28.0*p20)+(-18.0*p10*p20))*p212))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((-3.0/6160.0)*((p112*(37.0+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(2.0*p11*(41.0+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202))*p21)+((68.0+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20))))*p212))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((1.0/6160.0)*((3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-2.0*p11*(110.0+(p20*(165.0+(-63.0*p21)))+(p202*(99.0+(-27.0*p21)))+(3.0*p10*(22.0+(33.0*p20)+(-6.0*p21)))+(-42.0*p21)+(27.0*p102*p21)))+(p21*(-770.0+(-330.0*p20)+(18.0*p10*p20*(11.0+(-3.0*p21)))+(126.0*p21)+(84.0*p20*p21)+(-9.0*p102*(-22.0+(5.0*p21))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((1.0/6160.0)*((-3.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(2.0*p11*(-352.0+(3.0*p10*(22.0+(33.0*p20)+(-39.0*p21)))+(p202*(99.0+(-27.0*p21)))+(123.0*p21)+(27.0*p102*p21)+(9.0*p20*(-33.0+(4.0*p21)))))+(p21*(-1078.0+(p20*(330.0+(-84.0*p21)))+(204.0*p21)+(9.0*p102*(-22.0+(5.0*p21)))+(6.0*p10*(154.0+(-33.0*p21)+(p20*(-33.0+(9.0*p21))))))))*sign);
propagate[x][y] += 0.0;
break;
case 20:
p002 = p00*p00;
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-63.0*p202)+(-162.0*p11*p202)+(-135.0*p112*p202)+(-90.0*p202*p21)+(-162.0*p11*p202*p21)+(21.0*p002*(-110.0+(18.0*p112)+(12.0*p21)+(3.0*p212)+(12.0*p11*(5.0+p21))))+(9.0*p102*(-42.0+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(-30.0*p20*p30)+(-120.0*p11*p20*p30)+(-162.0*p112*p20*p30)+(-108.0*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(-252.0*p20*p212*p30)+(-5.0*p302)+(-30.0*p11*p302)+(-63.0*p112*p302)+(-42.0*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(2.0*p00*((3.0*p20*(-42.0+(-28.0*p11)+(18.0*p11*p21)+(15.0*p212)))+(3.0*p10*(-210.0+(42.0*p112)+(28.0*p21)+(63.0*p11*p21)+(27.0*p212)))+(-1.0*(14.0+(21.0*p11)+(9.0*p112)+(6.0*p21)+(-9.0*p212))*p30)))+(-6.0*p10*((3.0*p20*(14.0+(21.0*p11)+(9.0*p112)+(6.0*p21)+(-9.0*p212)))+((7.0+(15.0*p112)+(10.0*p21)+(18.0*p11*(1.0+p21)))*p30))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-660.0*p20)+(-990.0*p11*p20)+(-594.0*p112*p20)+(63.0*p202)+(162.0*p11*p202)+(135.0*p112*p202)+(-396.0*p20*p21)+(-594.0*p11*p20*p21)+(90.0*p202*p21)+(162.0*p11*p202*p21)+(-21.0*p002*(-110.0+(18.0*p112)+(12.0*p21)+(3.0*p212)+(12.0*p11*(5.0+p21))))+(-9.0*p102*(-42.0+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(-110.0*p30)+(-330.0*p11*p30)+(-396.0*p112*p30)+(30.0*p20*p30)+(120.0*p11*p20*p30)+(162.0*p112*p20*p30)+(-264.0*p21*p30)+(-990.0*p11*p21*p30)+(108.0*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(-990.0*p212*p30)+(252.0*p20*p212*p30)+(5.0*p302)+(30.0*p11*p302)+(63.0*p112*p302)+(42.0*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(6.0*p10*(-385.0+(99.0*p212)+(p20*(42.0+(18.0*p21)+(-27.0*p212)))+(7.0*p30)+(10.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(3.0*p11*(-55.0+(21.0*p20)+(33.0*p21)+(6.0*p30)+(6.0*p21*p30)))))+(-2.0*p00*(3080.0+(-126.0*p20)+(-330.0*p21)+(-198.0*p212)+(45.0*p20*p212)+(3.0*p10*(-210.0+(42.0*p112)+(28.0*p21)+(63.0*p11*p21)+(27.0*p212)))+(-14.0*p30)+(-6.0*p21*p30)+(9.0*p212*p30)+(-9.0*p112*(55.0+p30))+(3.0*p11*((-165.0*p21)+(2.0*p20*(-14.0+(9.0*p21)))+(-7.0*(55.0+p30)))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*(770.0+(378.0*p112)+(-408.0*p21)+(63.0*p212)+(42.0*p11*(-25.0+(6.0*p21)))))+(-2.0*p00*((3.0*p20*(68.0+(-28.0*p11)+(-66.0*p21)+(18.0*p11*p21)+(15.0*p212)))+(3.0*p10*(175.0+(42.0*p112)+(-137.0*p21)+(27.0*p212)+(3.0*p11*(-55.0+(21.0*p21)))))+((41.0+(12.0*p11)+(-9.0*p112)+(-39.0*p21)+(9.0*p212))*p30)))+(-3.0*((-3.0*p202*(-37.0+(15.0*p112)+(10.0*p21)+(6.0*p11*(-8.0+(3.0*p21)))))+(3.0*p102*(68.0+(-66.0*p21)+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(-2.0*p20*(-50.0+(27.0*p112)+(-147.0*p21)+(42.0*p212)+(p11*(-145.0+(63.0*p21))))*p30)+(-7.0*(-5.0+(3.0*p112)+(-108.0*p21)+(18.0*p212)+(6.0*p11*(-5.0+(2.0*p21))))*p302)+(p10*((-6.0*p20*(-41.0+(-12.0*p11)+(9.0*p112)+(39.0*p21)+(-9.0*p212)))+(-2.0*(-37.0+(15.0*p112)+(10.0*p21)+(6.0*p11*(-8.0+(3.0*p21))))*p30))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-2112.0*p20)+(-1782.0*p11*p20)+(594.0*p112*p20)+(333.0*p202)+(432.0*p11*p202)+(-135.0*p112*p202)+(396.0*p20*p21)+(594.0*p11*p20*p21)+(-90.0*p202*p21)+(-162.0*p11*p202*p21)+(p002*(770.0+(378.0*p112)+(-408.0*p21)+(63.0*p212)+(42.0*p11*(-25.0+(6.0*p21)))))+(9.0*p102*(68.0+(-66.0*p21)+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(-814.0*p30)+(-2442.0*p11*p30)+(396.0*p112*p30)+(300.0*p20*p30)+(870.0*p11*p20*p30)+(-162.0*p112*p20*p30)+(-5280.0*p21*p30)+(990.0*p11*p21*p30)+(882.0*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(990.0*p212*p30)+(-252.0*p20*p212*p30)+(105.0*p302)+(630.0*p11*p302)+(-63.0*p112*p302)+(2268.0*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(2.0*p00*(-1540.0+(204.0*p20)+(1056.0*p21)+(-198.0*p20*p21)+(-198.0*p212)+(45.0*p20*p212)+(3.0*p10*(175.0+(42.0*p112)+(-137.0*p21)+(27.0*p212)+(3.0*p11*(-55.0+(21.0*p21)))))+(41.0*p30)+(-39.0*p21*p30)+(9.0*p212*p30)+(-9.0*p112*(55.0+p30))+(3.0*p11*(539.0+(-28.0*p20)+(-165.0*p21)+(18.0*p20*p21)+(4.0*p30)))))+(-6.0*p10*(539.0+(-462.0*p21)+(99.0*p212)+(-3.0*p20*(41.0+(-39.0*p21)+(9.0*p212)))+(-37.0*p30)+(10.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(-3.0*p11*(55.0+(12.0*p20)+(-33.0*p21)+(16.0*p30)+(-6.0*p21*p30))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 21:
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(7.0+(15.0*p112)+(10.0*p21)+(18.0*p11*(1.0+p21))))+(9.0*p102*(-42.0+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(-6.0*p20*(5.0+(27.0*p112)+(18.0*p21)+(42.0*p212)+(p11*(20.0+(63.0*p21))))*p30)+(-1.0*(5.0+(63.0*p112)+(42.0*p21)+(378.0*p212)+(6.0*p11*(5.0+(42.0*p21))))*p302)+(-6.0*p10*((3.0*p20*(14.0+(21.0*p11)+(9.0*p112)+(6.0*p21)+(-9.0*p212)))+((7.0+(15.0*p112)+(10.0*p21)+(18.0*p11*(1.0+p21)))*p30))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(7.0+(15.0*p112)+(10.0*p21)+(18.0*p11*(1.0+p21))))+(-9.0*p102*(-42.0+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(6.0*p10*(-385.0+(99.0*p212)+(p20*(42.0+(18.0*p21)+(-27.0*p212)))+(7.0*p30)+(10.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(3.0*p11*(-55.0+(21.0*p20)+(33.0*p21)+(6.0*p30)+(6.0*p21*p30)))))+(p30*((5.0*(-22.0+p30))+(9.0*p112*(-44.0+(7.0*p30)))+(6.0*p21*(-44.0+(7.0*p30)))+(18.0*p212*(-55.0+(21.0*p30)))+(6.0*p11*(-55.0+(-165.0*p21)+(5.0*p30)+(42.0*p21*p30)))))+(6.0*p20*((5.0*(-22.0+p30))+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(6.0*p21*(-11.0+(3.0*p30)))+(p11*(-165.0+(-99.0*p21)+(20.0*p30)+(63.0*p21*p30))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/6160.0)*((3.0*p202*(-37.0+(15.0*p112)+(10.0*p21)+(6.0*p11*(-8.0+(3.0*p21)))))+(-3.0*p102*(68.0+(-66.0*p21)+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(2.0*p20*(-50.0+(27.0*p112)+(-147.0*p21)+(42.0*p212)+(p11*(-145.0+(63.0*p21))))*p30)+(7.0*(-5.0+(3.0*p112)+(-108.0*p21)+(18.0*p212)+(6.0*p11*(-5.0+(2.0*p21))))*p302)+(p10*((6.0*p20*(-41.0+(-12.0*p11)+(9.0*p112)+(39.0*p21)+(-9.0*p212)))+(2.0*(-37.0+(15.0*p112)+(10.0*p21)+(6.0*p11*(-8.0+(3.0*p21))))*p30))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(-37.0+(15.0*p112)+(10.0*p21)+(6.0*p11*(-8.0+(3.0*p21)))))+(9.0*p102*(68.0+(-66.0*p21)+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(-6.0*p10*(539.0+(-462.0*p21)+(99.0*p212)+(-3.0*p20*(41.0+(-39.0*p21)+(9.0*p212)))+(-37.0*p30)+(10.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(-3.0*p11*(55.0+(12.0*p20)+(-33.0*p21)+(16.0*p30)+(-6.0*p21*p30)))))+(-6.0*p20*(352.0+(-50.0*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(-3.0*p21*(22.0+(49.0*p30)))+(p11*(297.0+(-145.0*p30)+(9.0*p21*(-11.0+(7.0*p30)))))))+(p30*(-814.0+(p212*(990.0+(-378.0*p30)))+(105.0*p30)+(-9.0*p112*(-44.0+(7.0*p30)))+(12.0*p21*(-440.0+(189.0*p30)))+(-6.0*p11*(407.0+(-105.0*p30)+(3.0*p21*(-55.0+(14.0*p30))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 22:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*(-5.0+(-30.0*p11)+(-63.0*p112)+(-30.0*p20)+(-120.0*p11*p20)+(-162.0*p112*p20)+(-63.0*p202)+(-162.0*p11*p202)+(-135.0*p112*p202)+(-42.0*p21)+(-252.0*p11*p21)+(-108.0*p20*p21)+(-378.0*p11*p20*p21)+(-90.0*p202*p21)+(-162.0*p11*p202*p21)+(-378.0*p212)+(-252.0*p20*p212)+(21.0*p002*(-110.0+(18.0*p112)+(12.0*p21)+(3.0*p212)+(12.0*p11*(5.0+p21))))+(9.0*p102*(-42.0+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(-6.0*p10*(7.0+(3.0*p112*(5.0+(9.0*p20)))+(10.0*p21)+(9.0*p11*(2.0+(7.0*p20)+(2.0*p21)))+(p20*(42.0+(18.0*p21)+(-27.0*p212)))))+(2.0*p00*(-14.0+(-9.0*p112)+(-126.0*p20)+(-6.0*p21)+(9.0*p212)+(45.0*p20*p212)+(3.0*p11*(-7.0+(-28.0*p20)+(18.0*p20*p21)))+(3.0*p10*(-210.0+(42.0*p112)+(28.0*p21)+(63.0*p11*p21)+(27.0*p212))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/6160.0)*(-35.0+(-100.0*p11)+(-111.0*p112)+(-210.0*p20)+(-290.0*p11*p20)+(-144.0*p112*p20)+(21.0*p202)+(54.0*p11*p202)+(45.0*p112*p202)+(-74.0*p21)+(-246.0*p11*p21)+(-96.0*p20*p21)+(-72.0*p11*p20*p21)+(30.0*p202*p21)+(54.0*p11*p202*p21)+(-204.0*p212)+(84.0*p20*p212)+(p102*(126.0+(p11*(84.0+(-54.0*p21)))+(-45.0*p212)))+(-7.0*p002*(-110.0+(18.0*p112)+(12.0*p21)+(3.0*p212)+(12.0*p11*(5.0+p21))))+(2.0*p10*(-378.0+(3.0*p112*(5.0+(9.0*p20)))+(10.0*p21)+(99.0*p212)+(3.0*p11*(-49.0+(21.0*p20)+(39.0*p21)))+(p20*(42.0+(18.0*p21)+(-27.0*p212)))))+(p00*(-2044.0+(336.0*p112)+(84.0*p20)+(224.0*p21)+(126.0*p212)+(-30.0*p20*p212)+(p11*(784.0+(56.0*p20)+(330.0*p21)+(-36.0*p20*p21)))+(-2.0*p10*(-210.0+(42.0*p112)+(28.0*p21)+(63.0*p11*p21)+(27.0*p212))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*(770.0+(378.0*p112)+(-408.0*p21)+(63.0*p212)+(42.0*p11*(-25.0+(6.0*p21)))))+(-2.0*p00*(41.0+(-9.0*p112)+(204.0*p20)+(-39.0*p21)+(-198.0*p20*p21)+(9.0*p212)+(45.0*p20*p212)+(6.0*p11*(2.0+(-14.0*p20)+(9.0*p20*p21)))+(3.0*p10*(175.0+(42.0*p112)+(-137.0*p21)+(27.0*p212)+(3.0*p11*(-55.0+(21.0*p21)))))))+(-3.0*(35.0+(100.0*p20)+(111.0*p202)+(-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(756.0*p21)+(294.0*p20*p21)+(-30.0*p202*p21)+(-126.0*p212)+(-84.0*p20*p212)+(3.0*p102*(68.0+(-66.0*p21)+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(-2.0*p11*((21.0*(-5.0+(2.0*p21)))+(9.0*p202*(-8.0+(3.0*p21)))+(p20*(-145.0+(63.0*p21)))))+(p10*(74.0+(-6.0*p112*(5.0+(9.0*p20)))+(12.0*p11*(8.0+(6.0*p20)+(-3.0*p21)))+(-20.0*p21)+(6.0*p20*(41.0+(-39.0*p21)+(9.0*p212))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*(-709.0+(-1812.0*p11)+(333.0*p112)+(-1812.0*p20)+(-912.0*p11*p20)+(432.0*p112*p20)+(333.0*p202)+(432.0*p11*p202)+(-135.0*p112*p202)+(-3012.0*p21)+(738.0*p11*p21)+(1278.0*p20*p21)+(216.0*p11*p20*p21)+(-90.0*p202*p21)+(-162.0*p11*p202*p21)+(612.0*p212)+(-252.0*p20*p212)+(p002*(770.0+(378.0*p112)+(-408.0*p21)+(63.0*p212)+(42.0*p11*(-25.0+(6.0*p21)))))+(9.0*p102*(68.0+(-66.0*p21)+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(-6.0*p10*(502.0+(3.0*p112*(5.0+(9.0*p20)))+(-3.0*p11*(71.0+(12.0*p20)+(-39.0*p21)))+(-452.0*p21)+(99.0*p212)+(-3.0*p20*(41.0+(-39.0*p21)+(9.0*p212)))))+(2.0*p00*(-1499.0+(-504.0*p112)+(204.0*p20)+(1017.0*p21)+(-198.0*p20*p21)+(-189.0*p212)+(45.0*p20*p212)+(3.0*p11*(543.0+(-28.0*p20)+(-165.0*p21)+(18.0*p20*p21)))+(3.0*p10*(175.0+(42.0*p112)+(-137.0*p21)+(27.0*p212)+(3.0*p11*(-55.0+(21.0*p21))))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 23:
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*(-5.0+(-30.0*p20)+(-63.0*p202)+(-9.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-42.0*p21)+(-108.0*p20*p21)+(-90.0*p202*p21)+(-378.0*p212)+(-252.0*p20*p212)+(9.0*p102*(-42.0+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(-6.0*p11*(5.0+(42.0*p21)+(27.0*p202*(1.0+p21))+(p20*(20.0+(63.0*p21)))))+(-6.0*p10*(7.0+(3.0*p112*(5.0+(9.0*p20)))+(10.0*p21)+(9.0*p11*(2.0+(7.0*p20)+(2.0*p21)))+(p20*(42.0+(18.0*p21)+(-27.0*p212))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/6160.0)*(-35.0+(-210.0*p20)+(21.0*p202)+(3.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)))+(-74.0*p21)+(-96.0*p20*p21)+(30.0*p202*p21)+(-204.0*p212)+(84.0*p20*p212)+(p102*(126.0+(p11*(84.0+(-54.0*p21)))+(-45.0*p212)))+(2.0*p11*(-50.0+(-123.0*p21)+(27.0*p202*(1.0+p21))+(-1.0*p20*(145.0+(36.0*p21)))))+(2.0*p10*(-378.0+(3.0*p112*(5.0+(9.0*p20)))+(10.0*p21)+(99.0*p212)+(3.0*p11*(-49.0+(21.0*p20)+(39.0*p21)))+(p20*(42.0+(18.0*p21)+(-27.0*p212))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/6160.0)*(-35.0+(-100.0*p20)+(-111.0*p202)+(3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-756.0*p21)+(-294.0*p20*p21)+(30.0*p202*p21)+(126.0*p212)+(84.0*p20*p212)+(-3.0*p102*(68.0+(-66.0*p21)+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(2.0*p11*((21.0*(-5.0+(2.0*p21)))+(9.0*p202*(-8.0+(3.0*p21)))+(p20*(-145.0+(63.0*p21)))))+(p10*(-74.0+(6.0*p112*(5.0+(9.0*p20)))+(-12.0*p11*(8.0+(6.0*p20)+(-3.0*p21)))+(20.0*p21)+(-6.0*p20*(41.0+(-39.0*p21)+(9.0*p212))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*(-709.0+(-1812.0*p20)+(333.0*p202)+(-9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)))+(-3012.0*p21)+(1278.0*p20*p21)+(-90.0*p202*p21)+(612.0*p212)+(-252.0*p20*p212)+(-6.0*p11*(302.0+(-123.0*p21)+(9.0*p202*(-8.0+(3.0*p21)))+(-4.0*p20*(-38.0+(9.0*p21)))))+(9.0*p102*(68.0+(-66.0*p21)+(15.0*p212)+(2.0*p11*(-14.0+(9.0*p21)))))+(-6.0*p10*(502.0+(3.0*p112*(5.0+(9.0*p20)))+(-3.0*p11*(71.0+(12.0*p20)+(-39.0*p21)))+(-452.0*p21)+(99.0*p212)+(-3.0*p20*(41.0+(-39.0*p21)+(9.0*p212))))))*sign);
propagate[x][y] += ((-1.0/4.0)*sign);
break;
case 24:
p002 = p00*p00;
p012 = p01*p01;
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((p002*(5.0+(-2310.0*p012)+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))+(28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(-2.0*p00*((3.0*p10*(-5.0+(210.0*p012)+(-18.0*p11)+(-42.0*p112)+(-20.0*p21)+(-63.0*p11*p21)+(-27.0*p212)+(-7.0*p01*(1.0+(4.0*p21)))))+(3.0*p20*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+((-14.0+(14.0*p012)+(-6.0*p11)+(21.0*p01*p11)+(9.0*p112)+(-21.0*p21)+(6.0*p01*p21)+(-9.0*p212))*p30)))+(-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30)))))+(-2.0*p01*((9.0*p102*(-1.0+(14.0*p11)))+(9.0*p202*(1.0+(9.0*p11)+(5.0*p21)))+(3.0*p20*(7.0+(20.0*p11)+(18.0*p21))*p30)+((14.0+(15.0*p11)+(21.0*p21))*p302)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+p30+(5.0*p21*p30))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*(5.0+(-2310.0*p012)+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))+(28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(2.0*p00*((3.0*p10*(-5.0+(210.0*p012)+(-18.0*p11)+(-42.0*p112)+(-20.0*p21)+(-63.0*p11*p21)+(-27.0*p212)+(-7.0*p01*(1.0+(4.0*p21)))))+(3.0*p20*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+((-14.0+(14.0*p012)+(-6.0*p11)+(21.0*p01*p11)+(9.0*p112)+(-21.0*p21)+(6.0*p01*p21)+(-9.0*p212))*p30)))+(p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-22.0*((-30.0*p10)+(-18.0*p10*p11)+(-105.0*p20)+(27.0*p112*p20)+(-45.0*p10*p21)+(-27.0*p10*p11*p21)+(-45.0*p20*p21)+(27.0*p11*p20*p21)+(-27.0*p10*p212)+(p00*(-5.0+(280.0*p012)+(-12.0*p11)+(-45.0*p112)+(-15.0*p21)+(-45.0*p11*p21)+(-18.0*p212)+(-5.0*p01*(1.0+(21.0*p11)+(6.0*p21)))))+(-280.0*p30)+(30.0*p11*p30)+(18.0*p112*p30)+(105.0*p21*p30)+(45.0*p11*p21*p30)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(p01*((p10*(-3.0+(45.0*p11)))+(3.0*p20*(1.0+(15.0*p11)+(6.0*p21)))+((5.0+(15.0*p11)+(12.0*p21))*p30)))))+(-3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30)))))+(2.0*p01*((9.0*p102*(-1.0+(14.0*p11)))+(9.0*p202*(1.0+(9.0*p11)+(5.0*p21)))+(3.0*p20*(7.0+(20.0*p11)+(18.0*p21))*p30)+((14.0+(15.0*p11)+(21.0*p21))*p302)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+p30+(5.0*p21*p30))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((612.0*p202)+(-378.0*p01*p202)+(63.0*p012*p202)+(-594.0*p11*p202)+(162.0*p01*p11*p202)+(135.0*p112*p202)+(-252.0*p202*p21)+(90.0*p01*p202*p21)+(162.0*p11*p202*p21)+(21.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(1050.0*p20*p30)+(-288.0*p01*p20*p30)+(30.0*p012*p20*p30)+(-822.0*p11*p20*p30)+(120.0*p01*p11*p20*p30)+(162.0*p112*p20*p30)+(-990.0*p20*p21*p30)+(108.0*p01*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(252.0*p20*p212*p30)+(770.0*p302)+(-82.0*p01*p302)+(5.0*p012*p302)+(-408.0*p11*p302)+(30.0*p01*p11*p302)+(63.0*p112*p302)+(-1050.0*p21*p302)+(42.0*p01*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(6.0*p10*((3.0*p20*(41.0+(14.0*p012)+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)+(p01*(-55.0+(21.0*p11)+(6.0*p21)))))+((68.0+(7.0*p012)+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21))))*p30)))+(2.0*p00*((3.0*p10*(50.0+(210.0*p012)+(-42.0*p112)+(145.0*p21)+(-27.0*p212)+(-28.0*p01*(14.0+p21))+(-21.0*p11*(-7.0+(3.0*p21)))))+(3.0*p20*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+((41.0+(14.0*p012)+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)+(p01*(-55.0+(21.0*p11)+(6.0*p21))))*p30))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((3234.0*p20)+(-2706.0*p01*p20)+(660.0*p012*p20)+(-2772.0*p11*p20)+(990.0*p01*p11*p20)+(594.0*p112*p20)+(-612.0*p202)+(378.0*p01*p202)+(-63.0*p012*p202)+(594.0*p11*p202)+(-162.0*p01*p11*p202)+(-135.0*p112*p202)+(-990.0*p20*p21)+(396.0*p01*p20*p21)+(594.0*p11*p20*p21)+(252.0*p202*p21)+(-90.0*p01*p202*p21)+(-162.0*p11*p202*p21)+(-21.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(-9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(3080.0*p30)+(-814.0*p01*p30)+(110.0*p012*p30)+(-2112.0*p11*p30)+(330.0*p01*p11*p30)+(396.0*p112*p30)+(-1050.0*p20*p30)+(288.0*p01*p20*p30)+(-30.0*p012*p20*p30)+(822.0*p11*p20*p30)+(-120.0*p01*p11*p20*p30)+(-162.0*p112*p20*p30)+(-3234.0*p21*p30)+(264.0*p01*p21*p30)+(990.0*p11*p21*p30)+(990.0*p20*p21*p30)+(-108.0*p01*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(990.0*p212*p30)+(-252.0*p20*p212*p30)+(-770.0*p302)+(82.0*p01*p302)+(-5.0*p012*p302)+(408.0*p11*p302)+(-30.0*p01*p11*p302)+(-63.0*p112*p302)+(1050.0*p21*p302)+(-42.0*p01*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(-2.0*p00*(-407.0+(-2640.0*p11)+(495.0*p112)+(111.0*p20)+(-30.0*p11*p20)+(-1221.0*p21)+(495.0*p11*p21)+(144.0*p20*p21)+(-54.0*p11*p20*p21)+(198.0*p212)+(-45.0*p20*p212)+(-3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(41.0*p30)+(-39.0*p11*p30)+(9.0*p112*p30)+(12.0*p21*p30)+(-9.0*p212*p30)+(14.0*p012*(-220.0+(45.0*p10)+(9.0*p20)+p30))+(p01*(4675.0+(-336.0*p20)+(330.0*p21)+(-84.0*p10*(14.0+p21))+(-55.0*p30)+(6.0*p21*p30)+(21.0*p11*(55.0+(4.0*p20)+p30))))))+(-6.0*p10*(-352.0+(123.0*p20)+(-297.0*p21)+(36.0*p20*p21)+(99.0*p212)+(-27.0*p20*p212)+(68.0*p30)+(-28.0*p21*p30)+(7.0*p012*(-55.0+(6.0*p20)+p30))+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30)))+(p01*(935.0+(3.0*p20*(-55.0+(6.0*p21)))+(-42.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 25:
p012 = p01*p01;
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(-42.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(1.0+(9.0*p11)+(5.0*p21)))))+(-9.0*p102*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p20*((5.0*p012)+(27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(p01*(7.0+(20.0*p11)+(18.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-1.0*((5.0*p012)+(p01*(28.0+(30.0*p11)+(42.0*p21)))+(21.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))))*p302)+(-6.0*p10*((3.0*p20*(-14.0+(14.0*p012)+(-6.0*p11)+(21.0*p01*p11)+(9.0*p112)+(-21.0*p21)+(6.0*p01*p21)+(-9.0*p212)))+((-42.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(1.0+(9.0*p11)+(5.0*p21))))*p30))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(-42.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(1.0+(9.0*p11)+(5.0*p21)))))+(9.0*p102*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(6.0*p20*(385.0+(165.0*p21)+(5.0*p012*(-22.0+p30))+(-210.0*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(p11*((-99.0*p21)+(28.0*p30)+(63.0*p21*p30)))+(p01*(-11.0+(-66.0*p21)+(7.0*p30)+(18.0*p21*p30)+(5.0*p11*(-33.0+(4.0*p30)))))))+(6.0*p10*(110.0+(-42.0*p20)+(165.0*p21)+(-63.0*p20*p21)+(99.0*p212)+(-27.0*p20*p212)+(-42.0*p30)+(-28.0*p21*p30)+(7.0*p012*(-55.0+(6.0*p20)+p30))+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-18.0*p20)+(99.0*p21)+(18.0*p21*p30)))+(p01*(11.0+(18.0*p20*p21)+(2.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))))+(p30*(6160.0+(-2310.0*p21)+(-990.0*p212)+(5.0*p012*(-22.0+p30))+(-2310.0*p30)+(1260.0*p21*p30)+(378.0*p212*p30)+(9.0*p112*(-44.0+(7.0*p30)))+(2.0*p01*(-55.0+(15.0*p11*(-11.0+p30))+(14.0*p30)+(3.0*p21*(-44.0+(7.0*p30)))))+(6.0*p11*(-110.0+(42.0*p30)+(3.0*p21*(-55.0+(14.0*p30))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(68.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21)))+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21)))))+(9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(6.0*p20*(175.0+(5.0*p012)+(27.0*p112)+(-165.0*p21)+(42.0*p212)+(2.0*p01*(-24.0+(10.0*p11)+(9.0*p21)))+(p11*(-137.0+(63.0*p21))))*p30)+(((5.0*p012)+(63.0*p112)+(12.0*p11*(-34.0+(21.0*p21)))+(p01*(-82.0+(30.0*p11)+(42.0*p21)))+(14.0*(55.0+(-75.0*p21)+(27.0*p212))))*p302)+(6.0*p10*((3.0*p20*(41.0+(14.0*p012)+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)+(p01*(-55.0+(21.0*p11)+(6.0*p21)))))+((68.0+(7.0*p012)+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21))))*p30))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(68.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21)))+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21)))))+(-9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p20*(-539.0+(165.0*p21)+(5.0*p012*(-22.0+p30))+(175.0*p30)+(-165.0*p21*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(p11*(462.0+(-99.0*p21)+(-137.0*p30)+(63.0*p21*p30)))+(p01*(451.0+(-66.0*p21)+(-48.0*p30)+(18.0*p21*p30)+(5.0*p11*(-33.0+(4.0*p30)))))))+(-6.0*p10*(-352.0+(123.0*p20)+(-297.0*p21)+(36.0*p20*p21)+(99.0*p212)+(-27.0*p20*p212)+(68.0*p30)+(-28.0*p21*p30)+(7.0*p012*(-55.0+(6.0*p20)+p30))+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30)))+(p01*(935.0+(3.0*p20*(-55.0+(6.0*p21)))+(-42.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))))+(-1.0*p30*(-3080.0+(3234.0*p21)+(-990.0*p212)+(5.0*p012*(-22.0+p30))+(770.0*p30)+(-1050.0*p21*p30)+(378.0*p212*p30)+(9.0*p112*(-44.0+(7.0*p30)))+(p01*(814.0+(30.0*p11*(-11.0+p30))+(-82.0*p30)+(6.0*p21*(-44.0+(7.0*p30)))))+(6.0*p11*(352.0+(-68.0*p30)+(3.0*p21*(-55.0+(14.0*p30))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 26:
p002 = p00*p00;
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(p002*(5.0+(-2310.0*p012)+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))+(28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(2.0*p00*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p20)+(30.0*p11*p20)+(-14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(21.0*p21)+(54.0*p20*p21)+(54.0*p11*p20*p21)+(9.0*p212)+(45.0*p20*p212)+(3.0*p01*((-7.0*p11)+(2.0*p20)+(-28.0*p11*p20)+(-2.0*p21)+(7.0*p10*(1.0+(4.0*p21)))))+(3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))))+(-2.0*p01*(14.0+(9.0*p102*(-1.0+(14.0*p11)))+(21.0*p20)+(9.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*((9.0*p11*(2.0+(7.0*p20)))+(2.0*(1.0+(5.0*p21)+(9.0*p20*p21)))))))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(p10*(84.0+(-6.0*p112*(5.0+(9.0*p20)))+(36.0*p11*(p20+(-1.0*p21)))+(56.0*p21)+(6.0*p20*(14.0+(21.0*p21)+(9.0*p212))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*(3850.0+(408.0*p10)+(-63.0*p102)+(-408.0*p11)+(396.0*p10*p11)+(-90.0*p102*p11)+(-333.0*p112)+(90.0*p10*p112)+(1050.0*p20)+(-252.0*p10*p20)+(168.0*p11*p20)+(-108.0*p10*p11*p20)+(-432.0*p112*p20)+(162.0*p10*p112*p20)+(-378.0*p202)+(135.0*p112*p202)+(21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-1050.0*p21)+(822.0*p10*p21)+(-162.0*p102*p21)+(-738.0*p11*p21)+(702.0*p10*p11*p21)+(-162.0*p102*p11*p21)+(990.0*p20*p21)+(-378.0*p10*p20*p21)+(-216.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(-612.0*p212)+(594.0*p10*p212)+(-135.0*p102*p212)+(252.0*p20*p212)+(-162.0*p10*p20*p212)+(p002*(-5.0+(2310.0*p012)+(-42.0*p11)+(-378.0*p112)+(-30.0*p21)+(-252.0*p11*p21)+(-63.0*p212)+(-28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(2.0*p01*(-41.0+(9.0*p102*(-1.0+(14.0*p11)))+(-12.0*p20)+(9.0*p202)+(3.0*p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-111.0*p21)+(-144.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(13.0+(21.0*p11*(-7.0+(3.0*p20)))+(2.0*(5.0+(9.0*p20))*p21)))))+(2.0*p00*(41.0+(126.0*p11)+(504.0*p112)+(-21.0*p20)+(-30.0*p11*p20)+(42.0*p012*(-73.0+(15.0*p10)+(3.0*p20)))+(144.0*p21)+(495.0*p11*p21)+(-54.0*p20*p21)+(-54.0*p11*p20*p21)+(189.0*p212)+(-45.0*p20*p212)+(p01*(55.0+(-6.0*p20)+(84.0*p11*(14.0+p20))+(336.0*p21)+(-21.0*p10*(1.0+(4.0*p21)))))+(-3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*(770.0+(408.0*p10)+(333.0*p102)+(-408.0*p11)+(-396.0*p10*p11)+(-90.0*p102*p11)+(63.0*p112)+(90.0*p10*p112)+(1050.0*p20)+(738.0*p10*p20)+(-822.0*p11*p20)+(-702.0*p10*p11*p20)+(162.0*p112*p20)+(162.0*p10*p112*p20)+(612.0*p202)+(-594.0*p11*p202)+(135.0*p112*p202)+(p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-1050.0*p21)+(-168.0*p10*p21)+(432.0*p102*p21)+(252.0*p11*p21)+(108.0*p10*p11*p21)+(-162.0*p102*p11*p21)+(-990.0*p20*p21)+(216.0*p10*p20*p21)+(378.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(378.0*p212)+(-135.0*p102*p212)+(252.0*p20*p212)+(-162.0*p10*p20*p212)+(21.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(2.0*p00*(41.0+(-39.0*p11)+(9.0*p112)+(111.0*p20)+(-30.0*p11*p20)+(14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(12.0*p21)+(144.0*p20*p21)+(-54.0*p11*p20*p21)+(-9.0*p212)+(-45.0*p20*p212)+(p01*(-55.0+(21.0*p11)+(-336.0*p20)+(84.0*p11*p20)+(6.0*p21)+(-84.0*p10*(14.0+p21))))+(-3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))))+(2.0*p01*(-41.0+(126.0*p102*(-4.0+p11))+(-144.0*p20)+(-189.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(-42.0+(9.0*p11*(2.0+(7.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/6160.0)*(770.0+(568.0*p10)+(-111.0*p102)+(-568.0*p11)+(30.0*p102*p11)+(111.0*p112)+(-30.0*p10*p112)+(728.0*p20)+(-246.0*p10*p20)+(-650.0*p11*p20)+(234.0*p10*p11*p20)+(144.0*p112*p20)+(-54.0*p10*p112*p20)+(-204.0*p202)+(198.0*p11*p202)+(-45.0*p112*p202)+(-7.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-728.0*p21)+(650.0*p10*p21)+(-144.0*p102*p21)+(246.0*p11*p21)+(-234.0*p10*p11*p21)+(54.0*p102*p11*p21)+(-72.0*p10*p20*p21)+(72.0*p11*p20*p21)+(84.0*p202*p21)+(-54.0*p11*p202*p21)+(204.0*p212)+(-198.0*p10*p212)+(45.0*p102*p212)+(-84.0*p20*p212)+(54.0*p10*p20*p212)+(-7.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(-2.0*p00*(-122.0+(-893.0*p11)+(168.0*p112)+(37.0*p20)+(-10.0*p11*p20)+(14.0*p012*(-73.0+(15.0*p10)+(3.0*p20)))+(-403.0*p21)+(165.0*p11*p21)+(48.0*p20*p21)+(-18.0*p11*p20*p21)+(63.0*p212)+(-15.0*p20*p212)+(28.0*p01*(55.0+(-4.0*p20)+(p11*(14.0+p20))+(4.0*p21)+(-1.0*p10*(14.0+p21))))+(p10*(50.0+(-42.0*p112)+(145.0*p21)+(-27.0*p212)+(-21.0*p11*(-7.0+(3.0*p21)))))))+(-2.0*p01*(122.0+(42.0*p102*(-4.0+p11))+(403.0*p20)+(-63.0*p202)+(p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-37.0*p21)+(-48.0*p20*p21)+(15.0*p202*p21)+(p10*(893.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 27:
p012 = p01*p01;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-2.0*p01*(14.0+(9.0*p102*(-1.0+(14.0*p11)))+(21.0*p20)+(9.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*((9.0*p11*(2.0+(7.0*p20)))+(2.0*(1.0+(5.0*p21)+(9.0*p20*p21)))))))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(p10*(84.0+(-6.0*p112*(5.0+(9.0*p20)))+(36.0*p11*(p20+(-1.0*p21)))+(56.0*p21)+(6.0*p20*(14.0+(21.0*p21)+(9.0*p212))))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*(3850.0+(-408.0*p11)+(-333.0*p112)+(1050.0*p20)+(168.0*p11*p20)+(-432.0*p112*p20)+(-378.0*p202)+(135.0*p112*p202)+(21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-1050.0*p21)+(-738.0*p11*p21)+(990.0*p20*p21)+(-216.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(-612.0*p212)+(252.0*p20*p212)+(-9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(2.0*p01*(-41.0+(9.0*p102*(-1.0+(14.0*p11)))+(-12.0*p20)+(9.0*p202)+(3.0*p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-111.0*p21)+(-144.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(13.0+(21.0*p11*(-7.0+(3.0*p20)))+(2.0*(5.0+(9.0*p20))*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(137.0*p21)+(99.0*p212)+(p11*(66.0+(-18.0*p20)+(117.0*p21)))+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012))*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*(770.0+(-408.0*p11)+(63.0*p112)+(1050.0*p20)+(-822.0*p11*p20)+(162.0*p112*p20)+(612.0*p202)+(-594.0*p11*p202)+(135.0*p112*p202)+(p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-1050.0*p21)+(252.0*p11*p21)+(-990.0*p20*p21)+(378.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(378.0*p212)+(252.0*p20*p212)+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(-3.0*p11*(22.0+(39.0*p20)+(-6.0*p21)))+(-28.0*p21)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(2.0*p01*(-41.0+(126.0*p102*(-4.0+p11))+(-144.0*p20)+(-189.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(-42.0+(9.0*p11*(2.0+(7.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/6160.0)*(770.0+(-568.0*p11)+(111.0*p112)+(728.0*p20)+(-650.0*p11*p20)+(144.0*p112*p20)+(-204.0*p202)+(198.0*p11*p202)+(-45.0*p112*p202)+(-7.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-728.0*p21)+(246.0*p11*p21)+(72.0*p11*p20*p21)+(84.0*p202*p21)+(-54.0*p11*p202*p21)+(204.0*p212)+(-84.0*p20*p212)+(3.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p10*(-284.0+(3.0*p112*(5.0+(9.0*p20)))+(-117.0*p11*(p20+(-1.0*p21)))+(-325.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(-2.0*p01*(122.0+(42.0*p102*(-4.0+p11))+(403.0*p20)+(-63.0*p202)+(p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-37.0*p21)+(-48.0*p20*p21)+(15.0*p202*p21)+(p10*(893.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01))*sign);
break;
case 28:
p002 = p00*p00;
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(2.0*p00*((3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))+(3.0*p20*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+((14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212))*p30)))+(3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(22.0*((105.0*p20)+(-27.0*p112*p20)+(45.0*p20*p21)+(-27.0*p11*p20*p21)+(3.0*p10*(10.0+(6.0*p11)+(15.0*p21)+(9.0*p11*p21)+(9.0*p212)))+(p00*(5.0+(45.0*p112)+(15.0*p21)+(18.0*p212)+(3.0*p11*(4.0+(15.0*p21)))))+(280.0*p30)+(-30.0*p11*p30)+(-18.0*p112*p30)+(-105.0*p21*p30)+(-45.0*p11*p21*p30)+(-45.0*p212*p30)))+(-2.0*p00*((3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))+(3.0*p20*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+((14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212))*p30)))+(-3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((612.0*p202)+(-594.0*p11*p202)+(135.0*p112*p202)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(-21.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(1050.0*p20*p30)+(-822.0*p11*p20*p30)+(162.0*p112*p20*p30)+(-990.0*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(252.0*p20*p212*p30)+(770.0*p302)+(-408.0*p11*p302)+(63.0*p112*p302)+(-1050.0*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(-2.0*p00*((3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(3.0*p20*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+((-41.0+(39.0*p11)+(-9.0*p112)+(-12.0*p21)+(9.0*p212))*p30)))+(6.0*p10*((3.0*p20*(41.0+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)))+((68.0+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21))))*p30))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((3234.0*p20)+(-2772.0*p11*p20)+(594.0*p112*p20)+(-612.0*p202)+(594.0*p11*p202)+(-135.0*p112*p202)+(-990.0*p20*p21)+(594.0*p11*p20*p21)+(252.0*p202*p21)+(-162.0*p11*p202*p21)+(21.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(3080.0*p30)+(-2112.0*p11*p30)+(396.0*p112*p30)+(-1050.0*p20*p30)+(822.0*p11*p20*p30)+(-162.0*p112*p20*p30)+(-3234.0*p21*p30)+(990.0*p11*p21*p30)+(990.0*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(990.0*p212*p30)+(-252.0*p20*p212*p30)+(-770.0*p302)+(408.0*p11*p302)+(-63.0*p112*p302)+(1050.0*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(2.0*p00*(407.0+(-111.0*p20)+(1221.0*p21)+(-144.0*p20*p21)+(-198.0*p212)+(45.0*p20*p212)+(3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(-41.0*p30)+(-12.0*p21*p30)+(9.0*p212*p30)+(-9.0*p112*(55.0+p30))+(3.0*p11*(880.0+(-165.0*p21)+(2.0*p20*(5.0+(9.0*p21)))+(13.0*p30)))))+(-6.0*p10*(-352.0+(-297.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))+(68.0*p30)+(-28.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 29:
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((1.0/6160.0)*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(-9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p10*(110.0+(165.0*p21)+(99.0*p212)+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))+(-42.0*p30)+(-28.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-18.0*p20)+(99.0*p21)+(18.0*p21*p30)))))+(6.0*p20*(385.0+(165.0*p21)+(-210.0*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(p11*((28.0*p30)+(9.0*p21*(-11.0+(7.0*p30)))))))+(p30*((-770.0*(-8.0+(3.0*p30)))+(210.0*p21*(-11.0+(6.0*p30)))+(9.0*p112*(-44.0+(7.0*p30)))+(18.0*p212*(-55.0+(21.0*p30)))+(6.0*p11*(-110.0+(42.0*p30)+(3.0*p21*(-55.0+(14.0*p30))))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(68.0+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p20*(175.0+(-137.0*p11)+(27.0*p112)+(-165.0*p21)+(63.0*p11*p21)+(42.0*p212))*p30)+((770.0+(-408.0*p11)+(63.0*p112)+(-1050.0*p21)+(252.0*p11*p21)+(378.0*p212))*p302)+(6.0*p10*((3.0*p20*(41.0+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)))+((68.0+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21))*p30))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(68.0+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21)))))+(9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p10*(-352.0+(-297.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))+(68.0*p30)+(-28.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30)))))+(-6.0*p20*((-165.0*p21*(-1.0+p30))+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(7.0*(-77.0+(25.0*p30)))+(p11*(462.0+(-137.0*p30)+(9.0*p21*(-11.0+(7.0*p30)))))))+(p30*((p212*(990.0+(-378.0*p30)))+(-770.0*(-4.0+p30))+(-9.0*p112*(-44.0+(7.0*p30)))+(42.0*p21*(-77.0+(25.0*p30)))+(-6.0*p11*(352.0+(-68.0*p30)+(3.0*p21*(-55.0+(14.0*p30))))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 30:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(2.0*p00*(14.0+(-9.0*p112)+(21.0*p20)+(21.0*p21)+(54.0*p20*p21)+(9.0*p212)+(45.0*p20*p212)+(6.0*p11*(1.0+(5.0*p20)+(9.0*p20*p21)))+(3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(p10*(84.0+(-6.0*p112*(5.0+(9.0*p20)))+(36.0*p11*(p20+(-1.0*p21)))+(56.0*p21)+(6.0*p20*(14.0+(21.0*p21)+(9.0*p212))))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*(3850.0+(-408.0*p11)+(-333.0*p112)+(1050.0*p20)+(168.0*p11*p20)+(-432.0*p112*p20)+(-378.0*p202)+(135.0*p112*p202)+(-1050.0*p21)+(-738.0*p11*p21)+(990.0*p20*p21)+(-216.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(-612.0*p212)+(252.0*p20*p212)+(-1.0*p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(-9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(137.0*p21)+(99.0*p212)+(p11*(66.0+(-18.0*p20)+(117.0*p21)))+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))+(-2.0*p00*(-41.0+(-504.0*p112)+(21.0*p20)+(-144.0*p21)+(54.0*p20*p21)+(-189.0*p212)+(45.0*p20*p212)+(3.0*p11*(-42.0+(10.0*p20)+(-165.0*p21)+(18.0*p20*p21)))+(3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*(770.0+(-408.0*p11)+(63.0*p112)+(1050.0*p20)+(-822.0*p11*p20)+(162.0*p112*p20)+(612.0*p202)+(-594.0*p11*p202)+(135.0*p112*p202)+(-1050.0*p21)+(252.0*p11*p21)+(-990.0*p20*p21)+(378.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(378.0*p212)+(252.0*p20*p212)+(-21.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(-3.0*p11*(22.0+(39.0*p20)+(-6.0*p21)))+(-28.0*p21)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(-2.0*p00*(-41.0+(-9.0*p112)+(-111.0*p20)+(-12.0*p21)+(-144.0*p20*p21)+(9.0*p212)+(45.0*p20*p212)+(3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(p11*(39.0+(6.0*p20*(5.0+(9.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/6160.0)*(770.0+(-568.0*p11)+(111.0*p112)+(728.0*p20)+(-650.0*p11*p20)+(144.0*p112*p20)+(-204.0*p202)+(198.0*p11*p202)+(-45.0*p112*p202)+(-728.0*p21)+(246.0*p11*p21)+(72.0*p11*p20*p21)+(84.0*p202*p21)+(-54.0*p11*p202*p21)+(204.0*p212)+(-84.0*p20*p212)+(7.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(3.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p10*(-284.0+(3.0*p112*(5.0+(9.0*p20)))+(-117.0*p11*(p20+(-1.0*p21)))+(-325.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(2.0*p00*(122.0+(-168.0*p112)+(-37.0*p20)+(403.0*p21)+(-48.0*p20*p21)+(-63.0*p212)+(15.0*p20*p212)+(p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(p11*(893.0+(-165.0*p21)+(2.0*p20*(5.0+(9.0*p21))))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 31:
x = 0; y = 0;
pixel[y][-x] += ((1.0/6160.0)*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(p10*(84.0+(-6.0*p112*(5.0+(9.0*p20)))+(36.0*p11*(p20+(-1.0*p21)))+(56.0*p21)+(6.0*p20*(14.0+(21.0*p21)+(9.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)))+(6.0*p11*(-68.0+(-123.0*p21)+(27.0*p202*p21)+(-4.0*p20*(-7.0+(9.0*p21)))))+(-9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(137.0*p21)+(99.0*p212)+(p11*(66.0+(-18.0*p20)+(117.0*p21)))+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))+(-2.0*(-1925.0+(525.0*p21)+(306.0*p212)+(63.0*p202*(3.0+(2.0*p21)))+(-3.0*p20*(175.0+(165.0*p21)+(42.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*(770.0+(1050.0*p20)+(612.0*p202)+(9.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-1050.0*p21)+(-990.0*p20*p21)+(-252.0*p202*p21)+(378.0*p212)+(252.0*p20*p212)+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p11*(-68.0+(42.0*p21)+(9.0*p202*(-11.0+(3.0*p21)))+(p20*(-137.0+(63.0*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(-3.0*p11*(22.0+(39.0*p20)+(-6.0*p21)))+(-28.0*p21)+(p20*(123.0+(36.0*p21)+(-27.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/6160.0)*(770.0+(728.0*p20)+(-204.0*p202)+(p112*(111.0+(144.0*p20)+(-45.0*p202)))+(-728.0*p21)+(84.0*p202*p21)+(204.0*p212)+(-84.0*p20*p212)+(3.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(p11*(-568.0+(p202*(198.0+(-54.0*p21)))+(246.0*p21)+(p20*(-650.0+(72.0*p21)))))+(-2.0*p10*(-284.0+(3.0*p112*(5.0+(9.0*p20)))+(-117.0*p11*(p20+(-1.0*p21)))+(-325.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212))))))*sign);
propagate[x][y] += ((1.0/4.0)*sign);
break;
case 32:
p002 = p00*p00;
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((1.0/6160.0)*((99.0*p202)+(-54.0*p11*p202)+(-45.0*p112*p202)+(54.0*p202*p21)+(-54.0*p11*p202*p21)+(9.0*p102*(-1.0+p21)*(11.0+(6.0*p11)+(5.0*p21)))+(p002*(-759.0+(126.0*p112)+(94.0*p21)+(21.0*p212)+(14.0*p11*(31.0+(6.0*p21)))))+(396.0*p20*p30)+(-96.0*p11*p20*p30)+(-54.0*p112*p20*p30)+(-36.0*p20*p21*p30)+(-126.0*p11*p20*p21*p30)+(-84.0*p20*p212*p30)+(759.0*p302)+(-94.0*p11*p302)+(-21.0*p112*p302)+(-434.0*p21*p302)+(-84.0*p11*p21*p302)+(-126.0*p212*p302)+(2.0*p00*((3.0*p20*(-1.0+p21)*(11.0+(6.0*p11)+(5.0*p21)))+(3.0*p10*(-66.0+(14.0*p112)+(16.0*p21)+(9.0*p212)+(3.0*p11*(2.0+(7.0*p21)))))+(((-5.0*p11)+(-3.0*p112)+(p21*(5.0+(3.0*p21))))*p30)))+(-6.0*p10*((-3.0*p20*p21*(5.0+(3.0*p21)))+(-1.0*(11.0+(6.0*p21))*p30)+(p112*((9.0*p20)+(5.0*p30)))+(3.0*p11*((5.0*p20)+(2.0*(1.0+p21)*p30))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((1.0/6160.0)*((528.0*p20)+(-330.0*p11*p20)+(-198.0*p112*p20)+(-99.0*p202)+(54.0*p11*p202)+(45.0*p112*p202)+(198.0*p20*p21)+(-198.0*p11*p20*p21)+(-54.0*p202*p21)+(54.0*p11*p202*p21)+(-9.0*p102*(-1.0+p21)*(11.0+(6.0*p11)+(5.0*p21)))+(-1.0*p002*(-759.0+(126.0*p112)+(94.0*p21)+(21.0*p212)+(14.0*p11*(31.0+(6.0*p21)))))+(1980.0*p30)+(-330.0*p11*p30)+(-132.0*p112*p30)+(-396.0*p20*p30)+(96.0*p11*p20*p30)+(54.0*p112*p20*p30)+(-858.0*p21*p30)+(-330.0*p11*p21*p30)+(36.0*p20*p21*p30)+(126.0*p11*p20*p21*p30)+(-330.0*p212*p30)+(84.0*p20*p212*p30)+(-759.0*p302)+(94.0*p11*p302)+(21.0*p112*p302)+(434.0*p21*p302)+(84.0*p11*p21*p302)+(126.0*p212*p302)+(-2.0*p00*(990.0+(-33.0*p20)+(-165.0*p21)+(18.0*p20*p21)+(-66.0*p212)+(15.0*p20*p212)+(3.0*p10*(-66.0+(14.0*p112)+(16.0*p21)+(9.0*p212)+(3.0*p11*(2.0+(7.0*p21)))))+(5.0*p21*p30)+(3.0*p212*p30)+(-3.0*p112*(55.0+p30))+(-1.0*p11*(429.0+(-18.0*p20*(-1.0+p21))+(165.0*p21)+(5.0*p30)))))+(6.0*p10*(((33.0+(-9.0*p20))*p212)+(p21*(55.0+(-15.0*p20)+(-6.0*p30)))+(-11.0*(8.0+p30))+(p112*((9.0*p20)+(5.0*p30)))+(3.0*p11*(-11.0+(5.0*p20)+(11.0*p21)+(2.0*p30)+(2.0*p21*p30))))))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((3.0/6160.0)*((33.0*p202)+(-48.0*p11*p202)+(15.0*p112*p202)+(-18.0*p202*p21)+(18.0*p11*p202*p21)+(-3.0*p102*(-1.0+p21)*(-11.0+(6.0*p11)+(5.0*p21)))+(-7.0*p002*(11.0+(6.0*p112)+(4.0*p11*(-4.0+p21))+(-6.0*p21)+p212))+(88.0*p20*p30)+(-78.0*p11*p20*p30)+(18.0*p112*p20*p30)+(-98.0*p20*p21*p30)+(42.0*p11*p20*p21*p30)+(28.0*p20*p212*p30)+(77.0*p302)+(-42.0*p11*p302)+(7.0*p112*p302)+(-112.0*p21*p302)+(28.0*p11*p21*p302)+(42.0*p212*p302)+(-2.0*p00*((p20*(-1.0+p21)*(-11.0+(6.0*p11)+(5.0*p21)))+(p10*(44.0+(14.0*p112)+(-39.0*p21)+(9.0*p212)+(7.0*p11*(-7.0+(3.0*p21)))))+(((2.0*p11)+(-1.0*p112)+((-2.0+p21)*p21))*p30)))+(2.0*p10*((-9.0*p20*(-2.0+p21)*p21)+((11.0+(-6.0*p21))*p30)+(p112*((9.0*p20)+(5.0*p30)))+(-2.0*p11*((9.0*p20)+((8.0+(-3.0*p21))*p30))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((3.0/6160.0)*((132.0*p20)+(-198.0*p11*p20)+(66.0*p112*p20)+(-33.0*p202)+(48.0*p11*p202)+(-15.0*p112*p202)+(-66.0*p20*p21)+(66.0*p11*p20*p21)+(18.0*p202*p21)+(-18.0*p11*p202*p21)+(3.0*p102*(-1.0+p21)*(-11.0+(6.0*p11)+(5.0*p21)))+(7.0*p002*(11.0+(6.0*p112)+(4.0*p11*(-4.0+p21))+(-6.0*p21)+p212))+(264.0*p30)+(-198.0*p11*p30)+(44.0*p112*p30)+(-88.0*p20*p30)+(78.0*p11*p20*p30)+(-18.0*p112*p20*p30)+(-330.0*p21*p30)+(110.0*p11*p21*p30)+(98.0*p20*p21*p30)+(-42.0*p11*p20*p21*p30)+(110.0*p212*p30)+(-28.0*p20*p212*p30)+(-77.0*p302)+(42.0*p11*p302)+(-7.0*p112*p302)+(112.0*p21*p302)+(-28.0*p11*p21*p302)+(-42.0*p212*p302)+(2.0*p00*(-132.0+(11.0*p20)+(99.0*p21)+(-16.0*p20*p21)+(-22.0*p212)+(5.0*p20*p212)+(p10*(44.0+(14.0*p112)+(-39.0*p21)+(9.0*p212)+(7.0*p11*(-7.0+(3.0*p21)))))+(-2.0*p21*p30)+(p212*p30)+(-1.0*p112*(55.0+p30))+(p11*(165.0+(6.0*p20*(-1.0+p21))+(-55.0*p21)+(2.0*p30)))))+(-2.0*p10*(((33.0+(-9.0*p20))*p212)+(3.0*p21*(-33.0+(6.0*p20)+(-2.0*p30)))+(11.0*(6.0+p30))+(p112*((9.0*p20)+(5.0*p30)))+(p11*(-33.0+(-18.0*p20)+(33.0*p21)+(-16.0*p30)+(6.0*p21*p30))))))*sign);
propagate[x][y] += 0.0;
break;
case 33:
p302 = p30*p30;
x = 0; y = 0;
pixel[y][-x] += ((1.0/6160.0)*((9.0*p102*(-1.0+p21)*(11.0+(6.0*p11)+(5.0*p21)))+(-9.0*(-1.0+p11)*p202*(11.0+(5.0*p11)+(6.0*p21)))+(-6.0*p20*(-66.0+(9.0*p112)+(6.0*p21)+(14.0*p212)+(p11*(16.0+(21.0*p21))))*p30)+(-1.0*(-759.0+(21.0*p112)+(434.0*p21)+(126.0*p212)+(p11*(94.0+(84.0*p21))))*p302)+(-6.0*p10*((-3.0*p20*p21*(5.0+(3.0*p21)))+(-1.0*(11.0+(6.0*p21))*p30)+(p112*((9.0*p20)+(5.0*p30)))+(3.0*p11*((5.0*p20)+(2.0*(1.0+p21)*p30))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((1.0/6160.0)*((-9.0*p102*(-1.0+p21)*(11.0+(6.0*p11)+(5.0*p21)))+(9.0*(-1.0+p11)*p202*(11.0+(5.0*p11)+(6.0*p21)))+(6.0*p10*(((33.0+(-9.0*p20))*p212)+(p21*(55.0+(-15.0*p20)+(-6.0*p30)))+(-11.0*(8.0+p30))+(p112*((9.0*p20)+(5.0*p30)))+(3.0*p11*(-11.0+(5.0*p20)+(11.0*p21)+(2.0*p30)+(2.0*p21*p30)))))+(6.0*p20*(88.0+(-66.0*p30)+(14.0*p212*p30)+(p21*(33.0+(6.0*p30)))+(p112*(-33.0+(9.0*p30)))+(p11*(-55.0+(16.0*p30)+(3.0*p21*(-11.0+(7.0*p30)))))))+(p30*(1980.0+(-759.0*p30)+(3.0*p112*(-44.0+(7.0*p30)))+(6.0*p212*(-55.0+(21.0*p30)))+(p21*(-858.0+(434.0*p30)))+(p11*(-330.0+(94.0*p30)+(6.0*p21*(-55.0+(14.0*p30))))))))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((-3.0/6160.0)*((3.0*p102*(-1.0+p21)*(-11.0+(6.0*p11)+(5.0*p21)))+(-3.0*(-1.0+p11)*p202*(-11.0+(5.0*p11)+(6.0*p21)))+(-2.0*p20*(44.0+(9.0*p112)+(-49.0*p21)+(14.0*p212)+(3.0*p11*(-13.0+(7.0*p21))))*p30)+(-7.0*(11.0+p112+(-16.0*p21)+(6.0*p212)+(p11*(-6.0+(4.0*p21))))*p302)+(-2.0*p10*((-9.0*p20*(-2.0+p21)*p21)+((11.0+(-6.0*p21))*p30)+(p112*((9.0*p20)+(5.0*p30)))+(-2.0*p11*((9.0*p20)+((8.0+(-3.0*p21))*p30))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((3.0/6160.0)*((3.0*p102*(-1.0+p21)*(-11.0+(6.0*p11)+(5.0*p21)))+(-3.0*(-1.0+p11)*p202*(-11.0+(5.0*p11)+(6.0*p21)))+(-2.0*p10*(((33.0+(-9.0*p20))*p212)+(3.0*p21*(-33.0+(6.0*p20)+(-2.0*p30)))+(11.0*(6.0+p30))+(p112*((9.0*p20)+(5.0*p30)))+(p11*(-33.0+(-18.0*p20)+(33.0*p21)+(-16.0*p30)+(6.0*p21*p30)))))+(-2.0*p20*(-66.0+(p21*(33.0+(-49.0*p30)))+(44.0*p30)+(14.0*p212*p30)+(p112*(-33.0+(9.0*p30)))+(3.0*p11*(33.0+(-13.0*p30)+(p21*(-11.0+(7.0*p30)))))))+(p30*(264.0+(p212*(110.0+(-42.0*p30)))+(p112*(44.0+(-7.0*p30)))+(-77.0*p30)+(2.0*p21*(-165.0+(56.0*p30)))+(-2.0*p11*(99.0+(-21.0*p30)+(p21*(-55.0+(14.0*p30))))))))*sign);
propagate[x][y] += 0.0;
break;
case 34:
p002 = p00*p00;
x = 0; y = 0;
pixel[y][-x] += ((1.0/6160.0)*(759.0+(-94.0*p11)+(-21.0*p112)+(396.0*p20)+(-96.0*p11*p20)+(-54.0*p112*p20)+(99.0*p202)+(-54.0*p11*p202)+(-45.0*p112*p202)+(-434.0*p21)+(-84.0*p11*p21)+(-36.0*p20*p21)+(-126.0*p11*p20*p21)+(54.0*p202*p21)+(-54.0*p11*p202*p21)+(-126.0*p212)+(-84.0*p20*p212)+(9.0*p102*(-1.0+p21)*(11.0+(6.0*p11)+(5.0*p21)))+(-6.0*p10*(-11.0+(p112*(5.0+(9.0*p20)))+(-3.0*(2.0+(5.0*p20))*p21)+(-9.0*p20*p212)+(3.0*p11*(2.0+(5.0*p20)+(2.0*p21)))))+(p002*(-759.0+(126.0*p112)+(94.0*p21)+(21.0*p212)+(14.0*p11*(31.0+(6.0*p21)))))+(2.0*p00*((-3.0*p112)+(-33.0*p20)+(p11*(-5.0+(18.0*p20*(-1.0+p21))))+(5.0*p21)+(18.0*p20*p21)+(3.0*p212)+(15.0*p20*p212)+(3.0*p10*(-66.0+(14.0*p112)+(16.0*p21)+(9.0*p212)+(3.0*p11*(2.0+(7.0*p21))))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((1.0/6160.0)*(1221.0+(-236.0*p11)+(-111.0*p112)+(132.0*p20)+(-234.0*p11*p20)+(-144.0*p112*p20)+(-99.0*p202)+(54.0*p11*p202)+(45.0*p112*p202)+(-424.0*p21)+(-246.0*p11*p21)+(234.0*p20*p21)+(-72.0*p11*p20*p21)+(-54.0*p202*p21)+(54.0*p11*p202*p21)+(-204.0*p212)+(84.0*p20*p212)+(-9.0*p102*(-1.0+p21)*(11.0+(6.0*p11)+(5.0*p21)))+(-1.0*p002*(-759.0+(126.0*p112)+(94.0*p21)+(21.0*p212)+(14.0*p11*(31.0+(6.0*p21)))))+(6.0*p10*(-99.0+(p112*(5.0+(9.0*p20)))+((49.0+(-15.0*p20))*p21)+((33.0+(-9.0*p20))*p212)+(3.0*p11*(-9.0+(5.0*p20)+(13.0*p21)))))+(-2.0*p00*(990.0+(-168.0*p112)+(-33.0*p20)+(-160.0*p21)+(18.0*p20*p21)+(-63.0*p212)+(15.0*p20*p212)+(-1.0*p11*(434.0+(-18.0*p20*(-1.0+p21))+(165.0*p21)))+(3.0*p10*(-66.0+(14.0*p112)+(16.0*p21)+(9.0*p212)+(3.0*p11*(2.0+(7.0*p21))))))))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((3.0/6160.0)*(77.0+(-42.0*p11)+(7.0*p112)+(88.0*p20)+(-78.0*p11*p20)+(18.0*p112*p20)+(33.0*p202)+(-48.0*p11*p202)+(15.0*p112*p202)+(-112.0*p21)+(28.0*p11*p21)+(-98.0*p20*p21)+(42.0*p11*p20*p21)+(-18.0*p202*p21)+(18.0*p11*p202*p21)+(42.0*p212)+(28.0*p20*p212)+(-3.0*p102*(-1.0+p21)*(-11.0+(6.0*p11)+(5.0*p21)))+(-7.0*p002*(11.0+(6.0*p112)+(4.0*p11*(-4.0+p21))+(-6.0*p21)+p212))+(2.0*p10*(11.0+(p112*(5.0+(9.0*p20)))+(-2.0*p11*(8.0+(9.0*p20)+(-3.0*p21)))+(6.0*(-1.0+(3.0*p20))*p21)+(-9.0*p20*p212)))+(-2.0*p00*((-1.0*p112)+(11.0*p20)+(p11*(2.0+(6.0*p20*(-1.0+p21))))+(-2.0*p21)+(-16.0*p20*p21)+p212+(5.0*p20*p212)+(p10*(44.0+(14.0*p112)+(-39.0*p21)+(9.0*p212)+(7.0*p11*(-7.0+(3.0*p21))))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((3.0/6160.0)*(187.0+(-156.0*p11)+(37.0*p112)+(44.0*p20)+(-120.0*p11*p20)+(48.0*p112*p20)+(-33.0*p202)+(48.0*p11*p202)+(-15.0*p112*p202)+(-218.0*p21)+(82.0*p11*p21)+(32.0*p20*p21)+(24.0*p11*p20*p21)+(18.0*p202*p21)+(-18.0*p11*p202*p21)+(68.0*p212)+(-28.0*p20*p212)+(3.0*p102*(-1.0+p21)*(-11.0+(6.0*p11)+(5.0*p21)))+(7.0*p002*(11.0+(6.0*p112)+(4.0*p11*(-4.0+p21))+(-6.0*p21)+p212))+(-2.0*p10*(77.0+(p112*(5.0+(9.0*p20)))+(3.0*(-35.0+(6.0*p20))*p21)+((33.0+(-9.0*p20))*p212)+(p11*(-49.0+(-18.0*p20)+(39.0*p21)))))+(2.0*p00*(-132.0+(-56.0*p112)+(11.0*p20)+(p11*(167.0+(6.0*p20*(-1.0+p21))+(-55.0*p21)))+(97.0*p21)+(-16.0*p20*p21)+(-21.0*p212)+(5.0*p20*p212)+(p10*(44.0+(14.0*p112)+(-39.0*p21)+(9.0*p212)+(7.0*p11*(-7.0+(3.0*p21))))))))*sign);
propagate[x][y] += 0.0;
break;
case 35:
x = 0; y = 0;
pixel[y][-x] += ((1.0/6160.0)*(759.0+(396.0*p20)+(99.0*p202)+(-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-434.0*p21)+(-36.0*p20*p21)+(54.0*p202*p21)+(-126.0*p212)+(-84.0*p20*p212)+(9.0*p102*(-1.0+p21)*(11.0+(6.0*p11)+(5.0*p21)))+(-6.0*p10*(-11.0+(p112*(5.0+(9.0*p20)))+(-3.0*(2.0+(5.0*p20))*p21)+(-9.0*p20*p212)+(3.0*p11*(2.0+(5.0*p20)+(2.0*p21)))))+(-2.0*p11*(47.0+(42.0*p21)+(27.0*p202*(1.0+p21))+(p20*(48.0+(63.0*p21))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 0;
pixel[y][-x] += ((1.0/6160.0)*(1221.0+(132.0*p20)+(-99.0*p202)+(3.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)))+(-424.0*p21)+(234.0*p20*p21)+(-54.0*p202*p21)+(-204.0*p212)+(84.0*p20*p212)+(-9.0*p102*(-1.0+p21)*(11.0+(6.0*p11)+(5.0*p21)))+(p11*(-236.0+(-246.0*p21)+(54.0*p202*(1.0+p21))+(-18.0*p20*(13.0+(4.0*p21)))))+(6.0*p10*(-99.0+(p112*(5.0+(9.0*p20)))+((49.0+(-15.0*p20))*p21)+((33.0+(-9.0*p20))*p212)+(3.0*p11*(-9.0+(5.0*p20)+(13.0*p21))))))*sign);
propagate[x][y] += 0.0;
x = 0; y = 1;
pixel[y][-x] += ((3.0/6160.0)*(77.0+(88.0*p20)+(33.0*p202)+(p112*(7.0+(18.0*p20)+(15.0*p202)))+(-112.0*p21)+(-98.0*p20*p21)+(-18.0*p202*p21)+(42.0*p212)+(28.0*p20*p212)+(-3.0*p102*(-1.0+p21)*(-11.0+(6.0*p11)+(5.0*p21)))+(2.0*p10*(11.0+(p112*(5.0+(9.0*p20)))+(-2.0*p11*(8.0+(9.0*p20)+(-3.0*p21)))+(-6.0*p21)+(18.0*p20*p21)+(-9.0*p20*p212)))+(2.0*p11*((7.0*(-3.0+(2.0*p21)))+(3.0*p202*(-8.0+(3.0*p21)))+(3.0*p20*(-13.0+(7.0*p21))))))*sign);
propagate[x][y] += 0.0;
x = 1; y = 1;
pixel[y][-x] += ((3.0/6160.0)*(187.0+(44.0*p20)+(-33.0*p202)+(p112*(37.0+(48.0*p20)+(-15.0*p202)))+(-218.0*p21)+(32.0*p20*p21)+(18.0*p202*p21)+(68.0*p212)+(-28.0*p20*p212)+(3.0*p102*(-1.0+p21)*(-11.0+(6.0*p11)+(5.0*p21)))+(-2.0*p11*(78.0+(-12.0*p20*(-5.0+p21))+(-41.0*p21)+(3.0*p202*(-8.0+(3.0*p21)))))+(-2.0*p10*(77.0+(p112*(5.0+(9.0*p20)))+(3.0*(-35.0+(6.0*p20))*p21)+((33.0+(-9.0*p20))*p212)+(p11*(-49.0+(-18.0*p20)+(39.0*p21))))))*sign);
propagate[x][y] += 0.0;
break;
}
#elif 1
			int special = 0;
			if (c.flip_x != c.flip_y) special += 1;
			if (p00 == 0) special += 2;
			if (p01 == 0) special += 4;
			if (p30 == 1) special += 8;
			if (p31 == 1) special += 16;

			switch (special)
			{
			case 0:
				p002 = p00*p00;
				p012 = p01*p01;
				p302 = p30*p30;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(-2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-1.0*p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(-22.0*((27.0*p112*p20)+(-27.0*p10*p11*p21)+(27.0*p11*p20*p21)+(-27.0*p10*p212)+(18.0*p112*p30)+(45.0*p11*p21*p30)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(-18.0*p10*p11*p31)+(-45.0*p10*p21*p31)+(-45.0*p20*p21*p31)+(30.0*p11*p30*p31)+(105.0*p21*p30*p31)+(-30.0*p10*p312)+(-105.0*p20*p312)+(-280.0*p30*p312)+(p01*((45.0*p10*p11)+(45.0*p11*p20)+(18.0*p20*p21)+(15.0*p11*p30)+(12.0*p21*p30)+(-3.0*p10*p31)+(3.0*p20*p31)+(5.0*p30*p31)))+(p00*((280.0*p012)+(-45.0*p112)+(-45.0*p11*p21)+(-18.0*p212)+(-12.0*p11*p31)+(-15.0*p21*p31)+(-5.0*p312)+(-5.0*p01*((21.0*p11)+(6.0*p21)+p31))))))+(2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-1.0*p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-22.0*((45.0*p01*p102)+(45.0*p01*p10*p20)+(27.0*p10*p11*p20)+(18.0*p01*p202)+(27.0*p11*p202)+(-27.0*p102*p21)+(-27.0*p10*p20*p21)+(12.0*p01*p10*p30)+(18.0*p10*p11*p30)+(15.0*p01*p20*p30)+(45.0*p11*p20*p30)+(45.0*p20*p21*p30)+(5.0*p01*p302)+(30.0*p11*p302)+(105.0*p21*p302)+(5.0*p002*((28.0*p01)+(-21.0*p11)+(-6.0*p21)+(-1.0*p31)))+(-18.0*p102*p31)+(-45.0*p10*p20*p31)+(-45.0*p202*p31)+(-30.0*p10*p30*p31)+(-105.0*p20*p30*p31)+(-140.0*p302*p31)+(p00*((-18.0*p20*p21)+(3.0*p11*p30)+(-3.0*p21*p30)+(5.0*p01*((21.0*p10)+(6.0*p20)+p30))+(-12.0*p20*p31)+(-5.0*p30*p31)+(-15.0*p10*((3.0*p11)+(3.0*p21)+p31))))))+(2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-22.0*((126.0*p11*p20)+(-27.0*p10*p11*p20)+(-27.0*p11*p202)+(-126.0*p10*p21)+(27.0*p102*p21)+(27.0*p10*p20*p21)+(126.0*p11*p30)+(-18.0*p10*p11*p30)+(-45.0*p11*p20*p30)+(252.0*p21*p30)+(-45.0*p20*p21*p30)+(-30.0*p11*p302)+(-105.0*p21*p302)+(-1.0*p01*((45.0*p102)+(18.0*p202)+(3.0*p10*(-84.0+(15.0*p20)+(4.0*p30)))+(3.0*p20*(-42.0+(5.0*p30)))+(p30*(-42.0+(5.0*p30)))))+(-126.0*p10*p31)+(18.0*p102*p31)+(-252.0*p20*p31)+(45.0*p10*p20*p31)+(45.0*p202*p31)+(-420.0*p30*p31)+(30.0*p10*p30*p31)+(105.0*p20*p30*p31)+(140.0*p302*p31)+(5.0*p002*((-28.0*p01)+(21.0*p11)+(6.0*p21)+p31))+(p00*((-126.0*p21)+(45.0*p10*p21)+(18.0*p20*p21)+(3.0*p11*(-84.0+(15.0*p10)+(-1.0*p30)))+(3.0*p21*p30)+(-5.0*p01*(-84.0+(21.0*p10)+(6.0*p20)+p30))+(-42.0*p31)+(15.0*p10*p31)+(12.0*p20*p31)+(5.0*p30*p31)))))+(-2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(22.0*((27.0*p112*p20)+(-27.0*p10*p11*p21)+(27.0*p11*p20*p21)+(-27.0*p10*p212)+(18.0*p112*p30)+(45.0*p11*p21*p30)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(-18.0*p10*p11*p31)+(-45.0*p10*p21*p31)+(-45.0*p20*p21*p31)+(30.0*p11*p30*p31)+(105.0*p21*p30*p31)+(-30.0*p10*p312)+(-105.0*p20*p312)+(-280.0*p30*p312)+(p01*((45.0*p10*p11)+(45.0*p11*p20)+(18.0*p20*p21)+(15.0*p11*p30)+(12.0*p21*p30)+(-3.0*p10*p31)+(3.0*p20*p31)+(5.0*p30*p31)))+(p00*((280.0*p012)+(-45.0*p112)+(-45.0*p11*p21)+(-18.0*p212)+(-12.0*p11*p31)+(-15.0*p21*p31)+(-5.0*p312)+(-5.0*p01*((21.0*p11)+(6.0*p21)+p31))))))+(-2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31)));
				break;
			case 1:
				p002 = p00*p00;
				p012 = p01*p01;
				p302 = p30*p30;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-1.0*p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*(p012+(-1.0*p312)));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(22.0*((27.0*p112*p20)+(-27.0*p10*p11*p21)+(27.0*p11*p20*p21)+(-27.0*p10*p212)+(18.0*p112*p30)+(45.0*p11*p21*p30)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(-18.0*p10*p11*p31)+(-45.0*p10*p21*p31)+(-45.0*p20*p21*p31)+(30.0*p11*p30*p31)+(105.0*p21*p30*p31)+(-30.0*p10*p312)+(-105.0*p20*p312)+(-280.0*p30*p312)+(p01*((45.0*p10*p11)+(45.0*p11*p20)+(18.0*p20*p21)+(15.0*p11*p30)+(12.0*p21*p30)+(-3.0*p10*p31)+(3.0*p20*p31)+(5.0*p30*p31)))+(p00*((280.0*p012)+(-45.0*p112)+(-45.0*p11*p21)+(-18.0*p212)+(-12.0*p11*p31)+(-15.0*p21*p31)+(-5.0*p312)+(-5.0*p01*((21.0*p11)+(6.0*p21)+p31))))))+(-2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*(p012+(-1.0*p312)));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(22.0*((45.0*p01*p102)+(45.0*p01*p10*p20)+(27.0*p10*p11*p20)+(18.0*p01*p202)+(27.0*p11*p202)+(-27.0*p102*p21)+(-27.0*p10*p20*p21)+(12.0*p01*p10*p30)+(18.0*p10*p11*p30)+(15.0*p01*p20*p30)+(45.0*p11*p20*p30)+(45.0*p20*p21*p30)+(5.0*p01*p302)+(30.0*p11*p302)+(105.0*p21*p302)+(5.0*p002*((28.0*p01)+(-21.0*p11)+(-6.0*p21)+(-1.0*p31)))+(-18.0*p102*p31)+(-45.0*p10*p20*p31)+(-45.0*p202*p31)+(-30.0*p10*p30*p31)+(-105.0*p20*p30*p31)+(-140.0*p302*p31)+(p00*((-18.0*p20*p21)+(3.0*p11*p30)+(-3.0*p21*p30)+(5.0*p01*((21.0*p10)+(6.0*p20)+p30))+(-12.0*p20*p31)+(-5.0*p30*p31)+(-15.0*p10*((3.0*p11)+(3.0*p21)+p31))))))+(-2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(-2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*((2.0*p01)+(-1.0*p012)+((-2.0+p31)*p31)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-1.0*p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(22.0*((126.0*p11*p20)+(-27.0*p10*p11*p20)+(-27.0*p11*p202)+(-126.0*p10*p21)+(27.0*p102*p21)+(27.0*p10*p20*p21)+(126.0*p11*p30)+(-18.0*p10*p11*p30)+(-45.0*p11*p20*p30)+(252.0*p21*p30)+(-45.0*p20*p21*p30)+(-30.0*p11*p302)+(-105.0*p21*p302)+(-1.0*p01*((45.0*p102)+(18.0*p202)+(3.0*p10*(-84.0+(15.0*p20)+(4.0*p30)))+(3.0*p20*(-42.0+(5.0*p30)))+(p30*(-42.0+(5.0*p30)))))+(-126.0*p10*p31)+(18.0*p102*p31)+(-252.0*p20*p31)+(45.0*p10*p20*p31)+(45.0*p202*p31)+(-420.0*p30*p31)+(30.0*p10*p30*p31)+(105.0*p20*p30*p31)+(140.0*p302*p31)+(5.0*p002*((-28.0*p01)+(21.0*p11)+(6.0*p21)+p31))+(p00*((-126.0*p21)+(45.0*p10*p21)+(18.0*p20*p21)+(3.0*p11*(-84.0+(15.0*p10)+(-1.0*p30)))+(3.0*p21*p30)+(-5.0*p01*(-84.0+(21.0*p10)+(6.0*p20)+p30))+(-42.0*p31)+(15.0*p10*p31)+(12.0*p20*p31)+(5.0*p30*p31)))))+(2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(-22.0*((27.0*p112*p20)+(-27.0*p10*p11*p21)+(27.0*p11*p20*p21)+(-27.0*p10*p212)+(18.0*p112*p30)+(45.0*p11*p21*p30)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(-18.0*p10*p11*p31)+(-45.0*p10*p21*p31)+(-45.0*p20*p21*p31)+(30.0*p11*p30*p31)+(105.0*p21*p30*p31)+(-30.0*p10*p312)+(-105.0*p20*p312)+(-280.0*p30*p312)+(p01*((45.0*p10*p11)+(45.0*p11*p20)+(18.0*p20*p21)+(15.0*p11*p30)+(12.0*p21*p30)+(-3.0*p10*p31)+(3.0*p20*p31)+(5.0*p30*p31)))+(p00*((280.0*p012)+(-45.0*p112)+(-45.0*p11*p21)+(-18.0*p212)+(-12.0*p11*p31)+(-15.0*p21*p31)+(-5.0*p312)+(-5.0*p01*((21.0*p11)+(6.0*p21)+p31))))))+(2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*((2.0*p01)+(-1.0*p012)+((-2.0+p31)*p31)));
				break;
			case 2:
				p012 = p01*p01;
				p302 = p30*p30;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-22.0*((-27.0*p10*p212)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(9.0*p112*((3.0*p20)+(2.0*p30)))+(-45.0*p10*p21*p31)+(-45.0*p20*p21*p31)+(105.0*p21*p30*p31)+(-30.0*p10*p312)+(-105.0*p20*p312)+(-280.0*p30*p312)+(p01*((18.0*p20*p21)+(12.0*p21*p30)+(15.0*p11*((3.0*p20)+p30))+(p10*((45.0*p11)+(-3.0*p31)))+(3.0*p20*p31)+(5.0*p30*p31)))+(3.0*p11*((-9.0*p10*p21)+(9.0*p20*p21)+(15.0*p21*p30)+(-6.0*p10*p31)+(10.0*p30*p31)))))+(2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-594.0*p11*p202)+(135.0*p112*p202)+(162.0*p11*p202*p21)+(-990.0*p11*p20*p30)+(162.0*p112*p20*p30)+(-990.0*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(252.0*p20*p212*p30)+(-660.0*p11*p302)+(63.0*p112*p302)+(-2310.0*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(990.0*p202*p31)+(-252.0*p202*p21*p31)+(2310.0*p20*p30*p31)+(168.0*p11*p20*p30*p31)+(3080.0*p302*p31)+(252.0*p11*p302*p31)+(1260.0*p21*p302*p31)+(-378.0*p202*p312)+(-1260.0*p20*p30*p312)+(-2310.0*p302*p312)+(-9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(2.0*p01*((9.0*p102*(-55.0+(14.0*p11)+(-1.0*p31)))+(9.0*p202*(-22.0+(9.0*p11)+(5.0*p21)+p31))+(3.0*p20*p30*(-55.0+(20.0*p11)+(18.0*p21)+(7.0*p31)))+(p302*(-55.0+(15.0*p11)+(21.0*p21)+(14.0*p31)))+(3.0*p10*((3.0*p20*(-55.0+(21.0*p11)+(6.0*p21)))+(2.0*p30*(-22.0+(9.0*p11)+(5.0*p21)+p31))))))+(6.0*p10*((3.0*p112*((9.0*p20)+(5.0*p30)))+(-2.0*p30*p31*(-55.0+(14.0*p21)+(21.0*p31)))+(-3.0*p11*((22.0*p30)+(-6.0*p21*p30)+(p20*(33.0+(6.0*p31)))))+(-3.0*p20*((9.0*p212)+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((2772.0*p10*p21)+(-594.0*p102*p21)+(-594.0*p10*p20*p21)+(-594.0*p10*p212)+(135.0*p102*p212)+(162.0*p10*p20*p212)+(-5544.0*p21*p30)+(990.0*p20*p21*p30)+(990.0*p212*p30)+(-252.0*p20*p212*p30)+(2310.0*p21*p302)+(-378.0*p212*p302)+(-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*(-22.0+p30))+(5.0*(-22.0+p30)*p30)+(42.0*p10*(-55.0+(6.0*p20)+p30))))+(-9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(2772.0*p10*p31)+(-396.0*p102*p31)+(5544.0*p20*p31)+(-990.0*p10*p20*p31)+(-990.0*p202*p31)+(-990.0*p10*p21*p31)+(162.0*p102*p21*p31)+(-990.0*p20*p21*p31)+(378.0*p10*p20*p21*p31)+(252.0*p202*p21*p31)+(9240.0*p30*p31)+(-660.0*p10*p30*p31)+(-2310.0*p20*p30*p31)+(2310.0*p21*p30*p31)+(168.0*p10*p21*p30*p31)+(-3080.0*p302*p31)+(-1260.0*p21*p302*p31)+(-660.0*p10*p312)+(63.0*p102*p312)+(-2310.0*p20*p312)+(252.0*p10*p20*p312)+(378.0*p202*p312)+(-6160.0*p30*p312)+(252.0*p10*p30*p312)+(1260.0*p20*p30*p312)+(2310.0*p302*p312)+(-2.0*p01*((9.0*p102*(-55.0+(14.0*p11)+(-1.0*p31)))+(9.0*p202*(-22.0+(9.0*p11)+(5.0*p21)+p31))+(3.0*p10*(924.0+(3.0*p20*(-55.0+(6.0*p21)))+(-44.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))+(11.0*p31)+(2.0*p30*p31)))+(3.0*p20*(462.0+(-55.0*p30)+(6.0*p21*(-11.0+(3.0*p30)))+(5.0*p11*(-33.0+(4.0*p30)))+(-11.0*p31)+(7.0*p30*p31)))+(p30*(462.0+(15.0*p11*(-11.0+p30))+(-55.0*p30)+(3.0*p21*(-44.0+(7.0*p30)))+(-55.0*p31)+(14.0*p30*p31)))))+(-6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)+(22.0*p31)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30)+(-110.0*p31)+(42.0*p30*p31)))+(p20*(462.0+(-99.0*p21)+(-165.0*p30)+(63.0*p21*p30)+(28.0*p30*p31)+(-9.0*p10*(11.0+(2.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31)));
				break;
			case 3:
				p012 = p01*p01;
				p302 = p30*p30;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*(p012+(-1.0*p312)));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(22.0*((-27.0*p10*p212)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(9.0*p112*((3.0*p20)+(2.0*p30)))+(-45.0*p10*p21*p31)+(-45.0*p20*p21*p31)+(105.0*p21*p30*p31)+(-30.0*p10*p312)+(-105.0*p20*p312)+(-280.0*p30*p312)+(p01*((18.0*p20*p21)+(12.0*p21*p30)+(15.0*p11*((3.0*p20)+p30))+(p10*((45.0*p11)+(-3.0*p31)))+(3.0*p20*p31)+(5.0*p30*p31)))+(3.0*p11*((-9.0*p10*p21)+(9.0*p20*p21)+(15.0*p21*p30)+(-6.0*p10*p31)+(10.0*p30*p31)))))+(-2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*(p012+(-1.0*p312)));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((594.0*p11*p202)+(-135.0*p112*p202)+(-162.0*p11*p202*p21)+(990.0*p11*p20*p30)+(-162.0*p112*p20*p30)+(990.0*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(-252.0*p20*p212*p30)+(660.0*p11*p302)+(-63.0*p112*p302)+(2310.0*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-990.0*p202*p31)+(252.0*p202*p21*p31)+(-2310.0*p20*p30*p31)+(-168.0*p11*p20*p30*p31)+(-3080.0*p302*p31)+(-252.0*p11*p302*p31)+(-1260.0*p21*p302*p31)+(378.0*p202*p312)+(1260.0*p20*p30*p312)+(2310.0*p302*p312)+(9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(-2.0*p01*((9.0*p102*(-55.0+(14.0*p11)+(-1.0*p31)))+(9.0*p202*(-22.0+(9.0*p11)+(5.0*p21)+p31))+(3.0*p20*p30*(-55.0+(20.0*p11)+(18.0*p21)+(7.0*p31)))+(p302*(-55.0+(15.0*p11)+(21.0*p21)+(14.0*p31)))+(3.0*p10*((3.0*p20*(-55.0+(21.0*p11)+(6.0*p21)))+(2.0*p30*(-22.0+(9.0*p11)+(5.0*p21)+p31))))))+(-6.0*p10*((3.0*p112*((9.0*p20)+(5.0*p30)))+(-2.0*p30*p31*(-55.0+(14.0*p21)+(21.0*p31)))+(-3.0*p11*((22.0*p30)+(-6.0*p21*p30)+(p20*(33.0+(6.0*p31)))))+(-3.0*p20*((9.0*p212)+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*((2.0*p01)+(-1.0*p012)+((-2.0+p31)*p31)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-2772.0*p10*p21)+(594.0*p102*p21)+(594.0*p10*p20*p21)+(594.0*p10*p212)+(-135.0*p102*p212)+(-162.0*p10*p20*p212)+(5544.0*p21*p30)+(-990.0*p20*p21*p30)+(-990.0*p212*p30)+(252.0*p20*p212*p30)+(-2310.0*p21*p302)+(378.0*p212*p302)+(p012*((378.0*p102)+(63.0*p202)+(30.0*p20*(-22.0+p30))+(5.0*(-22.0+p30)*p30)+(42.0*p10*(-55.0+(6.0*p20)+p30))))+(9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(-2772.0*p10*p31)+(396.0*p102*p31)+(-5544.0*p20*p31)+(990.0*p10*p20*p31)+(990.0*p202*p31)+(990.0*p10*p21*p31)+(-162.0*p102*p21*p31)+(990.0*p20*p21*p31)+(-378.0*p10*p20*p21*p31)+(-252.0*p202*p21*p31)+(-9240.0*p30*p31)+(660.0*p10*p30*p31)+(2310.0*p20*p30*p31)+(-2310.0*p21*p30*p31)+(-168.0*p10*p21*p30*p31)+(3080.0*p302*p31)+(1260.0*p21*p302*p31)+(660.0*p10*p312)+(-63.0*p102*p312)+(2310.0*p20*p312)+(-252.0*p10*p20*p312)+(-378.0*p202*p312)+(6160.0*p30*p312)+(-252.0*p10*p30*p312)+(-1260.0*p20*p30*p312)+(-2310.0*p302*p312)+(2.0*p01*((9.0*p102*(-55.0+(14.0*p11)+(-1.0*p31)))+(9.0*p202*(-22.0+(9.0*p11)+(5.0*p21)+p31))+(3.0*p10*(924.0+(3.0*p20*(-55.0+(6.0*p21)))+(-44.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))+(11.0*p31)+(2.0*p30*p31)))+(3.0*p20*(462.0+(-55.0*p30)+(6.0*p21*(-11.0+(3.0*p30)))+(5.0*p11*(-33.0+(4.0*p30)))+(-11.0*p31)+(7.0*p30*p31)))+(p30*(462.0+(15.0*p11*(-11.0+p30))+(-55.0*p30)+(3.0*p21*(-44.0+(7.0*p30)))+(-55.0*p31)+(14.0*p30*p31)))))+(6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)+(22.0*p31)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30)+(-110.0*p31)+(42.0*p30*p31)))+(p20*(462.0+(-99.0*p21)+(-165.0*p30)+(63.0*p21*p30)+(28.0*p30*p31)+(-9.0*p10*(11.0+(2.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*((2.0*p01)+(-1.0*p012)+((-2.0+p31)*p31)));
				break;
			case 4:
				p002 = p00*p00;
				p302 = p30*p30;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p002*((378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))))+(2.0*p00*((45.0*p20*p212)+(-9.0*p112*p30)+(9.0*p212*p30)+(54.0*p20*p21*p31)+(21.0*p21*p30*p31)+(21.0*p20*p312)+(14.0*p30*p312)+(6.0*p11*((9.0*p20*p21)+(5.0*p20*p31)+(p30*p31)))+(3.0*p10*((42.0*p112)+(27.0*p212)+(20.0*p21*p31)+(5.0*p312)+(9.0*p11*((7.0*p21)+(2.0*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((1.0/4.0)*p312);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((594.0*p10*p212)+(-135.0*p102*p212)+(-162.0*p10*p20*p212)+(-990.0*p212*p30)+(252.0*p20*p212*p30)+(378.0*p212*p302)+(9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(990.0*p10*p21*p31)+(-162.0*p102*p21*p31)+(990.0*p20*p21*p31)+(-378.0*p10*p20*p21*p31)+(-252.0*p202*p21*p31)+(-2310.0*p21*p30*p31)+(-168.0*p10*p21*p30*p31)+(1260.0*p21*p302*p31)+(660.0*p10*p312)+(-63.0*p102*p312)+(2310.0*p20*p312)+(-252.0*p10*p20*p312)+(-378.0*p202*p312)+(6160.0*p30*p312)+(-252.0*p10*p30*p312)+(-1260.0*p20*p30*p312)+(-2310.0*p302*p312)+(-1.0*p002*((378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))))+(-6.0*p11*((-27.0*p202*p21)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(-3.0*p10*((p21*(33.0+(6.0*p30)))+(22.0*p31)+(-6.0*p20*p31)))+(p30*((165.0*p21)+(-42.0*p21*p30)+(110.0*p31)+(-42.0*p30*p31)))+(p20*((99.0*p21)+(-63.0*p21*p30)+(-28.0*p30*p31)))))+(-2.0*p00*((9.0*p112*(-55.0+(14.0*p10)+(-1.0*p30)))+(9.0*p212*(-22.0+(9.0*p10)+(5.0*p20)+p30))+(3.0*p21*(-55.0+(20.0*p10)+(18.0*p20)+(7.0*p30))*p31)+((-55.0+(15.0*p10)+(21.0*p20)+(14.0*p30))*p312)+(3.0*p11*((3.0*(-55.0+(21.0*p10)+(6.0*p20))*p21)+(2.0*(-22.0+(9.0*p10)+(5.0*p20)+p30)*p31)))))));
				propagate[x][y] += ((1.0/4.0)*p312);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-594.0*p11*p202)+(135.0*p112*p202)+(162.0*p11*p202*p21)+(-990.0*p11*p20*p30)+(162.0*p112*p20*p30)+(-990.0*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(252.0*p20*p212*p30)+(-660.0*p11*p302)+(63.0*p112*p302)+(-2310.0*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(990.0*p202*p31)+(-252.0*p202*p21*p31)+(2310.0*p20*p30*p31)+(168.0*p11*p20*p30*p31)+(3080.0*p302*p31)+(252.0*p11*p302*p31)+(1260.0*p21*p302*p31)+(-378.0*p202*p312)+(-1260.0*p20*p30*p312)+(-2310.0*p302*p312)+(-1.0*p002*((378.0*p112)+(63.0*p212)+(30.0*p21*(-22.0+p31))+(5.0*(-22.0+p31)*p31)+(42.0*p11*(-55.0+(6.0*p21)+p31))))+(-9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(6.0*p10*((3.0*p112*((9.0*p20)+(5.0*p30)))+(-2.0*p30*p31*(-55.0+(14.0*p21)+(21.0*p31)))+(-3.0*p11*((22.0*p30)+(-6.0*p21*p30)+(p20*(33.0+(6.0*p31)))))+(-3.0*p20*((9.0*p212)+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31)))))))+(-2.0*p00*((3.0*p10*((42.0*p112)+(27.0*p212)+(5.0*(-11.0+p31)*p31)+(5.0*p21*(-33.0+(4.0*p31)))+(3.0*p11*(-55.0+(21.0*p21)+(6.0*p31)))))+(3.0*p20*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(p30*((-9.0*p112)+(9.0*p212)+(p11*(33.0+(6.0*p31)))+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((2772.0*p10*p21)+(-594.0*p102*p21)+(-594.0*p10*p20*p21)+(-594.0*p10*p212)+(135.0*p102*p212)+(162.0*p10*p20*p212)+(-5544.0*p21*p30)+(990.0*p20*p21*p30)+(990.0*p212*p30)+(-252.0*p20*p212*p30)+(2310.0*p21*p302)+(-378.0*p212*p302)+(-9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(2772.0*p10*p31)+(-396.0*p102*p31)+(5544.0*p20*p31)+(-990.0*p10*p20*p31)+(-990.0*p202*p31)+(-990.0*p10*p21*p31)+(162.0*p102*p21*p31)+(-990.0*p20*p21*p31)+(378.0*p10*p20*p21*p31)+(252.0*p202*p21*p31)+(9240.0*p30*p31)+(-660.0*p10*p30*p31)+(-2310.0*p20*p30*p31)+(2310.0*p21*p30*p31)+(168.0*p10*p21*p30*p31)+(-3080.0*p302*p31)+(-1260.0*p21*p302*p31)+(-660.0*p10*p312)+(63.0*p102*p312)+(-2310.0*p20*p312)+(252.0*p10*p20*p312)+(378.0*p202*p312)+(-6160.0*p30*p312)+(252.0*p10*p30*p312)+(1260.0*p20*p30*p312)+(2310.0*p302*p312)+(p002*((378.0*p112)+(63.0*p212)+(30.0*p21*(-22.0+p31))+(5.0*(-22.0+p31)*p31)+(42.0*p11*(-55.0+(6.0*p21)+p31))))+(-6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)+(22.0*p31)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30)+(-110.0*p31)+(42.0*p30*p31)))+(p20*(462.0+(-99.0*p21)+(-165.0*p30)+(63.0*p21*p30)+(28.0*p30*p31)+(-9.0*p10*(11.0+(2.0*p31)))))))+(2.0*p00*((9.0*p112*(-55.0+(14.0*p10)+(-1.0*p30)))+(9.0*p212*(-22.0+(9.0*p10)+(5.0*p20)+p30))+(3.0*p21*(462.0+(-11.0*p30)+(-55.0*p31)+(7.0*p30*p31)+(6.0*p20*(-11.0+(3.0*p31)))+(5.0*p10*(-33.0+(4.0*p31)))))+(3.0*p11*(924.0+(3.0*(-55.0+(6.0*p20))*p21)+(11.0*p30)+(-44.0*p31)+(10.0*p20*p31)+(2.0*p30*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31)))))+(p31*(462.0+(-55.0*p30)+(15.0*p10*(-11.0+p31))+(-55.0*p31)+(14.0*p30*p31)+(3.0*p20*(-44.0+(7.0*p31)))))))));
				propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31);
				break;
			case 5:
				p002 = p00*p00;
				p302 = p30*p30;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*((378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))))+(-2.0*p00*((45.0*p20*p212)+(-9.0*p112*p30)+(9.0*p212*p30)+(54.0*p20*p21*p31)+(21.0*p21*p30*p31)+(21.0*p20*p312)+(14.0*p30*p312)+(6.0*p11*((9.0*p20*p21)+(5.0*p20*p31)+(p30*p31)))+(3.0*p10*((42.0*p112)+(27.0*p212)+(20.0*p21*p31)+(5.0*p312)+(9.0*p11*((7.0*p21)+(2.0*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))))));
				propagate[x][y] += ((-1.0/4.0)*p312);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-594.0*p10*p212)+(135.0*p102*p212)+(162.0*p10*p20*p212)+(990.0*p212*p30)+(-252.0*p20*p212*p30)+(-378.0*p212*p302)+(-9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(-990.0*p10*p21*p31)+(162.0*p102*p21*p31)+(-990.0*p20*p21*p31)+(378.0*p10*p20*p21*p31)+(252.0*p202*p21*p31)+(2310.0*p21*p30*p31)+(168.0*p10*p21*p30*p31)+(-1260.0*p21*p302*p31)+(-660.0*p10*p312)+(63.0*p102*p312)+(-2310.0*p20*p312)+(252.0*p10*p20*p312)+(378.0*p202*p312)+(-6160.0*p30*p312)+(252.0*p10*p30*p312)+(1260.0*p20*p30*p312)+(2310.0*p302*p312)+(p002*((378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))))+(6.0*p11*((-27.0*p202*p21)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(-3.0*p10*((p21*(33.0+(6.0*p30)))+(22.0*p31)+(-6.0*p20*p31)))+(p30*((165.0*p21)+(-42.0*p21*p30)+(110.0*p31)+(-42.0*p30*p31)))+(p20*((99.0*p21)+(-63.0*p21*p30)+(-28.0*p30*p31)))))+(2.0*p00*((9.0*p112*(-55.0+(14.0*p10)+(-1.0*p30)))+(9.0*p212*(-22.0+(9.0*p10)+(5.0*p20)+p30))+(3.0*p21*(-55.0+(20.0*p10)+(18.0*p20)+(7.0*p30))*p31)+((-55.0+(15.0*p10)+(21.0*p20)+(14.0*p30))*p312)+(3.0*p11*((3.0*(-55.0+(21.0*p10)+(6.0*p20))*p21)+(2.0*(-22.0+(9.0*p10)+(5.0*p20)+p30)*p31)))))));
				propagate[x][y] += ((-1.0/4.0)*p312);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((594.0*p11*p202)+(-135.0*p112*p202)+(-162.0*p11*p202*p21)+(990.0*p11*p20*p30)+(-162.0*p112*p20*p30)+(990.0*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(-252.0*p20*p212*p30)+(660.0*p11*p302)+(-63.0*p112*p302)+(2310.0*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(-990.0*p202*p31)+(252.0*p202*p21*p31)+(-2310.0*p20*p30*p31)+(-168.0*p11*p20*p30*p31)+(-3080.0*p302*p31)+(-252.0*p11*p302*p31)+(-1260.0*p21*p302*p31)+(378.0*p202*p312)+(1260.0*p20*p30*p312)+(2310.0*p302*p312)+(p002*((378.0*p112)+(63.0*p212)+(30.0*p21*(-22.0+p31))+(5.0*(-22.0+p31)*p31)+(42.0*p11*(-55.0+(6.0*p21)+p31))))+(9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(-6.0*p10*((3.0*p112*((9.0*p20)+(5.0*p30)))+(-2.0*p30*p31*(-55.0+(14.0*p21)+(21.0*p31)))+(-3.0*p11*((22.0*p30)+(-6.0*p21*p30)+(p20*(33.0+(6.0*p31)))))+(-3.0*p20*((9.0*p212)+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31)))))))+(2.0*p00*((3.0*p10*((42.0*p112)+(27.0*p212)+(5.0*(-11.0+p31)*p31)+(5.0*p21*(-33.0+(4.0*p31)))+(3.0*p11*(-55.0+(21.0*p21)+(6.0*p31)))))+(3.0*p20*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(p30*((-9.0*p112)+(9.0*p212)+(p11*(33.0+(6.0*p31)))+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*(-2.0+p31)*p31);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-2772.0*p10*p21)+(594.0*p102*p21)+(594.0*p10*p20*p21)+(594.0*p10*p212)+(-135.0*p102*p212)+(-162.0*p10*p20*p212)+(5544.0*p21*p30)+(-990.0*p20*p21*p30)+(-990.0*p212*p30)+(252.0*p20*p212*p30)+(-2310.0*p21*p302)+(378.0*p212*p302)+(9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(-2772.0*p10*p31)+(396.0*p102*p31)+(-5544.0*p20*p31)+(990.0*p10*p20*p31)+(990.0*p202*p31)+(990.0*p10*p21*p31)+(-162.0*p102*p21*p31)+(990.0*p20*p21*p31)+(-378.0*p10*p20*p21*p31)+(-252.0*p202*p21*p31)+(-9240.0*p30*p31)+(660.0*p10*p30*p31)+(2310.0*p20*p30*p31)+(-2310.0*p21*p30*p31)+(-168.0*p10*p21*p30*p31)+(3080.0*p302*p31)+(1260.0*p21*p302*p31)+(660.0*p10*p312)+(-63.0*p102*p312)+(2310.0*p20*p312)+(-252.0*p10*p20*p312)+(-378.0*p202*p312)+(6160.0*p30*p312)+(-252.0*p10*p30*p312)+(-1260.0*p20*p30*p312)+(-2310.0*p302*p312)+(-1.0*p002*((378.0*p112)+(63.0*p212)+(30.0*p21*(-22.0+p31))+(5.0*(-22.0+p31)*p31)+(42.0*p11*(-55.0+(6.0*p21)+p31))))+(6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)+(22.0*p31)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30)+(-110.0*p31)+(42.0*p30*p31)))+(p20*(462.0+(-99.0*p21)+(-165.0*p30)+(63.0*p21*p30)+(28.0*p30*p31)+(-9.0*p10*(11.0+(2.0*p31)))))))+(-2.0*p00*((9.0*p112*(-55.0+(14.0*p10)+(-1.0*p30)))+(9.0*p212*(-22.0+(9.0*p10)+(5.0*p20)+p30))+(3.0*p21*(462.0+(-11.0*p30)+(-55.0*p31)+(7.0*p30*p31)+(6.0*p20*(-11.0+(3.0*p31)))+(5.0*p10*(-33.0+(4.0*p31)))))+(3.0*p11*(924.0+(3.0*(-55.0+(6.0*p20))*p21)+(11.0*p30)+(-44.0*p31)+(10.0*p20*p31)+(2.0*p30*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31)))))+(p31*(462.0+(-55.0*p30)+(15.0*p10*(-11.0+p31))+(-55.0*p31)+(14.0*p30*p31)+(3.0*p20*(-44.0+(7.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*(-2.0+p31)*p31);
				break;
			case 6:
				p302 = p30*p30;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/6160.0)*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))));
				propagate[x][y] += ((1.0/4.0)*p312);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(-9.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)))+(-6.0*p10*((9.0*(-11.0+(3.0*p20))*p212)+(p21*(-165.0+(63.0*p20)+(28.0*p30))*p31)+(2.0*(-55.0+(21.0*p20)+(21.0*p30))*p312)))+(-2.0*((-9.0*p212*p30*(-55.0+(14.0*p20)+(21.0*p30)))+(3.0*p21*((-165.0*p20)+(42.0*p202)+(35.0*(11.0+(-6.0*p30))*p30))*p31)+(7.0*((27.0*p202)+(55.0*p30*(-8.0+(3.0*p30)))+(15.0*p20*(-11.0+(6.0*p30))))*p312)))+(-6.0*p11*((-27.0*p202*p21)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(-3.0*p10*((p21*(33.0+(6.0*p30)))+(22.0*p31)+(-6.0*p20*p31)))+(p30*((165.0*p21)+(-42.0*p21*p30)+(110.0*p31)+(-42.0*p30*p31)))+(p20*((99.0*p21)+(-63.0*p21*p30)+(-28.0*p30*p31)))))));
				propagate[x][y] += ((1.0/4.0)*p312);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-990.0*p20*p21*p30)+(252.0*p20*p212*p30)+(-2310.0*p21*p302)+(378.0*p212*p302)+(9.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(990.0*p202*p31)+(-252.0*p202*p21*p31)+(2310.0*p20*p30*p31)+(3080.0*p302*p31)+(1260.0*p21*p302*p31)+(-378.0*p202*p312)+(-1260.0*p20*p30*p312)+(-2310.0*p302*p312)+(-9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(2.0*p302*(-55.0+(21.0*p21)+(21.0*p31)))+(p20*p30*(-165.0+(63.0*p21)+(28.0*p31)))))+(6.0*p10*((3.0*p112*((9.0*p20)+(5.0*p30)))+(-2.0*p30*p31*(-55.0+(14.0*p21)+(21.0*p31)))+(-3.0*p11*((22.0*p30)+(-6.0*p21*p30)+(p20*(33.0+(6.0*p31)))))+(-3.0*p20*((9.0*p212)+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p31)))+(p31*(-44.0+(7.0*p31)))))+(-6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)+(22.0*p31)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30)+(-110.0*p31)+(42.0*p30*p31)))+(p20*(462.0+(-99.0*p21)+(-165.0*p30)+(63.0*p21*p30)+(28.0*p30*p31)+(-9.0*p10*(11.0+(2.0*p31)))))))+(6.0*p10*((9.0*(-11.0+(3.0*p20))*p212)+(p31*(462.0+(-165.0*p20)+(-110.0*p30)+(-110.0*p31)+(42.0*p20*p31)+(42.0*p30*p31)))+(p21*(462.0+(-165.0*p31)+(28.0*p30*p31)+(9.0*p20*(-11.0+(7.0*p31)))))))+(2.0*((-9.0*p212*p30*(-55.0+(14.0*p20)+(21.0*p30)))+(-3.0*p21*((3.0*(55.0+(-14.0*p20))*p20*p31)+(35.0*p302*(-11.0+(6.0*p31)))+(-11.0*p30*(-84.0+(15.0*p20)+(35.0*p31)))))+(p31*((9.0*p202*(-55.0+(21.0*p31)))+(385.0*p30*(12.0+(-4.0*p30)+(-8.0*p31)+(3.0*p30*p31)))+(21.0*p20*(132.0+(-55.0*p30)+(-55.0*p31)+(30.0*p30*p31)))))))));
				propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31);
				break;
			case 7:
				p302 = p30*p30;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/6160.0)*((3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(-3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(-1.0*p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(-14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312)))))));
				propagate[x][y] += ((-1.0/4.0)*p312);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(9.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)))+(6.0*p10*((9.0*(-11.0+(3.0*p20))*p212)+(p21*(-165.0+(63.0*p20)+(28.0*p30))*p31)+(2.0*(-55.0+(21.0*p20)+(21.0*p30))*p312)))+(2.0*((-9.0*p212*p30*(-55.0+(14.0*p20)+(21.0*p30)))+(3.0*p21*((-165.0*p20)+(42.0*p202)+(35.0*(11.0+(-6.0*p30))*p30))*p31)+(7.0*((27.0*p202)+(55.0*p30*(-8.0+(3.0*p30)))+(15.0*p20*(-11.0+(6.0*p30))))*p312)))+(6.0*p11*((-27.0*p202*p21)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(-3.0*p10*((p21*(33.0+(6.0*p30)))+(22.0*p31)+(-6.0*p20*p31)))+(p30*((165.0*p21)+(-42.0*p21*p30)+(110.0*p31)+(-42.0*p30*p31)))+(p20*((99.0*p21)+(-63.0*p21*p30)+(-28.0*p30*p31)))))));
				propagate[x][y] += ((-1.0/4.0)*p312);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((990.0*p20*p21*p30)+(-252.0*p20*p212*p30)+(2310.0*p21*p302)+(-378.0*p212*p302)+(-9.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-990.0*p202*p31)+(252.0*p202*p21*p31)+(-2310.0*p20*p30*p31)+(-3080.0*p302*p31)+(-1260.0*p21*p302*p31)+(378.0*p202*p312)+(1260.0*p20*p30*p312)+(2310.0*p302*p312)+(9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p11)+(3.0*p31)))+(p31*(-44.0+(10.0*p11)+(7.0*p31)))))+(-6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(2.0*p302*(-55.0+(21.0*p21)+(21.0*p31)))+(p20*p30*(-165.0+(63.0*p21)+(28.0*p31)))))+(-6.0*p10*((3.0*p112*((9.0*p20)+(5.0*p30)))+(-2.0*p30*p31*(-55.0+(14.0*p21)+(21.0*p31)))+(-3.0*p11*((22.0*p30)+(-6.0*p21*p30)+(p20*(33.0+(6.0*p31)))))+(-3.0*p20*((9.0*p212)+(3.0*p21*(-11.0+(7.0*p31)))+(p31*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*(-2.0+p31)*p31);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*((15.0*p202)+(6.0*p20*(-11.0+(3.0*p10)+(3.0*p30)))+(p30*(-44.0+(10.0*p10)+(7.0*p30)))))+(-9.0*p102*((15.0*p212)+(6.0*p21*(-11.0+(3.0*p31)))+(p31*(-44.0+(7.0*p31)))))+(6.0*p11*((9.0*p202*(-11.0+(3.0*p21)))+(-3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*((33.0*p21)+(-22.0*p30)+(6.0*p21*p30)+(22.0*p31)))+(p30*(462.0+(-165.0*p21)+(-110.0*p30)+(42.0*p21*p30)+(-110.0*p31)+(42.0*p30*p31)))+(p20*(462.0+(-99.0*p21)+(-165.0*p30)+(63.0*p21*p30)+(28.0*p30*p31)+(-9.0*p10*(11.0+(2.0*p31)))))))+(-6.0*p10*((9.0*(-11.0+(3.0*p20))*p212)+(p31*(462.0+(-165.0*p20)+(-110.0*p30)+(-110.0*p31)+(42.0*p20*p31)+(42.0*p30*p31)))+(p21*(462.0+(-165.0*p31)+(28.0*p30*p31)+(9.0*p20*(-11.0+(7.0*p31)))))))+(2.0*((9.0*p212*p30*(-55.0+(14.0*p20)+(21.0*p30)))+(3.0*p21*((3.0*(55.0+(-14.0*p20))*p20*p31)+(35.0*p302*(-11.0+(6.0*p31)))+(-11.0*p30*(-84.0+(15.0*p20)+(35.0*p31)))))+(p31*((-9.0*p202*(-55.0+(21.0*p31)))+(-385.0*p30*(12.0+(-4.0*p30)+(-8.0*p31)+(3.0*p30*p31)))+(-21.0*p20*(132.0+(-55.0*p30)+(-55.0*p31)+(30.0*p30*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*(-2.0+p31)*p31);
				break;
			case 8:
				p002 = p00*p00;
				p012 = p01*p01;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))))+(-2.0*p00*((9.0*p112)+(-126.0*p10*p112)+(14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-9.0*p212)+(-81.0*p10*p212)+(-45.0*p20*p212)+(-6.0*p11*p31)+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-21.0*p21*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-14.0*p312)+(-15.0*p10*p312)+(-21.0*p20*p312)+(3.0*p01*((7.0*p11*(1.0+(4.0*p20)))+(2.0*p21)+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((21.0*p012*(-5.0+(110.0*p002)+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)+(4.0*p00*(-73.0+(15.0*p10)+(3.0*p20)))))+(-9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(-612.0*p212)+(378.0*p00*p212)+(-63.0*p002*p212)+(594.0*p10*p212)+(-162.0*p00*p10*p212)+(-135.0*p102*p212)+(252.0*p20*p212)+(-90.0*p00*p20*p212)+(-162.0*p10*p20*p212)+(-1050.0*p21*p31)+(288.0*p00*p21*p31)+(-30.0*p002*p21*p31)+(822.0*p10*p21*p31)+(-120.0*p00*p10*p21*p31)+(-162.0*p102*p21*p31)+(990.0*p20*p21*p31)+(-108.0*p00*p20*p21*p31)+(-378.0*p10*p20*p21*p31)+(-252.0*p202*p21*p31)+(3850.0*p312)+(82.0*p00*p312)+(-5.0*p002*p312)+(408.0*p10*p312)+(-30.0*p00*p10*p312)+(-63.0*p102*p312)+(1050.0*p20*p312)+(-42.0*p00*p20*p312)+(-252.0*p10*p20*p312)+(-378.0*p202*p312)+(-6.0*p11*((3.0*(41.0+(14.0*p002)+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202)+(p00*(-55.0+(21.0*p10)+(6.0*p20))))*p21)+((68.0+(7.0*p002)+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p31)))+(-2.0*p01*((3.0*p11*(50.0+(210.0*p002)+(-42.0*p102)+(145.0*p20)+(-27.0*p202)+(-28.0*p00*(14.0+p20))+(-21.0*p10*(-7.0+(3.0*p20)))))+(3.0*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20))))*p21)+((41.0+(14.0*p002)+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202)+(p00*(-55.0+(21.0*p10)+(6.0*p20))))*p31)))));
				propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(-1.0*p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))))+(-22.0*((5.0*p01)+(12.0*p01*p10)+(45.0*p01*p102)+(30.0*p11)+(18.0*p10*p11)+(15.0*p01*p20)+(45.0*p01*p10*p20)+(45.0*p11*p20)+(27.0*p10*p11*p20)+(18.0*p01*p202)+(27.0*p11*p202)+(105.0*p21)+(-27.0*p102*p21)+(45.0*p20*p21)+(-27.0*p10*p20*p21)+(5.0*p002*((28.0*p01)+(-21.0*p11)+(-6.0*p21)+(-1.0*p31)))+(-140.0*p31)+(-30.0*p10*p31)+(-18.0*p102*p31)+(-105.0*p20*p31)+(-45.0*p10*p20*p31)+(-45.0*p202*p31)+(p00*((3.0*p11)+(5.0*p01*(1.0+(21.0*p10)+(6.0*p20)))+(-3.0*p21)+(-18.0*p20*p21)+(-5.0*p31)+(-12.0*p20*p31)+(-15.0*p10*((3.0*p11)+(3.0*p21)+p31))))))+(2.0*p00*((9.0*p112)+(-126.0*p10*p112)+(14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-9.0*p212)+(-81.0*p10*p212)+(-45.0*p20*p212)+(-6.0*p11*p31)+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-21.0*p21*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-14.0*p312)+(-15.0*p10*p312)+(-21.0*p20*p312)+(3.0*p01*((7.0*p11*(1.0+(4.0*p20)))+(2.0*p21)+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))));
				propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-21.0*p012*(-5.0+(110.0*p002)+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)+(4.0*p00*(-73.0+(15.0*p10)+(3.0*p20)))))+(9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(-3234.0*p21)+(2706.0*p00*p21)+(-660.0*p002*p21)+(2772.0*p10*p21)+(-990.0*p00*p10*p21)+(-594.0*p102*p21)+(990.0*p20*p21)+(-396.0*p00*p20*p21)+(-594.0*p10*p20*p21)+(612.0*p212)+(-378.0*p00*p212)+(63.0*p002*p212)+(-594.0*p10*p212)+(162.0*p00*p10*p212)+(135.0*p102*p212)+(-252.0*p20*p212)+(90.0*p00*p20*p212)+(162.0*p10*p20*p212)+(6160.0*p31)+(814.0*p00*p31)+(-110.0*p002*p31)+(2112.0*p10*p31)+(-330.0*p00*p10*p31)+(-396.0*p102*p31)+(3234.0*p20*p31)+(-264.0*p00*p20*p31)+(-990.0*p10*p20*p31)+(-990.0*p202*p31)+(1050.0*p21*p31)+(-288.0*p00*p21*p31)+(30.0*p002*p21*p31)+(-822.0*p10*p21*p31)+(120.0*p00*p10*p21*p31)+(162.0*p102*p21*p31)+(-990.0*p20*p21*p31)+(108.0*p00*p20*p21*p31)+(378.0*p10*p20*p21*p31)+(252.0*p202*p21*p31)+(-3850.0*p312)+(-82.0*p00*p312)+(5.0*p002*p312)+(-408.0*p10*p312)+(30.0*p00*p10*p312)+(63.0*p102*p312)+(-1050.0*p20*p312)+(42.0*p00*p20*p312)+(252.0*p10*p20*p312)+(378.0*p202*p312)+(2.0*p01*(-407.0+(150.0*p11)+(-1221.0*p20)+(435.0*p11*p20)+(198.0*p202)+(-81.0*p11*p202)+(111.0*p21)+(144.0*p20*p21)+(-45.0*p202*p21)+(41.0*p31)+(12.0*p20*p31)+(-9.0*p202*p31)+(9.0*p102*(55.0+(-14.0*p11)+p31))+(14.0*p002*(110.0+(45.0*p11)+(9.0*p21)+p31))+(-3.0*p10*(880.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))+(13.0*p31)))+(p00*(-4565.0+(330.0*p20)+(-84.0*p11*(14.0+p20))+(-336.0*p21)+(-55.0*p31)+(6.0*p20*p31)+(21.0*p10*(55.0+(4.0*p21)+p31))))))+(6.0*p11*(-352.0+(-297.0*p20)+(99.0*p202)+(123.0*p21)+(36.0*p20*p21)+(-27.0*p202*p21)+(68.0*p31)+(-28.0*p20*p31)+(7.0*p002*(-55.0+(6.0*p21)+p31))+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31)))))+(p00*(935.0+(3.0*(-55.0+(6.0*p20))*p21)+(-42.0*p31)+(10.0*p20*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31)));
				break;
			case 9:
				p002 = p00*p00;
				p012 = p01*p01;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(-1.0*p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))))+(2.0*p00*((9.0*p112)+(-126.0*p10*p112)+(14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-9.0*p212)+(-81.0*p10*p212)+(-45.0*p20*p212)+(-6.0*p11*p31)+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-21.0*p21*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-14.0*p312)+(-15.0*p10*p312)+(-21.0*p20*p312)+(3.0*p01*((7.0*p11*(1.0+(4.0*p20)))+(2.0*p21)+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))));
				propagate[x][y] += ((1.0/4.0)*(p012+(-1.0*p312)));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-21.0*p012*(-5.0+(110.0*p002)+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)+(4.0*p00*(-73.0+(15.0*p10)+(3.0*p20)))))+(9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(612.0*p212)+(-378.0*p00*p212)+(63.0*p002*p212)+(-594.0*p10*p212)+(162.0*p00*p10*p212)+(135.0*p102*p212)+(-252.0*p20*p212)+(90.0*p00*p20*p212)+(162.0*p10*p20*p212)+(1050.0*p21*p31)+(-288.0*p00*p21*p31)+(30.0*p002*p21*p31)+(-822.0*p10*p21*p31)+(120.0*p00*p10*p21*p31)+(162.0*p102*p21*p31)+(-990.0*p20*p21*p31)+(108.0*p00*p20*p21*p31)+(378.0*p10*p20*p21*p31)+(252.0*p202*p21*p31)+(-3850.0*p312)+(-82.0*p00*p312)+(5.0*p002*p312)+(-408.0*p10*p312)+(30.0*p00*p10*p312)+(63.0*p102*p312)+(-1050.0*p20*p312)+(42.0*p00*p20*p312)+(252.0*p10*p20*p312)+(378.0*p202*p312)+(6.0*p11*((3.0*(41.0+(14.0*p002)+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202)+(p00*(-55.0+(21.0*p10)+(6.0*p20))))*p21)+((68.0+(7.0*p002)+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p31)))+(2.0*p01*((3.0*p11*(50.0+(210.0*p002)+(-42.0*p102)+(145.0*p20)+(-27.0*p202)+(-28.0*p00*(14.0+p20))+(-21.0*p10*(-7.0+(3.0*p20)))))+(3.0*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20))))*p21)+((41.0+(14.0*p002)+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202)+(p00*(-55.0+(21.0*p10)+(6.0*p20))))*p31)))));
				propagate[x][y] += ((1.0/4.0)*(p012+(-1.0*p312)));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))))+(22.0*((5.0*p01)+(12.0*p01*p10)+(45.0*p01*p102)+(30.0*p11)+(18.0*p10*p11)+(15.0*p01*p20)+(45.0*p01*p10*p20)+(45.0*p11*p20)+(27.0*p10*p11*p20)+(18.0*p01*p202)+(27.0*p11*p202)+(105.0*p21)+(-27.0*p102*p21)+(45.0*p20*p21)+(-27.0*p10*p20*p21)+(5.0*p002*((28.0*p01)+(-21.0*p11)+(-6.0*p21)+(-1.0*p31)))+(-140.0*p31)+(-30.0*p10*p31)+(-18.0*p102*p31)+(-105.0*p20*p31)+(-45.0*p10*p20*p31)+(-45.0*p202*p31)+(p00*((3.0*p11)+(5.0*p01*(1.0+(21.0*p10)+(6.0*p20)))+(-3.0*p21)+(-18.0*p20*p21)+(-5.0*p31)+(-12.0*p20*p31)+(-15.0*p10*((3.0*p11)+(3.0*p21)+p31))))))+(-2.0*p00*((9.0*p112)+(-126.0*p10*p112)+(14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-9.0*p212)+(-81.0*p10*p212)+(-45.0*p20*p212)+(-6.0*p11*p31)+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-21.0*p21*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-14.0*p312)+(-15.0*p10*p312)+(-21.0*p20*p312)+(3.0*p01*((7.0*p11*(1.0+(4.0*p20)))+(2.0*p21)+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))));
				propagate[x][y] += ((1.0/4.0)*((2.0*p01)+(-1.0*p012)+((-2.0+p31)*p31)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((21.0*p012*(-5.0+(110.0*p002)+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)+(4.0*p00*(-73.0+(15.0*p10)+(3.0*p20)))))+(-9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(3234.0*p21)+(-2706.0*p00*p21)+(660.0*p002*p21)+(-2772.0*p10*p21)+(990.0*p00*p10*p21)+(594.0*p102*p21)+(-990.0*p20*p21)+(396.0*p00*p20*p21)+(594.0*p10*p20*p21)+(-612.0*p212)+(378.0*p00*p212)+(-63.0*p002*p212)+(594.0*p10*p212)+(-162.0*p00*p10*p212)+(-135.0*p102*p212)+(252.0*p20*p212)+(-90.0*p00*p20*p212)+(-162.0*p10*p20*p212)+(-6160.0*p31)+(-814.0*p00*p31)+(110.0*p002*p31)+(-2112.0*p10*p31)+(330.0*p00*p10*p31)+(396.0*p102*p31)+(-3234.0*p20*p31)+(264.0*p00*p20*p31)+(990.0*p10*p20*p31)+(990.0*p202*p31)+(-1050.0*p21*p31)+(288.0*p00*p21*p31)+(-30.0*p002*p21*p31)+(822.0*p10*p21*p31)+(-120.0*p00*p10*p21*p31)+(-162.0*p102*p21*p31)+(990.0*p20*p21*p31)+(-108.0*p00*p20*p21*p31)+(-378.0*p10*p20*p21*p31)+(-252.0*p202*p21*p31)+(3850.0*p312)+(82.0*p00*p312)+(-5.0*p002*p312)+(408.0*p10*p312)+(-30.0*p00*p10*p312)+(-63.0*p102*p312)+(1050.0*p20*p312)+(-42.0*p00*p20*p312)+(-252.0*p10*p20*p312)+(-378.0*p202*p312)+(-2.0*p01*(-407.0+(150.0*p11)+(-1221.0*p20)+(435.0*p11*p20)+(198.0*p202)+(-81.0*p11*p202)+(111.0*p21)+(144.0*p20*p21)+(-45.0*p202*p21)+(41.0*p31)+(12.0*p20*p31)+(-9.0*p202*p31)+(9.0*p102*(55.0+(-14.0*p11)+p31))+(14.0*p002*(110.0+(45.0*p11)+(9.0*p21)+p31))+(-3.0*p10*(880.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))+(13.0*p31)))+(p00*(-4565.0+(330.0*p20)+(-84.0*p11*(14.0+p20))+(-336.0*p21)+(-55.0*p31)+(6.0*p20*p31)+(21.0*p10*(55.0+(4.0*p21)+p31))))))+(-6.0*p11*(-352.0+(-297.0*p20)+(99.0*p202)+(123.0*p21)+(36.0*p20*p21)+(-27.0*p202*p21)+(68.0*p31)+(-28.0*p20*p31)+(7.0*p002*(-55.0+(6.0*p21)+p31))+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31)))))+(p00*(935.0+(3.0*(-55.0+(6.0*p20))*p21)+(-42.0*p31)+(10.0*p20*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*((2.0*p01)+(-1.0*p012)+((-2.0+p31)*p31)));
				break;
			case 10:
				p012 = p01*p01;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-612.0*p212)+(594.0*p10*p212)+(-135.0*p102*p212)+(252.0*p20*p212)+(-162.0*p10*p20*p212)+(-1050.0*p21*p31)+(822.0*p10*p21*p31)+(-162.0*p102*p21*p31)+(990.0*p20*p21*p31)+(-378.0*p10*p20*p21*p31)+(-252.0*p202*p21*p31)+(3850.0*p312)+(408.0*p10*p312)+(-63.0*p102*p312)+(1050.0*p20*p312)+(-252.0*p10*p20*p312)+(-378.0*p202*p312)+(2.0*p01*((3.0*p11*(-50.0+(42.0*p102)+(-145.0*p20)+(27.0*p202)+(21.0*p10*(-7.0+(3.0*p20)))))+(3.0*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((-41.0+(39.0*p10)+(-9.0*p102)+(-12.0*p20)+(9.0*p202))*p31)))+(-6.0*p11*((3.0*(41.0+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202))*p21)+((68.0+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20))))*p31)))));
				propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-22.0*((3.0*p11*(10.0+(6.0*p10)+(15.0*p20)+(9.0*p10*p20)+(9.0*p202)))+(p01*(5.0+(45.0*p102)+(15.0*p20)+(18.0*p202)+(3.0*p10*(4.0+(15.0*p20)))))+(105.0*p21)+(-27.0*p102*p21)+(45.0*p20*p21)+(-27.0*p10*p20*p21)+(-140.0*p31)+(-30.0*p10*p31)+(-18.0*p102*p31)+(-105.0*p20*p31)+(-45.0*p10*p20*p31)+(-45.0*p202*p31)))+(2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(-3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))))));
				propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-3234.0*p21)+(2772.0*p10*p21)+(-594.0*p102*p21)+(990.0*p20*p21)+(-594.0*p10*p20*p21)+(612.0*p212)+(-594.0*p10*p212)+(135.0*p102*p212)+(-252.0*p20*p212)+(162.0*p10*p20*p212)+(6160.0*p31)+(2112.0*p10*p31)+(-396.0*p102*p31)+(3234.0*p20*p31)+(-990.0*p10*p20*p31)+(-990.0*p202*p31)+(1050.0*p21*p31)+(-822.0*p10*p21*p31)+(162.0*p102*p21*p31)+(-990.0*p20*p21*p31)+(378.0*p10*p20*p21*p31)+(252.0*p202*p21*p31)+(-3850.0*p312)+(-408.0*p10*p312)+(63.0*p102*p312)+(-1050.0*p20*p312)+(252.0*p10*p20*p312)+(378.0*p202*p312)+(-2.0*p01*(407.0+(1221.0*p20)+(-198.0*p202)+(3.0*p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-111.0*p21)+(-144.0*p20*p21)+(45.0*p202*p21)+(9.0*p102*(-55.0+(14.0*p11)+(-1.0*p31)))+(-41.0*p31)+(-12.0*p20*p31)+(9.0*p202*p31)+(3.0*p10*(880.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))+(13.0*p31)))))+(6.0*p11*(-352.0+(p202*(99.0+(-27.0*p21)))+(123.0*p21)+(p20*(-297.0+(36.0*p21)+(-28.0*p31)))+(68.0*p31)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31)));
				break;
			case 11:
				p012 = p01*p01;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(-3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))))));
				propagate[x][y] += ((1.0/4.0)*(p012+(-1.0*p312)));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(612.0*p212)+(-594.0*p10*p212)+(135.0*p102*p212)+(-252.0*p20*p212)+(162.0*p10*p20*p212)+(1050.0*p21*p31)+(-822.0*p10*p21*p31)+(162.0*p102*p21*p31)+(-990.0*p20*p21*p31)+(378.0*p10*p20*p21*p31)+(252.0*p202*p21*p31)+(-3850.0*p312)+(-408.0*p10*p312)+(63.0*p102*p312)+(-1050.0*p20*p312)+(252.0*p10*p20*p312)+(378.0*p202*p312)+(-2.0*p01*((3.0*p11*(-50.0+(42.0*p102)+(-145.0*p20)+(27.0*p202)+(21.0*p10*(-7.0+(3.0*p20)))))+(3.0*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((-41.0+(39.0*p10)+(-9.0*p102)+(-12.0*p20)+(9.0*p202))*p31)))+(6.0*p11*((3.0*(41.0+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202))*p21)+((68.0+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20))))*p31)))));
				propagate[x][y] += ((1.0/4.0)*(p012+(-1.0*p312)));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(22.0*((3.0*p11*(10.0+(6.0*p10)+(15.0*p20)+(9.0*p10*p20)+(9.0*p202)))+(p01*(5.0+(45.0*p102)+(15.0*p20)+(18.0*p202)+(3.0*p10*(4.0+(15.0*p20)))))+(105.0*p21)+(-27.0*p102*p21)+(45.0*p20*p21)+(-27.0*p10*p20*p21)+(-140.0*p31)+(-30.0*p10*p31)+(-18.0*p102*p31)+(-105.0*p20*p31)+(-45.0*p10*p20*p31)+(-45.0*p202*p31)))+(-2.0*p01*((3.0*p11*(5.0+(42.0*p102)+(20.0*p20)+(27.0*p202)+(9.0*p10*(2.0+(7.0*p20)))))+(3.0*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20))))*p21)+((14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p31)))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))))));
				propagate[x][y] += ((1.0/4.0)*((2.0*p01)+(-1.0*p012)+((-2.0+p31)*p31)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3234.0*p21)+(-2772.0*p10*p21)+(594.0*p102*p21)+(-990.0*p20*p21)+(594.0*p10*p20*p21)+(-612.0*p212)+(594.0*p10*p212)+(-135.0*p102*p212)+(252.0*p20*p212)+(-162.0*p10*p20*p212)+(-6160.0*p31)+(-2112.0*p10*p31)+(396.0*p102*p31)+(-3234.0*p20*p31)+(990.0*p10*p20*p31)+(990.0*p202*p31)+(-1050.0*p21*p31)+(822.0*p10*p21*p31)+(-162.0*p102*p21*p31)+(990.0*p20*p21*p31)+(-378.0*p10*p20*p21*p31)+(-252.0*p202*p21*p31)+(3850.0*p312)+(408.0*p10*p312)+(-63.0*p102*p312)+(1050.0*p20*p312)+(-252.0*p10*p20*p312)+(-378.0*p202*p312)+(2.0*p01*(407.0+(1221.0*p20)+(-198.0*p202)+(3.0*p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-111.0*p21)+(-144.0*p20*p21)+(45.0*p202*p21)+(9.0*p102*(-55.0+(14.0*p11)+(-1.0*p31)))+(-41.0*p31)+(-12.0*p20*p31)+(9.0*p202*p31)+(3.0*p10*(880.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))+(13.0*p31)))))+(-6.0*p11*(-352.0+(p202*(99.0+(-27.0*p21)))+(123.0*p21)+(p20*(-297.0+(36.0*p21)+(-28.0*p31)))+(68.0*p31)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*((2.0*p01)+(-1.0*p012)+((-2.0+p31)*p31)));
				break;
			case 12:
				p002 = p00*p00;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(-7.0+(42.0*p002)+(p00*(-2.0+(28.0*p10)))+(-18.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(9.0*(-42.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(1.0+(9.0*p10)+(5.0*p20))))*p212)+(6.0*((5.0*p002)+(27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(p00*(7.0+(20.0*p10)+(18.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(((5.0*p002)+(p00*(28.0+(30.0*p10)+(42.0*p20)))+(21.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))))*p312)+(6.0*p11*((3.0*(-14.0+(14.0*p002)+(-6.0*p10)+(21.0*p00*p10)+(9.0*p102)+(-21.0*p20)+(6.0*p00*p20)+(-9.0*p202))*p21)+((-42.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(1.0+(9.0*p10)+(5.0*p20))))*p31)))));
				propagate[x][y] += ((1.0/4.0)*p312);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(-9.0*(68.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20)))+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p212)+(-6.0*(175.0+(5.0*p002)+(27.0*p102)+(-165.0*p20)+(42.0*p202)+(2.0*p00*(-24.0+(10.0*p10)+(9.0*p20)))+(p10*(-137.0+(63.0*p20))))*p21*p31)+(-1.0*((5.0*p002)+(63.0*p102)+(12.0*p10*(-34.0+(21.0*p20)))+(p00*(-82.0+(30.0*p10)+(42.0*p20)))+(14.0*(-275.0+(-75.0*p20)+(27.0*p202))))*p312)+(-6.0*p11*((3.0*(41.0+(14.0*p002)+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202)+(p00*(-55.0+(21.0*p10)+(6.0*p20))))*p21)+((68.0+(7.0*p002)+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p31)))));
				propagate[x][y] += ((1.0/4.0)*p312);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*(-7.0+(42.0*p002)+(p00*(-2.0+(28.0*p10)))+(-18.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(-9.0*(-42.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(1.0+(9.0*p10)+(5.0*p20))))*p212)+(-6.0*p21*(385.0+(165.0*p20)+(5.0*p002*(-22.0+p31))+(-210.0*p31)+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(p10*((-99.0*p20)+(28.0*p31)+(63.0*p20*p31)))+(p00*(-11.0+(-66.0*p20)+(7.0*p31)+(18.0*p20*p31)+(5.0*p10*(-33.0+(4.0*p31)))))))+(-6.0*p11*(110.0+(165.0*p20)+(99.0*p202)+(-42.0*p21)+(-63.0*p20*p21)+(-27.0*p202*p21)+(-42.0*p31)+(-28.0*p20*p31)+(7.0*p002*(-55.0+(6.0*p21)+p31))+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(33.0*p20)+(-6.0*p21)+(6.0*p20*p31)))+(p00*(11.0+(18.0*p20*p21)+(2.0*p31)+(10.0*p20*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31)))))))+(-1.0*p31*(-3080.0+(-2310.0*p20)+(-990.0*p202)+(5.0*p002*(-22.0+p31))+(2310.0*p31)+(1260.0*p20*p31)+(378.0*p202*p31)+(9.0*p102*(-44.0+(7.0*p31)))+(2.0*p00*(-55.0+(15.0*p10*(-11.0+p31))+(14.0*p31)+(3.0*p20*(-44.0+(7.0*p31)))))+(6.0*p10*(-110.0+(42.0*p31)+(3.0*p20*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(9.0*(68.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20)))+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p212)+(6.0*p21*(-539.0+(165.0*p20)+(5.0*p002*(-22.0+p31))+(175.0*p31)+(-165.0*p20*p31)+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(p10*(462.0+(-99.0*p20)+(-137.0*p31)+(63.0*p20*p31)))+(p00*(451.0+(-66.0*p20)+(-48.0*p31)+(18.0*p20*p31)+(5.0*p10*(-33.0+(4.0*p31)))))))+(6.0*p11*(-352.0+(-297.0*p20)+(99.0*p202)+(123.0*p21)+(36.0*p20*p21)+(-27.0*p202*p21)+(68.0*p31)+(-28.0*p20*p31)+(7.0*p002*(-55.0+(6.0*p21)+p31))+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31)))))+(p00*(935.0+(3.0*(-55.0+(6.0*p20))*p21)+(-42.0*p31)+(10.0*p20*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31)))))))+(p31*(6160.0+(3234.0*p20)+(-990.0*p202)+(5.0*p002*(-22.0+p31))+(-3850.0*p31)+(-1050.0*p20*p31)+(378.0*p202*p31)+(9.0*p102*(-44.0+(7.0*p31)))+(p00*(814.0+(30.0*p10*(-11.0+p31))+(-82.0*p31)+(6.0*p20*(-44.0+(7.0*p31)))))+(6.0*p10*(352.0+(-68.0*p31)+(3.0*p20*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31);
				break;
			case 13:
				p002 = p00*p00;
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*(-7.0+(42.0*p002)+(p00*(-2.0+(28.0*p10)))+(-18.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(-9.0*(-42.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(1.0+(9.0*p10)+(5.0*p20))))*p212)+(-6.0*((5.0*p002)+(27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(p00*(7.0+(20.0*p10)+(18.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(-1.0*((5.0*p002)+(p00*(28.0+(30.0*p10)+(42.0*p20)))+(21.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))))*p312)+(-6.0*p11*((3.0*(-14.0+(14.0*p002)+(-6.0*p10)+(21.0*p00*p10)+(9.0*p102)+(-21.0*p20)+(6.0*p00*p20)+(-9.0*p202))*p21)+((-42.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(1.0+(9.0*p10)+(5.0*p20))))*p31)))));
				propagate[x][y] += ((-1.0/4.0)*p312);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(9.0*(68.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20)))+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p212)+(6.0*(175.0+(5.0*p002)+(27.0*p102)+(-165.0*p20)+(42.0*p202)+(2.0*p00*(-24.0+(10.0*p10)+(9.0*p20)))+(p10*(-137.0+(63.0*p20))))*p21*p31)+(((5.0*p002)+(63.0*p102)+(12.0*p10*(-34.0+(21.0*p20)))+(p00*(-82.0+(30.0*p10)+(42.0*p20)))+(14.0*(-275.0+(-75.0*p20)+(27.0*p202))))*p312)+(6.0*p11*((3.0*(41.0+(14.0*p002)+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202)+(p00*(-55.0+(21.0*p10)+(6.0*p20))))*p21)+((68.0+(7.0*p002)+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p31)))));
				propagate[x][y] += ((-1.0/4.0)*p312);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(-7.0+(42.0*p002)+(p00*(-2.0+(28.0*p10)))+(-18.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(9.0*(-42.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20)+(2.0*p00*(1.0+(9.0*p10)+(5.0*p20))))*p212)+(6.0*p21*(385.0+(165.0*p20)+(5.0*p002*(-22.0+p31))+(-210.0*p31)+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(p10*((-99.0*p20)+(28.0*p31)+(63.0*p20*p31)))+(p00*(-11.0+(-66.0*p20)+(7.0*p31)+(18.0*p20*p31)+(5.0*p10*(-33.0+(4.0*p31)))))))+(6.0*p11*(110.0+(165.0*p20)+(99.0*p202)+(-42.0*p21)+(-63.0*p20*p21)+(-27.0*p202*p21)+(-42.0*p31)+(-28.0*p20*p31)+(7.0*p002*(-55.0+(6.0*p21)+p31))+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(33.0*p20)+(-6.0*p21)+(6.0*p20*p31)))+(p00*(11.0+(18.0*p20*p21)+(2.0*p31)+(10.0*p20*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31)))))))+(p31*(-3080.0+(-2310.0*p20)+(-990.0*p202)+(5.0*p002*(-22.0+p31))+(2310.0*p31)+(1260.0*p20*p31)+(378.0*p202*p31)+(9.0*p102*(-44.0+(7.0*p31)))+(2.0*p00*(-55.0+(15.0*p10*(-11.0+p31))+(14.0*p31)+(3.0*p20*(-44.0+(7.0*p31)))))+(6.0*p10*(-110.0+(42.0*p31)+(3.0*p20*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*(-2.0+p31)*p31);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*(37.0+(42.0*p002)+(28.0*p00*(-4.0+p10))+(48.0*p20)+(-15.0*p202)+(-2.0*p10*(5.0+(9.0*p20)))))+(-9.0*(68.0+(7.0*p002)+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20)))+(2.0*p00*(-21.0+(9.0*p10)+(5.0*p20))))*p212)+(-6.0*p21*(-539.0+(165.0*p20)+(5.0*p002*(-22.0+p31))+(175.0*p31)+(-165.0*p20*p31)+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(p10*(462.0+(-99.0*p20)+(-137.0*p31)+(63.0*p20*p31)))+(p00*(451.0+(-66.0*p20)+(-48.0*p31)+(18.0*p20*p31)+(5.0*p10*(-33.0+(4.0*p31)))))))+(-6.0*p11*(-352.0+(-297.0*p20)+(99.0*p202)+(123.0*p21)+(36.0*p20*p21)+(-27.0*p202*p21)+(68.0*p31)+(-28.0*p20*p31)+(7.0*p002*(-55.0+(6.0*p21)+p31))+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31)))))+(p00*(935.0+(3.0*(-55.0+(6.0*p20))*p21)+(-42.0*p31)+(10.0*p20*p31)+(3.0*p10*(-55.0+(21.0*p21)+(6.0*p31)))))))+(-1.0*p31*(6160.0+(3234.0*p20)+(-990.0*p202)+(5.0*p002*(-22.0+p31))+(-3850.0*p31)+(-1050.0*p20*p31)+(378.0*p202*p31)+(9.0*p102*(-44.0+(7.0*p31)))+(p00*(814.0+(30.0*p10*(-11.0+p31))+(-82.0*p31)+(6.0*p20*(-44.0+(7.0*p31)))))+(6.0*p10*(352.0+(-68.0*p31)+(3.0*p20*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*(-2.0+p31)*p31);
				break;
			case 14:
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/6160.0)*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(2.0*p11*((3.0*(-14.0+(-6.0*p10)+(9.0*p102)+(-21.0*p20)+(-9.0*p202))*p21)+((-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))));
				propagate[x][y] += ((1.0/4.0)*p312);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-9.0*(68.0+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(-6.0*(175.0+(-137.0*p10)+(27.0*p102)+(-165.0*p20)+(63.0*p10*p20)+(42.0*p202))*p21*p31)+((3850.0+(408.0*p10)+(-63.0*p102)+(1050.0*p20)+(-252.0*p10*p20)+(-378.0*p202))*p312)+(-6.0*p11*((3.0*(41.0+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202))*p21)+((68.0+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))));
				propagate[x][y] += ((1.0/4.0)*p312);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-9.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(-6.0*p11*(110.0+(p202*(99.0+(-27.0*p21)))+(-42.0*p21)+(p20*(165.0+(-63.0*p21)+(-28.0*p31)))+(-42.0*p31)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-6.0*p21)+(p20*(33.0+(6.0*p31)))))))+(-6.0*p21*(385.0+(165.0*p20)+(-210.0*p31)+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(p10*((28.0*p31)+(9.0*p20*(-11.0+(7.0*p31)))))))+(-1.0*p31*((770.0*(-4.0+(3.0*p31)))+(210.0*p20*(-11.0+(6.0*p31)))+(9.0*p102*(-44.0+(7.0*p31)))+(18.0*p202*(-55.0+(21.0*p31)))+(6.0*p10*(-110.0+(42.0*p31)+(3.0*p20*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(9.0*(68.0+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20))))*p212)+(6.0*p11*(-352.0+(p202*(99.0+(-27.0*p21)))+(123.0*p21)+(p20*(-297.0+(36.0*p21)+(-28.0*p31)))+(68.0*p31)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31)))))))+(6.0*p21*((-165.0*p20*(-1.0+p31))+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(7.0*(-77.0+(25.0*p31)))+(p10*(462.0+(-137.0*p31)+(9.0*p20*(-11.0+(7.0*p31)))))))+(p31*((-770.0*(-8.0+(5.0*p31)))+(9.0*p102*(-44.0+(7.0*p31)))+(18.0*p202*(-55.0+(21.0*p31)))+(-42.0*p20*(-77.0+(25.0*p31)))+(6.0*p10*(352.0+(-68.0*p31)+(3.0*p20*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((-1.0/4.0)*(-2.0+p31)*p31);
				break;
			case 15:
				p312 = p31*p31;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/6160.0)*((3.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-3.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(-2.0*((27.0*p102)+(7.0*p10*(4.0+(9.0*p20)))+(42.0*(-5.0+p202)))*p21*p31)+(-7.0*((3.0*p102)+(12.0*p10*(1.0+p20))+(2.0*(55.0+(30.0*p20)+(9.0*p202))))*p312)+(p11*((6.0*(14.0+(6.0*p10)+(-9.0*p102)+(21.0*p20)+(9.0*p202))*p21)+(2.0*(42.0+(-15.0*p102)+(28.0*p20)+(-18.0*p10*p20))*p31)))));
				propagate[x][y] += ((-1.0/4.0)*p312);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(9.0*(68.0+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(6.0*(175.0+(-137.0*p10)+(27.0*p102)+(-165.0*p20)+(63.0*p10*p20)+(42.0*p202))*p21*p31)+((-3850.0+(-408.0*p10)+(63.0*p102)+(-1050.0*p20)+(252.0*p10*p20)+(378.0*p202))*p312)+(6.0*p11*((3.0*(41.0+(-39.0*p10)+(9.0*p102)+(12.0*p20)+(-9.0*p202))*p21)+((68.0+(-66.0*p10)+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p31)))));
				propagate[x][y] += ((-1.0/4.0)*p312);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*(7.0+(18.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(9.0*(-42.0+(15.0*p102)+(-28.0*p20)+(18.0*p10*p20))*p212)+(6.0*p11*(110.0+(p202*(99.0+(-27.0*p21)))+(-42.0*p21)+(p20*(165.0+(-63.0*p21)+(-28.0*p31)))+(-42.0*p31)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-6.0*p21)+(p20*(33.0+(6.0*p31)))))))+(6.0*p21*(385.0+(165.0*p20)+(-210.0*p31)+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(p10*((28.0*p31)+(9.0*p20*(-11.0+(7.0*p31)))))))+(p31*((770.0*(-4.0+(3.0*p31)))+(210.0*p20*(-11.0+(6.0*p31)))+(9.0*p102*(-44.0+(7.0*p31)))+(18.0*p202*(-55.0+(21.0*p31)))+(6.0*p10*(-110.0+(42.0*p31)+(3.0*p20*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*(-2.0+p31)*p31);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)+(2.0*p10*(5.0+(9.0*p20)))))+(-9.0*(68.0+(15.0*p102)+(-28.0*p20)+(6.0*p10*(-11.0+(3.0*p20))))*p212)+(-6.0*p11*(-352.0+(p202*(99.0+(-27.0*p21)))+(123.0*p21)+(p20*(-297.0+(36.0*p21)+(-28.0*p31)))+(68.0*p31)+(3.0*p102*((9.0*p21)+(5.0*p31)))+(3.0*p10*(22.0+(-39.0*p21)+(-22.0*p31)+(p20*(33.0+(6.0*p31)))))))+(-6.0*p21*((-165.0*p20*(-1.0+p31))+(42.0*p202*p31)+(9.0*p102*(-11.0+(3.0*p31)))+(7.0*(-77.0+(25.0*p31)))+(p10*(462.0+(-137.0*p31)+(9.0*p20*(-11.0+(7.0*p31)))))))+(p31*((p202*(990.0+(-378.0*p31)))+(770.0*(-8.0+(5.0*p31)))+(-9.0*p102*(-44.0+(7.0*p31)))+(42.0*p20*(-77.0+(25.0*p31)))+(-6.0*p10*(352.0+(-68.0*p31)+(3.0*p20*(-55.0+(14.0*p31)))))))));
				propagate[x][y] += ((1.0/4.0)*(-2.0+p31)*p31);
				break;
			case 16:
				p002 = p00*p00;
				p012 = p01*p01;
				p302 = p30*p30;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p002*(5.0+(-2310.0*p012)+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))+(28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(-2.0*p00*((3.0*p10*(-5.0+(210.0*p012)+(-18.0*p11)+(-42.0*p112)+(-20.0*p21)+(-63.0*p11*p21)+(-27.0*p212)+(-7.0*p01*(1.0+(4.0*p21)))))+(3.0*p20*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+((-14.0+(14.0*p012)+(-6.0*p11)+(21.0*p01*p11)+(9.0*p112)+(-21.0*p21)+(6.0*p01*p21)+(-9.0*p212))*p30)))+(-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30)))))+(-2.0*p01*((9.0*p102*(-1.0+(14.0*p11)))+(9.0*p202*(1.0+(9.0*p11)+(5.0*p21)))+(3.0*p20*(7.0+(20.0*p11)+(18.0*p21))*p30)+((14.0+(15.0*p11)+(21.0*p21))*p302)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+p30+(5.0*p21*p30)))))))));
				propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012)));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*(5.0+(-2310.0*p012)+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))+(28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(2.0*p00*((3.0*p10*(-5.0+(210.0*p012)+(-18.0*p11)+(-42.0*p112)+(-20.0*p21)+(-63.0*p11*p21)+(-27.0*p212)+(-7.0*p01*(1.0+(4.0*p21)))))+(3.0*p20*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+((-14.0+(14.0*p012)+(-6.0*p11)+(21.0*p01*p11)+(9.0*p112)+(-21.0*p21)+(6.0*p01*p21)+(-9.0*p212))*p30)))+(p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-22.0*((-30.0*p10)+(-18.0*p10*p11)+(-105.0*p20)+(27.0*p112*p20)+(-45.0*p10*p21)+(-27.0*p10*p11*p21)+(-45.0*p20*p21)+(27.0*p11*p20*p21)+(-27.0*p10*p212)+(p00*(-5.0+(280.0*p012)+(-12.0*p11)+(-45.0*p112)+(-15.0*p21)+(-45.0*p11*p21)+(-18.0*p212)+(-5.0*p01*(1.0+(21.0*p11)+(6.0*p21)))))+(-280.0*p30)+(30.0*p11*p30)+(18.0*p112*p30)+(105.0*p21*p30)+(45.0*p11*p21*p30)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(p01*((p10*(-3.0+(45.0*p11)))+(3.0*p20*(1.0+(15.0*p11)+(6.0*p21)))+((5.0+(15.0*p11)+(12.0*p21))*p30)))))+(-3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30)))))+(2.0*p01*((9.0*p102*(-1.0+(14.0*p11)))+(9.0*p202*(1.0+(9.0*p11)+(5.0*p21)))+(3.0*p20*(7.0+(20.0*p11)+(18.0*p21))*p30)+((14.0+(15.0*p11)+(21.0*p21))*p302)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+p30+(5.0*p21*p30)))))))));
				propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012)));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((612.0*p202)+(-378.0*p01*p202)+(63.0*p012*p202)+(-594.0*p11*p202)+(162.0*p01*p11*p202)+(135.0*p112*p202)+(-252.0*p202*p21)+(90.0*p01*p202*p21)+(162.0*p11*p202*p21)+(21.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(1050.0*p20*p30)+(-288.0*p01*p20*p30)+(30.0*p012*p20*p30)+(-822.0*p11*p20*p30)+(120.0*p01*p11*p20*p30)+(162.0*p112*p20*p30)+(-990.0*p20*p21*p30)+(108.0*p01*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(252.0*p20*p212*p30)+(770.0*p302)+(-82.0*p01*p302)+(5.0*p012*p302)+(-408.0*p11*p302)+(30.0*p01*p11*p302)+(63.0*p112*p302)+(-1050.0*p21*p302)+(42.0*p01*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(6.0*p10*((3.0*p20*(41.0+(14.0*p012)+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)+(p01*(-55.0+(21.0*p11)+(6.0*p21)))))+((68.0+(7.0*p012)+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21))))*p30)))+(2.0*p00*((3.0*p10*(50.0+(210.0*p012)+(-42.0*p112)+(145.0*p21)+(-27.0*p212)+(-28.0*p01*(14.0+p21))+(-21.0*p11*(-7.0+(3.0*p21)))))+(3.0*p20*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+((41.0+(14.0*p012)+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)+(p01*(-55.0+(21.0*p11)+(6.0*p21))))*p30)))));
				propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((3234.0*p20)+(-2706.0*p01*p20)+(660.0*p012*p20)+(-2772.0*p11*p20)+(990.0*p01*p11*p20)+(594.0*p112*p20)+(-612.0*p202)+(378.0*p01*p202)+(-63.0*p012*p202)+(594.0*p11*p202)+(-162.0*p01*p11*p202)+(-135.0*p112*p202)+(-990.0*p20*p21)+(396.0*p01*p20*p21)+(594.0*p11*p20*p21)+(252.0*p202*p21)+(-90.0*p01*p202*p21)+(-162.0*p11*p202*p21)+(-21.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(-9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(3080.0*p30)+(-814.0*p01*p30)+(110.0*p012*p30)+(-2112.0*p11*p30)+(330.0*p01*p11*p30)+(396.0*p112*p30)+(-1050.0*p20*p30)+(288.0*p01*p20*p30)+(-30.0*p012*p20*p30)+(822.0*p11*p20*p30)+(-120.0*p01*p11*p20*p30)+(-162.0*p112*p20*p30)+(-3234.0*p21*p30)+(264.0*p01*p21*p30)+(990.0*p11*p21*p30)+(990.0*p20*p21*p30)+(-108.0*p01*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(990.0*p212*p30)+(-252.0*p20*p212*p30)+(-770.0*p302)+(82.0*p01*p302)+(-5.0*p012*p302)+(408.0*p11*p302)+(-30.0*p01*p11*p302)+(-63.0*p112*p302)+(1050.0*p21*p302)+(-42.0*p01*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(-2.0*p00*(-407.0+(-2640.0*p11)+(495.0*p112)+(111.0*p20)+(-30.0*p11*p20)+(-1221.0*p21)+(495.0*p11*p21)+(144.0*p20*p21)+(-54.0*p11*p20*p21)+(198.0*p212)+(-45.0*p20*p212)+(-3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(41.0*p30)+(-39.0*p11*p30)+(9.0*p112*p30)+(12.0*p21*p30)+(-9.0*p212*p30)+(14.0*p012*(-220.0+(45.0*p10)+(9.0*p20)+p30))+(p01*(4675.0+(-336.0*p20)+(330.0*p21)+(-84.0*p10*(14.0+p21))+(-55.0*p30)+(6.0*p21*p30)+(21.0*p11*(55.0+(4.0*p20)+p30))))))+(-6.0*p10*(-352.0+(123.0*p20)+(-297.0*p21)+(36.0*p20*p21)+(99.0*p212)+(-27.0*p20*p212)+(68.0*p30)+(-28.0*p21*p30)+(7.0*p012*(-55.0+(6.0*p20)+p30))+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30)))+(p01*(935.0+(3.0*p20*(-55.0+(6.0*p21)))+(-42.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				break;
			case 17:
				p002 = p00*p00;
				p012 = p01*p01;
				p302 = p30*p30;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*(5.0+(-2310.0*p012)+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))+(28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(2.0*p00*((3.0*p10*(-5.0+(210.0*p012)+(-18.0*p11)+(-42.0*p112)+(-20.0*p21)+(-63.0*p11*p21)+(-27.0*p212)+(-7.0*p01*(1.0+(4.0*p21)))))+(3.0*p20*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+((-14.0+(14.0*p012)+(-6.0*p11)+(21.0*p01*p11)+(9.0*p112)+(-21.0*p21)+(6.0*p01*p21)+(-9.0*p212))*p30)))+(p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30)))))+(2.0*p01*((9.0*p102*(-1.0+(14.0*p11)))+(9.0*p202*(1.0+(9.0*p11)+(5.0*p21)))+(3.0*p20*(7.0+(20.0*p11)+(18.0*p21))*p30)+((14.0+(15.0*p11)+(21.0*p21))*p302)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+p30+(5.0*p21*p30)))))))));
				propagate[x][y] += ((1.0/4.0)*(-1.0+p012));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p002*(5.0+(-2310.0*p012)+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))+(28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(-2.0*p00*((3.0*p10*(-5.0+(210.0*p012)+(-18.0*p11)+(-42.0*p112)+(-20.0*p21)+(-63.0*p11*p21)+(-27.0*p212)+(-7.0*p01*(1.0+(4.0*p21)))))+(3.0*p20*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+((-14.0+(14.0*p012)+(-6.0*p11)+(21.0*p01*p11)+(9.0*p112)+(-21.0*p21)+(6.0*p01*p21)+(-9.0*p212))*p30)))+(-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(22.0*((-30.0*p10)+(-18.0*p10*p11)+(-105.0*p20)+(27.0*p112*p20)+(-45.0*p10*p21)+(-27.0*p10*p11*p21)+(-45.0*p20*p21)+(27.0*p11*p20*p21)+(-27.0*p10*p212)+(p00*(-5.0+(280.0*p012)+(-12.0*p11)+(-45.0*p112)+(-15.0*p21)+(-45.0*p11*p21)+(-18.0*p212)+(-5.0*p01*(1.0+(21.0*p11)+(6.0*p21)))))+(-280.0*p30)+(30.0*p11*p30)+(18.0*p112*p30)+(105.0*p21*p30)+(45.0*p11*p21*p30)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(p01*((p10*(-3.0+(45.0*p11)))+(3.0*p20*(1.0+(15.0*p11)+(6.0*p21)))+((5.0+(15.0*p11)+(12.0*p21))*p30)))))+(3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30)))))+(-2.0*p01*((9.0*p102*(-1.0+(14.0*p11)))+(9.0*p202*(1.0+(9.0*p11)+(5.0*p21)))+(3.0*p20*(7.0+(20.0*p11)+(18.0*p21))*p30)+((14.0+(15.0*p11)+(21.0*p21))*p302)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+p30+(5.0*p21*p30)))))))));
				propagate[x][y] += ((1.0/4.0)*(-1.0+p012));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-612.0*p202)+(378.0*p01*p202)+(-63.0*p012*p202)+(594.0*p11*p202)+(-162.0*p01*p11*p202)+(-135.0*p112*p202)+(252.0*p202*p21)+(-90.0*p01*p202*p21)+(-162.0*p11*p202*p21)+(-21.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(-9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(-1050.0*p20*p30)+(288.0*p01*p20*p30)+(-30.0*p012*p20*p30)+(822.0*p11*p20*p30)+(-120.0*p01*p11*p20*p30)+(-162.0*p112*p20*p30)+(990.0*p20*p21*p30)+(-108.0*p01*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(-252.0*p20*p212*p30)+(-770.0*p302)+(82.0*p01*p302)+(-5.0*p012*p302)+(408.0*p11*p302)+(-30.0*p01*p11*p302)+(-63.0*p112*p302)+(1050.0*p21*p302)+(-42.0*p01*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(-6.0*p10*((3.0*p20*(41.0+(14.0*p012)+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)+(p01*(-55.0+(21.0*p11)+(6.0*p21)))))+((68.0+(7.0*p012)+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21))))*p30)))+(-2.0*p00*((3.0*p10*(50.0+(210.0*p012)+(-42.0*p112)+(145.0*p21)+(-27.0*p212)+(-28.0*p01*(14.0+p21))+(-21.0*p11*(-7.0+(3.0*p21)))))+(3.0*p20*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+((41.0+(14.0*p012)+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)+(p01*(-55.0+(21.0*p11)+(6.0*p21))))*p30)))));
				propagate[x][y] += ((-1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-3234.0*p20)+(2706.0*p01*p20)+(-660.0*p012*p20)+(2772.0*p11*p20)+(-990.0*p01*p11*p20)+(-594.0*p112*p20)+(612.0*p202)+(-378.0*p01*p202)+(63.0*p012*p202)+(-594.0*p11*p202)+(162.0*p01*p11*p202)+(135.0*p112*p202)+(990.0*p20*p21)+(-396.0*p01*p20*p21)+(-594.0*p11*p20*p21)+(-252.0*p202*p21)+(90.0*p01*p202*p21)+(162.0*p11*p202*p21)+(21.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(-3080.0*p30)+(814.0*p01*p30)+(-110.0*p012*p30)+(2112.0*p11*p30)+(-330.0*p01*p11*p30)+(-396.0*p112*p30)+(1050.0*p20*p30)+(-288.0*p01*p20*p30)+(30.0*p012*p20*p30)+(-822.0*p11*p20*p30)+(120.0*p01*p11*p20*p30)+(162.0*p112*p20*p30)+(3234.0*p21*p30)+(-264.0*p01*p21*p30)+(-990.0*p11*p21*p30)+(-990.0*p20*p21*p30)+(108.0*p01*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(-990.0*p212*p30)+(252.0*p20*p212*p30)+(770.0*p302)+(-82.0*p01*p302)+(5.0*p012*p302)+(-408.0*p11*p302)+(30.0*p01*p11*p302)+(63.0*p112*p302)+(-1050.0*p21*p302)+(42.0*p01*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(2.0*p00*(-407.0+(-2640.0*p11)+(495.0*p112)+(111.0*p20)+(-30.0*p11*p20)+(-1221.0*p21)+(495.0*p11*p21)+(144.0*p20*p21)+(-54.0*p11*p20*p21)+(198.0*p212)+(-45.0*p20*p212)+(-3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(41.0*p30)+(-39.0*p11*p30)+(9.0*p112*p30)+(12.0*p21*p30)+(-9.0*p212*p30)+(14.0*p012*(-220.0+(45.0*p10)+(9.0*p20)+p30))+(p01*(4675.0+(-336.0*p20)+(330.0*p21)+(-84.0*p10*(14.0+p21))+(-55.0*p30)+(6.0*p21*p30)+(21.0*p11*(55.0+(4.0*p20)+p30))))))+(6.0*p10*(-352.0+(123.0*p20)+(-297.0*p21)+(36.0*p20*p21)+(99.0*p212)+(-27.0*p20*p212)+(68.0*p30)+(-28.0*p21*p30)+(7.0*p012*(-55.0+(6.0*p20)+p30))+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30)))+(p01*(935.0+(3.0*p20*(-55.0+(6.0*p21)))+(-42.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))))));
				propagate[x][y] += ((-1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				break;
			case 18:
				p012 = p01*p01;
				p302 = p30*p30;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(-42.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(1.0+(9.0*p11)+(5.0*p21)))))+(-9.0*p102*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p20*((5.0*p012)+(27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(p01*(7.0+(20.0*p11)+(18.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-1.0*((5.0*p012)+(p01*(28.0+(30.0*p11)+(42.0*p21)))+(21.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))))*p302)+(-6.0*p10*((3.0*p20*(-14.0+(14.0*p012)+(-6.0*p11)+(21.0*p01*p11)+(9.0*p112)+(-21.0*p21)+(6.0*p01*p21)+(-9.0*p212)))+((-42.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(1.0+(9.0*p11)+(5.0*p21))))*p30)))));
				propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012)));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(-42.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(1.0+(9.0*p11)+(5.0*p21)))))+(9.0*p102*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(6.0*p20*(385.0+(165.0*p21)+(5.0*p012*(-22.0+p30))+(-210.0*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(p11*((-99.0*p21)+(28.0*p30)+(63.0*p21*p30)))+(p01*(-11.0+(-66.0*p21)+(7.0*p30)+(18.0*p21*p30)+(5.0*p11*(-33.0+(4.0*p30)))))))+(6.0*p10*(110.0+(-42.0*p20)+(165.0*p21)+(-63.0*p20*p21)+(99.0*p212)+(-27.0*p20*p212)+(-42.0*p30)+(-28.0*p21*p30)+(7.0*p012*(-55.0+(6.0*p20)+p30))+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-18.0*p20)+(99.0*p21)+(18.0*p21*p30)))+(p01*(11.0+(18.0*p20*p21)+(2.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))))+(p30*(6160.0+(-2310.0*p21)+(-990.0*p212)+(5.0*p012*(-22.0+p30))+(-2310.0*p30)+(1260.0*p21*p30)+(378.0*p212*p30)+(9.0*p112*(-44.0+(7.0*p30)))+(2.0*p01*(-55.0+(15.0*p11*(-11.0+p30))+(14.0*p30)+(3.0*p21*(-44.0+(7.0*p30)))))+(6.0*p11*(-110.0+(42.0*p30)+(3.0*p21*(-55.0+(14.0*p30)))))))));
				propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012)));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(68.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21)))+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21)))))+(9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(6.0*p20*(175.0+(5.0*p012)+(27.0*p112)+(-165.0*p21)+(42.0*p212)+(2.0*p01*(-24.0+(10.0*p11)+(9.0*p21)))+(p11*(-137.0+(63.0*p21))))*p30)+(((5.0*p012)+(63.0*p112)+(12.0*p11*(-34.0+(21.0*p21)))+(p01*(-82.0+(30.0*p11)+(42.0*p21)))+(14.0*(55.0+(-75.0*p21)+(27.0*p212))))*p302)+(6.0*p10*((3.0*p20*(41.0+(14.0*p012)+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)+(p01*(-55.0+(21.0*p11)+(6.0*p21)))))+((68.0+(7.0*p012)+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21))))*p30)))));
				propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(68.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21)))+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21)))))+(-9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p20*(-539.0+(165.0*p21)+(5.0*p012*(-22.0+p30))+(175.0*p30)+(-165.0*p21*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(p11*(462.0+(-99.0*p21)+(-137.0*p30)+(63.0*p21*p30)))+(p01*(451.0+(-66.0*p21)+(-48.0*p30)+(18.0*p21*p30)+(5.0*p11*(-33.0+(4.0*p30)))))))+(-6.0*p10*(-352.0+(123.0*p20)+(-297.0*p21)+(36.0*p20*p21)+(99.0*p212)+(-27.0*p20*p212)+(68.0*p30)+(-28.0*p21*p30)+(7.0*p012*(-55.0+(6.0*p20)+p30))+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30)))+(p01*(935.0+(3.0*p20*(-55.0+(6.0*p21)))+(-42.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))))+(-1.0*p30*(-3080.0+(3234.0*p21)+(-990.0*p212)+(5.0*p012*(-22.0+p30))+(770.0*p30)+(-1050.0*p21*p30)+(378.0*p212*p30)+(9.0*p112*(-44.0+(7.0*p30)))+(p01*(814.0+(30.0*p11*(-11.0+p30))+(-82.0*p30)+(6.0*p21*(-44.0+(7.0*p30)))))+(6.0*p11*(352.0+(-68.0*p30)+(3.0*p21*(-55.0+(14.0*p30)))))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				break;
			case 19:
				p012 = p01*p01;
				p302 = p30*p30;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(-42.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(1.0+(9.0*p11)+(5.0*p21)))))+(9.0*p102*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(6.0*p20*((5.0*p012)+(27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(p01*(7.0+(20.0*p11)+(18.0*p21)))+(42.0*(-5.0+p212)))*p30)+(((5.0*p012)+(p01*(28.0+(30.0*p11)+(42.0*p21)))+(21.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))))*p302)+(6.0*p10*((3.0*p20*(-14.0+(14.0*p012)+(-6.0*p11)+(21.0*p01*p11)+(9.0*p112)+(-21.0*p21)+(6.0*p01*p21)+(-9.0*p212)))+((-42.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(1.0+(9.0*p11)+(5.0*p21))))*p30)))));
				propagate[x][y] += ((1.0/4.0)*(-1.0+p012));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(-42.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(1.0+(9.0*p11)+(5.0*p21)))))+(-9.0*p102*(-7.0+(42.0*p012)+(p01*(-2.0+(28.0*p11)))+(-18.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p20*(385.0+(165.0*p21)+(5.0*p012*(-22.0+p30))+(-210.0*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(p11*((-99.0*p21)+(28.0*p30)+(63.0*p21*p30)))+(p01*(-11.0+(-66.0*p21)+(7.0*p30)+(18.0*p21*p30)+(5.0*p11*(-33.0+(4.0*p30)))))))+(-6.0*p10*(110.0+(-42.0*p20)+(165.0*p21)+(-63.0*p20*p21)+(99.0*p212)+(-27.0*p20*p212)+(-42.0*p30)+(-28.0*p21*p30)+(7.0*p012*(-55.0+(6.0*p20)+p30))+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-18.0*p20)+(99.0*p21)+(18.0*p21*p30)))+(p01*(11.0+(18.0*p20*p21)+(2.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))))+(-1.0*p30*(6160.0+(-2310.0*p21)+(-990.0*p212)+(5.0*p012*(-22.0+p30))+(-2310.0*p30)+(1260.0*p21*p30)+(378.0*p212*p30)+(9.0*p112*(-44.0+(7.0*p30)))+(2.0*p01*(-55.0+(15.0*p11*(-11.0+p30))+(14.0*p30)+(3.0*p21*(-44.0+(7.0*p30)))))+(6.0*p11*(-110.0+(42.0*p30)+(3.0*p21*(-55.0+(14.0*p30)))))))));
				propagate[x][y] += ((1.0/4.0)*(-1.0+p012));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(68.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21)))+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21)))))+(-9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p20*(175.0+(5.0*p012)+(27.0*p112)+(-165.0*p21)+(42.0*p212)+(2.0*p01*(-24.0+(10.0*p11)+(9.0*p21)))+(p11*(-137.0+(63.0*p21))))*p30)+(-1.0*((5.0*p012)+(63.0*p112)+(12.0*p11*(-34.0+(21.0*p21)))+(p01*(-82.0+(30.0*p11)+(42.0*p21)))+(14.0*(55.0+(-75.0*p21)+(27.0*p212))))*p302)+(-6.0*p10*((3.0*p20*(41.0+(14.0*p012)+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)+(p01*(-55.0+(21.0*p11)+(6.0*p21)))))+((68.0+(7.0*p012)+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21))))*p30)))));
				propagate[x][y] += ((-1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(68.0+(7.0*p012)+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21)))+(2.0*p01*(-21.0+(9.0*p11)+(5.0*p21)))))+(9.0*p102*(37.0+(42.0*p012)+(28.0*p01*(-4.0+p11))+(48.0*p21)+(-15.0*p212)+(-2.0*p11*(5.0+(9.0*p21)))))+(6.0*p20*(-539.0+(165.0*p21)+(5.0*p012*(-22.0+p30))+(175.0*p30)+(-165.0*p21*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(p11*(462.0+(-99.0*p21)+(-137.0*p30)+(63.0*p21*p30)))+(p01*(451.0+(-66.0*p21)+(-48.0*p30)+(18.0*p21*p30)+(5.0*p11*(-33.0+(4.0*p30)))))))+(6.0*p10*(-352.0+(123.0*p20)+(-297.0*p21)+(36.0*p20*p21)+(99.0*p212)+(-27.0*p20*p212)+(68.0*p30)+(-28.0*p21*p30)+(7.0*p012*(-55.0+(6.0*p20)+p30))+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30)))+(p01*(935.0+(3.0*p20*(-55.0+(6.0*p21)))+(-42.0*p30)+(10.0*p21*p30)+(3.0*p11*(-55.0+(21.0*p20)+(6.0*p30)))))))+(p30*(-3080.0+(3234.0*p21)+(-990.0*p212)+(5.0*p012*(-22.0+p30))+(770.0*p30)+(-1050.0*p21*p30)+(378.0*p212*p30)+(9.0*p112*(-44.0+(7.0*p30)))+(p01*(814.0+(30.0*p11*(-11.0+p30))+(-82.0*p30)+(6.0*p21*(-44.0+(7.0*p30)))))+(6.0*p11*(352.0+(-68.0*p30)+(3.0*p21*(-55.0+(14.0*p30)))))))));
				propagate[x][y] += ((-1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				break;
			case 20:
				p002 = p00*p00;
				p302 = p30*p30;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(2.0*p00*((3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))+(3.0*p20*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+((14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212))*p30)))+(3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30)))))));
				propagate[x][y] += (1.0/4.0);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(22.0*((105.0*p20)+(-27.0*p112*p20)+(45.0*p20*p21)+(-27.0*p11*p20*p21)+(3.0*p10*(10.0+(6.0*p11)+(15.0*p21)+(9.0*p11*p21)+(9.0*p212)))+(p00*(5.0+(45.0*p112)+(15.0*p21)+(18.0*p212)+(3.0*p11*(4.0+(15.0*p21)))))+(280.0*p30)+(-30.0*p11*p30)+(-18.0*p112*p30)+(-105.0*p21*p30)+(-45.0*p11*p21*p30)+(-45.0*p212*p30)))+(-2.0*p00*((3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))+(3.0*p20*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+((14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212))*p30)))+(-3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30)))))));
				propagate[x][y] += (1.0/4.0);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((612.0*p202)+(-594.0*p11*p202)+(135.0*p112*p202)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(-21.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(1050.0*p20*p30)+(-822.0*p11*p20*p30)+(162.0*p112*p20*p30)+(-990.0*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(252.0*p20*p212*p30)+(770.0*p302)+(-408.0*p11*p302)+(63.0*p112*p302)+(-1050.0*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(-2.0*p00*((3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(3.0*p20*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+((-41.0+(39.0*p11)+(-9.0*p112)+(-12.0*p21)+(9.0*p212))*p30)))+(6.0*p10*((3.0*p20*(41.0+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)))+((68.0+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21))))*p30)))));
				propagate[x][y] += (1.0/4.0);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((3234.0*p20)+(-2772.0*p11*p20)+(594.0*p112*p20)+(-612.0*p202)+(594.0*p11*p202)+(-135.0*p112*p202)+(-990.0*p20*p21)+(594.0*p11*p20*p21)+(252.0*p202*p21)+(-162.0*p11*p202*p21)+(21.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(3080.0*p30)+(-2112.0*p11*p30)+(396.0*p112*p30)+(-1050.0*p20*p30)+(822.0*p11*p20*p30)+(-162.0*p112*p20*p30)+(-3234.0*p21*p30)+(990.0*p11*p21*p30)+(990.0*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(990.0*p212*p30)+(-252.0*p20*p212*p30)+(-770.0*p302)+(408.0*p11*p302)+(-63.0*p112*p302)+(1050.0*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(2.0*p00*(407.0+(-111.0*p20)+(1221.0*p21)+(-144.0*p20*p21)+(-198.0*p212)+(45.0*p20*p212)+(3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(-41.0*p30)+(-12.0*p21*p30)+(9.0*p212*p30)+(-9.0*p112*(55.0+p30))+(3.0*p11*(880.0+(-165.0*p21)+(2.0*p20*(5.0+(9.0*p21)))+(13.0*p30)))))+(-6.0*p10*(-352.0+(-297.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))+(68.0*p30)+(-28.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30)))))));
				propagate[x][y] += (1.0/4.0);
				break;
			case 21:
				p002 = p00*p00;
				p302 = p30*p30;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(-2.0*p00*((3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))+(3.0*p20*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+((14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212))*p30)))+(-3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30)))))));
				propagate[x][y] += (-1.0/4.0);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(-22.0*((105.0*p20)+(-27.0*p112*p20)+(45.0*p20*p21)+(-27.0*p11*p20*p21)+(3.0*p10*(10.0+(6.0*p11)+(15.0*p21)+(9.0*p11*p21)+(9.0*p212)))+(p00*(5.0+(45.0*p112)+(15.0*p21)+(18.0*p212)+(3.0*p11*(4.0+(15.0*p21)))))+(280.0*p30)+(-30.0*p11*p30)+(-18.0*p112*p30)+(-105.0*p21*p30)+(-45.0*p11*p21*p30)+(-45.0*p212*p30)))+(2.0*p00*((3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))+(3.0*p20*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+((14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212))*p30)))+(3.0*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30)))))));
				propagate[x][y] += (-1.0/4.0);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-612.0*p202)+(594.0*p11*p202)+(-135.0*p112*p202)+(252.0*p202*p21)+(-162.0*p11*p202*p21)+(21.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-1050.0*p20*p30)+(822.0*p11*p20*p30)+(-162.0*p112*p20*p30)+(990.0*p20*p21*p30)+(-378.0*p11*p20*p21*p30)+(-252.0*p20*p212*p30)+(-770.0*p302)+(408.0*p11*p302)+(-63.0*p112*p302)+(1050.0*p21*p302)+(-252.0*p11*p21*p302)+(-378.0*p212*p302)+(2.0*p00*((3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(3.0*p20*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+((-41.0+(39.0*p11)+(-9.0*p112)+(-12.0*p21)+(9.0*p212))*p30)))+(-6.0*p10*((3.0*p20*(41.0+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)))+((68.0+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21))))*p30)))));
				propagate[x][y] += (-1.0/4.0);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-3234.0*p20)+(2772.0*p11*p20)+(-594.0*p112*p20)+(612.0*p202)+(-594.0*p11*p202)+(135.0*p112*p202)+(990.0*p20*p21)+(-594.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(-21.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-3080.0*p30)+(2112.0*p11*p30)+(-396.0*p112*p30)+(1050.0*p20*p30)+(-822.0*p11*p20*p30)+(162.0*p112*p20*p30)+(3234.0*p21*p30)+(-990.0*p11*p21*p30)+(-990.0*p20*p21*p30)+(378.0*p11*p20*p21*p30)+(-990.0*p212*p30)+(252.0*p20*p212*p30)+(770.0*p302)+(-408.0*p11*p302)+(63.0*p112*p302)+(-1050.0*p21*p302)+(252.0*p11*p21*p302)+(378.0*p212*p302)+(-2.0*p00*(407.0+(-111.0*p20)+(1221.0*p21)+(-144.0*p20*p21)+(-198.0*p212)+(45.0*p20*p212)+(3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(-41.0*p30)+(-12.0*p21*p30)+(9.0*p212*p30)+(-9.0*p112*(55.0+p30))+(3.0*p11*(880.0+(-165.0*p21)+(2.0*p20*(5.0+(9.0*p21)))+(13.0*p30)))))+(6.0*p10*(-352.0+(-297.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))+(68.0*p30)+(-28.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30)))))));
				propagate[x][y] += (-1.0/4.0);
				break;
			case 22:
				p302 = p30*p30;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/6160.0)*((-3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(-7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(p10*((6.0*p20*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p21)+(9.0*p212)))+(2.0*(42.0+(-15.0*p112)+(28.0*p21)+(-18.0*p11*p21))*p30)))));
				propagate[x][y] += (1.0/4.0);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(-9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p10*(110.0+(165.0*p21)+(99.0*p212)+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))+(-42.0*p30)+(-28.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-18.0*p20)+(99.0*p21)+(18.0*p21*p30)))))+(6.0*p20*(385.0+(165.0*p21)+(-210.0*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(p11*((28.0*p30)+(9.0*p21*(-11.0+(7.0*p30)))))))+(p30*((-770.0*(-8.0+(3.0*p30)))+(210.0*p21*(-11.0+(6.0*p30)))+(9.0*p112*(-44.0+(7.0*p30)))+(18.0*p212*(-55.0+(21.0*p30)))+(6.0*p11*(-110.0+(42.0*p30)+(3.0*p21*(-55.0+(14.0*p30)))))))));
				propagate[x][y] += (1.0/4.0);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(68.0+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p20*(175.0+(-137.0*p11)+(27.0*p112)+(-165.0*p21)+(63.0*p11*p21)+(42.0*p212))*p30)+((770.0+(-408.0*p11)+(63.0*p112)+(-1050.0*p21)+(252.0*p11*p21)+(378.0*p212))*p302)+(6.0*p10*((3.0*p20*(41.0+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)))+((68.0+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21))*p30)))));
				propagate[x][y] += (1.0/4.0);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(68.0+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21)))))+(9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p10*(-352.0+(-297.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))+(68.0*p30)+(-28.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30)))))+(-6.0*p20*((-165.0*p21*(-1.0+p30))+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(7.0*(-77.0+(25.0*p30)))+(p11*(462.0+(-137.0*p30)+(9.0*p21*(-11.0+(7.0*p30)))))))+(p30*((p212*(990.0+(-378.0*p30)))+(-770.0*(-4.0+p30))+(-9.0*p112*(-44.0+(7.0*p30)))+(42.0*p21*(-77.0+(25.0*p30)))+(-6.0*p11*(352.0+(-68.0*p30)+(3.0*p21*(-55.0+(14.0*p30)))))))));
				propagate[x][y] += (1.0/4.0);
				break;
			case 23:
				p302 = p30*p30;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/6160.0)*((3.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(-3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(2.0*p20*((27.0*p112)+(7.0*p11*(4.0+(9.0*p21)))+(42.0*(-5.0+p212)))*p30)+(7.0*((3.0*p112)+(12.0*p11*(1.0+p21))+(2.0*(-55.0+(30.0*p21)+(9.0*p212))))*p302)+(2.0*p10*((3.0*p20*(-14.0+(-6.0*p11)+(9.0*p112)+(-21.0*p21)+(-9.0*p212)))+((-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21))*p30)))));
				propagate[x][y] += (-1.0/4.0);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(-42.0+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p10*(110.0+(165.0*p21)+(99.0*p212)+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))+(-42.0*p30)+(-28.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-18.0*p20)+(99.0*p21)+(18.0*p21*p30)))))+(-6.0*p20*(385.0+(165.0*p21)+(-210.0*p30)+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(p11*((28.0*p30)+(9.0*p21*(-11.0+(7.0*p30)))))))+(-1.0*p30*((-770.0*(-8.0+(3.0*p30)))+(210.0*p21*(-11.0+(6.0*p30)))+(9.0*p112*(-44.0+(7.0*p30)))+(18.0*p212*(-55.0+(21.0*p30)))+(6.0*p11*(-110.0+(42.0*p30)+(3.0*p21*(-55.0+(14.0*p30)))))))));
				propagate[x][y] += (-1.0/4.0);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p202*(68.0+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21)))+(9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p20*(175.0+(-137.0*p11)+(27.0*p112)+(-165.0*p21)+(63.0*p11*p21)+(42.0*p212))*p30)+(-1.0*(770.0+(-408.0*p11)+(63.0*p112)+(-1050.0*p21)+(252.0*p11*p21)+(378.0*p212))*p302)+(-6.0*p10*((3.0*p20*(41.0+(-39.0*p11)+(9.0*p112)+(12.0*p21)+(-9.0*p212)))+((68.0+(-66.0*p11)+(15.0*p112)+(-28.0*p21)+(18.0*p11*p21))*p30)))));
				propagate[x][y] += (-1.0/4.0);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p202*(68.0+(15.0*p112)+(-28.0*p21)+(6.0*p11*(-11.0+(3.0*p21)))))+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p10*(-352.0+(-297.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))+(68.0*p30)+(-28.0*p21*p30)+(3.0*p112*((9.0*p20)+(5.0*p30)))+(p11*(66.0+(-117.0*p20)+(99.0*p21)+(-66.0*p30)+(18.0*p21*p30)))))+(6.0*p20*((-165.0*p21*(-1.0+p30))+(42.0*p212*p30)+(9.0*p112*(-11.0+(3.0*p30)))+(7.0*(-77.0+(25.0*p30)))+(p11*(462.0+(-137.0*p30)+(9.0*p21*(-11.0+(7.0*p30)))))))+(p30*((770.0*(-4.0+p30))+(9.0*p112*(-44.0+(7.0*p30)))+(18.0*p212*(-55.0+(21.0*p30)))+(-42.0*p21*(-77.0+(25.0*p30)))+(6.0*p11*(352.0+(-68.0*p30)+(3.0*p21*(-55.0+(14.0*p30)))))))));
				propagate[x][y] += (-1.0/4.0);
				break;
			case 24:
				p002 = p00*p00;
				p012 = p01*p01;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(p002*(5.0+(-2310.0*p012)+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))+(28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(2.0*p00*(14.0+(6.0*p11)+(-9.0*p112)+(21.0*p20)+(30.0*p11*p20)+(-14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(21.0*p21)+(54.0*p20*p21)+(54.0*p11*p20*p21)+(9.0*p212)+(45.0*p20*p212)+(3.0*p01*((-7.0*p11)+(2.0*p20)+(-28.0*p11*p20)+(-2.0*p21)+(7.0*p10*(1.0+(4.0*p21)))))+(3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))))+(-2.0*p01*(14.0+(9.0*p102*(-1.0+(14.0*p11)))+(21.0*p20)+(9.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*((9.0*p11*(2.0+(7.0*p20)))+(2.0*(1.0+(5.0*p21)+(9.0*p20*p21)))))))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(p10*(84.0+(-6.0*p112*(5.0+(9.0*p20)))+(36.0*p11*(p20+(-1.0*p21)))+(56.0*p21)+(6.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))))));
				propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012)));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*(3850.0+(408.0*p10)+(-63.0*p102)+(-408.0*p11)+(396.0*p10*p11)+(-90.0*p102*p11)+(-333.0*p112)+(90.0*p10*p112)+(1050.0*p20)+(-252.0*p10*p20)+(168.0*p11*p20)+(-108.0*p10*p11*p20)+(-432.0*p112*p20)+(162.0*p10*p112*p20)+(-378.0*p202)+(135.0*p112*p202)+(21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-1050.0*p21)+(822.0*p10*p21)+(-162.0*p102*p21)+(-738.0*p11*p21)+(702.0*p10*p11*p21)+(-162.0*p102*p11*p21)+(990.0*p20*p21)+(-378.0*p10*p20*p21)+(-216.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(-612.0*p212)+(594.0*p10*p212)+(-135.0*p102*p212)+(252.0*p20*p212)+(-162.0*p10*p20*p212)+(p002*(-5.0+(2310.0*p012)+(-42.0*p11)+(-378.0*p112)+(-30.0*p21)+(-252.0*p11*p21)+(-63.0*p212)+(-28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(2.0*p01*(-41.0+(9.0*p102*(-1.0+(14.0*p11)))+(-12.0*p20)+(9.0*p202)+(3.0*p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-111.0*p21)+(-144.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(13.0+(21.0*p11*(-7.0+(3.0*p20)))+(2.0*(5.0+(9.0*p20))*p21)))))+(2.0*p00*(41.0+(126.0*p11)+(504.0*p112)+(-21.0*p20)+(-30.0*p11*p20)+(42.0*p012*(-73.0+(15.0*p10)+(3.0*p20)))+(144.0*p21)+(495.0*p11*p21)+(-54.0*p20*p21)+(-54.0*p11*p20*p21)+(189.0*p212)+(-45.0*p20*p212)+(p01*(55.0+(-6.0*p20)+(84.0*p11*(14.0+p20))+(336.0*p21)+(-21.0*p10*(1.0+(4.0*p21)))))+(-3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))))));
				propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012)));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*(770.0+(408.0*p10)+(333.0*p102)+(-408.0*p11)+(-396.0*p10*p11)+(-90.0*p102*p11)+(63.0*p112)+(90.0*p10*p112)+(1050.0*p20)+(738.0*p10*p20)+(-822.0*p11*p20)+(-702.0*p10*p11*p20)+(162.0*p112*p20)+(162.0*p10*p112*p20)+(612.0*p202)+(-594.0*p11*p202)+(135.0*p112*p202)+(p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-1050.0*p21)+(-168.0*p10*p21)+(432.0*p102*p21)+(252.0*p11*p21)+(108.0*p10*p11*p21)+(-162.0*p102*p11*p21)+(-990.0*p20*p21)+(216.0*p10*p20*p21)+(378.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(378.0*p212)+(-135.0*p102*p212)+(252.0*p20*p212)+(-162.0*p10*p20*p212)+(21.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(2.0*p00*(41.0+(-39.0*p11)+(9.0*p112)+(111.0*p20)+(-30.0*p11*p20)+(14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(12.0*p21)+(144.0*p20*p21)+(-54.0*p11*p20*p21)+(-9.0*p212)+(-45.0*p20*p212)+(p01*(-55.0+(21.0*p11)+(-336.0*p20)+(84.0*p11*p20)+(6.0*p21)+(-84.0*p10*(14.0+p21))))+(-3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))))+(2.0*p01*(-41.0+(126.0*p102*(-4.0+p11))+(-144.0*p20)+(-189.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(-42.0+(9.0*p11*(2.0+(7.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/6160.0)*(770.0+(568.0*p10)+(-111.0*p102)+(-568.0*p11)+(30.0*p102*p11)+(111.0*p112)+(-30.0*p10*p112)+(728.0*p20)+(-246.0*p10*p20)+(-650.0*p11*p20)+(234.0*p10*p11*p20)+(144.0*p112*p20)+(-54.0*p10*p112*p20)+(-204.0*p202)+(198.0*p11*p202)+(-45.0*p112*p202)+(-7.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-728.0*p21)+(650.0*p10*p21)+(-144.0*p102*p21)+(246.0*p11*p21)+(-234.0*p10*p11*p21)+(54.0*p102*p11*p21)+(-72.0*p10*p20*p21)+(72.0*p11*p20*p21)+(84.0*p202*p21)+(-54.0*p11*p202*p21)+(204.0*p212)+(-198.0*p10*p212)+(45.0*p102*p212)+(-84.0*p20*p212)+(54.0*p10*p20*p212)+(-7.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(-2.0*p00*(-122.0+(-893.0*p11)+(168.0*p112)+(37.0*p20)+(-10.0*p11*p20)+(14.0*p012*(-73.0+(15.0*p10)+(3.0*p20)))+(-403.0*p21)+(165.0*p11*p21)+(48.0*p20*p21)+(-18.0*p11*p20*p21)+(63.0*p212)+(-15.0*p20*p212)+(28.0*p01*(55.0+(-4.0*p20)+(p11*(14.0+p20))+(4.0*p21)+(-1.0*p10*(14.0+p21))))+(p10*(50.0+(-42.0*p112)+(145.0*p21)+(-27.0*p212)+(-21.0*p11*(-7.0+(3.0*p21)))))))+(-2.0*p01*(122.0+(42.0*p102*(-4.0+p11))+(403.0*p20)+(-63.0*p202)+(p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-37.0*p21)+(-48.0*p20*p21)+(15.0*p202*p21)+(p10*(893.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				break;
			case 25:
				p002 = p00*p00;
				p012 = p01*p01;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(p002*(-5.0+(2310.0*p012)+(-42.0*p11)+(-378.0*p112)+(-30.0*p21)+(-252.0*p11*p21)+(-63.0*p212)+(-28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(2.0*p00*(-14.0+(-6.0*p11)+(9.0*p112)+(-21.0*p20)+(-30.0*p11*p20)+(14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(-21.0*p21)+(-54.0*p20*p21)+(-54.0*p11*p20*p21)+(-9.0*p212)+(-45.0*p20*p212)+(3.0*p01*((7.0*p11)+(-2.0*p20)+(28.0*p11*p20)+(2.0*p21)+(-7.0*p10*(1.0+(4.0*p21)))))+(-3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))))+(2.0*p01*(14.0+(9.0*p102*(-1.0+(14.0*p11)))+(21.0*p20)+(9.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*((9.0*p11*(2.0+(7.0*p20)))+(2.0*(1.0+(5.0*p21)+(9.0*p20*p21)))))))+(-3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(p10*(84.0+(-6.0*p112*(5.0+(9.0*p20)))+(36.0*p11*(p20+(-1.0*p21)))+(56.0*p21)+(6.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))))));
				propagate[x][y] += ((1.0/4.0)*(-1.0+p012));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*(-3850.0+(-408.0*p10)+(63.0*p102)+(408.0*p11)+(-396.0*p10*p11)+(90.0*p102*p11)+(333.0*p112)+(-90.0*p10*p112)+(-1050.0*p20)+(252.0*p10*p20)+(-168.0*p11*p20)+(108.0*p10*p11*p20)+(432.0*p112*p20)+(-162.0*p10*p112*p20)+(378.0*p202)+(-135.0*p112*p202)+(-21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(1050.0*p21)+(-822.0*p10*p21)+(162.0*p102*p21)+(738.0*p11*p21)+(-702.0*p10*p11*p21)+(162.0*p102*p11*p21)+(-990.0*p20*p21)+(378.0*p10*p20*p21)+(216.0*p11*p20*p21)+(252.0*p202*p21)+(-162.0*p11*p202*p21)+(612.0*p212)+(-594.0*p10*p212)+(135.0*p102*p212)+(-252.0*p20*p212)+(162.0*p10*p20*p212)+(p002*(5.0+(-2310.0*p012)+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))+(28.0*p01*(1.0+(45.0*p11)+(9.0*p21)))))+(-2.0*p01*(-41.0+(9.0*p102*(-1.0+(14.0*p11)))+(-12.0*p20)+(9.0*p202)+(3.0*p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-111.0*p21)+(-144.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(13.0+(21.0*p11*(-7.0+(3.0*p20)))+(2.0*(5.0+(9.0*p20))*p21)))))+(-2.0*p00*(41.0+(126.0*p11)+(504.0*p112)+(-21.0*p20)+(-30.0*p11*p20)+(42.0*p012*(-73.0+(15.0*p10)+(3.0*p20)))+(144.0*p21)+(495.0*p11*p21)+(-54.0*p20*p21)+(-54.0*p11*p20*p21)+(189.0*p212)+(-45.0*p20*p212)+(p01*(55.0+(-6.0*p20)+(84.0*p11*(14.0+p20))+(336.0*p21)+(-21.0*p10*(1.0+(4.0*p21)))))+(-3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))))));
				propagate[x][y] += ((1.0/4.0)*(-1.0+p012));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*(-770.0+(-408.0*p10)+(-333.0*p102)+(408.0*p11)+(396.0*p10*p11)+(90.0*p102*p11)+(-63.0*p112)+(-90.0*p10*p112)+(-1050.0*p20)+(-738.0*p10*p20)+(822.0*p11*p20)+(702.0*p10*p11*p20)+(-162.0*p112*p20)+(-162.0*p10*p112*p20)+(-612.0*p202)+(594.0*p11*p202)+(-135.0*p112*p202)+(-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(1050.0*p21)+(168.0*p10*p21)+(-432.0*p102*p21)+(-252.0*p11*p21)+(-108.0*p10*p11*p21)+(162.0*p102*p11*p21)+(990.0*p20*p21)+(-216.0*p10*p20*p21)+(-378.0*p11*p20*p21)+(252.0*p202*p21)+(-162.0*p11*p202*p21)+(-378.0*p212)+(135.0*p102*p212)+(-252.0*p20*p212)+(162.0*p10*p20*p212)+(-21.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(-2.0*p00*(41.0+(-39.0*p11)+(9.0*p112)+(111.0*p20)+(-30.0*p11*p20)+(14.0*p012*(1.0+(45.0*p10)+(9.0*p20)))+(12.0*p21)+(144.0*p20*p21)+(-54.0*p11*p20*p21)+(-9.0*p212)+(-45.0*p20*p212)+(p01*(-55.0+(21.0*p11)+(-336.0*p20)+(84.0*p11*p20)+(6.0*p21)+(-84.0*p10*(14.0+p21))))+(-3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))))+(-2.0*p01*(-41.0+(126.0*p102*(-4.0+p11))+(-144.0*p20)+(-189.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(-42.0+(9.0*p11*(2.0+(7.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))))))));
				propagate[x][y] += ((-1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/6160.0)*(-770.0+(-568.0*p10)+(111.0*p102)+(568.0*p11)+(-30.0*p102*p11)+(-111.0*p112)+(30.0*p10*p112)+(-728.0*p20)+(246.0*p10*p20)+(650.0*p11*p20)+(-234.0*p10*p11*p20)+(-144.0*p112*p20)+(54.0*p10*p112*p20)+(204.0*p202)+(-198.0*p11*p202)+(45.0*p112*p202)+(7.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(728.0*p21)+(-650.0*p10*p21)+(144.0*p102*p21)+(-246.0*p11*p21)+(234.0*p10*p11*p21)+(-54.0*p102*p11*p21)+(72.0*p10*p20*p21)+(-72.0*p11*p20*p21)+(-84.0*p202*p21)+(54.0*p11*p202*p21)+(-204.0*p212)+(198.0*p10*p212)+(-45.0*p102*p212)+(84.0*p20*p212)+(-54.0*p10*p20*p212)+(7.0*p002*(5.0+(110.0*p012)+(-18.0*p112)+(-12.0*p11*(-9.0+p21))+(30.0*p21)+(-3.0*p212)+(-4.0*p01*(37.0+(15.0*p11)+(3.0*p21)))))+(2.0*p00*(-122.0+(-893.0*p11)+(168.0*p112)+(37.0*p20)+(-10.0*p11*p20)+(14.0*p012*(-73.0+(15.0*p10)+(3.0*p20)))+(-403.0*p21)+(165.0*p11*p21)+(48.0*p20*p21)+(-18.0*p11*p20*p21)+(63.0*p212)+(-15.0*p20*p212)+(28.0*p01*(55.0+(-4.0*p20)+(p11*(14.0+p20))+(4.0*p21)+(-1.0*p10*(14.0+p21))))+(p10*(50.0+(-42.0*p112)+(145.0*p21)+(-27.0*p212)+(-21.0*p11*(-7.0+(3.0*p21)))))))+(2.0*p01*(122.0+(42.0*p102*(-4.0+p11))+(403.0*p20)+(-63.0*p202)+(p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-37.0*p21)+(-48.0*p20*p21)+(15.0*p202*p21)+(p10*(893.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))))))));
				propagate[x][y] += ((-1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				break;
			case 26:
				p012 = p01*p01;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-2.0*p01*(14.0+(9.0*p102*(-1.0+(14.0*p11)))+(21.0*p20)+(9.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*((9.0*p11*(2.0+(7.0*p20)))+(2.0*(1.0+(5.0*p21)+(9.0*p20*p21)))))))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(p10*(84.0+(-6.0*p112*(5.0+(9.0*p20)))+(36.0*p11*(p20+(-1.0*p21)))+(56.0*p21)+(6.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))))));
				propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012)));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*(3850.0+(-408.0*p11)+(-333.0*p112)+(1050.0*p20)+(168.0*p11*p20)+(-432.0*p112*p20)+(-378.0*p202)+(135.0*p112*p202)+(21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-1050.0*p21)+(-738.0*p11*p21)+(990.0*p20*p21)+(-216.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(-612.0*p212)+(252.0*p20*p212)+(-9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(2.0*p01*(-41.0+(9.0*p102*(-1.0+(14.0*p11)))+(-12.0*p20)+(9.0*p202)+(3.0*p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-111.0*p21)+(-144.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(13.0+(21.0*p11*(-7.0+(3.0*p20)))+(2.0*(5.0+(9.0*p20))*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(137.0*p21)+(99.0*p212)+(p11*(66.0+(-18.0*p20)+(117.0*p21)))+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))));
				propagate[x][y] += ((1.0/4.0)*(1.0+(-1.0*p012)));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*(770.0+(-408.0*p11)+(63.0*p112)+(1050.0*p20)+(-822.0*p11*p20)+(162.0*p112*p20)+(612.0*p202)+(-594.0*p11*p202)+(135.0*p112*p202)+(p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(-1050.0*p21)+(252.0*p11*p21)+(-990.0*p20*p21)+(378.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(378.0*p212)+(252.0*p20*p212)+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(-3.0*p11*(22.0+(39.0*p20)+(-6.0*p21)))+(-28.0*p21)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(2.0*p01*(-41.0+(126.0*p102*(-4.0+p11))+(-144.0*p20)+(-189.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(-42.0+(9.0*p11*(2.0+(7.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/6160.0)*(770.0+(-568.0*p11)+(111.0*p112)+(728.0*p20)+(-650.0*p11*p20)+(144.0*p112*p20)+(-204.0*p202)+(198.0*p11*p202)+(-45.0*p112*p202)+(-7.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(-728.0*p21)+(246.0*p11*p21)+(72.0*p11*p20*p21)+(84.0*p202*p21)+(-54.0*p11*p202*p21)+(204.0*p212)+(-84.0*p20*p212)+(3.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p10*(-284.0+(3.0*p112*(5.0+(9.0*p20)))+(-117.0*p11*(p20+(-1.0*p21)))+(-325.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(-2.0*p01*(122.0+(42.0*p102*(-4.0+p11))+(403.0*p20)+(-63.0*p202)+(p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-37.0*p21)+(-48.0*p20*p21)+(15.0*p202*p21)+(p10*(893.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))))))));
				propagate[x][y] += ((1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				break;
			case 27:
				p012 = p01*p01;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(2.0*p01*(14.0+(9.0*p102*(-1.0+(14.0*p11)))+(21.0*p20)+(9.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*((9.0*p11*(2.0+(7.0*p20)))+(2.0*(1.0+(5.0*p21)+(9.0*p20*p21)))))))+(-3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(p10*(84.0+(-6.0*p112*(5.0+(9.0*p20)))+(36.0*p11*(p20+(-1.0*p21)))+(56.0*p21)+(6.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))))));
				propagate[x][y] += ((1.0/4.0)*(-1.0+p012));
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*(-3850.0+(408.0*p11)+(333.0*p112)+(-1050.0*p20)+(-168.0*p11*p20)+(432.0*p112*p20)+(378.0*p202)+(-135.0*p112*p202)+(-21.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(1050.0*p21)+(738.0*p11*p21)+(-990.0*p20*p21)+(216.0*p11*p20*p21)+(252.0*p202*p21)+(-162.0*p11*p202*p21)+(612.0*p212)+(-252.0*p20*p212)+(9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p01*(-41.0+(9.0*p102*(-1.0+(14.0*p11)))+(-12.0*p20)+(9.0*p202)+(3.0*p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-111.0*p21)+(-144.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(13.0+(21.0*p11*(-7.0+(3.0*p20)))+(2.0*(5.0+(9.0*p20))*p21)))))+(-6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(137.0*p21)+(99.0*p212)+(p11*(66.0+(-18.0*p20)+(117.0*p21)))+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))));
				propagate[x][y] += ((1.0/4.0)*(-1.0+p012));
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*(-770.0+(408.0*p11)+(-63.0*p112)+(-1050.0*p20)+(822.0*p11*p20)+(-162.0*p112*p20)+(-612.0*p202)+(594.0*p11*p202)+(-135.0*p112*p202)+(-1.0*p012*(5.0+(378.0*p102)+(30.0*p20)+(63.0*p202)+(42.0*p10*(1.0+(6.0*p20)))))+(1050.0*p21)+(-252.0*p11*p21)+(990.0*p20*p21)+(-378.0*p11*p20*p21)+(252.0*p202*p21)+(-162.0*p11*p202*p21)+(-378.0*p212)+(-252.0*p20*p212)+(9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(-3.0*p11*(22.0+(39.0*p20)+(-6.0*p21)))+(-28.0*p21)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(-2.0*p01*(-41.0+(126.0*p102*(-4.0+p11))+(-144.0*p20)+(-189.0*p202)+(3.0*p11*(5.0+(20.0*p20)+(27.0*p202)))+(21.0*p21)+(54.0*p20*p21)+(45.0*p202*p21)+(3.0*p10*(-42.0+(9.0*p11*(2.0+(7.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))))))));
				propagate[x][y] += ((-1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/6160.0)*(-770.0+(568.0*p11)+(-111.0*p112)+(-728.0*p20)+(650.0*p11*p20)+(-144.0*p112*p20)+(204.0*p202)+(-198.0*p11*p202)+(45.0*p112*p202)+(7.0*p012*(-5.0+(18.0*p102)+(12.0*p10*(-9.0+p20))+(-30.0*p20)+(3.0*p202)))+(728.0*p21)+(-246.0*p11*p21)+(-72.0*p11*p20*p21)+(-84.0*p202*p21)+(54.0*p11*p202*p21)+(-204.0*p212)+(84.0*p20*p212)+(-3.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(2.0*p10*(-284.0+(3.0*p112*(5.0+(9.0*p20)))+(-117.0*p11*(p20+(-1.0*p21)))+(-325.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(2.0*p01*(122.0+(42.0*p102*(-4.0+p11))+(403.0*p20)+(-63.0*p202)+(p11*(-50.0+(-145.0*p20)+(27.0*p202)))+(-37.0*p21)+(-48.0*p20*p21)+(15.0*p202*p21)+(p10*(893.0+(21.0*p11*(-7.0+(3.0*p20)))+(10.0*p21)+(3.0*p20*(-55.0+(6.0*p21)))))))));
				propagate[x][y] += ((-1.0/4.0)*((-1.0+p01)*(-1.0+p01)));
				break;
			case 28:
				p002 = p00*p00;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(2.0*p00*(14.0+(-9.0*p112)+(21.0*p20)+(21.0*p21)+(54.0*p20*p21)+(9.0*p212)+(45.0*p20*p212)+(6.0*p11*(1.0+(5.0*p20)+(9.0*p20*p21)))+(3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))))+(3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(p10*(84.0+(-6.0*p112*(5.0+(9.0*p20)))+(36.0*p11*(p20+(-1.0*p21)))+(56.0*p21)+(6.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))))));
				propagate[x][y] += (1.0/4.0);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*(3850.0+(-408.0*p11)+(-333.0*p112)+(1050.0*p20)+(168.0*p11*p20)+(-432.0*p112*p20)+(-378.0*p202)+(135.0*p112*p202)+(-1050.0*p21)+(-738.0*p11*p21)+(990.0*p20*p21)+(-216.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(-612.0*p212)+(252.0*p20*p212)+(-1.0*p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(-9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(137.0*p21)+(99.0*p212)+(p11*(66.0+(-18.0*p20)+(117.0*p21)))+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))+(-2.0*p00*(-41.0+(-504.0*p112)+(21.0*p20)+(-144.0*p21)+(54.0*p20*p21)+(-189.0*p212)+(45.0*p20*p212)+(3.0*p11*(-42.0+(10.0*p20)+(-165.0*p21)+(18.0*p20*p21)))+(3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))))));
				propagate[x][y] += (1.0/4.0);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*(770.0+(-408.0*p11)+(63.0*p112)+(1050.0*p20)+(-822.0*p11*p20)+(162.0*p112*p20)+(612.0*p202)+(-594.0*p11*p202)+(135.0*p112*p202)+(-1050.0*p21)+(252.0*p11*p21)+(-990.0*p20*p21)+(378.0*p11*p20*p21)+(-252.0*p202*p21)+(162.0*p11*p202*p21)+(378.0*p212)+(252.0*p20*p212)+(-21.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(-3.0*p11*(22.0+(39.0*p20)+(-6.0*p21)))+(-28.0*p21)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(-2.0*p00*(-41.0+(-9.0*p112)+(-111.0*p20)+(-12.0*p21)+(-144.0*p20*p21)+(9.0*p212)+(45.0*p20*p212)+(3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(p11*(39.0+(6.0*p20*(5.0+(9.0*p21)))))))));
				propagate[x][y] += (1.0/4.0);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/6160.0)*(770.0+(-568.0*p11)+(111.0*p112)+(728.0*p20)+(-650.0*p11*p20)+(144.0*p112*p20)+(-204.0*p202)+(198.0*p11*p202)+(-45.0*p112*p202)+(-728.0*p21)+(246.0*p11*p21)+(72.0*p11*p20*p21)+(84.0*p202*p21)+(-54.0*p11*p202*p21)+(204.0*p212)+(-84.0*p20*p212)+(7.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(3.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-2.0*p10*(-284.0+(3.0*p112*(5.0+(9.0*p20)))+(-117.0*p11*(p20+(-1.0*p21)))+(-325.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(2.0*p00*(122.0+(-168.0*p112)+(-37.0*p20)+(403.0*p21)+(-48.0*p20*p21)+(-63.0*p212)+(15.0*p20*p212)+(p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(p11*(893.0+(-165.0*p21)+(2.0*p20*(5.0+(9.0*p21)))))))));
				propagate[x][y] += (1.0/4.0);
				break;
			case 29:
				p002 = p00*p00;
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-1.0*p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(-2.0*p00*(14.0+(-9.0*p112)+(21.0*p20)+(21.0*p21)+(54.0*p20*p21)+(9.0*p212)+(45.0*p20*p212)+(6.0*p11*(1.0+(5.0*p20)+(9.0*p20*p21)))+(3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))))+(-3.0*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(p10*(84.0+(-6.0*p112*(5.0+(9.0*p20)))+(36.0*p11*(p20+(-1.0*p21)))+(56.0*p21)+(6.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))))));
				propagate[x][y] += (-1.0/4.0);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*(-3850.0+(408.0*p11)+(333.0*p112)+(-1050.0*p20)+(-168.0*p11*p20)+(432.0*p112*p20)+(378.0*p202)+(-135.0*p112*p202)+(1050.0*p21)+(738.0*p11*p21)+(-990.0*p20*p21)+(216.0*p11*p20*p21)+(252.0*p202*p21)+(-162.0*p11*p202*p21)+(612.0*p212)+(-252.0*p20*p212)+(p002*(5.0+(378.0*p112)+(30.0*p21)+(63.0*p212)+(42.0*p11*(1.0+(6.0*p21)))))+(9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(137.0*p21)+(99.0*p212)+(p11*(66.0+(-18.0*p20)+(117.0*p21)))+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))+(2.0*p00*(-41.0+(-504.0*p112)+(21.0*p20)+(-144.0*p21)+(54.0*p20*p21)+(-189.0*p212)+(45.0*p20*p212)+(3.0*p11*(-42.0+(10.0*p20)+(-165.0*p21)+(18.0*p20*p21)))+(3.0*p10*(5.0+(42.0*p112)+(20.0*p21)+(27.0*p212)+(9.0*p11*(2.0+(7.0*p21)))))))));
				propagate[x][y] += (-1.0/4.0);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*(-770.0+(408.0*p11)+(-63.0*p112)+(-1050.0*p20)+(822.0*p11*p20)+(-162.0*p112*p20)+(-612.0*p202)+(594.0*p11*p202)+(-135.0*p112*p202)+(1050.0*p21)+(-252.0*p11*p21)+(990.0*p20*p21)+(-378.0*p11*p20*p21)+(252.0*p202*p21)+(-162.0*p11*p202*p21)+(-378.0*p212)+(-252.0*p20*p212)+(21.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(-3.0*p11*(22.0+(39.0*p20)+(-6.0*p21)))+(-28.0*p21)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(2.0*p00*(-41.0+(-9.0*p112)+(-111.0*p20)+(-12.0*p21)+(-144.0*p20*p21)+(9.0*p212)+(45.0*p20*p212)+(3.0*p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(p11*(39.0+(6.0*p20*(5.0+(9.0*p21)))))))));
				propagate[x][y] += (-1.0/4.0);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/6160.0)*(-770.0+(568.0*p11)+(-111.0*p112)+(-728.0*p20)+(650.0*p11*p20)+(-144.0*p112*p20)+(204.0*p202)+(-198.0*p11*p202)+(45.0*p112*p202)+(728.0*p21)+(-246.0*p11*p21)+(-72.0*p11*p20*p21)+(-84.0*p202*p21)+(54.0*p11*p202*p21)+(-204.0*p212)+(84.0*p20*p212)+(-7.0*p002*(-5.0+(18.0*p112)+(12.0*p11*(-9.0+p21))+(-30.0*p21)+(3.0*p212)))+(-3.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(2.0*p10*(-284.0+(3.0*p112*(5.0+(9.0*p20)))+(-117.0*p11*(p20+(-1.0*p21)))+(-325.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))+(-2.0*p00*(122.0+(-168.0*p112)+(-37.0*p20)+(403.0*p21)+(-48.0*p20*p21)+(-63.0*p212)+(15.0*p20*p212)+(p10*(-50.0+(42.0*p112)+(-145.0*p21)+(27.0*p212)+(21.0*p11*(-7.0+(3.0*p21)))))+(p11*(893.0+(-165.0*p21)+(2.0*p20*(5.0+(9.0*p21)))))))));
				propagate[x][y] += (-1.0/4.0);
				break;
			case 30:
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/6160.0)*((-3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(p10*(84.0+(-6.0*p112*(5.0+(9.0*p20)))+(36.0*p11*(p20+(-1.0*p21)))+(56.0*p21)+(6.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))));
				propagate[x][y] += (1.0/4.0);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)))+(6.0*p11*(-68.0+(-123.0*p21)+(27.0*p202*p21)+(-4.0*p20*(-7.0+(9.0*p21)))))+(-9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(137.0*p21)+(99.0*p212)+(p11*(66.0+(-18.0*p20)+(117.0*p21)))+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))+(-2.0*(-1925.0+(525.0*p21)+(306.0*p212)+(63.0*p202*(3.0+(2.0*p21)))+(-3.0*p20*(175.0+(165.0*p21)+(42.0*p212)))))));
				propagate[x][y] += (1.0/4.0);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*(770.0+(1050.0*p20)+(612.0*p202)+(9.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(-1050.0*p21)+(-990.0*p20*p21)+(-252.0*p202*p21)+(378.0*p212)+(252.0*p20*p212)+(-9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(6.0*p11*(-68.0+(42.0*p21)+(9.0*p202*(-11.0+(3.0*p21)))+(p20*(-137.0+(63.0*p21)))))+(6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(-3.0*p11*(22.0+(39.0*p20)+(-6.0*p21)))+(-28.0*p21)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))));
				propagate[x][y] += (1.0/4.0);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/6160.0)*(770.0+(728.0*p20)+(-204.0*p202)+(p112*(111.0+(144.0*p20)+(-45.0*p202)))+(-728.0*p21)+(84.0*p202*p21)+(204.0*p212)+(-84.0*p20*p212)+(3.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(p11*(-568.0+(p202*(198.0+(-54.0*p21)))+(246.0*p21)+(p20*(-650.0+(72.0*p21)))))+(-2.0*p10*(-284.0+(3.0*p112*(5.0+(9.0*p20)))+(-117.0*p11*(p20+(-1.0*p21)))+(-325.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))));
				propagate[x][y] += (1.0/4.0);
				break;
			case 31:
				x = 0; y = 0;
				pixel[y][-x] += ((1.0/6160.0)*((3.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(2.0*p11*((27.0*p202*p21)+(42.0*(1.0+p21))+(7.0*p20*(4.0+(9.0*p21)))))+(-3.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-14.0*(55.0+(-30.0*p21)+(-9.0*p212)+(p202*(9.0+(6.0*p21)))+(-6.0*p20*(-5.0+p212))))+(2.0*p10*(-42.0+(3.0*p112*(5.0+(9.0*p20)))+(-18.0*p11*(p20+(-1.0*p21)))+(-28.0*p21)+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))));
				propagate[x][y] += (-1.0/4.0);
				x = 1; y = 0;
				pixel[y][-x] += ((1.0/18480.0)*((-9.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)))+(-6.0*p11*(-68.0+(-123.0*p21)+(27.0*p202*p21)+(-4.0*p20*(-7.0+(9.0*p21)))))+(9.0*p102*(7.0+(18.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(137.0*p21)+(99.0*p212)+(p11*(66.0+(-18.0*p20)+(117.0*p21)))+(-3.0*p20*(14.0+(21.0*p21)+(9.0*p212)))))+(2.0*(-1925.0+(525.0*p21)+(306.0*p212)+(63.0*p202*(3.0+(2.0*p21)))+(-3.0*p20*(175.0+(165.0*p21)+(42.0*p212)))))));
				propagate[x][y] += (-1.0/4.0);
				x = 0; y = 1;
				pixel[y][-x] += ((1.0/18480.0)*(-770.0+(-1050.0*p20)+(-612.0*p202)+(-9.0*p112*(7.0+(18.0*p20)+(15.0*p202)))+(1050.0*p21)+(990.0*p20*p21)+(252.0*p202*p21)+(-378.0*p212)+(-252.0*p20*p212)+(9.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(-6.0*p11*(-68.0+(42.0*p21)+(9.0*p202*(-11.0+(3.0*p21)))+(p20*(-137.0+(63.0*p21)))))+(-6.0*p10*(68.0+(3.0*p112*(5.0+(9.0*p20)))+(-3.0*p11*(22.0+(39.0*p20)+(-6.0*p21)))+(-28.0*p21)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))));
				propagate[x][y] += (-1.0/4.0);
				x = 1; y = 1;
				pixel[y][-x] += ((1.0/6160.0)*(-770.0+(-728.0*p20)+(204.0*p202)+(3.0*p112*(-37.0+(-48.0*p20)+(15.0*p202)))+(728.0*p21)+(-84.0*p202*p21)+(-204.0*p212)+(84.0*p20*p212)+(2.0*p11*(284.0+(p20*(325.0+(-36.0*p21)))+(-123.0*p21)+(9.0*p202*(-11.0+(3.0*p21)))))+(-3.0*p102*(-37.0+(-48.0*p21)+(15.0*p212)+(2.0*p11*(5.0+(9.0*p21)))))+(2.0*p10*(-284.0+(3.0*p112*(5.0+(9.0*p20)))+(-117.0*p11*(p20+(-1.0*p21)))+(-325.0*p21)+(99.0*p212)+(p20*(123.0+(36.0*p21)+(-27.0*p212)))))));
				propagate[x][y] += (-1.0/4.0);
				break;
			}
#else
			int sign = c.flip_x != c.flip_y ? -1 : 1;
p002 = p00*p00;
p012 = p01*p01;
p302 = p30*p30;
p312 = p31*p31;
x = 0; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(-2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312)*sign);
x = 1; y = 0;
pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-1.0*p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(-22.0*((27.0*p112*p20)+(-27.0*p10*p11*p21)+(27.0*p11*p20*p21)+(-27.0*p10*p212)+(18.0*p112*p30)+(45.0*p11*p21*p30)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(-18.0*p10*p11*p31)+(-45.0*p10*p21*p31)+(-45.0*p20*p21*p31)+(30.0*p11*p30*p31)+(105.0*p21*p30*p31)+(-30.0*p10*p312)+(-105.0*p20*p312)+(-280.0*p30*p312)+(p01*((45.0*p10*p11)+(45.0*p11*p20)+(18.0*p20*p21)+(15.0*p11*p30)+(12.0*p21*p30)+(-3.0*p10*p31)+(3.0*p20*p31)+(5.0*p30*p31)))+(p00*((280.0*p012)+(-45.0*p112)+(-45.0*p11*p21)+(-18.0*p212)+(-12.0*p11*p31)+(-15.0*p21*p31)+(-5.0*p312)+(-5.0*p01*((21.0*p11)+(6.0*p21)+p31))))))+(2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-1.0*p012)+p312)*sign);
x = 0; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(-1.0*p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-22.0*((45.0*p01*p102)+(45.0*p01*p10*p20)+(27.0*p10*p11*p20)+(18.0*p01*p202)+(27.0*p11*p202)+(-27.0*p102*p21)+(-27.0*p10*p20*p21)+(12.0*p01*p10*p30)+(18.0*p10*p11*p30)+(15.0*p01*p20*p30)+(45.0*p11*p20*p30)+(45.0*p20*p21*p30)+(5.0*p01*p302)+(30.0*p11*p302)+(105.0*p21*p302)+(5.0*p002*((28.0*p01)+(-21.0*p11)+(-6.0*p21)+(-1.0*p31)))+(-18.0*p102*p31)+(-45.0*p10*p20*p31)+(-45.0*p202*p31)+(-30.0*p10*p30*p31)+(-105.0*p20*p30*p31)+(-140.0*p302*p31)+(p00*((-18.0*p20*p21)+(3.0*p11*p30)+(-3.0*p21*p30)+(5.0*p01*((21.0*p10)+(6.0*p20)+p30))+(-12.0*p20*p31)+(-5.0*p30*p31)+(-15.0*p10*((3.0*p11)+(3.0*p21)+p31))))))+(2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(-3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31))*sign);
x = 1; y = 1;
pixel[y][-x] += ((1.0/18480.0)*((-1.0*p012*((378.0*p102)+(63.0*p202)+(30.0*p20*p30)+(5.0*p302)+(42.0*p10*((6.0*p20)+p30))))+(p002*((-2310.0*p012)+(378.0*p112)+(63.0*p212)+(30.0*p21*p31)+(5.0*p312)+(42.0*p11*((6.0*p21)+p31))+(28.0*p01*((45.0*p11)+(9.0*p21)+p31))))+(-22.0*((126.0*p11*p20)+(-27.0*p10*p11*p20)+(-27.0*p11*p202)+(-126.0*p10*p21)+(27.0*p102*p21)+(27.0*p10*p20*p21)+(126.0*p11*p30)+(-18.0*p10*p11*p30)+(-45.0*p11*p20*p30)+(252.0*p21*p30)+(-45.0*p20*p21*p30)+(-30.0*p11*p302)+(-105.0*p21*p302)+(-1.0*p01*((45.0*p102)+(18.0*p202)+(3.0*p10*(-84.0+(15.0*p20)+(4.0*p30)))+(3.0*p20*(-42.0+(5.0*p30)))+(p30*(-42.0+(5.0*p30)))))+(-126.0*p10*p31)+(18.0*p102*p31)+(-252.0*p20*p31)+(45.0*p10*p20*p31)+(45.0*p202*p31)+(-420.0*p30*p31)+(30.0*p10*p30*p31)+(105.0*p20*p30*p31)+(140.0*p302*p31)+(5.0*p002*((-28.0*p01)+(21.0*p11)+(6.0*p21)+p31))+(p00*((-126.0*p21)+(45.0*p10*p21)+(18.0*p20*p21)+(3.0*p11*(-84.0+(15.0*p10)+(-1.0*p30)))+(3.0*p21*p30)+(-5.0*p01*(-84.0+(21.0*p10)+(6.0*p20)+p30))+(-42.0*p31)+(15.0*p10*p31)+(12.0*p20*p31)+(5.0*p30*p31)))))+(-2.0*p00*((-126.0*p10*p112)+(-189.0*p10*p11*p21)+(-54.0*p11*p20*p21)+(-81.0*p10*p212)+(-45.0*p20*p212)+(9.0*p112*p30)+(-9.0*p212*p30)+(14.0*p012*((45.0*p10)+(9.0*p20)+p30))+(-54.0*p10*p11*p31)+(-30.0*p11*p20*p31)+(-60.0*p10*p21*p31)+(-54.0*p20*p21*p31)+(-6.0*p11*p30*p31)+(-21.0*p21*p30*p31)+(-15.0*p10*p312)+(-21.0*p20*p312)+(-14.0*p30*p312)+(3.0*p01*((2.0*p21*p30)+(7.0*p11*((4.0*p20)+p30))+(-2.0*p20*p31)+(-7.0*p10*((4.0*p21)+p31))))))+(22.0*((27.0*p112*p20)+(-27.0*p10*p11*p21)+(27.0*p11*p20*p21)+(-27.0*p10*p212)+(18.0*p112*p30)+(45.0*p11*p21*p30)+(45.0*p212*p30)+(5.0*p012*((21.0*p10)+(6.0*p20)+p30))+(-18.0*p10*p11*p31)+(-45.0*p10*p21*p31)+(-45.0*p20*p21*p31)+(30.0*p11*p30*p31)+(105.0*p21*p30*p31)+(-30.0*p10*p312)+(-105.0*p20*p312)+(-280.0*p30*p312)+(p01*((45.0*p10*p11)+(45.0*p11*p20)+(18.0*p20*p21)+(15.0*p11*p30)+(12.0*p21*p30)+(-3.0*p10*p31)+(3.0*p20*p31)+(5.0*p30*p31)))+(p00*((280.0*p012)+(-45.0*p112)+(-45.0*p11*p21)+(-18.0*p212)+(-12.0*p11*p31)+(-15.0*p21*p31)+(-5.0*p312)+(-5.0*p01*((21.0*p11)+(6.0*p21)+p31))))))+(-2.0*p01*((45.0*p202*p21)+(54.0*p20*p21*p30)+(21.0*p21*p302)+(3.0*p11*((27.0*p202)+(20.0*p20*p30)+(5.0*p302)))+(9.0*p102*((14.0*p11)+(-1.0*p31)))+(9.0*p202*p31)+(21.0*p20*p30*p31)+(14.0*p302*p31)+(3.0*p10*((9.0*p11*((7.0*p20)+(2.0*p30)))+(2.0*((9.0*p20*p21)+(5.0*p21*p30)+(p30*p31)))))))+(3.0*((-3.0*p112*((15.0*p202)+(18.0*p20*p30)+(7.0*p302)))+(-2.0*p11*((27.0*p202*p21)+(42.0*p302*(p21+p31))+(7.0*p20*p30*((9.0*p21)+(4.0*p31)))))+(3.0*p102*((15.0*p212)+(18.0*p21*p31)+(7.0*p312)+(2.0*p11*((9.0*p21)+(5.0*p31)))))+(p10*((-6.0*p112*((9.0*p20)+(5.0*p30)))+(28.0*p30*p31*((2.0*p21)+(3.0*p31)))+(-36.0*p11*((p21*p30)+(-1.0*p20*p31)))+(6.0*p20*((9.0*p212)+(21.0*p21*p31)+(14.0*p312)))))+(14.0*((3.0*p202*p31*((2.0*p21)+(3.0*p31)))+(-6.0*p20*p30*(p212+(-5.0*p312)))+(p302*((-9.0*p212)+(-30.0*p21*p31)+(55.0*p312))))))))*sign);
propagate[x][y] += ((1.0/4.0)*((-2.0*p01)+p012+(-1.0*(-2.0+p31)*p31))*sign);
#endif

#endif
#endif
		}
	}
};
#endif


#ifdef FILTER2
struct FilterBSpline2
{
	static const int w = 3, h = 3;
	static double precalc[h];
	static const double threshhold;

	void add_to_vals(double *pixel[h], double (&propagate)[w][h], LineRaster &c);
};
#endif


#ifdef FILTER3
struct FilterCubicLanczos2
{
	static const int w = 4, h = 4;
	static double precalc[h];
	static const double threshhold;

	void add_to_vals(double *pixel[h], double (&propagate)[w][h], LineRaster &c);
};
#endif


#ifdef FILTER4
struct FilterMitchellNetravali
{
	static const int w = 4, h = 4;
	static double precalc[h];
	static const double threshhold;

	void add_to_vals(double *pixel[h], double (&propagate)[w][h], LineRaster &c);
};
#endif


#ifdef FILTER5
struct FilterCubicLanczos3
{
	static const int w = 6, h = 6;
	static double precalc[h];
	static const double threshhold;

	void add_to_vals(double *pixel[h], double (&propagate)[w][h], LineRaster &c);
};
#endif




#ifdef FILTER6
struct FilterRadialSincSinTotal2
{
	static const int w = 4, h = 4;
	static double precalc[h];
	static const double threshhold;

	void add_to_vals(double *pixel[h], double (&propagate)[w][h], LineRaster &c);
};
#endif




#ifdef FILTER7
struct FilterRadialSincSinTotal3
{
	static const int w = 6, h = 6;
	static double precalc[h];
	static const double threshhold;

	void add_to_vals(double *pixel[h], double (&propagate)[w][h], LineRaster &c);
};
#endif
